/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef archivectrl_INCLUDED
#define archivectrl_INCLUDED

#include "archiver-common.h"

#include "CommonUtil.h"

#include <list>
#include <string>
#include <memory>

#include "mapi_ptr.h"

#include "archiver.h"

class Session;
class ECConfig;
class ECLogger;
class IArchiveOperation;


/**
 * This class is the entry point to the archiving system. It's responsible for executing the actual archive
 * operation.
 */
class ArchiveControlImpl : public ArchiveControl
{
public:
	static eResult Create(Session *lpSession, ECConfig *lpConfig, ECLogger *lpLogger, ArchiveControlPtr *lpptrArchiveControl);

	eResult ArchiveAll(bool bLocalOnly);
	eResult Archive(const char *lpszUser);

	~ArchiveControlImpl();
	
private:
	ArchiveControlImpl(Session *lpSession, ECConfig *lpConfig, ECLogger *lpLogger);
	HRESULT Init();
	
	HRESULT ProcessFolder(MAPIFolderPtr &ptrFolder, IArchiveOperation *lpArchiveOperation, int ulAge, bool bProcessUnread);
	HRESULT ArchiveAllFromContainer(ABContainerPtr &ptrABContainer, bool bLocalOnly);

	HRESULT PurgeArchives(const ObjectEntryList &lstArchives);
	HRESULT PurgeArchiveFolder(MsgStorePtr &ptrArchive, const entryid_t &folderEntryID, const LPSRestriction lpRestriction);

private:
	Session	*m_lpSession;
	ECConfig *m_lpConfig;
	ECLogger *m_lpLogger;

	FILETIME m_ftCurrent;
	bool m_bArchiveEnable;
	int m_ulArchiveAfter;

	bool m_bDeleteEnable;
	bool m_bDeleteUnread;
	int m_ulDeleteAfter;

	bool m_bStubEnable;
	bool m_bStubUnread;
	int m_ulStubAfter;

	bool m_bPurgeEnable;
	int m_ulPurgeAfter;
	
	PROPMAP_START
	PROPMAP_DEF_NAMED_ID(ARCHIVE_STORE_ENTRYIDS)
	PROPMAP_DEF_NAMED_ID(ARCHIVE_ITEM_ENTRYIDS)
	PROPMAP_DEF_NAMED_ID(STUBBED)
};

#endif // ndef archivectrl_INCLUDED
