/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include "Http.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/**
 * Default constructor
 * @param[in]	lpChannel	The channel to communicate with client
 * @param[in]	lpLogger	The Logger to log errors and information 
 */
Http::Http(ECChannel *lpChannel, ECLogger *lpLogger)
{
	m_lpChannel = lpChannel;
	m_lpLogger = lpLogger;

	m_ulContLength = 0;
	m_ulKeepAlive = 0;

	m_bIsFolder = false;

	m_strRespHeader.clear();
	m_bHeadersSent = false;
}

/**
 * Default destructor
 */
Http::~Http()
{
	
}
/**
 * Reads the http headers from the channel and Parses them
 *
 * @return	HRESULT
 * @retval	MAPI_E_INVALID_PARAMETER	The http hearders are invalid
 */
HRESULT Http::HrReadHeaders()
{
	HRESULT hr = hrSuccess;
	std::string strBuffer;

	m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Receiving headers:");
	do
	{
		hr = m_lpChannel->HrReadLine(&strBuffer);
		if (hr != hrSuccess)
			goto exit;

		if (strBuffer.empty())
			break;

		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "< "+strBuffer);

		m_strReqHeaders.append(strBuffer+"\n");
	} while(hr == hrSuccess);

	m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "-- headers done --");

	hr = HrParseHeaders();

exit:
	return hr;
}

/**
 * Parse the http headers
 * @return	HRESULT
 * @retval	MAPI_E_INVALID_PARAMETER	The http headers are invalid
 */
HRESULT Http::HrParseHeaders()
{
	HRESULT hr = hrSuccess;
	std::string strAuthdata;
	std::string strBase64Data;
	std::string strSubPath;
	std::string strDepth;
	std::string strElout;
	std::string strPath;
	std::string strTrail;
	std::string strHttpVer;
	size_t ulStart = 0;
	size_t ulEnd = 0;
	size_t ulLength = 0;
	ECIConv *lpConvU2W = new ECIConv("windows-1252", "utf-8" );
	ECIConv *lpConvW2U = new ECIConv("utf-8", "windows-1252");
	char *lpURI = NULL;

	// find the http method
	m_strMethod = m_strReqHeaders.substr(0, m_strReqHeaders.find(" "));

	// find the path
	ulStart = m_strReqHeaders.find("/") + 1;
	if(ulStart == string::npos)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	ulEnd = m_strReqHeaders.find("HTTP");
	if(ulEnd == string::npos)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	m_strHttpVer = m_strReqHeaders.substr(ulEnd + 5, 3);

	ulLength = ulEnd - ulStart - 1;
	strPath = m_strReqHeaders.substr(ulStart, ulLength);
	m_strPath = "/" + strPath;
	
	// find the service ical/caldav
	ulEnd = strPath.find("/");
	if(ulEnd == string::npos)
	{
		m_strService = strPath;
		strPath.clear();
	}
	else
	{
		m_strService = strPath.substr(0, ulEnd);
		strPath = strPath.erase(0, m_strService.length()+1);
	}

	m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Request Method: " + m_strMethod);
	m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Request Service: "+ m_strService);

	//change case of Service name ICAL -> ical CALDaV ->caldav
	std::transform(m_strService.begin(), m_strService.end(), m_strService.begin(), (int(*)(int)) std::tolower);

	if(!m_strService.compare("ical") || !m_strService.compare("caldav"))
	{
		//find the CalPath (remaining is the Calendar path.)
		m_strCalPath = strPath;
	}

	lpURI = xmlURIUnescapeString(( const char *)m_strCalPath.c_str(), m_strCalPath.size(), NULL);
	if (lpURI) {
		m_strCalPath = lpURI;
		xmlFree(lpURI);
		lpURI = NULL;
	}
	if (!m_strCalPath.empty())
	{
		// Folder names of the format %C3%43%A2 are converted to utf-8 char
		m_strCalPath = lpConvU2W->convert(m_strCalPath);
		m_strCalPath = lpConvW2U->convert(m_strCalPath);		
	}

	m_strCalName = m_strCalPath;
	
	//Trim Calendar Path.
	if(!m_strCalPath.empty())
		strTrail = m_strCalName.at(m_strCalPath.length() - 1 );

	if(!strTrail.compare("/")) {
		m_strCalPath.erase(m_strCalPath.end()-1);
		m_bIsFolder = true;
	}

	if(!m_strCalPath.empty())
	{
		ulStart = m_strCalPath.find("/");
		
		if(ulStart != string::npos)
		{
			
			m_strPublic = m_strCalPath.substr(0, ulStart);
			m_strCalPath = m_strPublic.compare("public")? m_strCalPath : m_strCalPath.erase(0, m_strPublic.length() + 1);
			ulStart = m_strCalPath.find("/");
			m_strFldOwner = m_strPublic.compare("public")? m_strCalPath.substr(0 , ulStart): string();
			m_strCalPath = m_strFldOwner.empty()? m_strCalPath : m_strCalPath.erase( 0 , m_strFldOwner.size() + 1);
		}
		else
		{
			ulStart = m_strCalPath.find(".ics");
			if(ulStart == string::npos)
			{
				m_strPublic = m_strCalPath.compare("public")? string(): m_strCalPath;
				m_strFldOwner = m_strCalPath.compare("public")? m_strCalPath : string();
				m_strCalPath = string ();
			}
		}
	}
	// find the host
	hr = HrGetElementValue("Host: ", m_strReqHeaders, &m_strHost);
	if (hr != hrSuccess)
		goto exit;

	// Find the User-agent
	HrGetElementValue("User-Agent: ", m_strReqHeaders, &m_strUserAgent);

	//Find The connection type
	HrGetElementValue("Connection: ", m_strReqHeaders, &m_strConnection);

	// find the content-type
	// Content-Type: text/xml;charset=UTF-8
	hr = HrGetElementValue("Content-Type: ", m_strReqHeaders, &m_strCharSet);
	if (hr == hrSuccess && m_strCharSet.find("charset") != std::string::npos)
		m_strCharSet = m_strCharSet.substr(m_strCharSet.find("charset")+ strlen("charset") + 1, m_strCharSet.length());
	else
		m_strCharSet.clear();

	// Find If-Match or If Condition
	HrGetElementValue("If-Match: ", m_strReqHeaders, &m_strIfMatch);

	if(m_strIfMatch.empty())
	{
		HrGetElementValue("If: ", m_strReqHeaders, &m_strIfMatch);
		if(!m_strIfMatch.empty())
		{
			ulStart = m_strIfMatch.find("[") + 1;
			ulEnd = m_strIfMatch.find("]");
			m_strIfMatch = m_strIfMatch.substr(ulStart, ulEnd - ulStart);
		}
	}
	
	if (!m_strIfMatch.empty() && 
		((m_strIfMatch.at(0) == '\"' && m_strIfMatch.at(m_strIfMatch.size()-1) == '\"') ||
		(m_strIfMatch.at(0) == '\'' && m_strIfMatch.at(m_strIfMatch.size()-1) == '\'')))
		m_strIfMatch = m_strIfMatch.substr( 1, m_strIfMatch.size() - 2);

	// find the Content-Length
	hr = HrGetElementValue("Content-Length: ", m_strReqHeaders, &strElout);
	if (hr != hrSuccess)
		m_ulContLength = 0;
	else
		m_ulContLength = atoi( (char*)strElout.c_str());

	// find the Depth
	hr = HrGetElementValue("Depth: ", m_strReqHeaders, &strDepth);
	if (hr != hrSuccess)
		strDepth = "0";		// default is no subfolders

	m_ulDepth = atoi(strDepth.c_str());

	// find the Authorisation data (Authorization: Basic wr8y273yr2y3r87y23ry7=)
	hr = HrGetElementValue("Authorization: ", m_strReqHeaders, &strAuthdata);
	if (hr != hrSuccess)
	{
		hr = HrGetElementValue("WWW-Authenticate: ", m_strReqHeaders, &strAuthdata);
		//ignore empty Authorization
		if (hr != hrSuccess)
		{
			hr = S_OK;
			goto exit;
		}
	}

	// we only support basic authentication
	hr = HrGetElementValue("Basic ", strAuthdata, &strBase64Data);
	if (hr != hrSuccess)
		goto exit;

	strAuthdata = base64_decode(strBase64Data);
	m_strUser = strAuthdata.substr(0, strAuthdata.find(":"));
	m_strPass = strAuthdata.substr(strAuthdata.find(":")+1, strAuthdata.find("^\n"));

	m_lpLogger->Log(EC_LOGLEVEL_INFO, "Request from User : %s", m_strUser.c_str());

exit:

	if(lpConvU2W)
		delete lpConvU2W;

	if(lpConvW2U)
		delete lpConvW2U;

	return hr;
}

/**
 * Returns the Element Value in Header
 * 
 * for e.g. Host: zarafa.com:8080
 * HrGetElementValue("Host: ",strInput,strReturn) returns zarafa.com:8080
 *
 * @param[in]	strElement	The name of the element whose value is needed
 * @param[in]	strInput	The Input headers which has to be searched for element
 * @param[out]	strValue	Return string for the value of the element
 *
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	The element not in the input headers
 */
HRESULT Http::HrGetElementValue(std::string strElement, std::string strInput, std::string *strValue)
{
	HRESULT hr = hrSuccess;
	int ulEnd = -1;//initialized with -1 as 0 indicates first string element
	int ulStart = -1;//initialized with -1 as 0 indicates first string element
	int ulLength = 0;

	ulStart = strInput.find(strElement);
	if(ulStart < 0) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	ulStart = strInput.find(strElement) + strElement.length();
	ulEnd = strInput.find("\n", strInput.find(strElement));
	ulLength = ulEnd - ulStart ;

	strValue->assign(strInput.substr(ulStart, ulLength));

exit:
	return hr;
}
/**
 * Returns the user name set in the request
 * @param[in]	strUser		Return string for username in request
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	No username set in the request
 */
HRESULT Http::HrGetUser(std::string *strUser)
{
	HRESULT hr = hrSuccess;
	if (!m_strUser.empty())
		strUser->assign(m_strUser);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}

/**
 * Check if the request is for accessing public folders
 * @return		bool	True if the request is for public folders
 */
bool Http::IsPublic()
{
	return (!m_strPublic.compare("public"));
}

/** 
 * Returns true if the request is on a folder. False if it is for a message.
 * 
 * This is determained by the fact that the URL ended with a /
 *
 * @return request for a folder or not
 */
bool Http::IsFolderRequest()
{
	return m_bIsFolder;
}

/**
 * Returns the method set in the url
 * @param[out]	strMethod	Return string for method set in request
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	Empty method in request
 */
HRESULT Http::HrGetMethod(std::string *strMethod)
{
	HRESULT hr = hrSuccess;

	if (!m_strMethod.empty())
		strMethod->assign(m_strMethod);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}
/**
 * Returns the password sent by user
 * @param[out]	strPass		The password is returned
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	Empty password in request
 */
HRESULT Http::HrGetPass(std::string *strPass)
{
	HRESULT hr = hrSuccess;
	if (!m_strPass.empty())
        strPass->assign(m_strPass);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}

/**
 * Returns the full path of request(e.g. /caldav/user/folder)
 * @param[out]	strReqPath		Return string for path
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	Empty path in request
 */
HRESULT Http::HrGetReqPath(std::string *strReqPath)
{
	HRESULT hr = hrSuccess;

	if (!m_strPath.empty())
        strReqPath->assign(m_strPath);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}
/**
 * Returns the full path of request
 * @param[out]	strPath		Return string for path
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	Empty path in request
 */
HRESULT Http::HrGetFldPath(std::string *strPath)
{
	HRESULT hr = hrSuccess;
	if (!m_strCalPath.empty())
		strPath->assign(m_strCalPath);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}

/**
 * Returns the folder owner name, only present while accessing single folders or shared folder
 * @param[out]	strFldOwner		Return string  for folder's owner name
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	Empty owners name in request
 */
HRESULT Http::HrGetFldOwner(std::string *strFldOwner)
{
	HRESULT hr = hrSuccess;
	if (!m_strFldOwner.empty())
		strFldOwner->assign(m_strFldOwner);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}

/**
 * Returns the Folders full name 
 * @param[out]	strCalName		Return full name of the folder
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	Empty folder name in request
 */
HRESULT Http::HrGetCalName(std::string *strCalName)
{
	HRESULT hr = hrSuccess;
	if (!m_strCalName.empty())
		strCalName->assign(m_strCalName);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}
/**
 * Returns the User agent of the request
 * @param[out]	strAgent	Return user-agent
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	No user-agent string present in request
 */
HRESULT Http::HrGetUserAgent(std::string *strAgent)
{
	HRESULT hr = hrSuccess;
	if (!m_strUserAgent.empty())
		strAgent->assign(m_strUserAgent);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}

/**
 * Returns service type i.e caldav or ical request
 *
 * service type is identified fromt the url prefix,
 * caldav url	-> http://server/caldav/....
 * ical url		-> http://server/ical/....
 *
 * @param[out]	strService	Return string for service
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	No service is set in the request
 */
HRESULT Http::HrGetService(std::string *strService)
{
	HRESULT hr = hrSuccess;
	if (!m_strService.empty())
		strService->assign(m_strService);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}

/**
 * Returns body of the request
 * @param[in]	strBody		Return string for  body of the request
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	No body present in request
 */
HRESULT Http::HrGetBody(std::string *strBody)
{
	HRESULT hr = hrSuccess;
	if (!m_strReqBody.empty())
		strBody->assign(m_strReqBody);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}

/**
 * Returns the Depth set in request
 * @param[out]	ulDepth		Return string for depth set in request
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	No depth value set in request
 */
HRESULT Http::HrGetDepth(ULONG *ulDepth)
{
	HRESULT hr = hrSuccess;
	if(m_ulDepth >= 0)
		*ulDepth = m_ulDepth;
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}
/**
 * Returns Match, If-Match value of request
 * 
 * Used to compare with etag values
 *
 * @param[out]	strIfMatch	Return string for Match, If-match value set in request
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	No Match, If-match value set in request
 */
HRESULT Http::HrGetIfMatch(std::string *strIfMatch)
{
	HRESULT hr = hrSuccess;
	if(!m_strIfMatch.empty())
		strIfMatch->assign(m_strIfMatch);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}
/**
 * Returns Charset of the request
 * @param[out]	strCharset	Return string Charset of the request
 *
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	No charset set in request
 */
HRESULT Http::HrGetCharSet(std::string *strCharset)
{
	HRESULT hr = hrSuccess;

	if(!m_strCharSet.empty())
		strCharset->assign(m_strCharSet);
	else
		hr = MAPI_E_NOT_FOUND;

	return hr;
}
/**
 * Reads request body from the channel
 * @return		HRESULT
 * @retval		MAPI_E_NOT_FOUND	Empty body
 */
HRESULT Http::HrReadBody()
{
	HRESULT hr = hrSuccess;

	if(m_ulContLength != 0)
	{
		hr = m_lpChannel->HrReadBytes(&m_strReqBody, m_ulContLength);
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Request body:\n%s\n", m_strReqBody.c_str());
		
	}
	else
		hr = MAPI_E_NOT_FOUND;
	
	return hr;
}

/**
 * Check for errors in the http request
 * @return		HRESULT
 * @retval		MAPI_E_INVALID_PARAMETER	Unsupported http request
 */
HRESULT Http::HrValidateReq()
{
	HRESULT hr = hrSuccess;
	char *lpszMethods[] = {"HEAD", "ACL","GET","POST","PUT","DELETE","TRACE","OPTIONS","CONNECT","PROPFIND","REPORT","MKCALENDAR" ,"PROPPATCH" , NULL};
	bool bFound = false;
	int i = 0;

	for(i = 0; lpszMethods[i] != NULL; i++)
	{
		if (m_strMethod.compare(lpszMethods[i]) == 0)
		{
			bFound = true;
			break;
		}
	}

	if(bFound == false)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// validate authentication data
	if (m_strUser.empty() || m_strPass.empty())
	{
		// hr still success, since http request is valid
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Request Missing Authorization Data");
	}

exit:
	return hr;
}

/**
 * Sets keep-alive time for the http response
 *
 * @param[in]	ulKeepAlive		Numerical value set as keep-alive time of http connection
 * @return		HRESULT			Always set as hrSuccess
 */
HRESULT Http::HrSetKeepAlive(int ulKeepAlive)
{
	m_ulKeepAlive = ulKeepAlive;
	return hrSuccess;
}

/**
 * Flush all headers and body to client(i.e Send all data to client)
 *
 * @return	HRESULT
 * @retval	MAPI_E_END_OF_SESSION	States that client as set connection type as closed
 */
HRESULT Http::HrFinalize()
{
	HRESULT hr = hrSuccess;
	std::string strBodyChunked;
	std::string strBodyTemp;
	ULONG ulLength = HTTP_CHUNK_SIZE;

	// force chunked http for long size response
	if (!m_bHeadersSent && (m_strRespBody.size() < HTTP_CHUNK_SIZE || m_strHttpVer.compare("1.1") != 0) ) {
		HrResponseHeader("Content-Length", stringify(m_strRespBody.length()));
		hr = HrFlushHeaders();
		if (hr != hrSuccess && hr != MAPI_E_END_OF_SESSION)
			goto exit;
		
		if (!m_strRespBody.empty()) {
			m_lpChannel->HrWriteLine(m_strRespBody);
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Response body:\n%s", m_strRespBody.c_str());
		} else {
			m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Response body is empty");
		}
	}
	else 
	{
		while (!m_strRespBody.empty())
		{
			strBodyChunked = m_strRespBody.substr(0, ulLength);
			m_strRespBody = m_strRespBody.substr(ulLength, m_strRespBody.length());
			if(m_strRespBody.length() <= ulLength && !m_strRespBody.empty())
				ulLength = m_strRespBody.length();
			HrResponseChunkedBody(strBodyChunked);
		}
	}

exit:
	return hr;
}

/**
 * Sets http response headers
 * @param[in]	ulCode			Http header status code
 * @param[in]	strResponse		Http header status string
 * 
 * @return		HRESULT
 * @retval		MAPI_E_CALL_FAILED	The status is already set
 * 
 */
HRESULT Http::HrResponseHeader(unsigned int ulCode, std::string strResponse)
{
	HRESULT hr = hrSuccess;
	
	// do not set headers if once set
	if (m_strRespHeader.empty()) 
		m_strRespHeader = "HTTP/1.1 " + stringify(ulCode) + " " + strResponse;
	else
		hr = MAPI_E_CALL_FAILED;

	return hr;
}
/**
 * Adds response header to the list of headers
 * @param[in]	strHeader	Name of the header eg. Connection, Host, Date
 * @param[in]	strValue	Value of the header to be set
 * @return		HRESULT		Always set to hrSuccess
 */
HRESULT Http::HrResponseHeader(std::string strHeader, std::string strValue)
{
	HRESULT hr = hrSuccess;
	std::string header;

	header = strHeader + ": " + strValue;
	m_lstHeaders.push_back(header);

	return hr;
}

/**
 * Sends chunked http response to the client
 * @param[in]	strChnkBody		Body to be sent to client in chunked http
 * @return		HRESULT
 * @retval		MAPI_E_END_OF_SESSION	States that the client has set connection type as closed
 */
HRESULT Http::HrResponseChunkedBody(std::string strChnkBody)
{
	HRESULT hr = hrSuccess;
	std::string strBody;
	
	if (!m_bHeadersSent) {
		HrResponseHeader("Transfer-Encoding", "chunked");

		hr = HrFlushHeaders();
		if (hr != hrSuccess && hr != MAPI_E_END_OF_SESSION)
			goto exit;
	}
	
	strBody.resize(strChnkBody.length() + 20); // 20 for body size & CRLF
	
	if(strChnkBody.length() < HTTP_CHUNK_SIZE)
		sprintf((char*)strBody.c_str(), "%X\r\n%s\r\n0\r\n", strChnkBody.size(), strChnkBody.c_str()); // end of response
	else
		sprintf((char*)strBody.c_str(), "%X\r\n%s", strChnkBody.size(), strChnkBody.c_str());
	
	m_lpChannel->HrWriteLine(strBody.c_str());
	m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "%s", strBody.c_str());	

exit:
	return hr;
}
/**
 * Add string to http response body
 * @param[in]	strResponse		The string to be added to http response body
 * return		HRESULT			Always set to hrSuccess
 */
HRESULT Http::HrResponseBody(std::string strResponse)
{
	HRESULT hr = hrSuccess;

	m_strRespBody += strResponse;
	// data send in HrFinalize()

	return hr;
}

/**
 * Request authorization information from the client
 *
 * @param[in]	strMsg	Message to be shown to the client
 * @return		HRESULT 
 */
HRESULT Http::HrRequestAuth(std::string strMsg)
{
	HRESULT hr = hrSuccess;

	hr = HrResponseHeader(401, "Unauthorized");
	if (hr != hrSuccess)
		goto exit;

	strMsg = "Basic realm=\"" + strMsg + "\"";
	hr = HrResponseHeader("WWW-Authenticate", strMsg);
	if (hr != hrSuccess)
		goto exit;

exit:
	return hr;
}
/**
 * Write all http headers to ECChannel
 * @retrun	HRESULT
 * @retval	MAPI_E_END_OF_SESSION	If Connection type is set to close then the mapi session is ended
 */
HRESULT Http::HrFlushHeaders()
{
	HRESULT hr = hrSuccess;
	std::list<std::string>::iterator h;
	std::string strOutput;
	char lpszChar[128];
	time_t tmCurrenttime = time(NULL);

	if (m_bHeadersSent)
		return hrSuccess;

	// Add misc. headers
	HrResponseHeader("Server","Zarafa");
	strftime(lpszChar, 127, "%a, %d %b %Y %H:%M:%S GMT", gmtime(&tmCurrenttime));
	HrResponseHeader("Date", lpszChar);
	HrResponseHeader("DAV", "1, 2, 3, access-control, calendar-access, calendar-schedule, calendarserver-principal-property-search");
	if (m_ulKeepAlive != 0 && stricmp(m_strConnection.c_str(), "keep-alive") == 0) {
		HrResponseHeader("Connection", "Keep-Alive");
		HrResponseHeader("Keep-Alive", stringify(m_ulKeepAlive, false));
	}
	else
	{
		HrResponseHeader("Connection", "close");
		hr = MAPI_E_END_OF_SESSION;
	}

	// create headers packet

	if (!m_strRespHeader.empty()) {
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "Response header: " + m_strRespHeader);
		strOutput += m_strRespHeader + "\r\n";
		m_strRespHeader.clear();
	}

	for (h = m_lstHeaders.begin(); h != m_lstHeaders.end(); h++) {
		m_lpLogger->Log(EC_LOGLEVEL_DEBUG, "> " + *h);
		strOutput += *h + "\r\n";
	}
	m_lstHeaders.clear();

	//as last line has a CRLF. The HrWriteLine adds one more CRLF.
	//this means the End of headder.
	m_lpChannel->HrWriteLine(strOutput);
	m_bHeadersSent = true;

	return hr;
}
