/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"

// Damn windows header defines max which break C++ header files
#undef max

#include "VMIMEToMAPI.h"
#include "ECGuid.h"

#include <string>
#include <fstream>
#include <iostream>
#include <stdlib.h>
#include <string.h>

// vmime
#include "vmime/vmime.hpp"
#include "vmime/platforms/posix/posixHandler.hpp"

// mapi
#include <mapi.h>
#include <mapix.h>
#include <mapiutil.h>
#include <mapiext.h>
#include <mapiguidext.h>
#include <edkmdb.h>

#include "EMSAbTag.h"
#include "tnef.h"
#include "codepage.h"
#include "Util.h"
#include "CommonUtil.h"
#include "namedprops.h"
#include "str1252.h"

// inetmapi
#include "ECMapiUtils.h"
#include "ECVMIMEUtils.h"
#include "outputStreamMAPIAdapter.h"

// vcal support
#include "ICalToMAPI.h"

using namespace std;

VMIMEToMAPI::VMIMEToMAPI()
{
	lpLogger = new ECLogger_File(EC_LOGLEVEL_ANY, 1, "-");	// TODO: remove fixed defaults
	delete_logger = true;
	imopt_default_delivery_options(&m_dopt);
	m_dopt.use_received_date = false; // use Date header
	m_lpAdrBook = NULL;
	m_lpDefaultDir = NULL;
}

VMIMEToMAPI::VMIMEToMAPI(LPADRBOOK lpAdrBook, ECLogger *newlogger, delivery_options dopt)
{
	lpLogger = newlogger;
	m_dopt = dopt;
	if (!lpLogger) {
		lpLogger = new ECLogger_File(EC_LOGLEVEL_ANY, 1, "-");	// TODO: remove fixed defaults
		delete_logger = true;
	} else {
		delete_logger = false;
	}
	m_lpAdrBook = lpAdrBook;
	m_lpDefaultDir = NULL;
}

VMIMEToMAPI::~VMIMEToMAPI()
{
	if (delete_logger)
		delete lpLogger;
		
    if (m_lpDefaultDir)
        m_lpDefaultDir->Release();
}


HRESULT VMIMEToMAPI::convertVMIMEToMAPI(const string &input, IMessage *lpMessage) {
	HRESULT hr = hrSuccess;
	// signature variables
	ULONG ulAttNr;
	LPATTACH lpAtt = NULL;
	LPSTREAM lpStream = NULL;
	ULONG nProps = 0;
	SPropValue	attProps[3];
	SPropValue sPropSMIMEClass;


	try {
		if (m_mailState.ulMsgInMsg == 0)
			m_mailState.reset();

		// get raw headers
		size_t pos = input.find("\r\n\r\n");
		if(pos != std::string::npos) {
		    SPropValue sPropHeaders;
		    std::string strHeaders = input.substr(0, pos);
		    
		    sPropHeaders.ulPropTag = PR_TRANSPORT_MESSAGE_HEADERS;
		    sPropHeaders.Value.lpszA = (char *) strHeaders.c_str();
		    
		    HrSetOneProp(lpMessage, &sPropHeaders);
		}

		// turn buffer into a message
		vmime::message VMMessage;
		VMMessage.parse(input);

		hr = fillMAPIMail(&VMMessage, lpMessage);
		if (hr != hrSuccess)
			goto exit;

		if (m_mailState.bAttachSignature) {
			// TODO: mark all currently created attachments hidden, so
			// you see in outlook only the attachments from the parsed
			// signed message.  we would need a workaround in the
			// webaccess to still show these hidden messages.

			// smime signature was found, so attach 
			vmime::header* VMHeader = VMMessage.getHeader();
			string::size_type pos;

			hr = lpMessage->CreateAttach(NULL, 0, &ulAttNr, &lpAtt);
			if (hr != hrSuccess)
				goto exit;

			// open stream
			hr = lpAtt->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream, STGM_WRITE|STGM_TRANSACTED,
									 MAPI_CREATE|MAPI_MODIFY, (LPUNKNOWN *)&lpStream);
			if (hr != hrSuccess)
				goto exit;


			outputStreamMAPIAdapter os(lpStream);
			// get the content-type string from the headers
			VMHeader->ContentType().generate(os);

			// find the original received body
			// vmime re-generates different headers and spacings, so we can't use this.
			pos = input.find("\r\n\r\n");
			if (pos != string::npos) {
				os.write(input.c_str() + pos, input.size() - pos);
			}

			hr = lpStream->Commit(0);
			if (hr != hrSuccess)
				goto exit;

			attProps[nProps].ulPropTag = PR_ATTACH_METHOD;
			attProps[nProps++].Value.ul = ATTACH_BY_VALUE;

			attProps[nProps].ulPropTag = PR_ATTACH_MIME_TAG;
			attProps[nProps++].Value.lpszA = "multipart/signed";
			
			attProps[nProps].ulPropTag = PR_RENDERING_POSITION;
			attProps[nProps++].Value.i = -1;

			hr = lpAtt->SetProps(nProps, attProps, NULL);
			if (hr != hrSuccess)
				goto exit;

			hr = lpAtt->SaveChanges(0);
			if (hr != hrSuccess)
				goto exit;

			// saved, so mark the message so outlook knows how to find the encoded message
			sPropSMIMEClass.ulPropTag = PR_MESSAGE_CLASS;
			sPropSMIMEClass.Value.lpszA = "IPM.Note.SMIME.MultipartSigned";

			hr = lpMessage->SetProps(1, &sPropSMIMEClass, NULL);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to set message class");
				goto exit;
			}

			m_mailState.attachLevel = ATTACH_NORMAL;
		}
		if (m_mailState.attachLevel == ATTACH_INLINE && m_mailState.bodyLevel == BODY_HTML) {
			// add hidden flag, to the attachment icon for the mail does not show up
			MAPINAMEID sNameID;
			LPMAPINAMEID lpNameID = &sNameID;
			LPSPropTagArray lpPropTag = NULL;

			sNameID.lpguid = (GUID*)&PSETID_Common;
			sNameID.ulKind = MNID_ID;
			sNameID.Kind.lID = dispidSmartNoAttach;

			hr = lpMessage->GetIDsFromNames(1, &lpNameID, MAPI_CREATE, &lpPropTag);
			if (hr != hrSuccess) {
				hr = hrSuccess;
				goto exit;
			}

			attProps[0].ulPropTag = CHANGE_PROP_TYPE(lpPropTag->aulPropTag[0], PT_BOOLEAN);
			attProps[0].Value.b = TRUE;

			MAPIFreeBuffer(lpPropTag);

			hr = lpMessage->SetProps(1, attProps, NULL);
			if (hr != hrSuccess) {
				hr = hrSuccess;
				goto exit;
			}
		}
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (lpAtt)
		lpAtt->Release();

	if (lpStream)
		lpStream->Release();

	return hr;
}

HRESULT VMIMEToMAPI::fillMAPIMail(vmime::message* VMMessage, IMessage *lpMessage) {
	HRESULT	hr = hrSuccess;
	SPropValue sPropDefaults[2];

	sPropDefaults[0].ulPropTag = PR_MESSAGE_CLASS;
	sPropDefaults[0].Value.lpszA = "IPM.Note";
	sPropDefaults[1].ulPropTag = PR_MESSAGE_FLAGS;
	sPropDefaults[1].Value.ul = (m_dopt.mark_as_read ? MSGFLAG_READ : 0) | MSGFLAG_UNMODIFIED;

	hr = lpMessage->SetProps(2, sPropDefaults, NULL);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to set default mail properties");
		goto exit;
	}
	
	try {
		// turn buffer into a message

		// get the part header and find out what it is...
		vmime::header* VMHeader = VMMessage->getHeader();
		vmime::body* VMBody = VMMessage->getBody();
		vmime::mediaType mt = VMHeader->ContentType().getValue();
		bool bMixed = false;

		if (mt.getType() == "multipart" && mt.getSubType() == "mixed")
			bMixed = true;

		// pass recipients somewhere else 
		hr = handleRecipients(VMHeader, lpMessage);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse mail recipients");
			goto exit;
		}

		// Headers
		hr = handleHeaders(VMHeader, lpMessage);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse mail headers");
			goto exit;
		}

		if(vmime::mdn::MDNHelper::isMDN(VMMessage) == true)
		{
			vmime::mdn::receivedMDNInfos receivedMDN = vmime::mdn::MDNHelper::getReceivedMDN(VMMessage);

			vmime::body* myBody = (vmime::body*)receivedMDN.getMessage()->getBody();
			// it is possible to get 3 bodyparts.
			// text/Plain, message/disposition-notification, text/rfc822-headers
			// Loop to get text/plain body.
			for (int i=0; i < myBody->getPartCount(); i++)
			{
				vmime::bodyPart* bPart = myBody->getPartAt(i);

				if(bPart->getHeader()->ContentType().getValue().getType() == vmime::mediaTypes::TEXT &&
					bPart->getHeader()->ContentType().getValue().getSubType() == vmime::mediaTypes::TEXT_PLAIN)
				{
					hr = disectBody(bPart->getHeader(), bPart->getBody(), lpMessage, true);
					if (hr != hrSuccess) {
						lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse MDN mail body");
						goto exit;
					}
				}
				
			}

			if(receivedMDN.getDisposition().getType() == vmime::dispositionTypes::DELETED)
			{
				sPropDefaults[0].ulPropTag = PR_MESSAGE_CLASS;
				sPropDefaults[0].Value.lpszA = "REPORT.IPM.Note.IPNNRN";				
			}else{
				sPropDefaults[0].ulPropTag = PR_MESSAGE_CLASS;
				sPropDefaults[0].Value.lpszA = "REPORT.IPM.Note.IPNRN";
			}

			string strId = "<"+receivedMDN.getOriginalMessageId().getId()+">";
			sPropDefaults[1].ulPropTag = 0x1046001E;//ptagOriginalInetMessageID
			sPropDefaults[1].Value.lpszA = (LPSTR)strId.c_str();

			hr = lpMessage->SetProps(2, sPropDefaults, NULL);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to set MDN mail properties");
				goto exit;
			}

		} else if (VMBody->getPartCount() > 0) {
			// multipart message, handle by part
			for (int i=0; i < VMBody->getPartCount(); i++) {
				vmime::bodyPart* VMBodyPart = VMBody->getPartAt(i);
				hr = disectBody(VMBodyPart->getHeader(), VMBodyPart->getBody(), lpMessage, false, false, bMixed);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse multipart %d of mail body", i);
					goto exit;
				}
			}
		} else {
			// one body part, always place in the body, no matter if the header says it's an _text_ attachment
			hr = disectBody(VMHeader, VMBody, lpMessage, true);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse mail body");
				goto exit;
			}
		}
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on create message: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on create message: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on create message");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	// ignore error/warings from fixup function: it's not critical for correct delivery
	postWriteFixups(lpMessage);

exit:
	return hr;
}

// put all headers where they belong..
HRESULT VMIMEToMAPI::handleHeaders(vmime::header* VMHeader, IMessage* lpMessage) {
	HRESULT			hr = hrSuccess;
	std::string		strHeaders, strInternetMessageId;
	std::string		strInReplyTos, strReferences;
	std::string		strSubject, strReplyTo;
	std::string		strClientTime;
	std::string		strFromEmail, strFromName, strFromSearchKey;
	std::string		strSenderEmail,	strSenderName, strSenderSearchKey;
	std::string		strReportName;
	ULONG			cbFromEntryID; // representing entry id
	LPENTRYID		lpFromEntryID = NULL;
	ULONG			cbSenderEntryID;
	LPENTRYID		lpSenderEntryID = NULL;
	LPSPropValue	lpPropNormalizedSubject = NULL;
	SPropValue		sConTopic;
	// setprops
	FLATENTRY		*lpEntry = NULL;
	FLATENTRYLIST	*lpEntryList = NULL;
	ULONG			cb = 0;
	int				nProps = 0;
	SPropValue		msgProps[22];
	// temp
	ULONG			cbEntryID;
	LPENTRYID		lpEntryID = NULL;

	LPSPropValue	lpRecipProps = NULL;
	ULONG			ulRecipProps;
	
	// order is important for modifyFromAddressBook()
	SizedSPropTagArray(7, sptaRecipPropsSentRepr) = { 7, {
		PR_SENT_REPRESENTING_ADDRTYPE, PR_SENT_REPRESENTING_NAME,
		PR_NULL /* PR_xxx_DISPLAY_TYPE not available */,
		PR_SENT_REPRESENTING_EMAIL_ADDRESS, PR_SENT_REPRESENTING_ENTRYID,
		PR_SENT_REPRESENTING_SEARCH_KEY, PR_NULL /* PR_xxx_SMTP_ADDRESS not available */
	} };
	SizedSPropTagArray(7, sptaRecipPropsSender) = { 7, {
		PR_SENDER_ADDRTYPE, PR_SENDER_NAME,
		PR_NULL /* PR_xxx_DISPLAY_TYPE not available */,
		PR_SENDER_EMAIL_ADDRESS, PR_SENDER_ENTRYID,
		PR_SENDER_SEARCH_KEY, PR_NULL /* PR_xxx_SMTP_ADDRESS not available */
	} };

	try { 
		int nIds, i;

		// internet message ID
		strInternetMessageId = "<"+VMHeader->MessageId().getValue().getId()+">";
		msgProps[nProps].ulPropTag = PR_INTERNET_MESSAGE_ID;
		msgProps[nProps++].Value.lpszA = (char*)strInternetMessageId.c_str();

		// In-Reply-To header
		strInReplyTos.clear();
		nIds = VMHeader->InReplyTo().getValue().getMessageIdCount();
		for (i = 0; i < nIds; i++) {
		    // space/enter/nothing? between id's?
		    strInReplyTos += "<"+VMHeader->InReplyTo().getValue().getMessageIdAt(i)->getId()+"> ";
		}
		if (!strInReplyTos.empty()) {
			msgProps[nProps].ulPropTag = PR_IN_REPLY_TO_ID;
			msgProps[nProps++].Value.lpszA = (char*)strInReplyTos.c_str();
		}

		// References header
		strReferences.clear();
		nIds = VMHeader->References().getValue().getMessageIdCount();
		for (i = 0; i < nIds; i++) {
		    strReferences += "<"+VMHeader->References().getValue().getMessageIdAt(i)->getId()+"> ";
		}
		if (!strReferences.empty()) {
			msgProps[nProps].ulPropTag = PR_INTERNET_REFERENCES;
			msgProps[nProps++].Value.lpszA = (char*)strInReplyTos.c_str();
		}

		// set subject
		strSubject = VMHeader->Subject().getValue().getConvertedText(MAPI_CHARSET);
		msgProps[nProps].ulPropTag = PR_SUBJECT;
		msgProps[nProps++].Value.lpszA = (char*)strSubject.c_str();

		// set ReplyTo
		if (!VMHeader->ReplyTo().getValue().isEmpty()) {
			// First, set PR_REPLY_RECIPIENT_NAMES
			strReplyTo = VMHeader->ReplyTo().getValue().getName().getConvertedText(MAPI_CHARSET);
			if (strReplyTo == "")
				strReplyTo = VMHeader->ReplyTo().getValue().getEmail();

			msgProps[nProps].ulPropTag = PR_REPLY_RECIPIENT_NAMES;
			msgProps[nProps++].Value.lpszA = (char*)strReplyTo.c_str();

			// Now, set PR_REPLY_RECIPIENT_ENTRIES (a FLATENTRYLIST)
			hr = ECCreateOneOff((char *)strReplyTo.c_str(), "SMTP",
								(char *)VMHeader->ReplyTo().getValue().getEmail().c_str(), MAPI_SEND_NO_RICH_INFO, &cbEntryID, &lpEntryID);
			if (hr != hrSuccess)
				goto exit;

			cb = CbNewFLATENTRY(cbEntryID);
			hr = MAPIAllocateBuffer(cb, (void**)&lpEntry);
			if (hr != hrSuccess)
				goto exit;

			memcpy(lpEntry->abEntry, lpEntryID, cbEntryID);
			lpEntry->cb = cbEntryID;

			cb = CbNewFLATENTRYLIST(cb);
			hr = MAPIAllocateBuffer(cb, (void **)&lpEntryList);
			if (hr != hrSuccess)
				goto exit;

			lpEntryList->cEntries = 1;
			lpEntryList->cbEntries = CbFLATENTRY(lpEntry);
			memcpy(&lpEntryList->abEntries, lpEntry, CbFLATENTRY(lpEntry));

			msgProps[nProps].ulPropTag = PR_REPLY_RECIPIENT_ENTRIES;
			msgProps[nProps].Value.bin.cb = CbFLATENTRYLIST(lpEntryList);
			msgProps[nProps++].Value.bin.lpb = (BYTE*)lpEntryList;

			MAPIFreeBuffer(lpEntryID);
			lpEntryID = NULL;
		}

		// setting sent time
		msgProps[nProps].ulPropTag = PR_CLIENT_SUBMIT_TIME;
		msgProps[nProps++].Value.ft = vmimeDatetimeToFiletime(VMHeader->Date().getValue());

		// set sent date (actual send date, disregarding timezone)
		vmime::datetime d = VMHeader->Date().getValue();
		d.setTime(0,0,0,0);
		msgProps[nProps].ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
		msgProps[nProps++].Value.ft = vmimeDatetimeToFiletime(d);

		vmime::datetime date;
		// setting receive date (now)
		// parse from Received header, if possible. Otherwise, use now()
		if (m_dopt.use_received_date) {
			try {
				vmime::relayField *recv = dynamic_cast <vmime::relayField*>(VMHeader->findField("Received"));
				date = recv->getValue().getDate();
			}
			catch (...) {
				if (m_mailState.ulMsgInMsg)
					date = VMHeader->Date().getValue();
				else
					date = vmime::datetime::now();
			}
		} else {
			if (m_mailState.ulMsgInMsg)
				date = VMHeader->Date().getValue();
			else
				date = vmime::datetime::now();
		}

		msgProps[nProps].ulPropTag = PR_MESSAGE_DELIVERY_TIME;
		msgProps[nProps++].Value.ft = vmimeDatetimeToFiletime(date);

		// Also save delivery DATE without timezone
		date.setTime(0,0,0,0);
		msgProps[nProps].ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
		msgProps[nProps++].Value.ft = vmimeDatetimeToFiletime(date);

		// The real sender of the mail
		strFromEmail = VMHeader->From().getValue().getEmail();
		if (!VMHeader->From().getValue().getName().isEmpty()) {
			strFromName = VMHeader->From().getValue().getName().getConvertedText(MAPI_CHARSET);
		}

		hr = modifyFromAddressBook(&lpRecipProps, &ulRecipProps, strFromEmail.c_str(), strFromName.c_str(), MAPI_ORIG, (LPSPropTagArray)&sptaRecipPropsSentRepr);
		if (hr == hrSuccess) {
			hr = lpMessage->SetProps(ulRecipProps, lpRecipProps, NULL);
			if (hr != hrSuccess)
				goto exit;

			MAPIFreeBuffer(lpRecipProps);
			lpRecipProps = NULL;
		} else {
			if (strFromName.empty())
				strFromName = strFromEmail;

			msgProps[nProps].ulPropTag = PR_SENT_REPRESENTING_NAME;
			msgProps[nProps++].Value.lpszA = (char*)strFromName.c_str();

			msgProps[nProps].ulPropTag = PR_SENT_REPRESENTING_EMAIL_ADDRESS;
			msgProps[nProps++].Value.lpszA = (char*)strFromEmail.c_str();

			strFromSearchKey = "SMTP:"+strFromEmail;
			transform(strFromSearchKey.begin(), strFromSearchKey.end(), strFromSearchKey.begin(), ::toupper);
			msgProps[nProps].ulPropTag = PR_SENT_REPRESENTING_SEARCH_KEY;
			msgProps[nProps].Value.bin.cb = strFromSearchKey.size()+1; // include string terminator
			msgProps[nProps++].Value.bin.lpb = (BYTE*)strFromSearchKey.c_str();

			msgProps[nProps].ulPropTag = PR_SENT_REPRESENTING_ADDRTYPE;
			msgProps[nProps++].Value.lpszA = "SMTP";
			
			hr = ECCreateOneOff((char *)strFromName.c_str(), "SMTP", (char*)strFromEmail.c_str(), MAPI_SEND_NO_RICH_INFO, &cbFromEntryID, &lpFromEntryID);
			if(hr != hrSuccess)
				goto exit;

			msgProps[nProps].ulPropTag = PR_SENT_REPRESENTING_ENTRYID;
			msgProps[nProps].Value.bin.cb = cbFromEntryID;
			msgProps[nProps++].Value.bin.lpb = (BYTE*)lpFromEntryID;

			// SetProps is later on...
		}

		// The original sender of the mail account (if non sender exist then the FROM)
		strSenderEmail = VMHeader->Sender().getValue().getEmail();
		if(VMHeader->Sender().getValue().getName().isEmpty() && strSenderEmail.empty()) {
			// Fallback on the original from address
			strSenderName = strFromName;
			strSenderEmail = strFromEmail;
		} else {
			if (!VMHeader->Sender().getValue().getName().isEmpty()) {
				strSenderName = VMHeader->Sender().getValue().getName().getConvertedText(MAPI_CHARSET);
			}
		}

		hr = modifyFromAddressBook(&lpRecipProps, &ulRecipProps, strSenderEmail.c_str(), strSenderName.c_str(), MAPI_ORIG, (LPSPropTagArray)&sptaRecipPropsSender);
		if (hr == hrSuccess) {
			hr = lpMessage->SetProps(ulRecipProps, lpRecipProps, NULL);
			if (hr != hrSuccess)
				goto exit;

			MAPIFreeBuffer(lpRecipProps);
			lpRecipProps = NULL;
		} else {
			if (strSenderName.empty())
				strSenderName.assign(strFromEmail);
			msgProps[nProps].ulPropTag = PR_SENDER_NAME;
			msgProps[nProps++].Value.lpszA = (char*)strSenderName.c_str();

			msgProps[nProps].ulPropTag = PR_SENDER_EMAIL_ADDRESS;
			msgProps[nProps++].Value.lpszA = (char*)strSenderEmail.c_str();

			strSenderSearchKey = "SMTP:"+strSenderEmail;
			transform(strSenderSearchKey.begin(), strSenderSearchKey.end(), strSenderSearchKey.begin(), ::toupper);
			msgProps[nProps].ulPropTag = PR_SENDER_SEARCH_KEY;
			msgProps[nProps].Value.bin.cb = strSenderSearchKey.size()+1; // include string terminator
			msgProps[nProps++].Value.bin.lpb = (BYTE*)strSenderSearchKey.c_str();

			msgProps[nProps].ulPropTag = PR_SENDER_ADDRTYPE;
			msgProps[nProps++].Value.lpszA = "SMTP";
			
			hr = ECCreateOneOff((char *)strSenderName.c_str(), "SMTP", (char*)strSenderEmail.c_str(), MAPI_SEND_NO_RICH_INFO, &cbSenderEntryID, &lpSenderEntryID);
			if(hr != hrSuccess)
				goto exit;

			msgProps[nProps].ulPropTag = PR_SENDER_ENTRYID;
			msgProps[nProps].Value.bin.cb = cbSenderEntryID;
			msgProps[nProps++].Value.bin.lpb = (BYTE*)lpSenderEntryID;
		}

		hr = lpMessage->SetProps(nProps, msgProps, NULL);
		if (hr != hrSuccess)
			goto exit;

		//Conversation topic
		if(VMHeader->hasField("Thread-Topic"))
		{
			// Thread-Topic is created as a vmime::defaultField, and is just a normal string, ignoring the text
			// so we convert the defaultField to a textField by parsing it again .. bit of a long road, but it seems to work
			vmime::defaultField *defTT = dynamic_cast <vmime::defaultField*>(VMHeader->findField("Thread-Topic"));
			vmime::textField *txtTT = dynamic_cast<vmime::textField*>(vmime::headerFieldFactory::registerer<vmime::textField>::creator());
			txtTT->parse(defTT->getValue());
			std::string convTT = txtTT->getValue().getConvertedText(MAPI_CHARSET);
			delete txtTT;

			sConTopic.Value.lpszA = (char*)convTT.c_str();
			sConTopic.ulPropTag = PR_CONVERSATION_TOPIC;

			hr = lpMessage->SetProps(1, &sConTopic, NULL);
			if (hr != hrSuccess)
				goto exit;

		} else if(HrGetOneProp(lpMessage, PR_NORMALIZED_SUBJECT, &lpPropNormalizedSubject) == hrSuccess) {
			
			sConTopic.ulPropTag = PR_CONVERSATION_TOPIC;
			sConTopic.Value.lpszA = lpPropNormalizedSubject->Value.lpszA;
			
			hr = lpMessage->SetProps(1, &sConTopic, NULL);
			if (hr != hrSuccess)
				goto exit;
		}

		// Thread-Index header
		if (VMHeader->hasField("Thread-Index"))
		{
			vmime::string outString;
			SPropValue sThreadIndex;

			string threadIndex = VMHeader->findField("Thread-Index")->getValue().generate();

			vmime::encoder* enc = vmime::encoderFactory::getInstance()->create("base64");

			vmime::utility::inputStreamStringAdapter in(threadIndex);			
			vmime::utility::outputStreamStringAdapter out(outString);

			enc->decode(in, out);
			delete enc;

			sThreadIndex.ulPropTag = PR_CONVERSATION_INDEX;
			sThreadIndex.Value.bin.cb = outString.size();
			sThreadIndex.Value.bin.lpb = (LPBYTE)outString.c_str();

			hr = lpMessage->SetProps(1, &sThreadIndex, NULL);
			if (hr != hrSuccess)
				goto exit;
		}

		// X-Priority header
		if (VMHeader->hasField("X-Priority")) {
			SPropValue sPriority[2];
			sPriority[0].ulPropTag = PR_PRIORITY;
			sPriority[1].ulPropTag = PR_IMPORTANCE;
			string xprio = VMHeader->findField("X-Priority")->getValue().generate();
			switch (xprio[0]) {
			case '1':
			case '2':
				sPriority[0].Value.ul = PRIO_URGENT;
				sPriority[1].Value.ul = IMPORTANCE_HIGH;
				break;
			case '4':
			case '5':
				sPriority[0].Value.ul = PRIO_NONURGENT;
				sPriority[1].Value.ul = IMPORTANCE_LOW;
				break;
			default:
			case '3':
				sPriority[0].Value.ul = PRIO_NORMAL;
				sPriority[1].Value.ul = IMPORTANCE_NORMAL;
				break;
			};
			hr = lpMessage->SetProps(2, sPriority, NULL);
			if (hr != hrSuccess)
				goto exit;
		}

		// X-Zarafa-Vacation header (TODO: other headers?)
		if (VMHeader->hasField("X-Zarafa-Vacation")) {
			SPropValue sIcon[1];
			sIcon[0].ulPropTag = PR_ICON_INDEX;
			sIcon[0].Value.l = ICON_MAIL_OOF;
			// exchange sets PR_MESSAGE_CLASS to IPM.Note.Rules.OofTemplate.Microsoft to get the icon
			hr = lpMessage->SetProps(1, sIcon, NULL);
			if (hr != hrSuccess)
				goto exit;
		}

		// Sensitivity header
		if (VMHeader->hasField("Sensitivity")) {
			SPropValue sSensitivity[1];
			sSensitivity[0].ulPropTag = PR_SENSITIVITY;
			string sensitivity = VMHeader->findField("Sensitivity")->getValue().generate();
			if (sensitivity == "Personal") {
				sSensitivity[0].Value.ul = SENSITIVITY_PERSONAL;
			} else if (sensitivity == "Private") {
				sSensitivity[0].Value.ul = SENSITIVITY_PRIVATE;
			} else if (sensitivity == "Company-Confidential") {
				sSensitivity[0].Value.ul = SENSITIVITY_COMPANY_CONFIDENTIAL;
			} else {
				sSensitivity[0].Value.ul = SENSITIVITY_NONE;
			}
			hr = lpMessage->SetProps(1, sSensitivity, NULL);
			if (hr != hrSuccess)
				goto exit;
		}

		// Expiry time header
		try {
			if (VMHeader->hasField("Expiry-Time")) {
				SPropValue sExpiryTime;

				// reparse string to datetime
				vmime::datetime expiry(VMHeader->findField("Expiry-Time")->getValue().generate());

				sExpiryTime.ulPropTag = PR_EXPIRY_TIME;
				sExpiryTime.Value.ft = vmimeDatetimeToFiletime(expiry);

				hr = lpMessage->SetProps(1, &sExpiryTime, NULL);
				if (hr != hrSuccess)
					goto exit;
			}
		}
		catch(...) {}

		// read receipt	request
		// note: vmime never checks if the given pos to getMailboxAt() and similar functions is valid, so we check if the list is empty before using it
		if(VMHeader->hasField("Disposition-Notification-To") && !VMHeader->DispositionNotificationTo().getValue().isEmpty()) {
			vmime::mailbox *mbReadReceipt = VMHeader->DispositionNotificationTo().getValue().getMailboxAt(0); // we only use the 1st
			if (mbReadReceipt && !mbReadReceipt->isEmpty()) {

				//FIXME: For zarafa use a addressbook entry?
				hr = ECCreateOneOff((LPSTR)mbReadReceipt->getName().getConvertedText(MAPI_CHARSET).c_str(), "SMTP", (LPSTR)mbReadReceipt->getEmail().c_str(), MAPI_SEND_NO_RICH_INFO, &cbEntryID, &lpEntryID);
				if (hr != hrSuccess)
					goto exit;

				SPropValue sRRProps[4];
				sRRProps[0].ulPropTag = PR_READ_RECEIPT_REQUESTED;
				sRRProps[0].Value.b = true;
				
				sRRProps[1].ulPropTag = PR_MESSAGE_FLAGS;
				sRRProps[1].Value.ul = MSGFLAG_UNMODIFIED | MSGFLAG_RN_PENDING | MSGFLAG_NRN_PENDING;

				sRRProps[2].ulPropTag = PR_REPORT_ENTRYID;
				sRRProps[2].Value.bin.cb = cbEntryID;
				sRRProps[2].Value.bin.lpb = (BYTE*)lpEntryID;

				strReportName = mbReadReceipt->getName().getConvertedText(MAPI_CHARSET);
				sRRProps[3].ulPropTag = PR_REPORT_NAME;
				sRRProps[3].Value.lpszA = (LPSTR)strReportName.c_str();

				hr = lpMessage->SetProps(4, sRRProps, NULL);
				if (hr != hrSuccess)
					goto exit;

				MAPIFreeBuffer(lpEntryID); lpEntryID = NULL;
			}
		}

		std::vector<vmime::headerField*> hf = VMHeader->getFieldList();
		std::vector<vmime::headerField*>::iterator hfi;
		for (hfi = hf.begin(); hfi != hf.end(); hfi++) {
			string value, name = (*hfi)->getName();
			
			if (name[0] != 'X')
				continue;

			// exclusion list?
			if (name == "X-Priority")
				continue;

			transform(name.begin(), name.end(), name.begin(), ::tolower);

			LPMAPINAMEID lpNameID = NULL;
			LPSPropTagArray lpPropTags = NULL;

			MAPIAllocateBuffer(sizeof(MAPINAMEID), (void**)&lpNameID);
			lpNameID->lpguid = (GUID*)&PS_INTERNET_HEADERS;
			lpNameID->ulKind = MNID_STRING;

			int vlen = mbstounicode(NULL, name.c_str(), 0) +1;
			MAPIAllocateMore(vlen*sizeof(WCHAR), lpNameID, (void**)&lpNameID->Kind.lpwstrName);
			mbstounicode(lpNameID->Kind.lpwstrName, name.c_str(), vlen);

			hr = lpMessage->GetIDsFromNames(1, &lpNameID, MAPI_CREATE, &lpPropTags);
			if (hr != hrSuccess) {
				hr = hrSuccess;
				goto next;
			}

			SPropValue sProp[1];
			value = (*hfi)->getValue().generate();
			sProp[0].ulPropTag = PROP_TAG(PT_STRING8, PROP_ID(lpPropTags->aulPropTag[0]));
			sProp[0].Value.lpszA = (char*)value.c_str();

			hr = lpMessage->SetProps(1, sProp, NULL);
			if (hr != hrSuccess) {
				hr = hrSuccess;	// ignore this x-header as named props then
				goto next;
			}

next:
			if (lpPropTags)
				MAPIFreeBuffer(lpPropTags);
			lpPropTags = NULL;

			if (lpNameID)
				MAPIFreeBuffer(lpNameID);
			lpNameID = NULL;
		}
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on parsing headers: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on parsing headers: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on parsing headers");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (lpRecipProps)
		MAPIFreeBuffer(lpRecipProps);

	if(lpEntry)
		MAPIFreeBuffer(lpEntry);

	if(lpEntryList)
		MAPIFreeBuffer(lpEntryList);

	if(lpFromEntryID)
		MAPIFreeBuffer(lpFromEntryID);

	if(lpSenderEntryID)
		MAPIFreeBuffer(lpSenderEntryID);

	if(lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	if(lpPropNormalizedSubject)
		MAPIFreeBuffer(lpPropNormalizedSubject);

	return hr;
}

// Sets PR_MESSAGE_TO_ME, PR_MESSAGE_CC_ME and PR_MESSAGE_RECIP_ME appropriately
HRESULT VMIMEToMAPI::handleMessageToMeProps(IMessage *lpMessage, LPADRLIST lpRecipients)  {
    HRESULT hr = hrSuccess;
    unsigned int i = 0;
    LPSPropValue lpRecipType = NULL; // non-free
    LPSPropValue lpEntryId = NULL;	// non-free
    
    bool bToMe = false;
    bool bCcMe = false;
    bool bRecipMe = false;
    
    SPropValue sProps[3];
    
    if(m_dopt.user_entryid == NULL) {
        hr = hrSuccess; // Not an error, but don't do any processing
        goto exit;
    }
    
    // Loop through all recipients of the message to find ourselves in the recipient list.
    for(i=0;i<lpRecipients->cEntries;i++) {
        lpRecipType = PpropFindProp(lpRecipients->aEntries[i].rgPropVals, lpRecipients->aEntries[i].cValues, PR_RECIPIENT_TYPE);
        lpEntryId = PpropFindProp(lpRecipients->aEntries[i].rgPropVals, lpRecipients->aEntries[i].cValues, PR_ENTRYID);
        
        if(lpRecipType == NULL)
            continue;
            
        if(lpEntryId == NULL)
            continue;

        // The user matches if the entryid of the recipient is equal to ours            
        if(lpEntryId->Value.bin.cb != m_dopt.user_entryid->cb)
            continue;
        
        if(memcmp(lpEntryId->Value.bin.lpb, m_dopt.user_entryid->lpb, lpEntryId->Value.bin.cb) != 0)
            continue;
            
        // Users match, check what type
        bRecipMe = true;
        
        if(lpRecipType->Value.ul == MAPI_TO)
            bToMe = true;
        else if(lpRecipType->Value.ul == MAPI_CC)
            bCcMe = true;
            
    }
    
    // Set the properties
    sProps[0].ulPropTag = PR_MESSAGE_RECIP_ME;
    sProps[0].Value.b = bRecipMe;
    sProps[1].ulPropTag = PR_MESSAGE_TO_ME;
    sProps[1].Value.b = bToMe;
    sProps[2].ulPropTag = PR_MESSAGE_CC_ME;
    sProps[2].Value.b = bCcMe;
    
    lpMessage->SetProps(3, sProps, NULL);
    
exit:    
    return hr;
}

HRESULT VMIMEToMAPI::handleRecipients(vmime::header* VMHeader, IMessage* lpMessage) {
	HRESULT		hr				= hrSuccess;
	LPADRLIST 	lpRecipients	= NULL;

	try {
		vmime::addressList lpVMAListRecip = VMHeader->To().getValue();
		vmime::addressList lpVMAListCopyRecip = VMHeader->Cc().getValue();
		vmime::addressList lpVMAListBlCpRecip = VMHeader->Bcc().getValue();

		int iAdresCount = lpVMAListRecip.getAddressCount() + lpVMAListCopyRecip.getAddressCount() + lpVMAListBlCpRecip.getAddressCount();

		if (iAdresCount == 0)
			goto exit;

		hr = MAPIAllocateBuffer(CbNewADRLIST(iAdresCount), (void **)&lpRecipients);
		if (hr != hrSuccess)
			goto exit;

		lpRecipients->cEntries = 0;

		if (!lpVMAListRecip.isEmpty()) {
			hr = modifyRecipientList(lpRecipients, &lpVMAListRecip, MAPI_TO);
			if (hr != hrSuccess)
				goto exit;
		}

		if (!lpVMAListCopyRecip.isEmpty()) {
			hr = modifyRecipientList(lpRecipients, &lpVMAListCopyRecip, MAPI_CC);
			if (hr != hrSuccess)
				goto exit;
		}

		if (!lpVMAListBlCpRecip.isEmpty()) {
			hr = modifyRecipientList(lpRecipients, &lpVMAListBlCpRecip, MAPI_BCC);
			if (hr != hrSuccess)
				goto exit;
		}
		
		// Handle PR_MESSAGE_*_ME props
		hr = handleMessageToMeProps(lpMessage, lpRecipients);
		if (hr != hrSuccess)
		    goto exit;

		// actually modify recipients in mapi object
		hr = lpMessage->ModifyRecipients(MODRECIP_ADD, lpRecipients);	
		if (hr != hrSuccess)
			goto exit;

	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on recipients: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on recipients: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on recipients");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (lpRecipients != NULL) {
		// Because the recipient list can be altered to delete 1 row, or add 1 row using ModifyRecipients()
		// we need to free this data with FreeProws, and not the default MAPIFreeBuffer() call.
		FreeProws((LPSRowSet) lpRecipients);
	}

	return hr;
}


// Modifies a recipientlist, starting at iOffset in that list
HRESULT VMIMEToMAPI::modifyRecipientList(LPADRLIST lpRecipients, const vmime::addressList* lpVMAddressList, ULONG ulRecipType) {
	HRESULT			hr				= hrSuccess;
	int				iAddressCount	= lpVMAddressList->getAddressCount();
	ULONG			cbEntryID		= 0;
	LPENTRYID		lpEntryID		= NULL;
	vmime::mailbox	*mbx			= NULL;
	const vmime::address* lpVMAddress;
	std::string	strName, strEmail, strSearch;
	unsigned int 	iRecipNum		= 0;

	// order is important for modifyFromAddressBook()
	SizedSPropTagArray(7, sptaRecipientProps) = { 7, { PR_ADDRTYPE, PR_DISPLAY_NAME, PR_DISPLAY_TYPE, PR_EMAIL_ADDRESS, PR_ENTRYID,
													   PR_SEARCH_KEY, PR_SMTP_ADDRESS } };

	// walk through all recipients
	for (int iRecip=0; iRecip < iAddressCount; iRecip++) {
		
		try {
			mbx = NULL;
			lpVMAddress = lpVMAddressList->getAddressAt(iRecip);
			
			// not sure what to do if it IS a mailbox group.... ( never saw one )
			if (!lpVMAddress->isGroup())
				mbx = (vmime::mailbox *)lpVMAddress;

			if (!mbx)
				goto exit;

   			strEmail = mbx->getEmail();
   			// Do we have a name ? 
   			const vmime::text vmText = mbx->getName();
   			if (!vmText.isEmpty()) {
   				strName = vmText.getConvertedText(MAPI_CHARSET);
   			} else {
   			    strName.clear();
            }
		}
		catch (vmime::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on modify recipient: "+e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (std::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on modify recipient: %s", e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (...) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on modify recipient");
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		iRecipNum = lpRecipients->cEntries;

		// use email address or fullname to find GAB entry, do not pass fullname to keep resolved addressbook fullname
		strSearch = strEmail;
		if (strSearch.empty())
			strSearch = strName;

		hr = modifyFromAddressBook(&lpRecipients->aEntries[iRecipNum].rgPropVals, &lpRecipients->aEntries[iRecipNum].cValues, strSearch.c_str(), NULL, ulRecipType, (LPSPropTagArray)&sptaRecipientProps);
		if (hr != hrSuccess) {
			// Fallback if the entry was not found (or errored) in the addressbook
			int iNumTags = 8;
	
			iRecipNum = lpRecipients->cEntries;

			if (strName.empty())
   				strName = strEmail;

			// will be cleaned up by caller.
			hr = MAPIAllocateBuffer(sizeof(SPropValue) * iNumTags, (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals);	
			if (hr != hrSuccess)
				goto exit;
			
			lpRecipients->aEntries[iRecipNum].cValues						= iNumTags;	
			lpRecipients->aEntries[iRecipNum].ulReserved1					= 0;
			
			lpRecipients->aEntries[iRecipNum].rgPropVals[0].ulPropTag		= PR_RECIPIENT_TYPE;
			lpRecipients->aEntries[iRecipNum].rgPropVals[0].Value.l			= ulRecipType;
		
			lpRecipients->aEntries[iRecipNum].rgPropVals[1].ulPropTag		= PR_DISPLAY_NAME;
			hr = MAPIAllocateMore(strName.size()+1, lpRecipients->aEntries[iRecipNum].rgPropVals,
								  (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals[1].Value.lpszA);
			if (hr != hrSuccess)
				goto exit;
			lstrcpy(lpRecipients->aEntries[iRecipNum].rgPropVals[1].Value.lpszA, strName.c_str());

			lpRecipients->aEntries[iRecipNum].rgPropVals[2].ulPropTag		= PR_SMTP_ADDRESS;
			hr = MAPIAllocateMore(strEmail.size()+1, lpRecipients->aEntries[iRecipNum].rgPropVals,
								  (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals[2].Value.lpszA);
			if (hr != hrSuccess)
				goto exit;
			lstrcpy(lpRecipients->aEntries[iRecipNum].rgPropVals[2].Value.lpszA, strEmail.c_str());

			lpRecipients->aEntries[iRecipNum].rgPropVals[3].ulPropTag		= PR_ENTRYID;
			hr = ECCreateOneOff((char *)strName.c_str(), "SMTP", (char *)strEmail.c_str(), MAPI_SEND_NO_RICH_INFO, &cbEntryID, &lpEntryID);
			if (hr != hrSuccess)
				goto exit;
				
			hr = MAPIAllocateMore(cbEntryID, lpRecipients->aEntries[iRecipNum].rgPropVals,
								  (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals[3].Value.bin.lpb);
			if (hr != hrSuccess)
				goto exit;

			lpRecipients->aEntries[iRecipNum].rgPropVals[3].Value.bin.cb = cbEntryID;
			memcpy(lpRecipients->aEntries[iRecipNum].rgPropVals[3].Value.bin.lpb, lpEntryID, cbEntryID);

			lpRecipients->aEntries[iRecipNum].rgPropVals[4].ulPropTag	= PR_ADDRTYPE;
			lpRecipients->aEntries[iRecipNum].rgPropVals[4].Value.lpszA	= "SMTP";

			strSearch = "SMTP:"+strEmail;
			transform(strSearch.begin(), strSearch.end(), strSearch.begin(), ::toupper);
			lpRecipients->aEntries[iRecipNum].rgPropVals[5].ulPropTag	= PR_SEARCH_KEY;
			lpRecipients->aEntries[iRecipNum].rgPropVals[5].Value.bin.cb = strSearch.size() + 1; // we include the trailing 0 as MS does this also
			hr = MAPIAllocateMore(strSearch.size()+1, lpRecipients->aEntries[iRecipNum].rgPropVals,
								  (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals[5].Value.bin.lpb);
			if (hr != hrSuccess)
				goto exit;
			memcpy(lpRecipients->aEntries[iRecipNum].rgPropVals[5].Value.bin.lpb, strSearch.c_str(), strSearch.size()+1);

			// Add Email address
			lpRecipients->aEntries[iRecipNum].rgPropVals[6].ulPropTag		= PR_EMAIL_ADDRESS;
			hr = MAPIAllocateMore(strEmail.size()+1, lpRecipients->aEntries[iRecipNum].rgPropVals,
								  (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals[6].Value.lpszA);
			if (hr != hrSuccess)
				goto exit;
			lstrcpy(lpRecipients->aEntries[iRecipNum].rgPropVals[6].Value.lpszA, strEmail.c_str());

			// Add display type
			lpRecipients->aEntries[iRecipNum].rgPropVals[7].ulPropTag = PR_DISPLAY_TYPE;
			lpRecipients->aEntries[iRecipNum].rgPropVals[7].Value.ul = DT_MAILUSER;			

			if (lpEntryID) {
				MAPIFreeBuffer(lpEntryID);
				lpEntryID = NULL;
			}

		}

		lpRecipients->cEntries++;
	} // end recipient loop

exit:
	if (lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	return hr;
}

// copies data from addressbook into lpRecipient
HRESULT VMIMEToMAPI::modifyFromAddressBook(LPSPropValue *lppPropVals, ULONG *lpulValues, const char *email, const char *fullname, ULONG ulRecipType, LPSPropTagArray lpPropsList) {
	HRESULT hr = hrSuccess;
	LPENTRYID lpDDEntryID = NULL;
	ULONG cbDDEntryID;
	ULONG ulObj = 0;
	ADRLIST *lpAdrList = NULL;
	FlagList *lpFlagList = NULL;
	LPSPropValue lpProp = NULL;
	SPropValue sRecipProps[9]; // 8 from addressbook + PR_RECIPIENT_TYPE == max
	ULONG cValues = 0;
	SizedSPropTagArray(8, sptaAddress) = { 8, { PR_SMTP_ADDRESS, PR_ADDRTYPE, PR_EMAIL_ADDRESS, PR_DISPLAY_TYPE, PR_DISPLAY_NAME, PR_ENTRYID, PR_SEARCH_KEY, PR_OBJECT_TYPE } };

	if (!m_lpAdrBook) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	if ((!email || *email == '\0') && (!fullname || *fullname == '\0')) {
		// we have no data to lookup
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}


	if (!m_lpDefaultDir) {
    	hr = m_lpAdrBook->GetDefaultDir(&cbDDEntryID, &lpDDEntryID);
    	if (hr != hrSuccess)
		    goto exit;

    	hr = m_lpAdrBook->OpenEntry(cbDDEntryID, lpDDEntryID, NULL, 0, &ulObj, (LPUNKNOWN*)&m_lpDefaultDir);
    	if (hr != hrSuccess)
	    	goto exit;
    }

    hr = MAPIAllocateBuffer(CbNewSRowSet(1), (void **) &lpAdrList);
    if (hr != hrSuccess)
        goto exit;
        
    lpAdrList->cEntries = 1;
    lpAdrList->aEntries[0].cValues = 1;
    
    hr = MAPIAllocateBuffer(sizeof(SPropValue), (void **) &lpAdrList->aEntries[0].rgPropVals);
    if (hr != hrSuccess)
        goto exit;
        
    lpAdrList->aEntries[0].rgPropVals[0].ulPropTag = PR_DISPLAY_NAME;
	// static reference is OK here
	if (!email || *email == '\0')
		lpAdrList->aEntries[0].rgPropVals[0].Value.lpszA = (char *)fullname; // try to find with fullname for groups without email addresses
	else
		lpAdrList->aEntries[0].rgPropVals[0].Value.lpszA = (char *)email; // normally resolve on email address

    hr = MAPIAllocateBuffer(CbNewFlagList(1), (void **) &lpFlagList);
    if (hr != hrSuccess)
        goto exit;
        
    lpFlagList->cFlags = 1;
    lpFlagList->ulFlag[0] = MAPI_UNRESOLVED;
        
    hr = m_lpDefaultDir->ResolveNames((LPSPropTagArray)&sptaAddress, EMS_AB_ADDRESS_LOOKUP, lpAdrList, lpFlagList);
    if (hr != hrSuccess)
        goto exit;
        
    if (lpFlagList->cFlags != 1 || lpFlagList->ulFlag[0] != MAPI_RESOLVED) {
        hr = MAPI_E_NOT_FOUND;
        goto exit;
    }

	if (PROP_TYPE(lpPropsList->aulPropTag[0]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_ADDRTYPE);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[0]; // PR_xxx_ADDRTYPE;
		sRecipProps[cValues].Value.lpszA = lpProp->Value.lpszA;
		cValues++;
	}

	if (PROP_TYPE(lpPropsList->aulPropTag[1]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_DISPLAY_NAME);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[1]; // PR_xxx_DISPLAY_NAME;
		if (!fullname || *fullname == '\0')
			sRecipProps[cValues].Value.lpszA = lpProp->Value.lpszA;	// use addressbook version
		else
			sRecipProps[cValues].Value.lpszA = (char*)fullname;		// use email version
		cValues++;
	}

	if (PROP_TYPE(lpPropsList->aulPropTag[2]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_DISPLAY_TYPE);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[2]; // PR_xxx_DISPLAY_TYPE;
		sRecipProps[cValues].Value.ul = lpProp->Value.ul;
		cValues++;
	}

	if (PROP_TYPE(lpPropsList->aulPropTag[3]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_EMAIL_ADDRESS);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[3]; // PR_xxx_EMAIL_ADDRESS;
		sRecipProps[cValues].Value.lpszA = lpProp->Value.lpszA;
		cValues++;
	}

	if (PROP_TYPE(lpPropsList->aulPropTag[4]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_ENTRYID);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[4]; // PR_xxx_ENTRYID;
		sRecipProps[cValues].Value.bin = lpProp->Value.bin;
		cValues++;
	}

	if (PROP_TYPE(lpPropsList->aulPropTag[5]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_SEARCH_KEY);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[5]; // PR_xxx_SEARCH_KEY;
		sRecipProps[cValues].Value.bin = lpProp->Value.bin;
		cValues++;
	}


	if (PROP_TYPE(lpPropsList->aulPropTag[6]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_SMTP_ADDRESS);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[6]; // PR_xxx_SMTP_ADDRESS;
		sRecipProps[cValues].Value.lpszA = lpProp->Value.lpszA;
		cValues++;
	}

	lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_OBJECT_TYPE);
	if (!lpProp) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	sRecipProps[cValues].ulPropTag = PR_OBJECT_TYPE;
	sRecipProps[cValues].Value.ul = lpProp->Value.ul;
	cValues++;

	if (ulRecipType != MAPI_ORIG) {
		sRecipProps[cValues].ulPropTag = PR_RECIPIENT_TYPE;
		sRecipProps[cValues].Value.ul = ulRecipType;
		cValues++;
	}

	hr = Util::HrCopyPropertyArray(sRecipProps, cValues, lppPropVals, &cValues);
	if (hr == hrSuccess && lpulValues)
		*lpulValues = cValues;

	
exit:
	if (lpAdrList)
		FreeProws((LPSRowSet)lpAdrList);

	if (lpFlagList)
	    MAPIFreeBuffer(lpFlagList);

	if (lpDDEntryID)
		MAPIFreeBuffer(lpDDEntryID);

	return hr;
}


/*		Disect Body

Here we are going to split the body into pieces and throw every part into its container.
We make decisions on the basis of Content Types...

Content Types...
http://www.faqs.org/rfcs/rfc2046.html

Top level				Subtypes

discrete:
	text				plain, html, richtext, enriched
	image				jpeg, gif, png.. etc
	audio				basic, wav, ai.. etc
	video				mpeg, avi.. etc
	application			octet-stream, postscript

composite:
	multipart			mixed, alternative, digest ( contains message ), paralell, 
	message				rfc822, partial ( please no fragmentation and reassembly ), external-body

*/

HRESULT VMIMEToMAPI::disectBody(vmime::header* VMHeader, vmime::body* VMBody, IMessage* lpMessage, bool onlyBody, bool filterDouble, bool bAppendBody) {
	HRESULT	hr = hrSuccess;
	IStream *lpStream = NULL;
	SPropValue sPropSMIMEClass;
	bool bFilterDouble = filterDouble;
	bool bMixed = bAppendBody;
	ICalToMapi *lpIcalMapi = NULL;

	try {
		vmime::mediaType mt = VMHeader->ContentType().getValue();

		// find body type
		if (mt.getType() == "multipart") {
			if (VMBody->getPartCount() > 0) {
				if (mt.getSubType() == "appledouble")
					bFilterDouble = true;
				else if (mt.getSubType() == "mixed")
					bMixed = true;
				// recursively process multipart message
				for (int i=0; i < VMBody->getPartCount(); i++) {
					vmime::bodyPart* VMBodyPart = VMBody->getPartAt(i);

					hr = disectBody(VMBodyPart->getHeader(), VMBodyPart->getBody(), lpMessage, onlyBody, bFilterDouble, bMixed);
					if (hr != hrSuccess) {
						lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse sub multipart %d of mail body", i);
						goto exit;
					}
				}
				bFilterDouble = filterDouble;
			} else {
				// a lonely attachment ?
				hr = handleAttachment(VMHeader, VMBody, lpMessage);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to set message smime attachment");
					goto exit;
				}

				if (mt.getSubType() == "signed") {
					// the attachment is a signed message, so mark the message container as such
					sPropSMIMEClass.ulPropTag = PR_MESSAGE_CLASS;
					sPropSMIMEClass.Value.lpszA = "IPM.Note.SMIME.MultipartSigned";
					hr = lpMessage->SetProps(1, &sPropSMIMEClass, NULL);
					if (hr != hrSuccess) {
						lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to set message class");
						goto exit;
					}
				}
			}


		// Only handle as inline text if no filename is specified and not specified as 'attachment'
		// or if the text part is the only body part in the mail
		} else if (	mt.getType() == "text" && (mt.getSubType() == "plain" || mt.getSubType() == "html") && (
					(VMHeader->ContentDisposition().getValue().getName() != "attachment" && 
					 !VMHeader->ContentDisposition().hasParameter("filename") && 
					 !VMHeader->ContentType().hasParameter("name")) || 
					 onlyBody)
					) {
			if (mt.getSubType() == "html") {
				hr = handleHTMLTextpart(VMHeader, VMBody, lpMessage, bMixed);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse mail HTML text");
					goto exit;
				}
			} else {
				hr = handleTextpart(VMHeader, VMBody, lpMessage, bMixed);
				if (hr != hrSuccess)
					goto exit;
			}
		
		} else if (mt.getType() == "message") {
			// Create Attach			
			ULONG		ulAttNr			= 0;
			LPATTACH	pAtt			= NULL;
			IMessage*	lpNewMessage	= NULL;
			LPSPropValue	lpSubject	= NULL;
			char *		lpszBody	= NULL;
			char *		lpszBodyOrig	= NULL;
			sMailState savedState;

			std::string newMessage;
			vmime::utility::outputStreamStringAdapter os(newMessage);
			VMBody->generate(os);
					
			lpszBodyOrig = lpszBody = (char *)newMessage.c_str();

			// Skip any leading newlines from the e-mail (attached messaged produced by Microsoft MimeOLE seem to do this)
			while (*lpszBody != '\0' && (*lpszBody == '\r' || *lpszBody == '\n'))
				lpszBody++;

			// and remove from string
			newMessage.erase(0, lpszBody - lpszBodyOrig);

			hr = lpMessage->CreateAttach(NULL, 0, &ulAttNr, &pAtt);
			if (hr != hrSuccess)
				goto next;

			hr = pAtt->OpenProperty(PR_ATTACH_DATA_OBJ, &IID_IMessage, 0, MAPI_CREATE | MAPI_MODIFY, (LPUNKNOWN *)&lpNewMessage);
			if (hr != hrSuccess)
				goto next;

			// handle message-in-message, save current state variables
			savedState = m_mailState;
			m_mailState.reset();
			m_mailState.ulMsgInMsg++;

			hr = convertVMIMEToMAPI(newMessage, lpNewMessage);

			// return to previous state
			m_mailState = savedState;

			if (hr != hrSuccess)
				goto next;

			hr = HrGetOneProp(lpNewMessage, PR_SUBJECT, &lpSubject);
			if (hr != hrSuccess)
				goto next;

			// Set PR_ATTACH_FILENAME of attachment to message subject, (WARNING: abuse of lpSubject variable)
			lpSubject->ulPropTag = PR_DISPLAY_NAME;
			pAtt->SetProps(1, lpSubject, NULL);

			lpSubject->ulPropTag = PR_ATTACH_METHOD;
			lpSubject->Value.ul = ATTACH_EMBEDDED_MSG;
			pAtt->SetProps(1, lpSubject, NULL);

			lpNewMessage->SaveChanges(0);
			pAtt->SaveChanges(0);

next:
			if (lpSubject) {
				MAPIFreeBuffer(lpSubject);
				lpSubject = NULL;
			}

			if (lpNewMessage) {
				lpNewMessage->Release();
				lpNewMessage = NULL;
			}

			if (pAtt) {
				pAtt->Release();
				pAtt = NULL;
			}
			
			if (lpStream) {
				lpStream->Release();
				lpStream = NULL;
			}

		} else if(mt.getType() == "application" && mt.getSubType() == "ms-tnef") {
			LARGE_INTEGER zero = {{0,0}};
			
			hr = CreateStreamOnHGlobal(NULL, TRUE, &lpStream);	
			if(hr != hrSuccess)
				goto exit;
				
			outputStreamMAPIAdapter str(lpStream);
			VMBody->getContents().extract(str);
			hr = lpStream->Seek(zero, STREAM_SEEK_SET, NULL);
			if(hr != hrSuccess)
				goto exit;
			
			ECTNEF tnef(TNEF_DECODE, lpMessage, lpStream);

			hr = tnef.ExtractProps(TNEF_PROP_EXCLUDE, NULL);
			if (hr == hrSuccess) {
				tnef.Finish();
			} else {
				lpLogger->Log(EC_LOGLEVEL_WARNING, "TNEF attachment parsing failed: 0x%08X", hr);
				hr = hrSuccess;
			}
		} else if (mt.getType() == "text" && mt.getSubType() == "calendar") {
			// ical file
			string icaldata;
			vmime::utility::outputStreamStringAdapter os(icaldata);
			std::string strCharset;
			VMBody->getContents().extract(os);

			CreateICalToMapi(lpMessage, m_lpAdrBook, true, &lpIcalMapi);

			// Some senders send utf-8 iCalendar information without a charset (Exchange does this). Default
			// to utf-8 if no charset was specified
			strCharset = VMBody->getCharset().getName();
			if(strCharset == "us-ascii") // We can safely upgrade from us-ascii to utf-8 since it is compatible
			    strCharset = "utf-8";

			hr = lpIcalMapi->ParseICal(icaldata, strCharset, "UTC" , NULL, 0);
			if (hr != hrSuccess || lpIcalMapi->GetItemCount() != 1) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse ical information: 0x%08X, adding as normal attachment", hr);

				hr = handleAttachment(VMHeader, VMBody, lpMessage);
				if (hr != hrSuccess)
					goto exit;
			} else {
				// TODO: create attachment and copy props on success not to crapout my message?
				hr = lpIcalMapi->GetItem(0, IC2M_NO_RECIPIENTS | IC2M_APPEND_ONLY, lpMessage);
				if (hr != hrSuccess)
					goto exit;
			}
		} else if (filterDouble && mt.getType() == "application" && mt.getSubType() == "applefile") {
		} else if (filterDouble && mt.getType() == "application" && mt.getSubType() == "mac-binhex40") {
				// ignore appledouble parts
				// mac-binhex40 is appledouble v1, applefile is v2
				// see: http://www.iana.org/assignments/media-types/multipart/appledouble			
		} else if (mt.getType() == "application" && (mt.getSubType() == "pkcs7-signature" || mt.getSubType() == "x-pkcs7-signature")) {
			// smime signature (smime.p7s)
			// must be handled a level above to get all headers and bodies beloning to the signed message
			m_mailState.bAttachSignature = true;
		} else if (mt.getType() == "application" && (mt.getSubType() == "pkcs7-mime" || mt.getSubType() == "x-pkcs7-mime")) {
			// smime encrypted message (smime.p7m)
			hr = handleAttachment(VMHeader, VMBody, lpMessage);
			if (hr != hrSuccess)
				goto exit;

			// Mark the message so outlook knows how to find the encoded message
			sPropSMIMEClass.ulPropTag = PR_MESSAGE_CLASS;
			sPropSMIMEClass.Value.lpszA = "IPM.Note.SMIME";

			hr = lpMessage->SetProps(1, &sPropSMIMEClass, NULL);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to set message class");
				goto exit;
			}
		} else if (mt.getType() == "application" && mt.getSubType() == "octet-stream") {
			if (VMHeader->ContentDisposition().hasParameter("filename") || VMHeader->ContentType().hasParameter("name")) {
				// should be attachment
				hr = handleAttachment(VMHeader, VMBody, lpMessage);
				if (hr != hrSuccess)
					goto exit;
			} else {
				// possibly text?
				hr = handleTextpart(VMHeader, VMBody, lpMessage, false);
				if (hr != hrSuccess)
					goto exit;
			}
		} else {
			hr = handleAttachment(VMHeader, VMBody, lpMessage);
			if (hr != hrSuccess)
				goto exit;
		}
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on parsing body: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on parsing body: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on parsing body");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if(lpStream)
		lpStream->Release();

	if (lpIcalMapi)
		delete lpIcalMapi;
		
	return hr;
}


HRESULT VMIMEToMAPI::handleTextpart(vmime::header* VMHeader, vmime::body* VMBody, IMessage* lpMessage, bool bAppendBody) {
	HRESULT hr = S_OK;
	std::string strBuffOut;
	IStream *lpStream = NULL;
	LARGE_INTEGER liZero = {{0,0}};
	ULONG ulFlags = MAPI_MODIFY;

	if (m_mailState.bodyLevel < BODY_PLAIN || (m_mailState.bodyLevel == BODY_PLAIN && bAppendBody)) {
		// we have no body up until now, not even an HTML body
		try {
			std::string strMAPIText;
			
			vmime::utility::outputStreamStringAdapter os(strBuffOut);
			VMBody->getContents().extract(os);
						
			// Convert the body to something MAPI is expecting to have in PR_BODY
			try {
				vmime::charset bodyCharset;

				// iso-8859-15 is compatible with us-ascii. We do this
				// so that some broken e-mail with no charset information
				// still come through with accented characters etc.
				bodyCharset = VMBody->getCharset();
				if(stricmp(bodyCharset.getName().c_str(), vmime::charsets::US_ASCII) == 0)
					bodyCharset = vmime::charset(vmime::charsets::ISO8859_15);

				vmime::charset::convert(strBuffOut, strMAPIText, bodyCharset, MAPI_CHARSET);
			}
			catch (vmime::exception&) {
				// vmime::charset::convert() threw up, paste 'raw' body in PR_BODY to deliver the e-mail
				strMAPIText = strBuffOut;
			}

			// create new or reset body
			if (m_mailState.bodyLevel < BODY_PLAIN || !bAppendBody)
				ulFlags |= MAPI_CREATE;

			hr = lpMessage->OpenProperty(PR_BODY, &IID_IStream, STGM_TRANSACTED, ulFlags, (LPUNKNOWN*)&lpStream);
			if (hr != hrSuccess)
				goto exit;

			if (bAppendBody) {
				hr = lpStream->Seek(liZero, SEEK_END, NULL);
				if (hr != hrSuccess)
					goto exit;
			}

			hr = lpStream->Write(strMAPIText.c_str(), strMAPIText.length(), NULL);
			if (hr != hrSuccess)
				goto exit;			

			hr = lpStream->Commit(0);
			if (hr != hrSuccess)
				goto exit;			
		}
		catch (vmime::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on text body: "+e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (std::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on text body: %s", e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (...) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on text body");
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		m_mailState.bodyLevel = BODY_PLAIN;
	} else {
		// we already had a plaintext or html body, so attach this text part
		hr = handleAttachment(VMHeader, VMBody, lpMessage);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse attached text mail");
			goto exit;
		}
	}

exit:
	if (lpStream)
		lpStream->Release();

	return hr;
}

HRESULT VMIMEToMAPI::handleHTMLTextpart(vmime::header* VMHeader, vmime::body* VMBody, IMessage* lpMessage, bool bAppendBody) {
	HRESULT		hr				= hrSuccess;
	IStream*	lpHTMLStream	= NULL; 
	ULONG		cbWritten		= 0;
	std::string strHTML;
	LARGE_INTEGER liZero = {{0,0}};
	ULONG ulFlags = MAPI_MODIFY;

	if (m_mailState.bodyLevel < BODY_HTML || (m_mailState.bodyLevel == BODY_HTML && bAppendBody)) {
		// we're overriding a plain text body, or setting a new body
		try {
			vmime::charset bodyCharset;
			std::basic_string<WCHAR> wstrHTML;

			vmime::utility::outputStreamStringAdapter os(strHTML);
			VMBody->getContents().extract(os);
		
			// iso-8859-15 is compatible with us-ascii. We do this
			// so that some broken e-mail with no charset information
			// still come through with accented characters etc.
			bodyCharset = VMBody->getCharset();
			if(bodyCharset.getName() == vmime::charsets::US_ASCII)
				bodyCharset = vmime::charset(vmime::charsets::ISO8859_15);
		
			// convert incoming HTML source into WCHARs
			if((hr = HrConvertToWString(strHTML, bodyCharset.getName(), wstrHTML)) != hrSuccess) {
				hr = HrConvertToWString(strHTML, vmime::charsets::ISO8859_15, wstrHTML);
			}
			
			if(hr != hrSuccess)
				goto exit;

			// convert wchar-HTML into single-byte us-ascii HTML with escaped chars
			hr = HrEscapeHTML(wstrHTML, strHTML);
			if(hr != hrSuccess)
				goto exit;
		}
		catch (vmime::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on html body: "+e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (std::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on html body: %s", e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (...) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on html body");
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		// create new or reset body
		if (m_mailState.bodyLevel < BODY_HTML || !bAppendBody)
			ulFlags |= MAPI_CREATE;

		hr = lpMessage->OpenProperty(PR_HTML, &IID_IStream, STGM_TRANSACTED, ulFlags, (LPUNKNOWN *)&lpHTMLStream);
		if (hr != hrSuccess)
			goto exit;

		if (bAppendBody) {
			hr = lpHTMLStream->Seek(liZero, SEEK_END, NULL);
			if (hr != hrSuccess)
				goto exit;
		}

		hr = lpHTMLStream->Write(strHTML.c_str(), strHTML.length(), &cbWritten);
		if (hr != hrSuccess)		// check cbWritten too?
			goto exit;

		hr = lpHTMLStream->Commit(0);
		if (hr != hrSuccess)
			goto exit;

		m_mailState.bodyLevel = BODY_HTML;
		if (bAppendBody)
			m_mailState.strHTMLBody.append(strHTML);
		else
			swap(strHTML, m_mailState.strHTMLBody);
	} else {
		// already found html as body, so this is an attachment
		hr = handleAttachment(VMHeader, VMBody, lpMessage);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse attached text mail");
			goto exit;
		}
	}

exit:
	if (lpHTMLStream)
		lpHTMLStream->Release(); 

	return hr;
}


// Handle Attachments.. Now works for inlines and attachments...
HRESULT VMIMEToMAPI::handleAttachment(vmime::header* VMHeader, vmime::body* VMBody, IMessage* lpMessage) {
	HRESULT		hr			= hrSuccess;
	IStream		*lpStream	= NULL; 
	LPATTACH	lpAtt		= NULL;
	ULONG		ulAttNr		= 0;
	std::string	strLongFilename, strFilename, strId, strMimeType, strLocation;
	int			nProps = 0;
	SPropValue	attProps[12];
	int reparse = 0;

	// Create Attach
	hr = lpMessage->CreateAttach(NULL, 0, &ulAttNr, &lpAtt);
	if (hr != hrSuccess)
		goto exit;

	// open stream
	hr = lpAtt->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream, STGM_WRITE|STGM_TRANSACTED,
							MAPI_CREATE|MAPI_MODIFY, (LPUNKNOWN *)&lpStream);
	if (hr != hrSuccess)
		goto exit;

	try {
		// attach adapter, generate in right encoding
		outputStreamMAPIAdapter osMAPI(lpStream);
		VMBody->getContents().generate(osMAPI, vmime::encoding(vmime::encodingTypes::BINARY));

		hr = lpStream->Commit(0);
		if (hr != hrSuccess)
			goto exit;
			
        // Free memory used by the stream
        lpStream->Release();
        lpStream = NULL;

		// set info on attachment
		attProps[nProps].ulPropTag = PR_ATTACH_METHOD;
		attProps[nProps++].Value.ul = ATTACH_BY_VALUE;

		strId = VMHeader->ContentId().getValue().getId();
		if (!strId.empty()) {
			// only set this property when string is present
			// otherwise, you don't get the 'save attachments' list in the main menu of outlook
			attProps[nProps].ulPropTag = PR_ATTACH_CONTENT_ID;
			attProps[nProps++].Value.lpszA = (char*)strId.c_str();
		}

		strLocation = VMHeader->ContentLocation().getValue();
		if (!strLocation.empty()) {
			attProps[nProps].ulPropTag = PR_ATTACH_CONTENT_LOCATION;
			attProps[nProps++].Value.lpszA = (char*)strLocation.c_str();
		}

		// make hidden when inline, is an image or text, has an content id or location, is an HTML mail,
		// has a CID reference in the HTML or has a location reference in the HTML.
		if (VMHeader->ContentDisposition().getValue().getName() == vmime::contentDispositionTypes::INLINE &&
			(VMHeader->ContentType().getValue().getType() == vmime::mediaTypes::IMAGE ||
			 VMHeader->ContentType().getValue().getType() == vmime::mediaTypes::TEXT) &&
			(!strId.empty() || !strLocation.empty()) &&
			m_mailState.bodyLevel == BODY_HTML &&
			((!strId.empty() && strcasestr_str1252((char*)m_mailState.strHTMLBody.c_str(), (char*)string("cid:"+strId).c_str())) ||
			 (!strLocation.empty() && strcasestr_str1252((char*)m_mailState.strHTMLBody.c_str(), (char*)strLocation.c_str()))	))
		{
			attProps[nProps].ulPropTag = PR_ATTACHMENT_HIDDEN;
			attProps[nProps++].Value.b = TRUE;

			attProps[nProps].ulPropTag = PR_ATTACH_FLAGS;
			attProps[nProps++].Value.ul = 4; // ATT_MHTML_REF

			attProps[nProps].ulPropTag = PR_ATTACHMENT_FLAGS;
			attProps[nProps++].Value.ul = 8; // unknown, for now

			if (m_mailState.attachLevel < ATTACH_NORMAL)
				m_mailState.attachLevel = ATTACH_INLINE;

		} else {
			attProps[nProps].ulPropTag = PR_ATTACHMENT_HIDDEN;
			attProps[nProps++].Value.b = FALSE;

			attProps[nProps].ulPropTag = PR_ATTACH_FLAGS;
			attProps[nProps++].Value.ul = 0;

			m_mailState.attachLevel = ATTACH_NORMAL;
		}

		// filenames
		if (VMHeader->ContentDisposition().hasParameter("filename")) {
			strLongFilename = VMHeader->ContentDisposition().getFilename().getConvertedText(MAPI_CHARSET);
			if (VMHeader->ContentDisposition().getFilename().getCharset() == vmime::charset(vmime::charsets::US_ASCII))
				reparse = 1;
		} else if (VMHeader->ContentType().hasParameter("name")) {
			strLongFilename = VMHeader->ContentType().getParameter("name")->getValue().generate();
			reparse = 1;		// never seen q-p here, but does not hurt to try
		} else if (VMHeader->ContentType().getValue().getType() == "text" && VMHeader->ContentType().getValue().getSubType() == "calendar") {
			strLongFilename = "calendar.ics";
		} else {
			// TODO: add guessFilenameFromContentType()
			strLongFilename = "inline.txt";
		}

		// quoted-printable is not available in vmime 0.7.1 on parameterizedHeaderField class
		// outlook express sends qp, icedove sends the encoded version (filename*=<charset>''%FF)
		if (reparse) {
			// text was in plain text, so convert possible quoted-printable
			vmime::word wFilename;
			string strParsed;
			string::size_type startPos = 0;
			string::size_type parsedPos = 0;

			// buffer may contain more than one "word"
			while (startPos < strLongFilename.length()) {
				wFilename.parse(strLongFilename, startPos, strLongFilename.length(), &parsedPos);
				strParsed += wFilename.getConvertedText(MAPI_CHARSET);
				// this is really nasty, but I don't know how to handle this in a 'correct' manner
				while (parsedPos < strLongFilename.length() &&
					   (strLongFilename[parsedPos] == '\n' || strLongFilename[parsedPos] == '\r' ||
						strLongFilename[parsedPos] == '\t' || strLongFilename[parsedPos] == ' '))
					parsedPos++;
				startPos = parsedPos;
			}
			strLongFilename = strParsed;
		}

		getShortFilename(strLongFilename, strFilename);

		attProps[nProps].ulPropTag = PR_ATTACH_LONG_FILENAME;
		attProps[nProps++].Value.lpszA = (char*)strLongFilename.c_str();
		attProps[nProps].ulPropTag = PR_ATTACH_FILENAME;
		attProps[nProps++].Value.lpszA = (char*)strFilename.c_str();

		// i get the impression this doesn't work... so do we really need to set this?
		attProps[nProps].ulPropTag = PR_RENDERING_POSITION;
		attProps[nProps++].Value.i = -1;

		try {
			if (!VMHeader->ContentType().getValue().getType().empty() && !VMHeader->ContentType().getValue().getSubType().empty()) {
				strMimeType = VMHeader->ContentType().getValue().getType() + "/" + VMHeader->ContentType().getValue().getSubType();
				// due to a bug in vmime 0.7, the continuation header text can be prefixed in the string, so strip it (easiest way to fix)
				while (strMimeType[0] == '\r' || strMimeType[0] == '\n' || strMimeType[0] == '\t' || strMimeType[0] == ' ')
					strMimeType.erase(0, 1);
				attProps[nProps].ulPropTag = PR_ATTACH_MIME_TAG;
				attProps[nProps++].Value.lpszA = (char*)strMimeType.c_str();
			}
		}
		catch (vmime::exceptions::no_such_field) {
		}

		hr = lpAtt->SetProps(nProps, attProps, NULL);
		if (hr != hrSuccess)
			goto exit;
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on attachment: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on attachment: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on attachment");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	hr = lpAtt->SaveChanges(0);
	if (hr != hrSuccess)
		goto exit;

exit:
	if (hr != hrSuccess)
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to create attachment");

	if (lpAtt)
		lpAtt->Release();

	if (lpStream)
		lpStream->Release();

	return hr;
}

// Do various fixups of missing or incorrect data
HRESULT VMIMEToMAPI::postWriteFixups(IMessage *lpMessage)
{
    HRESULT hr = hrSuccess;
    LPSPropValue lpMessageClass = NULL;
    ULONG cValues = 0;
    LPSPropValue lpProps = NULL;
    LPSPropValue lpRecProps = NULL;
    ULONG cRecProps = 0;
    ULONG cbConversationIndex = 0;
    LPBYTE lpConversationIndex = NULL;
    
    PROPMAP_START
     PROPMAP_NAMED_ID(RECURRENCESTATE,			PT_BINARY,	PSETID_Appointment, dispidRecurrenceState)
     
     PROPMAP_NAMED_ID(RESPONSESTATUS, 			PT_LONG, 	PSETID_Appointment, dispidResponseStatus)
     PROPMAP_NAMED_ID(RECURRING, 				PT_BOOLEAN, PSETID_Appointment, dispidRecurring)
     PROPMAP_NAMED_ID(ATTENDEECRITICALCHANGE, 	PT_SYSTIME, PSETID_Meeting, dispidAttendeeCriticalChange)
     PROPMAP_NAMED_ID(OWNERCRITICALCHANGE, 		PT_SYSTIME, PSETID_Meeting, dispidOwnerCriticalChange)

     PROPMAP_NAMED_ID(MEETING_RECURRING,		PT_BOOLEAN,	PSETID_Meeting, dispidIsRecurring)
     PROPMAP_NAMED_ID(MEETING_STARTRECDATE,		PT_LONG,	PSETID_Meeting, dispidStartRecurrenceDate)
     PROPMAP_NAMED_ID(MEETING_STARTRECTIME,		PT_LONG,	PSETID_Meeting, dispidStartRecurrenceTime)
     PROPMAP_NAMED_ID(MEETING_ENDRECDATE,		PT_LONG,	PSETID_Meeting, dispidEndRecurrenceDate)
     PROPMAP_NAMED_ID(MEETING_ENDRECTIME,		PT_LONG,	PSETID_Meeting, dispidEndRecurrenceTime)

     PROPMAP_NAMED_ID(MEETING_DAYINTERVAL,		PT_I2,		PSETID_Meeting, dispidDayInterval)
     PROPMAP_NAMED_ID(MEETING_WEEKINTERVAL,		PT_I2,		PSETID_Meeting, dispidWeekInterval)
     PROPMAP_NAMED_ID(MEETING_MONTHINTERVAL,	PT_I2,		PSETID_Meeting, dispidMonthInterval)
     PROPMAP_NAMED_ID(MEETING_YEARINTERVAL,		PT_I2,		PSETID_Meeting, dispidYearInterval)
     
     PROPMAP_NAMED_ID(MEETING_DOWMASK,			PT_LONG,	PSETID_Meeting, dispidDayOfWeekMask)
     PROPMAP_NAMED_ID(MEETING_DOMMASK,			PT_LONG,	PSETID_Meeting, dispidDayOfMonthMask)
     PROPMAP_NAMED_ID(MEETING_MOYMASK,			PT_LONG,	PSETID_Meeting, dispidMonthOfYearMask)

     PROPMAP_NAMED_ID(MEETING_RECURRENCETYPE,	PT_I2,		PSETID_Meeting, dispidOldRecurrenceType)
     PROPMAP_NAMED_ID(MEETING_DOWSTART,			PT_I2,		PSETID_Meeting, dispidDayOfWeekStart)
     
     PROPMAP_NAMED_ID(CLIPSTART,				PT_SYSTIME,	PSETID_Appointment, dispidClipStart)
     PROPMAP_NAMED_ID(CLIPEND,					PT_SYSTIME,	PSETID_Appointment, dispidClipEnd)
     
    PROPMAP_INIT(lpMessage)

    hr = HrGetOneProp(lpMessage, PR_MESSAGE_CLASS, &lpMessageClass);
    if(hr != hrSuccess)
        goto exit;

    if(strnicmp(lpMessageClass->Value.lpszA, "IPM.Schedule.Meeting.", strlen( "IPM.Schedule.Meeting." )) == 0) {
        // IPM.Schedule.Meeting.*

        /* Fixups for CDO (blackberry):
         * CDO requires the following properties to be available on meeting requests (defaults in brackets):
         * - responsestatus (0)
         * - recurring (false)
         * - ATTENDEECRITICALCHANGE (now())
         * - OWNERCRITICALCHANGE (now())
         * - PR_OWNER_APPT_ID (-1)
         */

        SizedSPropTagArray(6, sptaMeetingReqProps) = {6, {PROP_RESPONSESTATUS, PROP_RECURRING, PROP_ATTENDEECRITICALCHANGE, PROP_OWNERCRITICALCHANGE, PR_OWNER_APPT_ID, PR_CONVERSATION_INDEX }};
        
        hr = lpMessage->GetProps((LPSPropTagArray)&sptaMeetingReqProps, 0, &cValues, &lpProps);
        if(FAILED(hr))
            goto exit;
        
        // If hr is hrSuccess then all properties are available, and we don't need to do anything
        if(hr != hrSuccess) {
            hr = hrSuccess;
            
            if(lpProps[0].ulPropTag != PROP_RESPONSESTATUS) {
                lpProps[0].ulPropTag = PROP_RESPONSESTATUS;
                lpProps[0].Value.ul = 0;
            }
            
            if(lpProps[1].ulPropTag != PROP_RECURRING) {
                lpProps[1].ulPropTag = PROP_RECURRING;
                lpProps[1].Value.b = false;
            }
            
            if(lpProps[2].ulPropTag != PROP_ATTENDEECRITICALCHANGE) {
                lpProps[2].ulPropTag = PROP_ATTENDEECRITICALCHANGE;
                UnixTimeToFileTime(time(NULL), &lpProps[2].Value.ft);
            }
            
            if(lpProps[3].ulPropTag != PROP_OWNERCRITICALCHANGE) {
                lpProps[3].ulPropTag = PROP_OWNERCRITICALCHANGE;
                UnixTimeToFileTime(time(NULL), &lpProps[3].Value.ft);
            }
            
            if(lpProps[4].ulPropTag != PR_OWNER_APPT_ID) {
                lpProps[4].ulPropTag = PR_OWNER_APPT_ID;
                lpProps[4].Value.ul = -1;
            }
            
            if(lpProps[5].ulPropTag != PR_CONVERSATION_INDEX) {
                lpProps[5].ulPropTag = PR_CONVERSATION_INDEX;
                hr = ScCreateConversationIndex(0, NULL, &cbConversationIndex, &lpConversationIndex);
                if(hr != hrSuccess)
                    goto exit;
                
                lpProps[5].Value.bin.cb = cbConversationIndex;
                lpProps[5].Value.bin.lpb = lpConversationIndex;
            }
            
            hr = lpMessage->SetProps(6, lpProps, NULL);
            if(hr != hrSuccess)
                goto exit;
		}

		// @todo
		// this code should be in a separate function, which can easily
		// do 'goto exit', and we can continue here with other fixes.
		if(lpProps[1].Value.b)
		{
			// This is a recurring appointment. Generate the properties needed by CDO, which can be
			// found in the recurrence state. Since these properties are completely redundant we always
			// write them to correct any possible errors in the incoming message.
			SPropValue sMeetingProps[14];
			SizedSPropTagArray (3, sptaRecProps) =  { 3, { PROP_RECURRENCESTATE, PROP_CLIPSTART, PROP_CLIPEND } };
			RecurrenceState rec;

			// @todo, if all properties are not available: remove recurrence true marker
			hr = lpMessage->GetProps((LPSPropTagArray)&sptaRecProps, 0, &cRecProps, &lpRecProps);
			if(hr != hrSuccess) // Warnings not accepted
				goto exit;
			
			hr = rec.ParseBlob((char *)lpRecProps[0].Value.bin.lpb, (unsigned int)lpRecProps[0].Value.bin.cb, 0);
			if(FAILED(hr))
				goto exit;
			
			// Ignore warnings	
			hr = hrSuccess;
			
			sMeetingProps[0].ulPropTag = PROP_MEETING_STARTRECDATE;
			sMeetingProps[0].Value.ul = FileTimeToIntDate(lpRecProps[1].Value.ft);
			
			sMeetingProps[1].ulPropTag = PROP_MEETING_STARTRECTIME;
			sMeetingProps[1].Value.ul = SecondsToIntTime(rec.ulStartTimeOffset * 60);

			if(rec.ulEndType != ET_NEVER) {
				sMeetingProps[2].ulPropTag = PROP_MEETING_ENDRECDATE;
				sMeetingProps[2].Value.ul = FileTimeToIntDate(lpRecProps[2].Value.ft);
			} else {
				sMeetingProps[2].ulPropTag = PR_NULL;
			}
			
			sMeetingProps[3].ulPropTag = PROP_MEETING_ENDRECTIME;
			sMeetingProps[3].Value.ul = SecondsToIntTime(rec.ulEndTimeOffset * 60);

			// Default the following values to 0 and set them later if needed
			sMeetingProps[4].ulPropTag = PROP_MEETING_DAYINTERVAL;
			sMeetingProps[4].Value.i = 0;
			sMeetingProps[5].ulPropTag = PROP_MEETING_WEEKINTERVAL;
			sMeetingProps[5].Value.i = 0;
			sMeetingProps[6].ulPropTag = PROP_MEETING_MONTHINTERVAL;
			sMeetingProps[6].Value.i = 0;
			sMeetingProps[7].ulPropTag = PROP_MEETING_YEARINTERVAL;
			sMeetingProps[7].Value.i = 0;
			
			sMeetingProps[8].ulPropTag = PROP_MEETING_DOWMASK;
			sMeetingProps[8].Value.ul = 0 ;

			sMeetingProps[9].ulPropTag = PROP_MEETING_DOMMASK;
			sMeetingProps[9].Value.ul = 0;
			
			sMeetingProps[10].ulPropTag = PROP_MEETING_MOYMASK;
			sMeetingProps[10].Value.ul = 0;
			
			sMeetingProps[11].ulPropTag = PROP_MEETING_RECURRENCETYPE;
			sMeetingProps[11].Value.ul = 0;
			
			sMeetingProps[12].ulPropTag = PROP_MEETING_DOWSTART;
			sMeetingProps[12].Value.i = rec.ulFirstDOW;
			
			sMeetingProps[13].ulPropTag = PROP_MEETING_RECURRING;
			sMeetingProps[13].Value.b = true;

			// Set the values depending on the type
			switch(rec.ulRecurFrequency) {
				case RF_DAILY:
					if(rec.ulPatternType == 0) {
						// Daily
						sMeetingProps[4].Value.i = rec.ulPeriod / 1440; // DayInterval
						sMeetingProps[11].Value.i = 64;					// RecurrenceType
					} else {
						// Every workday, actually a weekly recurrence (weekly every workday)
						sMeetingProps[5].Value.i = 1;					// WeekInterval
						sMeetingProps[8].Value.ul = 62; // Mo-Fri
						sMeetingProps[11].Value.i = 48;	// Weekly	
					}
					break;
				case RF_WEEKLY:
					sMeetingProps[5].Value.i = rec.ulPeriod;			// WeekInterval
					sMeetingProps[8].Value.ul = rec.ulWeekDays;			// DayOfWeekMask
					sMeetingProps[11].Value.i = 48;						// RecurrenceType
					break;
				case RF_MONTHLY:
					sMeetingProps[6].Value.i = rec.ulPeriod;			// MonthInterval
					if(rec.ulPatternType == 3) { // Every Nth [weekday] of the month
						sMeetingProps[5].Value.ul = rec.ulWeekNumber;	// WeekInterval
						sMeetingProps[8].Value.ul = rec.ulWeekDays;		// DayOfWeekMask
						sMeetingProps[11].Value.i = 56;					// RecurrenceType
					} else {
						sMeetingProps[9].Value.ul = 1 << (rec.ulDayOfMonth-1); // day of month 1..31 mask
						sMeetingProps[11].Value.i = 12;					// RecurrenceType
					}
					break;
				case RF_YEARLY:
					sMeetingProps[6].Value.i = rec.ulPeriod;			// YearInterval
					sMeetingProps[7].Value.i = rec.ulPeriod / 12;		// MonthInterval
					sMeetingProps[10].Value.ul = 1 << ((rec.ulFirstDateTime/(24*60*29)) % 12); // month of year (minutes since beginning of the year)
					
					if(rec.ulPatternType == 3) { // Every Nth [weekday] in Month X
						sMeetingProps[5].Value.ul = rec.ulWeekNumber;	// WeekInterval
						sMeetingProps[8].Value.ul = rec.ulWeekDays;		// DayOfWeekMask
						sMeetingProps[11].Value.i = 51;					// RecurrenceType
					} else {
						sMeetingProps[9].Value.ul = 1 << (rec.ulDayOfMonth-1); // day of month 1..31 mask
						sMeetingProps[11].Value.i = 7;					// RecurrenceType
					}
					break;
				default:
					break;
			}
			
			hr = lpMessage->SetProps(14, sMeetingProps, NULL);
			if(hr != hrSuccess)
				goto exit;
		}
        
    }
    
exit:
	if(lpRecProps)
		MAPIFreeBuffer(lpRecProps);
		
    if(lpConversationIndex)
        MAPIFreeBuffer(lpConversationIndex);
        
    if(lpMessageClass)
        MAPIFreeBuffer(lpMessageClass);
        
    if(lpProps)
        MAPIFreeBuffer(lpProps);
        
    return hr;     
}

// options.h code
void imopt_default_delivery_options(delivery_options *dopt) {
	dopt->use_received_date = true;
	dopt->mark_as_read = false;
	dopt->user_entryid = NULL;
}

void imopt_default_sending_options(sending_options *sopt) {
	sopt->alternate_boundary = NULL;
	sopt->no_recipients_workaround = false;
	sopt->msg_in_msg = false;
	sopt->headers_only = false;
	sopt->add_received_date = false;
	sopt->force_tnef = false;
}
