/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include "ClientUtil.h"

#ifndef NO_GETTEXT
#include <libintl.h>
#define _(string) dcgettext ("zarafa", string, LC_MESSAGES)
#else
#define _(string) string
#endif

#include <mapi.h>
#include <mapidefs.h>
#include <mapiutil.h>

#include "CommonUtil.h"
#include "WSTransport.h"
#include "ECConfig.h"

#include "Zarafa.h"
#include "ECGuid.h"
#include "guid.h"
#include "mapiguidext.h"
#include "mapiext.h"

#include "Mem.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

HRESULT ClientUtil::HrInitializeStatusRow (const char * lpszProviderDisplay, ULONG ulResourceType, LPMAPISUP lpMAPISup, LPSPropValue lpspvIdentity, ULONG ulFlags)
{
	HRESULT			hResult = hrSuccess;
	LPSPropValue	lpspvStatusRow = NULL;
	ULONG			cCurVal = 0;
	unsigned int	size = 0;
	std::string		strSearchKey;

	hResult = MAPIAllocateBuffer(sizeof(SPropValue) * 13, (void**)&lpspvStatusRow);
	if(hResult != hrSuccess)
		goto exit;

	memset(lpspvStatusRow, 0, sizeof(SPropValue) * 13);

	if(lpszProviderDisplay)
	{
		size = strlen(lpszProviderDisplay)+1;

		// Set the PR_PROVIDER_DISPLAY property:
		lpspvStatusRow[cCurVal].ulPropTag = PR_PROVIDER_DISPLAY;
		hResult = MAPIAllocateMore(size, lpspvStatusRow, (void**)&lpspvStatusRow[cCurVal].Value.lpszA);
		if(hResult != hrSuccess)
			goto exit;
		memcpy(lpspvStatusRow[cCurVal].Value.lpszA, lpszProviderDisplay, size);
		cCurVal++;

	// Set the PR_DISPLAY_NAME property
		lpspvStatusRow[cCurVal].ulPropTag = PR_DISPLAY_NAME;
		hResult = MAPIAllocateMore(size, lpspvStatusRow, (void**)&lpspvStatusRow[cCurVal].Value.lpszA);
		if(hResult != hrSuccess)
			goto exit;
		memcpy(lpspvStatusRow[cCurVal].Value.lpszA, lpszProviderDisplay, size);
		cCurVal++;
	}

	// PR_PROVIDER_DLL_NAME
	lpspvStatusRow[cCurVal].ulPropTag = PR_PROVIDER_DLL_NAME;
	lpspvStatusRow[cCurVal++].Value.lpszA = ZARAFA_DLL_NAME;

	// Set the PR_STATUS_CODE property:
	lpspvStatusRow[cCurVal].ulPropTag = PR_STATUS_CODE;
	lpspvStatusRow[cCurVal++].Value.l = 1;

	// Set the PR_STATUS_STRING property
	lpspvStatusRow[cCurVal].ulPropTag = PR_STATUS_STRING;
	lpspvStatusRow[cCurVal++].Value.lpszA = "Available";

	// Set the PR_IDENTITY_ENTRYID property
	lpspvStatusRow[cCurVal].ulPropTag = PR_IDENTITY_ENTRYID;
	lpspvStatusRow[cCurVal++].Value.bin = lpspvIdentity[XPID_EID].Value.bin;

	// Set the PR_IDENTITY_DISPLAY property
	lpspvStatusRow[cCurVal].ulPropTag = PR_IDENTITY_DISPLAY;
	lpspvStatusRow[cCurVal++].Value.lpszA = lpspvIdentity[XPID_NAME].Value.lpszA;

	// Set the PR_IDENTITY_SEARCH_KEY property
	lpspvStatusRow[cCurVal].ulPropTag = PR_IDENTITY_SEARCH_KEY;

	strSearchKey = lpspvIdentity[XPID_ADDRTYPE].Value.lpszA;
	strSearchKey += ":";
	strSearchKey += lpspvIdentity[XPID_ADDRESS].Value.lpszA;
	lpspvStatusRow[cCurVal].Value.bin.cb = strSearchKey.size()+1;
	MAPIAllocateMore(lpspvStatusRow[cCurVal].Value.bin.cb, lpspvStatusRow, (void**)&lpspvStatusRow[cCurVal].Value.bin.lpb);
	memcpy(lpspvStatusRow[cCurVal].Value.bin.lpb, (unsigned char*)strSearchKey.c_str(), lpspvStatusRow[cCurVal].Value.bin.cb);
	cCurVal++;

	// Set the PR_OWN_STORE_ENTRYID property
	lpspvStatusRow[cCurVal].ulPropTag = PR_OWN_STORE_ENTRYID;
	lpspvStatusRow[cCurVal++].Value.bin = lpspvIdentity[XPID_STORE_EID].Value.bin;

	lpspvStatusRow[cCurVal].ulPropTag = PR_RESOURCE_METHODS;
	lpspvStatusRow[cCurVal++].Value.l = STATUS_VALIDATE_STATE;

	lpspvStatusRow[cCurVal].ulPropTag = PR_RESOURCE_TYPE;
	lpspvStatusRow[cCurVal++].Value.l = ulResourceType; //like MAPI_STORE_PROVIDER or MAPI_TRANSPORT_PROVIDER

	hResult = lpMAPISup->ModifyStatusRow(cCurVal, lpspvStatusRow, ulFlags);
exit:
	if(lpspvStatusRow)
		MAPIFreeBuffer(lpspvStatusRow);

	return hResult;
}

HRESULT ClientUtil::HrSetIdentity(WSTransport *lpTransport, LPMAPISUP lpMAPISup, LPSPropValue* lppIdentityProps)
{
	HRESULT			hr = hrSuccess;
	char			*lpszProfileFullName = NULL;
	char			*lpszProfileAddress = NULL;
	ULONG			cbEntryStore = 0;
	LPENTRYID		lpEntryStore = NULL;
	LPENTRYID		lpEID = NULL;
	ULONG			cbEID = 0;
	ULONG			cValues = 0;
	ULONG			ulSize = 0;
	ULONG			cbUserId = 0;
	LPENTRYID		lpUserId = NULL;

	std::string		strProfileDisplayName;
	std::string		strProfileSenderSearchKey;

	LPSPropValue	lpIdentityProps = NULL;

	// Get the username and email adress
	hr = lpTransport->HrGetUserProfile(&lpszProfileFullName, &lpszProfileAddress, &cbUserId, &lpUserId);
	if(hr != hrSuccess)
		goto exit;

	cValues = NUM_IDENTITY_PROPS;
	MAPIAllocateBuffer(sizeof(SPropValue) * cValues, (void**)&lpIdentityProps);
	memset(lpIdentityProps, 0, sizeof(SPropValue) * cValues);
	
	strProfileDisplayName = lpszProfileFullName;
	strProfileDisplayName+= " [";
	strProfileDisplayName+= lpszProfileAddress;
	strProfileDisplayName+= "]";

	strProfileSenderSearchKey = TRANSPORT_ADDRESS_TYPE_ZARAFA;
	strProfileSenderSearchKey+= ":";
	strProfileSenderSearchKey +=lpszProfileAddress;

	lpIdentityProps[XPID_EID].ulPropTag = PR_SENDER_ENTRYID;
	lpIdentityProps[XPID_EID].Value.bin.cb = cbUserId;
	MAPIAllocateMore(cbUserId, (LPVOID)lpIdentityProps, (void**)&lpIdentityProps[XPID_EID].Value.bin.lpb);
	memcpy(lpIdentityProps[XPID_EID].Value.bin.lpb, (LPBYTE)lpUserId, cbUserId);

	// Create the PR_SENDER_NAME property value.
	lpIdentityProps[XPID_NAME].ulPropTag = PR_SENDER_NAME;
	ulSize = strlen(lpszProfileFullName) + 1;
	MAPIAllocateMore(ulSize, (LPVOID)lpIdentityProps , (void**)&lpIdentityProps[XPID_NAME].Value.lpszA);
	memcpy(lpIdentityProps[XPID_NAME].Value.lpszA, lpszProfileFullName, ulSize);

	// Create the PR_SENDER_SEARCH_KEY value. 
    lpIdentityProps[XPID_SEARCH_KEY].ulPropTag = PR_SENDER_SEARCH_KEY;
	lpIdentityProps[XPID_SEARCH_KEY].Value.bin.cb=strProfileSenderSearchKey.size()+1;
	MAPIAllocateMore(lpIdentityProps[XPID_SEARCH_KEY].Value.bin.cb, (LPVOID)lpIdentityProps , (void**)&lpIdentityProps[XPID_SEARCH_KEY].Value.bin.lpb);
	memcpy(lpIdentityProps[XPID_SEARCH_KEY].Value.bin.lpb, strProfileSenderSearchKey.c_str(), lpIdentityProps[XPID_SEARCH_KEY].Value.bin.cb);

	// PR_SENDER_EMAIL_ADDRESS
	lpIdentityProps[XPID_ADDRESS].ulPropTag = PR_SENDER_EMAIL_ADDRESS;
	ulSize = strlen(lpszProfileAddress) + 1;
	MAPIAllocateMore(ulSize, (LPVOID)lpIdentityProps , (void**)&lpIdentityProps[XPID_ADDRESS].Value.lpszA);
	memcpy(lpIdentityProps[XPID_ADDRESS].Value.lpszA, lpszProfileAddress, ulSize);

	// PR_SENDER_ADDRTYPE
	lpIdentityProps[XPID_ADDRTYPE].ulPropTag = PR_SENDER_ADDRTYPE;
	ulSize = strlen(TRANSPORT_ADDRESS_TYPE_ZARAFA)+1;
	MAPIAllocateMore(ulSize, (LPVOID)lpIdentityProps , (void**)&lpIdentityProps[XPID_ADDRTYPE].Value.lpszA);
	memcpy(lpIdentityProps[XPID_ADDRTYPE].Value.lpszA, TRANSPORT_ADDRESS_TYPE_ZARAFA, ulSize);

	//PR_OWN_STORE_ENTRYID
	// Get the default store for this user, not an issue if it fails when not on home server
	if(lpTransport->HrGetStore(0, NULL, &cbEntryStore, &lpEntryStore, 0, NULL) == hrSuccess) 
	{
		hr = lpMAPISup->WrapStoreEntryID(cbEntryStore, lpEntryStore, &cbEID, (LPENTRYID*)&lpEID);
		if(hr != hrSuccess) 
			goto exit;

		lpIdentityProps[XPID_STORE_EID].ulPropTag = PR_OWN_STORE_ENTRYID;
		lpIdentityProps[XPID_STORE_EID].Value.bin.cb = cbEID;
		MAPIAllocateMore(cbEID, (LPVOID)lpIdentityProps , (void**)&lpIdentityProps[XPID_STORE_EID].Value.bin.lpb);
		memcpy(lpIdentityProps[XPID_STORE_EID].Value.bin.lpb, (LPBYTE) lpEID, cbEID);
	}
	
	// Set the identity in the global provider identity
	*lppIdentityProps = lpIdentityProps;

exit:
	if(hr != hrSuccess && lpIdentityProps != NULL) {
		MAPIFreeBuffer(lpIdentityProps);
		*lppIdentityProps = NULL;	// just to be sure...
	}

	if(lpEntryStore)
		MAPIFreeBuffer(lpEntryStore);

	if(lpEID)
		MAPIFreeBuffer(lpEID);

	if(lpszProfileFullName)
		MAPIFreeBuffer(lpszProfileFullName);
	
	if(lpszProfileAddress)
		MAPIFreeBuffer(lpszProfileAddress);

	if(lpUserId)
		MAPIFreeBuffer(lpUserId);

	return hr;
}

/*
	Readreceipt replace function of MAPI for windows
	Mapi for windows support only UNICODE properties but Zarafa don't support this.

	FIXME: Support UNICODE
*/
HRESULT ClientUtil::ReadReceipt(ULONG ulFlags, LPMESSAGE lpReadMessage, LPMESSAGE* lppEmptyMessage)
{
	HRESULT			hr = hrSuccess;
	LPSPropValue	lpSrcPropValue = NULL;
	LPSPropValue	lpDestPropValue = NULL;
	ULONG			ulMaxDestValues = 0;
	ULONG			ulCurDestValues = 0;
	ULONG			cSrcValues = 0;
	ULONG			cbTmp = 0;
	LPBYTE			lpByteTmp = NULL;
	char*			lpMsgClass = NULL;
	char*			lpReportText = NULL;
	char*			lpReadText = NULL;
	FILETIME		ft;	
	LPADRLIST		lpMods = NULL;
	string			strName;
	string			strType;
	string			strAddress;
	string			strBodyText;
	time_t			zero = 0;
	time_t			tt;
	struct tm*		tm;
	char			szTime[255];
	IStream*		lpBodyStream = NULL;

	// The same properties as under windows
	enum ePropReadReceipt{	RR_REPORT_TAG, RR_CONVERSATION_TOPIC, RR_CONVERSATION_INDEX, 
							RR_SEARCH_KEY, RR_MESSAGE_CLASS, RR_SENDER_SEARCH_KEY,
							RR_SUBJECT, RR_SUBJECT_PREFIX, RR_NORMALIZED_SUBJECT, 
							RR_SENDER_NAME, RR_SENDER_ENTRYID, RR_SENDER_ADDRTYPE,												
							RR_SENDER_EMAIL_ADDRESS, RR_REPORT_NAME, RR_REPORT_ENTRYID, 
							RR_READ_RECEIPT_ENTRYID, RR_RECEIVED_BY_NAME, RR_RECEIVED_BY_ENTRYID, 
							RR_PRIORITY, RR_IMPORTANCE, RR_SENT_REPRESENTING_NAME, 
							RR_SENT_REPRESENTING_ENTRYID, RR_SENT_REPRESENTING_SEARCH_KEY, RR_RCVD_REPRESENTING_NAME, RR_RCVD_REPRESENTING_ENTRYID, 
							RR_MESSAGE_DELIVERY_TIME, RR_CLIENT_SUBMIT_TIME, RR_DISPLAY_TO,
							RR_DISPLAY_CC, RR_DISPLAY_BCC, RR_SENSITIVITY, 
							RR_INTERNET_MESSAGE_ID, RR_DELIVER_TIME, RR_SENT_REPRESENTING_ADDRTYPE, RR_SENT_REPRESENTING_EMAIL_ADDRESS,
							RR_MDN_DISPOSITION_TYPE, RR_MDN_DISPOSITION_SENDINGMODE};

	SizedSPropTagArray(37, sPropReadReceipt) = {37, { PR_REPORT_TAG, PR_CONVERSATION_TOPIC, PR_CONVERSATION_INDEX, 
												PR_SEARCH_KEY, PR_MESSAGE_CLASS, PR_SENDER_SEARCH_KEY,
												PR_SUBJECT, PR_SUBJECT_PREFIX, PR_NORMALIZED_SUBJECT, 
												PR_SENDER_NAME, PR_SENDER_ENTRYID, PR_SENDER_ADDRTYPE,												
												PR_SENDER_EMAIL_ADDRESS, PR_REPORT_NAME, PR_REPORT_ENTRYID, 
												PR_READ_RECEIPT_ENTRYID, PR_RECEIVED_BY_NAME, PR_RECEIVED_BY_ENTRYID, 
												PR_PRIORITY, PR_IMPORTANCE, PR_SENT_REPRESENTING_NAME, 
												PR_SENT_REPRESENTING_ENTRYID, PR_SENT_REPRESENTING_SEARCH_KEY, PR_RCVD_REPRESENTING_NAME, PR_RCVD_REPRESENTING_ENTRYID, 
												PR_MESSAGE_DELIVERY_TIME, PR_CLIENT_SUBMIT_TIME, PR_DISPLAY_TO,
												PR_DISPLAY_CC, PR_DISPLAY_BCC, PR_SENSITIVITY, 
												PR_INTERNET_MESSAGE_ID, PR_DELIVER_TIME, PR_SENT_REPRESENTING_ADDRTYPE, PR_SENT_REPRESENTING_EMAIL_ADDRESS,
												PR_MDN_DISPOSITION_TYPE, PR_MDN_DISPOSITION_SENDINGMODE } };


	// Check incomming parameter
	if(lpReadMessage == NULL || lppEmptyMessage == NULL || *lppEmptyMessage == NULL) {
		hr = MAPI_E_INVALID_OBJECT;
		goto exit;
	}

	if((ulFlags &~ MAPI_NON_READ) != 0) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	GetSystemTimeAsFileTime(&ft);

	if((ulFlags&MAPI_NON_READ) == MAPI_NON_READ) {
		lpMsgClass = "REPORT.IPM.Note.IPNNRN";
		lpReadText = _("Not read:");
		lpReportText = _("was not read because it expired before reading at time");
	}else{
		lpMsgClass = "REPORT.IPM.Note.IPNRN";
		lpReadText = _("Read:");
		lpReportText = _("was read on");
	}

	hr = lpReadMessage->GetProps((LPSPropTagArray)&sPropReadReceipt, 0, &cSrcValues, &lpSrcPropValue);
	if(FAILED(hr) != hrSuccess)
		goto exit;

	// important properties
	if(lpSrcPropValue[RR_REPORT_ENTRYID].ulPropTag != PR_REPORT_ENTRYID)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}


	strBodyText = _("Your message");
	strBodyText+= "\r\n\r\n";

	if(lpSrcPropValue[RR_DISPLAY_TO].ulPropTag == PR_DISPLAY_TO) {
		strBodyText+= "\t";
		strBodyText+= _("To:");
		strBodyText+= " ";
		strBodyText+= lpSrcPropValue[RR_DISPLAY_TO].Value.lpszA;
		strBodyText+= "\r\n";
	}
	if(lpSrcPropValue[RR_DISPLAY_CC].ulPropTag == PR_DISPLAY_CC) {
		strBodyText+= "\t";
		strBodyText+= _("Cc:");
		strBodyText+= " ";
		strBodyText+= lpSrcPropValue[RR_DISPLAY_CC].Value.lpszA;
		strBodyText+= "\r\n";
	}
	
	if(lpSrcPropValue[RR_SUBJECT].ulPropTag == PR_SUBJECT) {
		strBodyText+= "\t";
		strBodyText+= _("Subject:");
		strBodyText+= " ";
		strBodyText+= lpSrcPropValue[RR_SUBJECT].Value.lpszA;
		strBodyText+= "\r\n";
	}

	if(lpSrcPropValue[RR_CLIENT_SUBMIT_TIME].ulPropTag == PR_CLIENT_SUBMIT_TIME) {
		strBodyText+= "\t";
		strBodyText+= _("Sent on:");
		strBodyText+= " ";

		FileTimeToUnixTime(lpSrcPropValue[RR_CLIENT_SUBMIT_TIME].Value.ft, &tt);
		tm = localtime(&tt);
		if(tm == NULL) {
			tm = localtime(&zero);
		}
		strftime(szTime, 255, "%c", tm);

		strBodyText+= szTime;
		strBodyText+= "\r\n";
	}


	strBodyText+= "\r\n";
	strBodyText+= lpReportText;
	strBodyText+= " ";
	
	FileTimeToUnixTime(ft, &tt);
	tm = localtime(&tt);
	if(tm == NULL) {
		tm = localtime(&zero);
	}
	strftime(szTime, 255, "%c", tm);

	strBodyText+= szTime;
	strBodyText+= "\r\n";


	ulMaxDestValues = cSrcValues + 4;//+ default properties
	hr = MAPIAllocateBuffer(sizeof(SPropValue)*ulMaxDestValues, (void**)&lpDestPropValue);
	if(hr != hrSuccess)
		goto exit;

	memset(lpDestPropValue, 0, sizeof(SPropValue)*ulMaxDestValues);

	// Default properties
	lpDestPropValue[ulCurDestValues].ulPropTag = PR_DELETE_AFTER_SUBMIT;
	lpDestPropValue[ulCurDestValues++].Value.b = true;

	lpDestPropValue[ulCurDestValues].ulPropTag = PR_READ_RECEIPT_REQUESTED;
	lpDestPropValue[ulCurDestValues++].Value.b = false;

	lpDestPropValue[ulCurDestValues].ulPropTag = PR_MESSAGE_FLAGS;
	lpDestPropValue[ulCurDestValues++].Value.ul = 0;

	lpDestPropValue[ulCurDestValues].ulPropTag = PR_MESSAGE_CLASS;
	lpDestPropValue[ulCurDestValues++].Value.lpszA = lpMsgClass;
	
	lpDestPropValue[ulCurDestValues].ulPropTag = PR_REPORT_TEXT;
	lpDestPropValue[ulCurDestValues++].Value.lpszA = lpReportText;

    lpDestPropValue[ulCurDestValues].ulPropTag = PR_REPORT_TIME;
	lpDestPropValue[ulCurDestValues++].Value.ft = ft;
	
	lpDestPropValue[ulCurDestValues].ulPropTag = PR_SUBJECT_PREFIX;
	lpDestPropValue[ulCurDestValues++].Value.lpszA = lpReadText;

	if(lpSrcPropValue[RR_SUBJECT].ulPropTag == PR_SUBJECT)
	{
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SUBJECT;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_SUBJECT].Value.lpszA;

		lpDestPropValue[ulCurDestValues].ulPropTag = PR_SUBJECT;
		MAPIAllocateMore(strlen(lpReadText) + 1 + strlen(lpSrcPropValue[RR_SUBJECT].Value.lpszA)+1, lpDestPropValue, (void**)&lpDestPropValue[ulCurDestValues].Value.lpszA);
		memcpy(lpDestPropValue[ulCurDestValues].Value.lpszA, lpReadText, strlen(lpReadText) );
		memcpy(lpDestPropValue[ulCurDestValues].Value.lpszA+strlen(lpReadText), " ", 1 );
		memcpy(lpDestPropValue[ulCurDestValues].Value.lpszA+strlen(lpReadText)+1, lpSrcPropValue[RR_SUBJECT].Value.lpszA, strlen(lpSrcPropValue[RR_SUBJECT].Value.lpszA)+1);
		ulCurDestValues++; 
	}else {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_SUBJECT;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpReadText;
	}

	if(lpSrcPropValue[RR_REPORT_TAG].ulPropTag == PR_REPORT_TAG) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_REPORT_TAG;
		lpDestPropValue[ulCurDestValues++].Value.bin = lpSrcPropValue[RR_REPORT_TAG].Value.bin;
	}

	if(lpSrcPropValue[RR_DISPLAY_TO].ulPropTag == PR_DISPLAY_TO) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_DISPLAY_TO;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_DISPLAY_TO].Value.lpszA;
	}

	if(lpSrcPropValue[RR_DISPLAY_CC].ulPropTag == PR_DISPLAY_CC) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_DISPLAY_CC;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_DISPLAY_CC].Value.lpszA;
	}

	if(lpSrcPropValue[RR_DISPLAY_BCC].ulPropTag == PR_DISPLAY_BCC) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_DISPLAY_BCC;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_DISPLAY_BCC].Value.lpszA;
	}

	if(lpSrcPropValue[RR_CLIENT_SUBMIT_TIME].ulPropTag == PR_CLIENT_SUBMIT_TIME) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SUBMIT_TIME;
		lpDestPropValue[ulCurDestValues++].Value.ft = lpSrcPropValue[RR_CLIENT_SUBMIT_TIME].Value.ft;
	}
	
	if(lpSrcPropValue[RR_DELIVER_TIME].ulPropTag == PR_DELIVER_TIME) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_DELIVERY_TIME;
		lpDestPropValue[ulCurDestValues++].Value.ft = lpSrcPropValue[RR_DELIVER_TIME].Value.ft;
	}	

	if(lpSrcPropValue[RR_CONVERSATION_TOPIC].ulPropTag == PR_CONVERSATION_TOPIC) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_CONVERSATION_TOPIC;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_CONVERSATION_TOPIC].Value.lpszA;
	}

	if(lpSrcPropValue[RR_CONVERSATION_INDEX].ulPropTag == PR_CONVERSATION_INDEX &&
		ScCreateConversationIndex(lpSrcPropValue[RR_CONVERSATION_INDEX].Value.bin.cb, lpSrcPropValue[RR_CONVERSATION_INDEX].Value.bin.lpb, &cbTmp, &lpByteTmp) == hrSuccess )
	{
		hr = MAPIAllocateMore(cbTmp, lpDestPropValue, (void**)&lpDestPropValue[ulCurDestValues].Value.bin.lpb);
		if(hr != hrSuccess)
			goto exit;
		
		lpDestPropValue[ulCurDestValues].Value.bin.cb = cbTmp;
		memcpy(lpDestPropValue[ulCurDestValues].Value.bin.lpb, lpByteTmp, cbTmp);

		lpDestPropValue[ulCurDestValues++].ulPropTag = PR_CONVERSATION_INDEX;

		if(lpByteTmp){ MAPIFreeBuffer(lpByteTmp); lpByteTmp = NULL;}
	}

	if(lpSrcPropValue[RR_IMPORTANCE].ulPropTag == PR_IMPORTANCE)
	{
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_IMPORTANCE;
		lpDestPropValue[ulCurDestValues++].Value.ul = lpSrcPropValue[RR_IMPORTANCE].Value.ul;
	}

	if(lpSrcPropValue[RR_PRIORITY].ulPropTag == PR_PRIORITY) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_PRIORITY;
		lpDestPropValue[ulCurDestValues++].Value.ul = lpSrcPropValue[RR_PRIORITY].Value.ul;
	}

	if(lpSrcPropValue[RR_SENDER_NAME].ulPropTag == PR_SENDER_NAME) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SENDER_NAME;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_SENDER_NAME].Value.lpszA;
	}
	
	if(lpSrcPropValue[RR_SENDER_ADDRTYPE].ulPropTag == PR_SENDER_ADDRTYPE) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SENDER_ADDRTYPE;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_SENDER_ADDRTYPE].Value.lpszA;
	}

	if(lpSrcPropValue[RR_SENDER_ENTRYID].ulPropTag == PR_SENDER_ENTRYID) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SENDER_ENTRYID;
		lpDestPropValue[ulCurDestValues++].Value.bin = lpSrcPropValue[RR_SENDER_ENTRYID].Value.bin;
	}

	if(lpSrcPropValue[RR_SENDER_SEARCH_KEY].ulPropTag == PR_SENDER_SEARCH_KEY) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SENDER_SEARCH_KEY;
		lpDestPropValue[ulCurDestValues++].Value.bin = lpSrcPropValue[RR_SENDER_SEARCH_KEY].Value.bin;
	}

	if(lpSrcPropValue[RR_SENDER_EMAIL_ADDRESS].ulPropTag == PR_SENDER_EMAIL_ADDRESS) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SENDER_EMAIL_ADDRESS;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_SENDER_EMAIL_ADDRESS].Value.lpszA;
	}
	
	if(lpSrcPropValue[RR_SENT_REPRESENTING_NAME].ulPropTag == PR_SENT_REPRESENTING_NAME) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SENT_REPRESENTING_NAME;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_SENT_REPRESENTING_NAME].Value.lpszA;
	}
	
	if(lpSrcPropValue[RR_SENT_REPRESENTING_ADDRTYPE].ulPropTag == PR_SENT_REPRESENTING_ADDRTYPE) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SENT_REPRESENTING_ADDRTYPE;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_SENT_REPRESENTING_ADDRTYPE].Value.lpszA;
	}

	if(lpSrcPropValue[RR_SENT_REPRESENTING_ENTRYID].ulPropTag == PR_SENT_REPRESENTING_ENTRYID) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SENT_REPRESENTING_ENTRYID;
		lpDestPropValue[ulCurDestValues++].Value.bin = lpSrcPropValue[RR_SENT_REPRESENTING_ENTRYID].Value.bin;
	}

	if(lpSrcPropValue[RR_SENT_REPRESENTING_SEARCH_KEY].ulPropTag == PR_SENT_REPRESENTING_SEARCH_KEY) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SENT_REPRESENTING_SEARCH_KEY;
		lpDestPropValue[ulCurDestValues++].Value.bin = lpSrcPropValue[RR_SENT_REPRESENTING_SEARCH_KEY].Value.bin;
	}

	if(lpSrcPropValue[RR_SENT_REPRESENTING_EMAIL_ADDRESS].ulPropTag == PR_SENT_REPRESENTING_EMAIL_ADDRESS) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_ORIGINAL_SENT_REPRESENTING_EMAIL_ADDRESS;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_SENT_REPRESENTING_EMAIL_ADDRESS].Value.lpszA;
	}

	if(lpSrcPropValue[RR_MDN_DISPOSITION_SENDINGMODE].ulPropTag == PR_MDN_DISPOSITION_SENDINGMODE) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_MDN_DISPOSITION_SENDINGMODE;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_MDN_DISPOSITION_SENDINGMODE].Value.lpszA;
	}

	if(lpSrcPropValue[RR_MDN_DISPOSITION_TYPE].ulPropTag == PR_MDN_DISPOSITION_TYPE) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_MDN_DISPOSITION_TYPE;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_MDN_DISPOSITION_TYPE].Value.lpszA;
	}

//	PR_RECEIVED_BY_NAME, PR_RECEIVED_BY_ENTRYID
//	PR_RCVD_REPRESENTING_NAME, PR_RCVD_REPRESENTING_ENTRYID	

	if(lpSrcPropValue[RR_INTERNET_MESSAGE_ID].ulPropTag == PR_INTERNET_MESSAGE_ID) {
		lpDestPropValue[ulCurDestValues].ulPropTag = PR_INTERNET_MESSAGE_ID;
		lpDestPropValue[ulCurDestValues++].Value.lpszA = lpSrcPropValue[RR_INTERNET_MESSAGE_ID].Value.lpszA;
	}

	hr = (*lppEmptyMessage)->OpenProperty(PR_BODY, &IID_IStream, 0, MAPI_CREATE | MAPI_MODIFY, (IUnknown**)&lpBodyStream);
	if (hr != hrSuccess)
		goto exit;

	hr = lpBodyStream->Write(strBodyText.c_str(), strBodyText.size(), NULL);
	if (hr != hrSuccess)
		goto exit;

	hr = lpBodyStream->Commit( 0 );//0 = STGC_DEFAULT
	if (hr != hrSuccess)
		goto exit;

	hr = MAPIAllocateBuffer(CbNewADRLIST(1), (void**)&lpMods);
	if (hr != hrSuccess)
		goto exit;

	lpMods->cEntries = 1;

	hr = MAPIAllocateBuffer(sizeof(SPropValue) * 8, (void**)&lpMods->aEntries->rgPropVals);
	if (hr != hrSuccess)
		goto exit;

	hr = ECParseOneOff((LPENTRYID)lpSrcPropValue[RR_REPORT_ENTRYID].Value.bin.lpb, lpSrcPropValue[RR_REPORT_ENTRYID].Value.bin.cb, strName, strType, strAddress);
	if (hr != hrSuccess)
		goto exit;

	lpMods->aEntries->rgPropVals[0].ulPropTag = PR_ENTRYID;
	lpMods->aEntries->rgPropVals[0].Value.bin = lpSrcPropValue[RR_REPORT_ENTRYID].Value.bin;
	lpMods->aEntries->rgPropVals[1].ulPropTag = PR_ADDRTYPE;
	lpMods->aEntries->rgPropVals[1].Value.lpszA = (LPSTR)strType.c_str();
	lpMods->aEntries->rgPropVals[2].ulPropTag = PR_DISPLAY_NAME;
	lpMods->aEntries->rgPropVals[2].Value.lpszA = (LPSTR)strName.c_str();
	lpMods->aEntries->rgPropVals[3].ulPropTag = PR_TRANSMITABLE_DISPLAY_NAME;
	lpMods->aEntries->rgPropVals[3].Value.lpszA = (LPSTR)strName.c_str();
	lpMods->aEntries->rgPropVals[4].ulPropTag = PR_SMTP_ADDRESS;
	lpMods->aEntries->rgPropVals[4].Value.lpszA = (LPSTR)strAddress.c_str();
	lpMods->aEntries->rgPropVals[5].ulPropTag = PR_EMAIL_ADDRESS;
	lpMods->aEntries->rgPropVals[5].Value.lpszA = (LPSTR)strAddress.c_str();
	
	hr = HrCreateEmailSearchKey((LPSTR)strType.c_str(), (LPSTR)strAddress.c_str(), &cbTmp, &lpByteTmp);
	if (hr != hrSuccess)
		goto exit;	

	lpMods->aEntries->rgPropVals[6].ulPropTag = PR_SEARCH_KEY;
	lpMods->aEntries->rgPropVals[6].Value.bin.cb = cbTmp;
	lpMods->aEntries->rgPropVals[6].Value.bin.lpb = lpByteTmp;

	lpMods->aEntries->rgPropVals[7].ulPropTag = PR_RECIPIENT_TYPE;
	lpMods->aEntries->rgPropVals[7].Value.ul = MAPI_TO;
	
	lpMods->aEntries->cValues = 8;

	hr = (*lppEmptyMessage)->ModifyRecipients(MODRECIP_ADD, lpMods);
	if (hr != hrSuccess)
		goto exit;

	hr = (*lppEmptyMessage)->SetProps(ulCurDestValues, lpDestPropValue, NULL);
	if (hr != hrSuccess)
		goto exit;

exit:
	if(lpBodyStream)
		lpBodyStream->Release();

	if(lpDestPropValue)
		MAPIFreeBuffer(lpDestPropValue);
	
	if(lpSrcPropValue)
		MAPIFreeBuffer(lpSrcPropValue);

	if(lpByteTmp)
		MAPIFreeBuffer(lpByteTmp);

	if(lpMods)
		FreePadrlist(lpMods);	

    return hr;
}

HRESULT ClientUtil::GetGlobalProfileProperties(LPMAPISUP lpMAPISup, struct sGlobalProfileProps* lpsProfileProps)
{
	HRESULT			hr = hrSuccess;
	LPPROFSECT		lpGlobalProfSect = NULL;
	LPSPropValue	lpPropName = NULL;
	bool			bIsEMS = false;
	ULONG			cGuids;
	GUID			*lpGuid;
	char			pbProfileGuid[] = "\x92\x07\xf3\xe0\xa3\xb1\x10\x19\x90\x8b\x08\x00\x2b\x2a\x56\xc2";

	hr = lpMAPISup->OpenProfileSection((LPMAPIUID)pbGlobalProfileSectionGuid, MAPI_MODIFY, &lpGlobalProfSect);
	if(hr != hrSuccess)
		goto exit;

	// PR_PROFILE_UNRESOLVED_NAME only available in EMS profiles
	if(HrGetOneProp(lpGlobalProfSect, PR_PROFILE_UNRESOLVED_NAME, &lpPropName) == hrSuccess)
		bIsEMS = true;

	hr = ClientUtil::GetGlobalProfileProperties(lpGlobalProfSect, lpsProfileProps);
	if(hr != hrSuccess)
		goto exit;

exit:
	if(lpGlobalProfSect)
		lpGlobalProfSect->Release();

	if(lpPropName)
		MAPIFreeBuffer(lpPropName);

	return hr;
}

HRESULT ClientUtil::GetGlobalProfileProperties(LPPROFSECT lpGlobalProfSect, struct sGlobalProfileProps* lpsProfileProps)
{
	HRESULT			hr = hrSuccess;
	LPSPropValue	lpsPropArray = NULL;
	ULONG			cValues = 0;
	LPSPropValue	lpsEMSPropArray = NULL;
	LPSPropValue	lpPropEMS = NULL;
	ULONG			cEMSValues = 0;
	LPSPropValue	lpProp = NULL;
	bool			bIsEMS = false;

	if(lpGlobalProfSect == NULL || lpsProfileProps == NULL)
	{
		hr = MAPI_E_INVALID_OBJECT;
		goto exit;
	}

	if(HrGetOneProp(lpGlobalProfSect, PR_PROFILE_UNRESOLVED_NAME, &lpPropEMS) == hrSuccess) {
		bIsEMS = true;
	}

	if(bIsEMS) {
		SizedSPropTagArray(2, sptaEMSProfile) = {2,{PR_PROFILE_UNRESOLVED_SERVER, PR_PROFILE_UNRESOLVED_NAME}};

		// This is an emulated MSEMS store. Get the properties we need and convert them to ZARAFA-style properties
		hr = lpGlobalProfSect->GetProps((LPSPropTagArray)&sptaEMSProfile, 0, &cEMSValues, &lpsEMSPropArray);
		if(FAILED(hr))
			goto exit;

		hr = ConvertMSEMSProps(cEMSValues, lpsEMSPropArray, &cValues, &lpsPropArray);
		if(FAILED(hr))
			goto exit;
	} else {
		// Get the properties we need directly from the global profile section
		hr = lpGlobalProfSect->GetProps((LPSPropTagArray)&sptaZarafaProfile, 0, &cValues, &lpsPropArray);
		if(FAILED(hr))
			goto exit;
	}


	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_PATH))
		lpsProfileProps->strServerPath = lpProp->Value.lpszA;

	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_PROFILE_NAME))
		lpsProfileProps->strProfileName = lpProp->Value.lpszA;

	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_USERNAME))
		lpsProfileProps->strUserName = lpProp->Value.lpszA;

	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_USERPASSWORD))
		lpsProfileProps->strPassword = lpProp->Value.lpszA;
	
	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_FLAGS))
		lpsProfileProps->ulProfileFlags = lpProp->Value.ul;
	else
		lpsProfileProps->ulProfileFlags = 0;

	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_SSLKEY_FILE))
		lpsProfileProps->strSSLKeyFile = lpProp->Value.lpszA;

	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_SSLKEY_PASS))
		lpsProfileProps->strSSLKeyPass = lpProp->Value.lpszA;

	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_PROXY_HOST))
		lpsProfileProps->strProxyHost = lpProp->Value.lpszA;
	
	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_PROXY_PORT))
		lpsProfileProps->ulProxyPort = lpProp->Value.ul;
	else
		lpsProfileProps->ulProxyPort = 0;

	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_PROXY_FLAGS))
		lpsProfileProps->ulProxyFlags = lpProp->Value.ul;
	else
		lpsProfileProps->ulProxyFlags = 0;

	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_PROXY_USERNAME))
		lpsProfileProps->strProxyUserName = lpProp->Value.lpszA;
	
	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_PROXY_PASSWORD))
		lpsProfileProps->strProxyPassword = lpProp->Value.lpszA;

	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_CONNECTION_TIMEOUT))
		lpsProfileProps->ulConnectionTimeOut = lpProp->Value.ul;
	else
		lpsProfileProps->ulConnectionTimeOut = 30;

	if(lpProp = PpropFindProp(lpsPropArray, cValues, PR_EC_OFFLINE_PATH))
		lpsProfileProps->strOfflinePath = lpProp->Value.lpszA;

	lpsProfileProps->bIsEMS = bIsEMS;

	hr = hrSuccess;

exit:
	if(lpPropEMS)
		MAPIFreeBuffer(lpPropEMS);

	if(lpsPropArray)
		MAPIFreeBuffer(lpsPropArray);

	if(lpsEMSPropArray)
		MAPIFreeBuffer(lpsEMSPropArray);

	return hr;
}

HRESULT ClientUtil::GetGlobalProfileDeligateStoresProp(LPPROFSECT lpGlobalProfSect, ULONG* lpcDeligates, LPBYTE* lppDeligateStores)
{
	HRESULT			hr = hrSuccess;
	LPSPropValue	lpsPropValue = NULL;
	ULONG			cValues = 0;
	SPropTagArray	sPropTagArray;
	LPBYTE			lpDeligateStores = NULL;

	if(lpGlobalProfSect == NULL || lpcDeligates == NULL || lppDeligateStores == NULL)
	{
		hr = MAPI_E_INVALID_OBJECT;
		goto exit;
	}
	
	sPropTagArray.cValues = 1;
	sPropTagArray.aulPropTag[0] =  PR_STORE_PROVIDERS;

	hr = lpGlobalProfSect->GetProps(&sPropTagArray, 0, &cValues, &lpsPropValue);
	if(hr != hrSuccess)
		goto exit;

	if(lpsPropValue[0].Value.bin.cb > 0){
		hr = MAPIAllocateBuffer(lpsPropValue[0].Value.bin.cb, (void**)&lpDeligateStores);
		if(hr != hrSuccess)
			goto exit;

		memcpy(lpDeligateStores, lpsPropValue[0].Value.bin.lpb, lpsPropValue[0].Value.bin.cb);
	}

	*lpcDeligates = lpsPropValue[0].Value.bin.cb;
	*lppDeligateStores = lpDeligateStores;

	hr = hrSuccess;

exit:
	if(lpsPropValue)
		MAPIFreeBuffer(lpsPropValue);

	return hr;
}

/*
 * Read registry key to discover the installation directory for the exchange redirector
 *
 * @param[out] lpConfigPath String containing full config path
 * @return HRESULT Result
 */
HRESULT ClientUtil::GetConfigPath(std::string *lpConfigPath)
{
	return MAPI_E_NO_SUPPORT;
}

/**
 * Convert incoming MSEMS profile properties to ZARAFA properties
 *
 * Basically we take the username and servername from the exchange properties and set all other properties
 * by reading a configuration file.
 *
 * @param cValues[in] Number of props in pValues
 * @param pValues[in] Incoming exchange properties (must contain PR_PROFILE_UNRESOLVED_{USER,SERVER})
 * @param lpcValues[out] Number of properties in lppProps
 * @param lppProps[out] New ZARAFA properties
 * @return HRESULT Success result
 */
HRESULT ClientUtil::ConvertMSEMSProps(ULONG cValues, LPSPropValue pValues, ULONG *lpcValues, LPSPropValue *lppProps)
{
	HRESULT hr = hrSuccess;
	LPSPropValue lpProps = NULL;
	char *szUsername;
	std::string strServerPath;
	ULONG cProps = 0;
	LPSPropValue lpServer = NULL;
	LPSPropValue lpUsername = NULL;
	static const configsetting_t settings[] = {
		{ "ssl_port", "237" },
		{ "ssl_key_file", "c:\\program files\\zarafa\\exchange-redirector.pem" },
		{ "ssl_key_pass", "zarafa" },
		{ "server_address", "" },
		{ "log_method","file" },
		{ "log_file","-" },
		{ "log_level","2", CONFIGSETTING_RELOADABLE },
		{ "log_timestamp","1" },
		{ NULL, NULL },
	};
	ECConfig *lpConfig = new ECConfig(settings);
	std::string strConfigPath;

	hr = GetConfigPath(&strConfigPath);
	if(hr != hrSuccess) {
		TRACE_RELEASE("Unable to find config file (registry key missing)", (char *)strConfigPath.c_str());
		goto exit;
	}

	// Remove trailing slash
	if(*(strConfigPath.end()-1) == '\\' )
		strConfigPath.resize(strConfigPath.size()-1);

	strConfigPath += "\\exchange-redirector.cfg";

	TRACE_RELEASE("Using config file '%s'", (char *)strConfigPath.c_str());

	if(!lpConfig->LoadSettings((char *)strConfigPath.c_str())) {
		hr = MAPI_E_NOT_FOUND;
		TRACE_RELEASE("Unable to load config file '%s'", (char *)strConfigPath.c_str());
		goto exit;
	}

	lpUsername = PpropFindProp(pValues, cValues, PR_PROFILE_UNRESOLVED_NAME);
	lpServer = PpropFindProp(pValues, cValues, PR_PROFILE_UNRESOLVED_SERVER);

	if(!lpServer || !lpUsername) {
		TRACE_RELEASE("PR_PROFILE_UNRESOLVED_NAME or PR_PROFILE_UNRESOLVED_SERVER not set", (char *)strConfigPath.c_str());
		hr = MAPI_E_UNCONFIGURED;
		goto exit;
	}

	hr = MAPIAllocateBuffer(sizeof(SPropValue) * 6, (LPVOID *)&lpProps);
	if (hr != hrSuccess)
		goto exit;

	if (strlen(lpConfig->GetSetting("server_address")) > 0)
		strServerPath = (std::string)"https://" + lpConfig->GetSetting("server_address") + ":" + lpConfig->GetSetting("ssl_port") + "/zarafa";
	else
		strServerPath = (std::string)"https://" + lpServer->Value.lpszA + ":" + lpConfig->GetSetting("ssl_port") + "/zarafa";
	szUsername = lpUsername->Value.lpszA;

	if(strrchr(szUsername, '='))
		szUsername = strrchr(szUsername, '=')+1;

	lpProps[cProps].ulPropTag = PR_EC_PATH;
	MAPIAllocateMore(strServerPath.size() + 1, lpProps, (void**)&lpProps[cProps].Value.lpszA);
	strcpy(lpProps[cProps++].Value.lpszA, strServerPath.c_str());

	lpProps[cProps].ulPropTag = PR_EC_USERNAME;
	MAPIAllocateMore(strlen(szUsername) + 1, lpProps, (void**)&lpProps[cProps].Value.lpszA);
	strcpy(lpProps[cProps++].Value.lpszA, szUsername);

	lpProps[cProps].ulPropTag = PR_EC_USERPASSWORD;
	MAPIAllocateMore(1, lpProps, (void**)&lpProps[cProps].Value.lpszA);
	strcpy(lpProps[cProps++].Value.lpszA, "");

	lpProps[cProps].ulPropTag = PR_EC_SSLKEY_FILE;
	MAPIAllocateMore(strlen(lpConfig->GetSetting("ssl_key_file")) + 1, lpProps, (void**)&lpProps[cProps].Value.lpszA);
	strcpy(lpProps[cProps++].Value.lpszA, lpConfig->GetSetting("ssl_key_file"));

	lpProps[cProps].ulPropTag = PR_EC_SSLKEY_PASS;
	MAPIAllocateMore(strlen(lpConfig->GetSetting("ssl_key_pass")) + 1, lpProps, (void**)&lpProps[cProps].Value.lpszA);
	strcpy(lpProps[cProps++].Value.lpszA, lpConfig->GetSetting("ssl_key_pass"));

	lpProps[cProps].ulPropTag = PR_EC_FLAGS; // Since we're emulating exchange, use 22-byte exchange-style sourcekeys
	lpProps[cProps++].Value.ul = EC_PROFILE_FLAGS_TRUNCATE_SOURCEKEY;

	TRACE_RELEASE("Redirecting to %s", (char *)strServerPath.c_str());

	*lpcValues = cProps;
	*lppProps = lpProps;

exit:
	if (hr != hrSuccess && lpProps)
		MAPIFreeBuffer(lpProps);

	if(lpConfig)
		delete lpConfig;

	return hr;
}

/* 
Zarafa entryid functions

*/

HRESULT HrCreateEntryId(GUID guidStore, unsigned int ulObjType, ULONG* lpcbEntryId, LPENTRYID* lppEntryId)
{
	HRESULT		hr = hrSuccess;
	EID			eid;
	ULONG		cbEntryId = 0;
	LPENTRYID	lpEntryId = NULL;

	if(lpcbEntryId == NULL || lppEntryId == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(CoCreateGuid(&eid.uniqueId) != hrSuccess) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	cbEntryId = CbNewEID("");

	hr = ECAllocateBuffer(cbEntryId, (void**)&lpEntryId); 
	if(hr != hrSuccess)
		goto exit;

	eid.guid = guidStore;
	eid.ulType = ulObjType;

	memcpy(lpEntryId, &eid, cbEntryId);

	*lpcbEntryId = cbEntryId;
	*lppEntryId = lpEntryId;

exit:

	return hr;
}

HRESULT HrGetServerPath(LPMAILUSER lpsMailUser, std::string *lpstrServerPath)
{
	HRESULT			hr = hrSuccess;
	ULONG			cbResult = 0;
	LPSPropValue	lpsResult = NULL;

	SizedSPropTagArray(1, sServerPathProp) = {1, {PR_EC_SERVERPATH}};

	if (!lpsMailUser || !lpstrServerPath)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = lpsMailUser->GetProps((LPSPropTagArray)&sServerPathProp, 0, &cbResult, &lpsResult);
	if (hr != hrSuccess)
		goto exit;

	if (cbResult != 1 || lpsResult->Value.lpszA == NULL)
	{
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	*lpstrServerPath = lpsResult->Value.lpszA;

exit:
	if (lpsResult)
		ECFreeBuffer(lpsResult);

	return hr;
}

/**
 * Extract the server URL from a store entryid.
 * @param[in]	cbEntryId		The length of the entryid pointed to by lpEntryId
 * @param[in]	lpEntryId		Pointer to the store entryid.
 * @param[out]	lppServerPath	Pointer to a char pointer that will be set to the server path
 *								extracted from the entry id. Memory for this path will be allocates
 *								with MAPIAllocateBuffer and must be freed by the caller.
 * @param[out]	lpbIsPseudoUrl	Pointer to a boolean that will be set to true if the extracted
 *								server path is a pseudo URL.
 * @retval	MAPI_E_INVALID_PARAMETER	lpEntryId, lppServerName or lpbIsPseudoUrl is NULL
 * @retval	MAPI_E_NOT_FOUND			The extracted server path does not start with http://, https://, file:// or pseudo://
 */
HRESULT HrGetServerURLFromStoreEntryId(ULONG cbEntryId, LPENTRYID lpEntryId, char** lppServerName, bool *lpbIsPseudoUrl)
{
	HRESULT		hr = hrSuccess;
	PEID		peid = (PEID)lpEntryId;
	EID_V0*		peid_V0 = NULL;

	ULONG		ulMaxSize = 0;
	ULONG		ulSize = 0;
	char*		lpServerName = NULL;
	char*		lpTmpServerName = NULL;
	bool		bIsPseudoUrl = false;

	if(lpEntryId == NULL || lppServerName == NULL || lpbIsPseudoUrl == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(peid->ulVersion == 0) 
	{
		peid_V0 = (EID_V0*)lpEntryId;

		ulMaxSize = cbEntryId - offsetof(EID_V0, szServer);
		ulSize = strnlen(peid_V0->szServer, ulMaxSize);
		lpTmpServerName = peid_V0->szServer;
	}else {
		ulMaxSize = cbEntryId - offsetof(EID, szServer);
		ulSize = strnlen(peid->szServer, ulMaxSize);
		lpTmpServerName = peid->szServer;
	}

	if (ulSize >= ulMaxSize) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	if (strnicmp(lpTmpServerName, "pseudo://", 9) == 0)
		bIsPseudoUrl = true;
	else if (strnicmp(lpTmpServerName, "http://", 7) && 
			 strnicmp(lpTmpServerName, "https://", 8) && 
			 strnicmp(lpTmpServerName, "file://", 7))
	{
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	ECAllocateBuffer(strlen(lpTmpServerName) + 1, (void**)&lpServerName);
	strcpy(lpServerName, lpTmpServerName);

	*lppServerName = lpServerName;
	*lpbIsPseudoUrl = bIsPseudoUrl;

exit:
	return hr;
}

HRESULT HrResolvePseudoUrl(WSTransport *lpTransport, char *lpszUrl, std::string *lpstrServerPath, bool *lpbIsPeer)
{
	HRESULT		hr = hrSuccess;
	char		*lpszServerPath = NULL;
	bool		bIsPeer = false;

	if (lpTransport == NULL || lpszUrl == NULL || lpstrServerPath == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if (strncmp(lpszUrl, "pseudo://", 9))
	{
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = lpTransport->HrResolvePseudoUrl(lpszUrl, &lpszServerPath, &bIsPeer);
	if (hr != hrSuccess)
		goto exit;

	*lpstrServerPath = lpszServerPath;
	if (lpbIsPeer)
		*lpbIsPeer = bIsPeer;

exit:
	if (lpszServerPath)
		ECFreeBuffer(lpszServerPath);

	return hr;
}

HRESULT HrCompareEntryIdWithStoreGuid(ULONG cbEntryID, LPENTRYID lpEntryID, LPCGUID guidStore)
{
	HRESULT hr = hrSuccess;

	if (lpEntryID == NULL || guidStore == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if (cbEntryID < 20) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	if (memcmp(lpEntryID->ab, guidStore, sizeof(GUID)) != 0) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

exit:

	return hr;
}

HRESULT HrUnWrapMAPIEntryID(ULONG cbOrigEntry, LPENTRYID lpOrigEntry,  ULONG *lpcbUnWrappedEntry,  LPENTRYID *lppUnWrappedEntry)
{
	HRESULT hr = hrSuccess;
	ULONG cbRemove = 0;
	ULONG cbDLLName = 0;
	LPENTRYID lpEntryId = NULL;

	if (lpOrigEntry == NULL|| lpcbUnWrappedEntry == NULL || lppUnWrappedEntry == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Check if this a wrapped store entryid
	if (cbOrigEntry < (4 + sizeof(GUID) + 3) || memcmp(lpOrigEntry->ab, &MAPI_ENTRYID, sizeof(GUID)) != 0) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	cbRemove = 4; // Flags
	cbRemove+= sizeof(GUID); //Wrapped identifier
	cbRemove+= 2; // Unknown

	// Dllname size
	cbDLLName = (ULONG)strlen((LPCSTR)lpOrigEntry+cbRemove) + 1;
	cbRemove+= cbDLLName;

	cbRemove += (4 - (cbRemove & 0x03)) & 0x03; // padding to 4byte block

	if (cbOrigEntry <= cbRemove) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	// Create Unwrap entryid
	hr = MAPIAllocateBuffer(cbOrigEntry - cbRemove, (void**)&lpEntryId);
	if (hr != hrSuccess)
		goto exit;

	memcpy(lpEntryId, ((LPBYTE)lpOrigEntry)+cbRemove, cbOrigEntry - cbRemove);

	*lpcbUnWrappedEntry = cbOrigEntry - cbRemove;
	*lppUnWrappedEntry = lpEntryId;

exit:
	if (hr!= hrSuccess && lpEntryId)
		MAPIFreeBuffer(lpEntryId);

	return hr;
}

HRESULT GetPublicEntryId(enumPublicEntryID ePublicEntryID, GUID guidStore, void *lpBase, ULONG *lpcbEntryID, LPENTRYID *lppEntryID)
{
	HRESULT hr = hrSuccess;
	ULONG cbEntryID = 0;
	LPENTRYID lpEntryID = NULL;

	GUID guidEmpty = {0};
	EID eid = EID(MAPI_FOLDER, guidStore, guidEmpty);

	switch (ePublicEntryID) {
		case ePE_IPMSubtree:
			eid.uniqueId.Data4[7] = 1;
			break;
		case ePE_Favorites:
			eid.uniqueId.Data4[7] = 2;
			break;
		case ePE_PublicFolders:
			eid.uniqueId.Data4[7] = 3;
			break;
		default:
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
			break;
	}


	if (lpcbEntryID == NULL || lppEntryID == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	cbEntryID = CbEID(&eid);

	if (lpBase)
		hr = MAPIAllocateMore(cbEntryID, lpBase, (void**)&lpEntryID);
	else
		hr = MAPIAllocateBuffer(cbEntryID, (void**)&lpEntryID);
	if (hr != hrSuccess)
		goto exit;

	memcpy(lpEntryID, &eid, cbEntryID);

	*lpcbEntryID = cbEntryID;
	*lppEntryID = lpEntryID;

exit:
	return hr;
}
