/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include <sys/un.h>
#include "WSUtil.h"
#include "ECIConv.h"
#include "guid.h"

#include "Mem.h"

#include "mapiext.h"

// For the static row getprop functions
#include "ECMAPIProp.h"
#include "ECMAPIFolder.h"
#include "ECMessage.h"

#include "ECMailUser.h"
#include "ECDistList.h"
#include "ECABContainer.h"

#include "SOAPUtils.h"
#include "EntryPoint.h"

#ifndef NO_GETTEXT
#include <libintl.h>
#define _(string) dcgettext ("zarafa", string, LC_MESSAGES)
#else
#define _(string) string
#endif


#include "threadutil.h"

using namespace std;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define	pbMUIDECSABGuid	"\xac\x21\xa9\x50\x40\xd3\xee\x48\xb3\x19\xfb\xa7\x53\x30\x44\x25"

HRESULT CopyMAPIPropValToSOAPPropVal(propVal *lpPropValDst, LPSPropValue lpPropValSrc)
{
	HRESULT hr = hrSuccess;
	int i = 0;

	lpPropValDst->ulPropTag = lpPropValSrc->ulPropTag;
	memset(&lpPropValDst->Value, 0, sizeof(propValData));

	switch(PROP_TYPE(lpPropValSrc->ulPropTag)) {
	case PT_I2:
		lpPropValDst->__union = SOAP_UNION_propValData_i;
		lpPropValDst->Value.i = lpPropValSrc->Value.i;
		break;
	case PT_LONG: // or PT_ULONG
		lpPropValDst->__union = SOAP_UNION_propValData_ul;
		lpPropValDst->Value.ul = lpPropValSrc->Value.ul;
		break;
	case PT_R4:
		lpPropValDst->__union = SOAP_UNION_propValData_flt;
		lpPropValDst->Value.flt = lpPropValSrc->Value.flt;
		break;
	case PT_DOUBLE:
		lpPropValDst->__union = SOAP_UNION_propValData_dbl;
		lpPropValDst->Value.dbl = lpPropValSrc->Value.dbl;
		break;
	case PT_CURRENCY:
		lpPropValDst->__union = SOAP_UNION_propValData_hilo;
		lpPropValDst->Value.hilo = new hiloLong;
		lpPropValDst->Value.hilo->hi = lpPropValSrc->Value.cur.Hi;
		lpPropValDst->Value.hilo->lo = lpPropValSrc->Value.cur.Lo;
		break;
	case PT_APPTIME:
		lpPropValDst->__union = SOAP_UNION_propValData_dbl;
		lpPropValDst->Value.dbl = lpPropValSrc->Value.at;
		break;
	case PT_ERROR:
		lpPropValDst->__union = SOAP_UNION_propValData_ul;
		lpPropValDst->Value.ul = lpPropValSrc->Value.err;
		break;
	case PT_BOOLEAN:
		lpPropValDst->__union = SOAP_UNION_propValData_b;
		lpPropValDst->Value.b = lpPropValSrc->Value.b == 0 ? false : true;
		break;
	case PT_OBJECT:
		// can never be transmitted over the wire!
		hr = MAPI_E_INVALID_TYPE;
		break;
	case PT_I8:
		lpPropValDst->__union = SOAP_UNION_propValData_li;
		lpPropValDst->Value.li = lpPropValSrc->Value.li.QuadPart;
		break;
	case PT_STRING8:
		lpPropValDst->__union = SOAP_UNION_propValData_lpszA;
		lpPropValDst->Value.lpszA = new char[lstrlen(lpPropValSrc->Value.lpszA)+1];
		lstrcpy(lpPropValDst->Value.lpszA, lpPropValSrc->Value.lpszA);
		break;
	case PT_UNICODE:
	    // For now, we are converting back to PT_STRING8 because the server does not support PT_UNICODE
	    lpPropValDst->ulPropTag = CHANGE_PROP_TYPE(lpPropValSrc->ulPropTag, PT_STRING8);
	    lpPropValDst->__union = SOAP_UNION_propValData_lpszA;
	    lpPropValDst->Value.lpszA = new char [unicodelen(lpPropValSrc->Value.lpszW)+1];
	    unicodetombs(lpPropValDst->Value.lpszA, lpPropValSrc->Value.lpszW, unicodelen(lpPropValSrc->Value.lpszW)+1);
		break;
	case PT_SYSTIME:
		lpPropValDst->__union = SOAP_UNION_propValData_hilo;
		lpPropValDst->Value.hilo = new hiloLong;
		lpPropValDst->Value.hilo->hi = lpPropValSrc->Value.ft.dwHighDateTime;
		lpPropValDst->Value.hilo->lo = lpPropValSrc->Value.ft.dwLowDateTime;
		break;
	case PT_CLSID:
		lpPropValDst->__union = SOAP_UNION_propValData_bin;
		lpPropValDst->Value.bin = new xsd__base64Binary;
		lpPropValDst->Value.bin->__ptr = new unsigned char[sizeof(GUID)];
		lpPropValDst->Value.bin->__size = sizeof(GUID);
		memcpy((void *)lpPropValDst->Value.bin->__ptr, lpPropValSrc->Value.lpguid, sizeof(GUID));
		break;
	case PT_BINARY:
		lpPropValDst->__union = SOAP_UNION_propValData_bin;
		lpPropValDst->Value.bin = new xsd__base64Binary;
		lpPropValDst->Value.bin->__ptr = new unsigned char[lpPropValSrc->Value.bin.cb];
		lpPropValDst->Value.bin->__size = lpPropValSrc->Value.bin.cb;
		memcpy((void *)lpPropValDst->Value.bin->__ptr, lpPropValSrc->Value.bin.lpb, lpPropValSrc->Value.bin.cb);
		break;
	case PT_MV_I2:
		lpPropValDst->__union = SOAP_UNION_propValData_mvi;
		lpPropValDst->Value.mvi.__size = lpPropValSrc->Value.MVi.cValues;
        lpPropValDst->Value.mvi.__ptr = new short int[lpPropValDst->Value.mvi.__size];
		memcpy(lpPropValDst->Value.mvi.__ptr, lpPropValSrc->Value.MVi.lpi, sizeof(short int) * lpPropValDst->Value.mvi.__size);
		break;
	case PT_MV_LONG:
		lpPropValDst->__union = SOAP_UNION_propValData_mvl;
		lpPropValDst->Value.mvl.__size = lpPropValSrc->Value.MVl.cValues;
        lpPropValDst->Value.mvl.__ptr = new unsigned int[lpPropValDst->Value.mvl.__size];
		memcpy(lpPropValDst->Value.mvl.__ptr, lpPropValSrc->Value.MVl.lpl, sizeof(unsigned int) * lpPropValDst->Value.mvl.__size);
		break;
	case PT_MV_R4:
		lpPropValDst->__union = SOAP_UNION_propValData_mvflt;
		lpPropValDst->Value.mvflt.__size = lpPropValSrc->Value.MVflt.cValues;
        lpPropValDst->Value.mvflt.__ptr = new float[lpPropValDst->Value.mvflt.__size];
		memcpy(lpPropValDst->Value.mvflt.__ptr, lpPropValSrc->Value.MVflt.lpflt, sizeof(float) * lpPropValDst->Value.mvflt.__size);
		break;
	case PT_MV_DOUBLE:
		lpPropValDst->__union = SOAP_UNION_propValData_mvdbl;
		lpPropValDst->Value.mvdbl.__size = lpPropValSrc->Value.MVdbl.cValues;
        lpPropValDst->Value.mvdbl.__ptr = new double[lpPropValDst->Value.mvdbl.__size];
		memcpy(lpPropValDst->Value.mvdbl.__ptr, lpPropValSrc->Value.MVdbl.lpdbl, sizeof(double) * lpPropValDst->Value.mvdbl.__size);
		break;
	case PT_MV_CURRENCY:
		lpPropValDst->__union = SOAP_UNION_propValData_mvhilo;
		lpPropValDst->Value.mvhilo.__size = lpPropValSrc->Value.MVcur.cValues;
        lpPropValDst->Value.mvhilo.__ptr = new hiloLong[lpPropValDst->Value.mvhilo.__size];
		for(i=0; i < lpPropValDst->Value.mvhilo.__size; i++)
		{
			lpPropValDst->Value.mvhilo.__ptr[i].hi = lpPropValSrc->Value.MVcur.lpcur[i].Hi;
			lpPropValDst->Value.mvhilo.__ptr[i].lo = lpPropValSrc->Value.MVcur.lpcur[i].Lo;
		}
		break;
	case PT_MV_APPTIME:
		lpPropValDst->__union = SOAP_UNION_propValData_mvdbl;
		lpPropValDst->Value.mvdbl.__size = lpPropValSrc->Value.MVat.cValues;
        lpPropValDst->Value.mvdbl.__ptr = new double[lpPropValDst->Value.mvdbl.__size];
		memcpy(lpPropValDst->Value.mvdbl.__ptr, lpPropValSrc->Value.MVat.lpat, sizeof(double) * lpPropValDst->Value.mvdbl.__size);
		break;
	case PT_MV_SYSTIME:
		lpPropValDst->__union = SOAP_UNION_propValData_mvhilo;
		lpPropValDst->Value.mvhilo.__size = lpPropValSrc->Value.MVft.cValues;
        lpPropValDst->Value.mvhilo.__ptr = new hiloLong[lpPropValDst->Value.mvhilo.__size];
		for(i=0; i < lpPropValDst->Value.mvhilo.__size; i++)
		{
			lpPropValDst->Value.mvhilo.__ptr[i].hi = lpPropValSrc->Value.MVft.lpft[i].dwHighDateTime;
			lpPropValDst->Value.mvhilo.__ptr[i].lo = lpPropValSrc->Value.MVft.lpft[i].dwLowDateTime;
		}
		break;
	case PT_MV_BINARY:
		lpPropValDst->__union = SOAP_UNION_propValData_mvbin;
		lpPropValDst->Value.mvbin.__size = lpPropValSrc->Value.MVbin.cValues;
		lpPropValDst->Value.mvbin.__ptr = new xsd__base64Binary[lpPropValDst->Value.mvbin.__size];
		for(i=0; i < lpPropValDst->Value.mvbin.__size; i++)
		{
			lpPropValDst->Value.mvbin.__ptr[i].__size = lpPropValSrc->Value.MVbin.lpbin[i].cb;
			lpPropValDst->Value.mvbin.__ptr[i].__ptr = new unsigned char[lpPropValDst->Value.mvbin.__ptr[i].__size];
			memcpy(lpPropValDst->Value.mvbin.__ptr[i].__ptr, lpPropValSrc->Value.MVbin.lpbin[i].lpb, lpPropValDst->Value.mvbin.__ptr[i].__size);
		}
		break;
	case PT_MV_STRING8:
		lpPropValDst->__union = SOAP_UNION_propValData_mvszA;
		lpPropValDst->Value.mvszA.__size = lpPropValSrc->Value.MVszA.cValues;
		lpPropValDst->Value.mvszA.__ptr = new char*[lpPropValDst->Value.mvszA.__size];
		for(i=0; i < lpPropValDst->Value.mvszA.__size; i++)
		{
			lpPropValDst->Value.mvszA.__ptr[i] = new char[lstrlen(lpPropValSrc->Value.MVszA.lppszA[i])+1];
			lstrcpy(lpPropValDst->Value.mvszA.__ptr[i], lpPropValSrc->Value.MVszA.lppszA[i]);
		}
		break;
	case PT_MV_UNICODE:
		// For now, we are converting back to PT_STRING8 because the server does not support PT_UNICODE
		lpPropValDst->ulPropTag = CHANGE_PROP_TYPE(lpPropValSrc->ulPropTag, PT_STRING8);
		lpPropValDst->__union = SOAP_UNION_propValData_mvszA;
		lpPropValDst->Value.mvszA.__size = lpPropValSrc->Value.MVszA.cValues;
		lpPropValDst->Value.mvszA.__ptr = new char*[lpPropValDst->Value.mvszA.__size];
		for(i=0; i < lpPropValDst->Value.mvszA.__size; i++)
		{
			lpPropValDst->Value.mvszA.__ptr[i] = new char[unicodelen(lpPropValSrc->Value.MVszW.lppszW[i])+1];
			unicodetombs(lpPropValDst->Value.mvszA.__ptr[i], lpPropValSrc->Value.MVszW.lppszW[i], unicodelen(lpPropValSrc->Value.MVszW.lppszW[i])+1);
		}
		break;
	case PT_MV_CLSID:
		lpPropValDst->__union = SOAP_UNION_propValData_mvbin;
		lpPropValDst->Value.mvbin.__size = lpPropValSrc->Value.MVguid.cValues;
		lpPropValDst->Value.mvbin.__ptr = new xsd__base64Binary[lpPropValDst->Value.mvbin.__size];
		for(i=0; i < lpPropValDst->Value.mvbin.__size; i++)
		{
			lpPropValDst->Value.mvbin.__ptr[i].__size = sizeof(GUID);
			lpPropValDst->Value.mvbin.__ptr[i].__ptr = new unsigned char[lpPropValDst->Value.mvbin.__ptr[i].__size];
			memcpy(lpPropValDst->Value.mvbin.__ptr[i].__ptr, &lpPropValSrc->Value.MVguid.lpguid[i], lpPropValDst->Value.mvbin.__ptr[i].__size);
		}
		break;
	case PT_MV_I8:
		lpPropValDst->__union = SOAP_UNION_propValData_mvli;
		lpPropValDst->Value.mvli.__size = lpPropValSrc->Value.MVli.cValues;
        lpPropValDst->Value.mvli.__ptr = new LONG64[lpPropValDst->Value.mvli.__size];
		for(i=0; i < lpPropValDst->Value.mvli.__size; i++) {
			lpPropValDst->Value.mvli.__ptr[i] = lpPropValSrc->Value.MVli.lpli[i].QuadPart;
		}
		break;
	case PT_SRESTRICTION:
		lpPropValDst->__union = SOAP_UNION_propValData_res;
		// NOTE: we placed the object pointer in lpszA to make sure it's on the same offset as Value.x on 32bit as 64bit machines
		hr = CopyMAPIRestrictionToSOAPRestriction(&lpPropValDst->Value.res, (LPSRestriction)lpPropValSrc->Value.lpszA);
		break;
	case PT_ACTIONS: {
		// NOTE: we placed the object pointer in lpszA to make sure it's on the same offset as Value.x on 32bit as 64bit machines
		ACTIONS *lpSrcActions = ((ACTIONS*)lpPropValSrc->Value.lpszA);
		lpPropValDst->__union = SOAP_UNION_propValData_actions;
		lpPropValDst->Value.actions = new struct actions;
		lpPropValDst->Value.actions->__ptr = new struct action [ lpSrcActions->cActions ];
		lpPropValDst->Value.actions->__size = lpSrcActions->cActions;

		for(unsigned int i=0; i<lpSrcActions->cActions;i++) {
			ACTION *lpSrcAction = &lpSrcActions->lpAction[i];
			struct action *lpDstAction = &lpPropValDst->Value.actions->__ptr[i];

			lpDstAction->acttype = lpSrcAction->acttype;
			lpDstAction->flavor = lpSrcAction->ulActionFlavor;
			lpDstAction->flags = lpSrcAction->ulFlags;

			switch(lpSrcActions->lpAction[i].acttype) {
				case OP_MOVE:
				case OP_COPY:
					lpDstAction->__union = SOAP_UNION__act_moveCopy;

					lpDstAction->act.moveCopy.store.__ptr = new unsigned char [lpSrcAction->actMoveCopy.cbStoreEntryId];
					memcpy(lpDstAction->act.moveCopy.store.__ptr, lpSrcAction->actMoveCopy.lpStoreEntryId, lpSrcAction->actMoveCopy.cbStoreEntryId);
					lpDstAction->act.moveCopy.store.__size = lpSrcAction->actMoveCopy.cbStoreEntryId;

					lpDstAction->act.moveCopy.folder.__ptr = new unsigned char [lpSrcAction->actMoveCopy.cbFldEntryId];
					memcpy(lpDstAction->act.moveCopy.folder.__ptr, lpSrcAction->actMoveCopy.lpFldEntryId, lpSrcAction->actMoveCopy.cbFldEntryId);
					lpDstAction->act.moveCopy.folder.__size = lpSrcAction->actMoveCopy.cbFldEntryId;

					break;
				case OP_REPLY:
				case OP_OOF_REPLY:
					lpDstAction->__union = SOAP_UNION__act_reply;
					lpDstAction->act.reply.message.__ptr = new unsigned char [lpSrcAction->actReply.cbEntryId];
					memcpy(lpDstAction->act.reply.message.__ptr, lpSrcAction->actReply.lpEntryId, lpSrcAction->actReply.cbEntryId);
					lpDstAction->act.reply.message.__size = lpSrcAction->actReply.cbEntryId;

					lpDstAction->act.reply.guid.__size = sizeof(GUID);
					lpDstAction->act.reply.guid.__ptr = new unsigned char [ sizeof(GUID) ];
					memcpy(lpDstAction->act.reply.guid.__ptr, (void *)&lpSrcAction->actReply.guidReplyTemplate, sizeof(GUID));

					break;
				case OP_DEFER_ACTION:
					lpDstAction->__union = SOAP_UNION__act_defer;
					lpDstAction->act.defer.bin.__ptr = new unsigned char [ lpSrcAction->actDeferAction.cbData ];
					lpDstAction->act.defer.bin.__size = lpSrcAction->actDeferAction.cbData;
					memcpy(lpDstAction->act.defer.bin.__ptr,lpSrcAction->actDeferAction.pbData, lpSrcAction->actDeferAction.cbData);
					break;
				case OP_BOUNCE:
					lpDstAction->__union = SOAP_UNION__act_bouncecode;
					lpDstAction->act.bouncecode = lpSrcAction->scBounceCode;
					break;
				case OP_FORWARD:
				case OP_DELEGATE:
					lpDstAction->__union = SOAP_UNION__act_adrlist;
					hr = CopyMAPIRowSetToSOAPRowSet((LPSRowSet)lpSrcAction->lpadrlist, &lpDstAction->act.adrlist);
					if(hr != hrSuccess)
						goto exit;
					break;
				case OP_TAG:
					lpDstAction->__union = SOAP_UNION__act_prop;
					lpDstAction->act.prop = new propVal;
					hr = CopyMAPIPropValToSOAPPropVal(lpDstAction->act.prop, &lpSrcAction->propTag);
					break;
				case OP_DELETE:
				case OP_MARK_AS_READ:
					// no other data needed
					break;
			}
		}
		break;
	}
	default:
		hr = MAPI_E_INVALID_TYPE;
	}

exit:
	return hr;
}

HRESULT CopySOAPPropValToMAPIPropVal(LPSPropValue lpPropValDst, struct propVal *lpPropValSrc, void *lpBase)
{
	HRESULT hr = hrSuccess;
	unsigned int size = 0;

	lpPropValDst->ulPropTag = lpPropValSrc->ulPropTag;
	lpPropValDst->dwAlignPad = 0;

	// FIXME check pointer is OK before using in (lpPropValSrc->Value.hilo may be NULL!)
	switch(PROP_TYPE(lpPropValSrc->ulPropTag)) {
	case PT_I2:
		lpPropValDst->Value.i = lpPropValSrc->Value.i;
		break;
	case PT_LONG:
		lpPropValDst->Value.ul = lpPropValSrc->Value.ul;
		break;
	case PT_R4:
		lpPropValDst->Value.flt = lpPropValSrc->Value.flt;
		break;
	case PT_DOUBLE:
		lpPropValDst->Value.dbl = lpPropValSrc->Value.dbl;
		break;
	case PT_CURRENCY:
		if(lpPropValSrc->__union && lpPropValSrc->Value.hilo) {
			lpPropValDst->Value.cur.Hi = lpPropValSrc->Value.hilo->hi;
			lpPropValDst->Value.cur.Lo = lpPropValSrc->Value.hilo->lo;
		} else {
			lpPropValDst->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropValSrc->ulPropTag));
			lpPropValDst->Value.err = MAPI_E_NOT_FOUND;
		}
		break;
	case PT_APPTIME:
		lpPropValDst->Value.at = lpPropValSrc->Value.dbl;
		break;
	case PT_ERROR:
		lpPropValDst->Value.err = ZarafaErrorToMAPIError(lpPropValSrc->Value.ul);
		break;
	case PT_BOOLEAN:
		lpPropValDst->Value.b = lpPropValSrc->Value.b;
		break;
	case PT_OBJECT:
		// can never be transmitted over the wire!
		hr = MAPI_E_INVALID_TYPE;
		break;

	case PT_I8:
		lpPropValDst->Value.li.QuadPart = lpPropValSrc->Value.li;
		break;
	case PT_STRING8:
		if(lpPropValSrc->__union && lpPropValSrc->Value.lpszA) {
			ECAllocateMore(lstrlen(lpPropValSrc->Value.lpszA)+1,lpBase, (void **) &lpPropValDst->Value.lpszA);
			lstrcpy(lpPropValDst->Value.lpszA, lpPropValSrc->Value.lpszA);
		} else {
			lpPropValDst->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropValSrc->ulPropTag));
			lpPropValDst->Value.err = MAPI_E_NOT_FOUND;
		}
		break;
	case PT_UNICODE:
		{
			ECIConv iconv("ucs-2le", "utf-8");
			std::string unicode = iconv.convert(lpPropValSrc->Value.lpszA);
			ECAllocateMore(unicode.size() * sizeof(WCHAR) + sizeof(WCHAR), lpBase, (void **) &lpPropValDst->Value.lpszW);
			memcpy(lpPropValDst->Value.lpszW, (WCHAR *)unicode.c_str(), unicode.size());
			lpPropValDst->Value.lpszW[unicode.size()/sizeof(WCHAR)] = 0;
			break;
		}
	case PT_SYSTIME:
		if(lpPropValSrc->__union && lpPropValSrc->Value.hilo) {
			lpPropValDst->Value.ft.dwHighDateTime = lpPropValSrc->Value.hilo->hi;
			lpPropValDst->Value.ft.dwLowDateTime = lpPropValSrc->Value.hilo->lo;
		} else {
			lpPropValDst->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropValSrc->ulPropTag));
			lpPropValDst->Value.err = MAPI_E_NOT_FOUND;
		}
		break;
	case PT_CLSID:
		if(lpPropValSrc->__union && lpPropValSrc->Value.bin && lpPropValSrc->Value.bin->__size == sizeof(MAPIUID)) {
			ECAllocateMore(lpPropValSrc->Value.bin->__size, lpBase, (void **) &lpPropValDst->Value.lpguid);
			memcpy((void *)lpPropValDst->Value.lpguid, lpPropValSrc->Value.bin->__ptr, lpPropValSrc->Value.bin->__size);
		} else {
			lpPropValDst->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropValSrc->ulPropTag));
			lpPropValDst->Value.err = MAPI_E_NOT_FOUND;
		}
		break;
	case PT_BINARY:
		if(lpPropValSrc->__union && lpPropValSrc->Value.bin) {
			ECAllocateMore(lpPropValSrc->Value.bin->__size, lpBase, (void **) &lpPropValDst->Value.bin.lpb);
			memcpy((void *)lpPropValDst->Value.bin.lpb, lpPropValSrc->Value.bin->__ptr, lpPropValSrc->Value.bin->__size);
			lpPropValDst->Value.bin.cb = lpPropValSrc->Value.bin->__size;
		}else if(lpPropValSrc->__union == 0) {
			lpPropValDst->Value.bin.lpb = NULL;
			lpPropValDst->Value.bin.cb = 0;
		}else {
			lpPropValDst->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropValSrc->ulPropTag));
			lpPropValDst->Value.err = MAPI_E_NOT_FOUND;
		}

		break;
	case PT_MV_I2:
		lpPropValDst->Value.MVi.cValues = lpPropValSrc->Value.mvi.__size;
		ECAllocateMore(sizeof(short int)*lpPropValDst->Value.MVi.cValues, lpBase, (void**)&lpPropValDst->Value.MVi.lpi);
		memcpy(lpPropValDst->Value.MVi.lpi, lpPropValSrc->Value.mvi.__ptr, sizeof(short int)*lpPropValDst->Value.MVi.cValues);
		break;
	case PT_MV_LONG:
		lpPropValDst->Value.MVl.cValues = lpPropValSrc->Value.mvl.__size;
		ECAllocateMore(sizeof(unsigned int)*lpPropValDst->Value.MVl.cValues, lpBase, (void**)&lpPropValDst->Value.MVl.lpl);
		memcpy(lpPropValDst->Value.MVl.lpl, lpPropValSrc->Value.mvl.__ptr, sizeof(unsigned int)*lpPropValDst->Value.MVl.cValues);
		break;
	case PT_MV_R4:
		lpPropValDst->Value.MVflt.cValues = lpPropValSrc->Value.mvflt.__size;
		ECAllocateMore(sizeof(float)*lpPropValDst->Value.MVflt.cValues, lpBase, (void**)&lpPropValDst->Value.MVflt.lpflt);
		memcpy(lpPropValDst->Value.MVflt.lpflt, lpPropValSrc->Value.mvflt.__ptr, sizeof(float)*lpPropValDst->Value.MVflt.cValues);
		break;
	case PT_MV_DOUBLE:
		lpPropValDst->Value.MVdbl.cValues = lpPropValSrc->Value.mvdbl.__size;
		ECAllocateMore(sizeof(double)*lpPropValDst->Value.MVdbl.cValues, lpBase, (void**)&lpPropValDst->Value.MVdbl.lpdbl);
		memcpy(lpPropValDst->Value.MVdbl.lpdbl, lpPropValSrc->Value.mvdbl.__ptr, sizeof(double)*lpPropValDst->Value.MVdbl.cValues);
		break;
	case PT_MV_CURRENCY:
		lpPropValDst->Value.MVcur.cValues = lpPropValSrc->Value.mvhilo.__size;
		ECAllocateMore(sizeof(hiloLong)*lpPropValDst->Value.MVcur.cValues, lpBase, (void**)&lpPropValDst->Value.MVcur.lpcur);
		for(unsigned int i=0; i < lpPropValDst->Value.MVcur.cValues; i++)
		{
			lpPropValDst->Value.MVcur.lpcur[i].Hi = lpPropValSrc->Value.mvhilo.__ptr[i].hi;
			lpPropValDst->Value.MVcur.lpcur[i].Lo = lpPropValSrc->Value.mvhilo.__ptr[i].lo;
		}
		break;
	case PT_MV_APPTIME:
		lpPropValDst->Value.MVat.cValues = lpPropValSrc->Value.mvdbl.__size;
		ECAllocateMore(sizeof(double)*lpPropValDst->Value.MVat.cValues, lpBase, (void**)&lpPropValDst->Value.MVat.lpat);
		memcpy(lpPropValDst->Value.MVat.lpat, lpPropValSrc->Value.mvdbl.__ptr, sizeof(double)*lpPropValDst->Value.MVat.cValues);
		break;
	case PT_MV_SYSTIME:
		lpPropValDst->Value.MVft.cValues = lpPropValSrc->Value.mvhilo.__size;
		ECAllocateMore(sizeof(hiloLong)*lpPropValDst->Value.MVft.cValues, lpBase, (void**)&lpPropValDst->Value.MVft.lpft);
		for(unsigned int i=0; i < lpPropValDst->Value.MVft.cValues; i++)
		{
			lpPropValDst->Value.MVft.lpft[i].dwHighDateTime = lpPropValSrc->Value.mvhilo.__ptr[i].hi;
			lpPropValDst->Value.MVft.lpft[i].dwLowDateTime = lpPropValSrc->Value.mvhilo.__ptr[i].lo;
		}
		break;
	case PT_MV_BINARY:
		lpPropValDst->Value.MVbin.cValues = lpPropValSrc->Value.mvbin.__size;
		ECAllocateMore(sizeof(SBinary)*lpPropValDst->Value.MVbin.cValues, lpBase, (void**)&lpPropValDst->Value.MVbin.lpbin);
		for(unsigned int i=0; i< lpPropValDst->Value.MVbin.cValues; i++)
		{
			lpPropValDst->Value.MVbin.lpbin[i].cb =  lpPropValSrc->Value.mvbin.__ptr[i].__size;
			if(lpPropValDst->Value.MVbin.lpbin[i].cb > 0)
			{
				ECAllocateMore(sizeof(unsigned char)*lpPropValDst->Value.MVbin.lpbin[i].cb, lpBase, (void**)&lpPropValDst->Value.MVbin.lpbin[i].lpb);
				memcpy(lpPropValDst->Value.MVbin.lpbin[i].lpb, lpPropValSrc->Value.mvbin.__ptr[i].__ptr, sizeof(unsigned char)*lpPropValDst->Value.MVbin.lpbin[i].cb);
			}
			else
				lpPropValDst->Value.MVbin.lpbin[i].lpb = NULL;
		}
		break;
	case PT_MV_STRING8:
		lpPropValDst->Value.MVszA.cValues = lpPropValSrc->Value.mvszA.__size;
		ECAllocateMore(sizeof(LPSTR)*lpPropValDst->Value.MVszA.cValues, lpBase, (void**)&lpPropValDst->Value.MVszA.lppszA);
		for(unsigned int i=0; i < lpPropValDst->Value.MVszA.cValues; i++)
		{
			size = (lpPropValSrc->Value.mvszA.__ptr[i] != NULL)? (lstrlen(lpPropValSrc->Value.mvszA.__ptr[i])+1) : 1;

			ECAllocateMore(size, lpBase, (void**)&lpPropValDst->Value.MVszA.lppszA[i]);
			memcpy(lpPropValDst->Value.MVszA.lppszA[i], (lpPropValSrc->Value.mvszA.__ptr[i]!=NULL)?lpPropValSrc->Value.mvszA.__ptr[i] : "", size);
		}
		break;
#if 0
	case PT_MV_UNICODE:// FIXME
		lpPropValDst->Value.MVszW.cValues = 1;
		ECAllocateMore(sizeof(LPWSTR)*lpPropValDst->Value.MVszW.cValues, lpBase, (void**)&lpPropValDst->Value.MVszW.lppszW);
		for(int i=0; i< lpPropValDst->Value.MVszW.cValues; i++)
		{
			//linux bug ?
			ECAllocateMore((unicodelen((WCHAR*)L"")+sizeof(WCHAR)), lpBase, (void**)&lpPropValDst->Value.MVszW.lppszW[i]);
			memcpy(lpPropValDst->Value.MVszW.lppszW[i], L"", (lstrlen(L"")+sizeof(WCHAR)));
		}
		break;// FIXME
#endif
	case PT_MV_CLSID:
		lpPropValDst->Value.MVguid.cValues = lpPropValSrc->Value.mvbin.__size;
		ECAllocateMore(sizeof(GUID)*lpPropValDst->Value.MVguid.cValues, lpBase, (void**)&lpPropValDst->Value.MVguid.lpguid);
		for(unsigned int i=0; i< lpPropValDst->Value.MVguid.cValues; i++)
		{
			memcpy(&lpPropValDst->Value.MVguid.lpguid[i], lpPropValSrc->Value.mvbin.__ptr[i].__ptr, sizeof(GUID));
		}
		break;
	case PT_MV_I8:
		lpPropValDst->Value.MVli.cValues = lpPropValSrc->Value.mvli.__size;
		ECAllocateMore(sizeof(LARGE_INTEGER)*lpPropValDst->Value.MVli.cValues, lpBase, (void**)&lpPropValDst->Value.MVli.lpli);
		for(unsigned int i=0; i< lpPropValDst->Value.MVli.cValues; i++)
		{
			lpPropValDst->Value.MVli.lpli[i].QuadPart = lpPropValSrc->Value.mvli.__ptr[i];
		}
		break;
	case PT_SRESTRICTION:
		// NOTE: we place the object pointer in lpszA to make sure it's on the same offset as Value.x on 32bit as 64bit machines
		ECAllocateMore(sizeof(SRestriction), lpBase, (void **) &lpPropValDst->Value.lpszA);
		hr = CopySOAPRestrictionToMAPIRestriction((LPSRestriction)lpPropValDst->Value.lpszA, lpPropValSrc->Value.res, lpBase);
		break;
	case PT_ACTIONS: {
		// NOTE: we place the object pointer in lpszA to make sure it's on the same offset as Value.x on 32bit as 64bit machines
		ACTIONS *lpDstActions;
		ECAllocateMore(sizeof(ACTIONS), lpBase, (void **)&lpPropValDst->Value.lpszA);
		lpDstActions = (ACTIONS *)lpPropValDst->Value.lpszA;

		lpDstActions->cActions = lpPropValSrc->Value.actions->__size;
		ECAllocateMore(sizeof(ACTION) * lpPropValSrc->Value.actions->__size, lpBase, (void **)&lpDstActions->lpAction);

		lpDstActions->ulVersion = EDK_RULES_VERSION;

		for(int i=0;i<lpPropValSrc->Value.actions->__size;i++) {
			ACTION *lpDstAction = &lpDstActions->lpAction[i];
			struct action *lpSrcAction = &lpPropValSrc->Value.actions->__ptr[i];

			lpDstAction->acttype = (ACTTYPE)lpSrcAction->acttype;
			lpDstAction->ulActionFlavor = lpSrcAction->flavor;
			lpDstAction->ulFlags = lpSrcAction->flags;
			lpDstAction->lpRes = NULL;
			lpDstAction->lpPropTagArray = NULL;

			switch(lpSrcAction->acttype) {
				case OP_MOVE:
				case OP_COPY:
					lpDstAction->actMoveCopy.cbStoreEntryId = lpSrcAction->act.moveCopy.store.__size;
					ECAllocateMore(lpSrcAction->act.moveCopy.store.__size, lpBase, (void **)&lpDstAction->actMoveCopy.lpStoreEntryId);
					memcpy(lpDstAction->actMoveCopy.lpStoreEntryId, lpSrcAction->act.moveCopy.store.__ptr, lpSrcAction->act.moveCopy.store.__size);

					lpDstAction->actMoveCopy.cbFldEntryId = lpSrcAction->act.moveCopy.folder.__size;
					ECAllocateMore(lpSrcAction->act.moveCopy.folder.__size, lpBase, (void **)&lpDstAction->actMoveCopy.lpFldEntryId);
					memcpy(lpDstAction->actMoveCopy.lpFldEntryId, lpSrcAction->act.moveCopy.folder.__ptr, lpSrcAction->act.moveCopy.folder.__size);
					break;
				case OP_REPLY:
				case OP_OOF_REPLY:
					lpDstAction->actReply.cbEntryId = lpSrcAction->act.reply.message.__size;
					ECAllocateMore(lpSrcAction->act.reply.message.__size, lpBase, (void **)&lpDstAction->actReply.lpEntryId);
					memcpy(lpDstAction->actReply.lpEntryId, lpSrcAction->act.reply.message.__ptr, lpSrcAction->act.reply.message.__size);

					if(lpSrcAction->act.reply.guid.__size != sizeof(GUID)) {
						hr = MAPI_E_CORRUPT_DATA;
						goto exit;
					}
					memcpy((void *)&lpDstAction->actReply.guidReplyTemplate, lpSrcAction->act.reply.guid.__ptr, lpSrcAction->act.reply.guid.__size);
					break;
				case OP_DEFER_ACTION:
					ECAllocateMore(lpSrcAction->act.defer.bin.__size, lpBase, (void **)&lpDstAction->actDeferAction.pbData);
					lpDstAction->actDeferAction.cbData = lpSrcAction->act.defer.bin.__size;
					memcpy(lpDstAction->actDeferAction.pbData, lpSrcAction->act.defer.bin.__ptr,lpSrcAction->act.defer.bin.__size);
					break;
				case OP_BOUNCE:
					lpDstAction->scBounceCode = lpSrcAction->act.bouncecode;
					break;
				case OP_FORWARD:
				case OP_DELEGATE:
					if(lpSrcAction->act.adrlist == NULL) {
						hr = MAPI_E_CORRUPT_DATA;
						goto exit;
					}

					ECAllocateMore(CbNewSRowSet(lpSrcAction->act.adrlist->__size), lpBase, (void**)&lpDstAction->lpadrlist);

					lpDstAction->lpadrlist->cEntries = lpSrcAction->act.adrlist->__size;

					for(int j=0; j < lpSrcAction->act.adrlist->__size; j++) {
						lpDstAction->lpadrlist->aEntries[j].ulReserved1 = 0;
						lpDstAction->lpadrlist->aEntries[j].cValues = lpSrcAction->act.adrlist->__ptr[j].__size;

						// new rowset allocate more on old rowset, so we can just call FreeProws once
						ECAllocateMore(sizeof(SPropValue) * lpSrcAction->act.adrlist->__ptr[j].__size, lpBase, (void **)&lpDstAction->lpadrlist->aEntries[j].rgPropVals);

						hr = CopySOAPRowToMAPIRow(&lpSrcAction->act.adrlist->__ptr[j], lpDstAction->lpadrlist->aEntries[j].rgPropVals, lpBase);
						if(hr != hrSuccess)
							goto exit;
					}
					// FIXME rowset is not coupled to action -> leaks!

					break;
				case OP_TAG:
					hr = CopySOAPPropValToMAPIPropVal(&lpDstAction->propTag, lpSrcAction->act.prop, lpBase);
					break;
			}
		}
		break;
	}
	default:
		lpPropValDst->ulPropTag = PROP_TAG(PT_ERROR, PROP_ID(lpPropValSrc->ulPropTag));
		lpPropValDst->Value.err = MAPI_E_NOT_FOUND;
		break;
	}

exit:
	return hr;
}

HRESULT CopySOAPRowToMAPIRow(void* lpProvider, struct propValArray *lpsRowSrc, LPSPropValue lpsRowDst, void **lpBase, ULONG ulType)
{
	HRESULT hr = hrSuccess;
	int j=0;

	for(j=0;j<lpsRowSrc->__size;j++) {
		// First, try the default TableRowGetProp from ECMAPIProp
		if((ulType == MAPI_STORE || ulType == MAPI_FOLDER || ulType == MAPI_MESSAGE || ulType == MAPI_ATTACH) &&
			ECMAPIProp::TableRowGetProp(lpProvider, &lpsRowSrc->__ptr[j], &lpsRowDst[j], lpBase, ulType) == erSuccess)
			continue;
		else if((ulType == MAPI_MAILUSER || ulType == MAPI_ABCONT || ulType == MAPI_DISTLIST)&&
			ECABProp::TableRowGetProp(lpProvider, &lpsRowSrc->__ptr[j], &lpsRowDst[j], lpBase, ulType) == erSuccess)
			continue;

		switch(ulType) {
		case MAPI_FOLDER:
			// Then, try the specialized TableRowGetProp for the type of table we're handling
			if(ECMAPIFolder::TableRowGetProp(lpProvider, &lpsRowSrc->__ptr[j], &lpsRowDst[j], lpBase, ulType) == erSuccess)
				continue;
			break;
		case MAPI_MESSAGE:
			if(ECMessage::TableRowGetProp(lpProvider, &lpsRowSrc->__ptr[j], &lpsRowDst[j], lpBase, ulType) == erSuccess)
				continue;
			break;
		case MAPI_MAILUSER:
			if(ECMailUser::TableRowGetProp(lpProvider, &lpsRowSrc->__ptr[j], &lpsRowDst[j], lpBase, ulType) == erSuccess)
				continue;
			break;
		case MAPI_DISTLIST:
			if(ECDistList::TableRowGetProp(lpProvider, &lpsRowSrc->__ptr[j], &lpsRowDst[j], lpBase, ulType) == erSuccess)
				continue;
			break;
		case MAPI_ABCONT:
			if(ECABContainer::TableRowGetProp(lpProvider, &lpsRowSrc->__ptr[j], &lpsRowDst[j], lpBase, ulType) == erSuccess)
				continue;
			break;
		}

		if (ECGenericProp::TableRowGetProp(lpProvider, &lpsRowSrc->__ptr[j], &lpsRowDst[j], lpBase, ulType) == erSuccess)
			continue;

		// If all fails, get the actual data from the server
		CopySOAPPropValToMAPIPropVal(&lpsRowDst[j], &lpsRowSrc->__ptr[j], lpBase);
	}

	return hr;
}

HRESULT CopySOAPEntryId(entryId *lpSrc, entryId* lpDst)
{
	HRESULT hr = hrSuccess;

	if (!lpSrc || !lpDst) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	lpDst->__size = lpSrc->__size;
	lpDst->__ptr = new unsigned char[lpDst->__size];
	memcpy(lpDst->__ptr, lpSrc->__ptr, lpDst->__size);

exit:
	return hr;
}

HRESULT CopyMAPIEntryIdToSOAPEntryId(ULONG cbEntryIdSrc, LPENTRYID lpEntryIdSrc, entryId** lppDest)
{
	HRESULT hr = hrSuccess;

	entryId* lpDest = new entryId;

	hr = CopyMAPIEntryIdToSOAPEntryId(cbEntryIdSrc, lpEntryIdSrc, lpDest, false);
	if(hr != hrSuccess)
		goto exit;

	*lppDest = lpDest;
exit:
	if(hr != hrSuccess)
		delete lpDest;

	return hr;
}

HRESULT CopyMAPIEntryIdToSOAPEntryId(ULONG cbEntryIdSrc, LPENTRYID lpEntryIdSrc, entryId* lpDest, bool bCheapCopy)
{
	HRESULT hr = hrSuccess;

	if((cbEntryIdSrc > 0 && lpEntryIdSrc == NULL) || lpDest == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(cbEntryIdSrc == 0)	{
		lpDest->__ptr = NULL;
		lpDest->__size = 0;
		goto exit;
	}

	if(bCheapCopy == false) {
		lpDest->__ptr = new unsigned char[cbEntryIdSrc];
		memcpy(lpDest->__ptr, lpEntryIdSrc, cbEntryIdSrc);
	}else{
		lpDest->__ptr = (LPBYTE)lpEntryIdSrc;
	}

	lpDest->__size = cbEntryIdSrc;

exit:
	return hr;
}

HRESULT CopySOAPEntryIdToMAPIEntryId(entryId* lpSrc, ULONG* lpcbDest, LPENTRYID* lppEntryIdDest, void* lpBase)
{
	HRESULT		hr = hrSuccess;
	LPENTRYID	lpEntryId = NULL;

	if(lpSrc == NULL || lpcbDest == NULL || lppEntryIdDest == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(lpSrc->__size == 0) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	if(lpBase)
		hr = ECAllocateMore(lpSrc->__size, lpBase, (void**)&lpEntryId);
	else
		hr = ECAllocateBuffer(lpSrc->__size, (void**)&lpEntryId);
	if(hr != hrSuccess)
		goto exit;

	memcpy(lpEntryId, lpSrc->__ptr, lpSrc->__size);

	*lppEntryIdDest = lpEntryId;
	*lpcbDest = lpSrc->__size;

exit:
	return hr;
}

HRESULT CopySOAPEntryIdToMAPIEntryId(entryId* lpSrc, ULONG ulObjId, ULONG ulType, ULONG* lpcbDest, LPENTRYID* lppEntryIdDest, void *lpBase)
{
	HRESULT		hr = hrSuccess;
	ULONG		cbEntryId = 0;
	LPENTRYID	lpEntryId = NULL;

	if(lpSrc == NULL || lpcbDest == NULL || lppEntryIdDest == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(lpSrc->__size < CbNewABEID("") || lpSrc->__ptr == NULL)
	{
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	if (lpBase != NULL)
		hr = MAPIAllocateMore(lpSrc->__size, lpBase, (void**)&lpEntryId);
	else
		hr = MAPIAllocateBuffer(lpSrc->__size, (void**)&lpEntryId);

	if (hr != hrSuccess)
		goto exit;

	memcpy(lpEntryId, lpSrc->__ptr, lpSrc->__size);
	cbEntryId = lpSrc->__size;

	*lppEntryIdDest = lpEntryId;
	*lpcbDest = cbEntryId;

exit:
	return hr;
}

HRESULT CopySOAPEntryIdToMAPIEntryId(entryId* lpSrc, ULONG ulObjId, ULONG* lpcbDest, LPENTRYID* lppEntryIdDest, void *lpBase)
{
	return CopySOAPEntryIdToMAPIEntryId(lpSrc, ulObjId, MAPI_MAILUSER, lpcbDest, lppEntryIdDest, lpBase);
}

HRESULT CopyMAPIEntryListToSOAPEntryList(ENTRYLIST *lpMsgList, struct entryList* lpsEntryList)
{
	HRESULT hr = hrSuccess;
	unsigned int i = 0;

	if(lpMsgList == NULL || lpsEntryList == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(lpMsgList->cValues == 0 || lpMsgList->lpbin == NULL)	{
		lpsEntryList->__ptr = NULL;
		lpsEntryList->__size = 0;
		goto exit;
	}

	lpsEntryList->__ptr = new entryId[lpMsgList->cValues];

	for(i=0; i < lpMsgList->cValues; i++) {
		lpsEntryList->__ptr[i].__ptr = new unsigned char[lpMsgList->lpbin[i].cb];

		memcpy(lpsEntryList->__ptr[i].__ptr, lpMsgList->lpbin[i].lpb, lpMsgList->lpbin[i].cb);

		lpsEntryList->__ptr[i].__size = lpMsgList->lpbin[i].cb;
	}

	lpsEntryList->__size = i;

exit:
	return hr;
}

HRESULT CopySOAPEntryListToMAPIEntryList(struct entryList* lpsEntryList, LPENTRYLIST* lppMsgList)
{
	HRESULT			hr = hrSuccess;
	unsigned int	i = 0;
	ENTRYLIST*		lpMsgList = NULL;

	if(lpsEntryList == NULL || lppMsgList == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = ECAllocateBuffer(sizeof(ENTRYLIST), (void**)&lpMsgList);
	if(hr != hrSuccess)
		goto exit;

	if(lpsEntryList->__size == 0) {
		lpMsgList->cValues = 0;
		lpMsgList->lpbin = NULL;
	} else {

		hr = ECAllocateMore(lpsEntryList->__size * sizeof(SBinary), lpMsgList, (void**)&lpMsgList->lpbin);
		if(hr != hrSuccess)
			goto exit;
	}

	for(i=0; i<lpsEntryList->__size; i++) {

		hr = ECAllocateMore(lpsEntryList->__ptr[i].__size, lpMsgList, (void**)&lpMsgList->lpbin[i].lpb);
		if(hr != hrSuccess)
			goto exit;

		memcpy(lpMsgList->lpbin[i].lpb, lpsEntryList->__ptr[i].__ptr, lpsEntryList->__ptr[i].__size);

		lpMsgList->lpbin[i].cb = lpsEntryList->__ptr[i].__size;

	}

	lpMsgList->cValues = i;

	*lppMsgList = lpMsgList;

exit:
	if(hr != hrSuccess && lpMsgList)
		ECFreeBuffer(lpMsgList);

	return hr;
}

HRESULT CopySOAPRowToMAPIRow(struct propValArray *lpsRowSrc, LPSPropValue lpsRowDst, void *lpBase)
{
	HRESULT hr = hrSuccess;
	int j=0;

	for(j=0;j<lpsRowSrc->__size;j++) {

		// If all fails, get the actual data from the server
		hr = CopySOAPPropValToMAPIPropVal(&lpsRowDst[j], &lpsRowSrc->__ptr[j], lpBase);
		if(hr != hrSuccess)
			goto exit;
	}

exit:
	return hr;
}

HRESULT CopyMAPIRowToSOAPRow(LPSRow lpRowSrc, struct propValArray *lpsRowDst)
{
	HRESULT hr = hrSuccess;

	struct propVal* lpPropVal = NULL;

	lpPropVal = new struct propVal[lpRowSrc->cValues];
	memset(lpPropVal, 0, sizeof(struct propVal) *lpRowSrc->cValues);

	for(unsigned int i=0; i<lpRowSrc->cValues; i++) {
		hr = CopyMAPIPropValToSOAPPropVal(&lpPropVal[i], &lpRowSrc->lpProps[i]);
		if(hr != hrSuccess)
			goto exit;
	}

	lpsRowDst->__ptr = lpPropVal;
	lpsRowDst->__size = lpRowSrc->cValues;

exit:
	//FIXME: remove memory on an error

	return hr;
}

HRESULT CopyMAPIRowSetToSOAPRowSet(LPSRowSet lpRowSetSrc, struct rowSet **lppsRowSetDst)
{
	HRESULT hr = hrSuccess;
	struct rowSet *lpsRowSetDst = NULL;

	lpsRowSetDst = new struct rowSet;

	lpsRowSetDst->__ptr = new propValArray[lpRowSetSrc->cRows];
	lpsRowSetDst->__size = lpRowSetSrc->cRows;

	for(unsigned int i=0; i < lpRowSetSrc->cRows; i++)
	{
		hr = CopyMAPIRowToSOAPRow(&lpRowSetSrc->aRow[i], &lpsRowSetDst->__ptr[i]);
		if(hr != hrSuccess)
			goto exit;
	}

	*lppsRowSetDst = lpsRowSetDst;

exit:
	//FIXME: remove memory on an error

	return hr;
}

// Copies a row set, filling in client-side generated values on the fly
HRESULT CopySOAPRowSetToMAPIRowSet(void* lpProvider, struct rowSet *lpsRowSetSrc, LPSRowSet *lppRowSetDst, ULONG ulType)
{
	HRESULT hr = hrSuccess;
	ULONG ulRows = 0;
	LPSRowSet lpRowSet = NULL;
	ULONG i=0;

	ulRows = lpsRowSetSrc->__size;

	// Allocate space for the rowset
	ECAllocateBuffer(CbNewSRowSet(ulRows), (void **)&lpRowSet);
	lpRowSet->cRows = ulRows;

	// Loop through all the rows and values, fill in any client-side generated values, or translate
	// some serverside values through TableRowGetProps

	for(i=0;i<lpRowSet->cRows;i++) {
		lpRowSet->aRow[i].ulAdrEntryPad = 0;
		lpRowSet->aRow[i].cValues = lpsRowSetSrc->__ptr[i].__size;
		ECAllocateBuffer(sizeof(SPropValue) * lpsRowSetSrc->__ptr[i].__size, (void **)&lpRowSet->aRow[i].lpProps);
		CopySOAPRowToMAPIRow(lpProvider, &lpsRowSetSrc->__ptr[i], lpRowSet->aRow[i].lpProps, (void **)lpRowSet->aRow[i].lpProps, ulType);
	}

	*lppRowSetDst = lpRowSet;

	return hr;
}

HRESULT CopySOAPRestrictionToMAPIRestriction(LPSRestriction lpDst, struct restrictTable *lpSrc, void *lpBase)
{
	HRESULT hr = hrSuccess;
	unsigned int i=0;

	if(lpSrc == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	memset(lpDst, 0, sizeof(SRestriction));
	lpDst->rt = lpSrc->ulType;

	switch(lpSrc->ulType) {
	case RES_OR:
		if(lpSrc->lpOr == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpDst->res.resOr.cRes = lpSrc->lpOr->__size;
		ECAllocateMore(sizeof(SRestriction) * lpSrc->lpOr->__size, lpBase, (void **) &lpDst->res.resOr.lpRes);

		for(i=0;i<lpSrc->lpOr->__size;i++) {
			hr = CopySOAPRestrictionToMAPIRestriction(&lpDst->res.resOr.lpRes[i], lpSrc->lpOr->__ptr[i], lpBase);

			if(hr != hrSuccess)
				break;
		}
		break;

	case RES_AND:
		if(lpSrc->lpAnd == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpDst->res.resAnd.cRes = lpSrc->lpAnd->__size;
		ECAllocateMore(sizeof(SRestriction) * lpSrc->lpAnd->__size, lpBase, (void **) &lpDst->res.resAnd.lpRes);

		for(unsigned i=0;i<lpSrc->lpAnd->__size;i++) {
			hr = CopySOAPRestrictionToMAPIRestriction(&lpDst->res.resAnd.lpRes[i], lpSrc->lpAnd->__ptr[i], lpBase);

			if(hr != hrSuccess)
				break;
		}
		break;

	case RES_BITMASK:
		if(lpSrc->lpBitmask == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpDst->res.resBitMask.relBMR = lpSrc->lpBitmask->ulType;
		lpDst->res.resBitMask.ulMask = lpSrc->lpBitmask->ulMask;
		lpDst->res.resBitMask.ulPropTag = lpSrc->lpBitmask->ulPropTag;
		break;

	case RES_COMMENT:
		if (lpSrc->lpComment == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		hr = ECAllocateMore(sizeof(SRestriction), lpBase, (void **) &lpDst->res.resComment.lpRes);
		if (hr != hrSuccess)
			goto exit;
		hr = CopySOAPRestrictionToMAPIRestriction(lpDst->res.resComment.lpRes, lpSrc->lpComment->lpResTable, lpBase);
		if (hr != hrSuccess)
			goto exit;

		lpDst->res.resComment.cValues = lpSrc->lpComment->sProps.__size;
		hr = ECAllocateMore(sizeof(SPropValue) * lpSrc->lpComment->sProps.__size, lpBase, (void **)&lpDst->res.resComment.lpProp);
		if (hr != hrSuccess)
			goto exit;
		for (int i=0; i<lpSrc->lpComment->sProps.__size; i++) {
			hr = CopySOAPPropValToMAPIPropVal(&lpDst->res.resComment.lpProp[i], &lpSrc->lpComment->sProps.__ptr[i], lpBase);
			if (hr != hrSuccess)
				break;
		}
		break;

	case RES_COMPAREPROPS:
		if(lpSrc->lpCompare == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpDst->res.resCompareProps.relop = lpSrc->lpCompare->ulType;
		lpDst->res.resCompareProps.ulPropTag1 = lpSrc->lpCompare->ulPropTag1;
		lpDst->res.resCompareProps.ulPropTag2 = lpSrc->lpCompare->ulPropTag2;
		break;

	case RES_CONTENT:
		if(lpSrc->lpContent == NULL || (lpSrc->lpContent->lpProp == NULL && lpSrc->lpContent->szSearchString == NULL))  {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpDst->res.resContent.ulFuzzyLevel = lpSrc->lpContent->ulFuzzyLevel;
		lpDst->res.resContent.ulPropTag = lpSrc->lpContent->ulPropTag;

		hr = ECAllocateMore(sizeof(SPropValue), lpBase, (void **) &lpDst->res.resContent.lpProp);
		if(hr != hrSuccess)
			goto exit;

		if(lpSrc->lpContent->lpProp != NULL)
		{
			hr = CopySOAPPropValToMAPIPropVal(lpDst->res.resContent.lpProp, lpSrc->lpContent->lpProp, lpBase);
			if(hr != hrSuccess)
				goto exit;
		}
		else // NOTE: This is a workaround for version <=4.1 backward compatibility
		{
			hr = ECAllocateMore(strlen(lpSrc->lpContent->szSearchString)+1, lpBase, (void **) &lpDst->res.resContent.lpProp->Value.lpszA);
			if(hr != hrSuccess)
				goto exit;

			strcpy(lpDst->res.resContent.lpProp->Value.lpszA, lpSrc->lpContent->szSearchString);
			lpDst->res.resContent.lpProp->ulPropTag = lpSrc->lpContent->ulPropTag&~MVI_FLAG;
		}
		break;

	case RES_EXIST:
		if(lpSrc->lpExist == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpDst->res.resExist.ulPropTag = lpSrc->lpExist->ulPropTag;
		break;

	case RES_NOT:
		if(lpSrc->lpNot == NULL || lpSrc->lpNot->lpNot == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		ECAllocateMore(sizeof(SRestriction), lpBase, (void **)&lpDst->res.resNot.lpRes);

		hr = CopySOAPRestrictionToMAPIRestriction(lpDst->res.resNot.lpRes, lpSrc->lpNot->lpNot, lpBase);

		break;

	case RES_PROPERTY:
		if(lpSrc->lpProp == NULL || lpSrc->lpProp->lpProp == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		ECAllocateMore(sizeof(SPropValue), lpBase, (void **)&lpDst->res.resProperty.lpProp);

		lpDst->res.resProperty.relop = lpSrc->lpProp->ulType;
		lpDst->res.resProperty.ulPropTag = lpSrc->lpProp->ulPropTag;

		hr = CopySOAPPropValToMAPIPropVal(lpDst->res.resProperty.lpProp, lpSrc->lpProp->lpProp, lpBase);

		break;

	case RES_SIZE:
		if(lpSrc->lpSize == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpDst->res.resSize.cb = lpSrc->lpSize->cb;
		lpDst->res.resSize.relop = lpSrc->lpSize->ulType;
		lpDst->res.resSize.ulPropTag = lpSrc->lpSize->ulPropTag;
		break;

	case RES_SUBRESTRICTION:
		if(lpSrc->lpSub == NULL || lpSrc->lpSub->lpSubObject == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpDst->res.resSub.ulSubObject = lpSrc->lpSub->ulSubObject;
		ECAllocateMore(sizeof(SRestriction), lpBase, (void **)&lpDst->res.resSub.lpRes);

		hr = CopySOAPRestrictionToMAPIRestriction(lpDst->res.resSub.lpRes, lpSrc->lpSub->lpSubObject, lpBase);
		break;

	default:
		hr = MAPI_E_INVALID_PARAMETER;
		break;
	}

exit:
	return hr;

}

HRESULT CopyMAPIRestrictionToSOAPRestriction(struct restrictTable **lppDst, LPSRestriction lpSrc)
{
	HRESULT hr = hrSuccess;
	struct restrictTable *lpDst = NULL;
	unsigned int i=0;

	lpDst = new struct restrictTable;
	memset(lpDst, 0, sizeof(restrictTable));
	lpDst->ulType = lpSrc->rt;

	switch(lpSrc->rt) {
	case RES_OR:
		lpDst->lpOr = new restrictOr;
		memset(lpDst->lpOr,0,sizeof(restrictOr));

		lpDst->lpOr->__ptr = new restrictTable *[lpSrc->res.resOr.cRes];
		memset(lpDst->lpOr->__ptr, 0, sizeof(restrictTable*) * lpSrc->res.resOr.cRes);
		lpDst->lpOr->__size = lpSrc->res.resOr.cRes;

		for(i=0;i<lpSrc->res.resOr.cRes;i++) {
			hr = CopyMAPIRestrictionToSOAPRestriction(&(lpDst->lpOr->__ptr[i]), &lpSrc->res.resOr.lpRes[i]);

			if(hr != hrSuccess)
				goto exit;
		}
		break;

	case RES_AND:
		lpDst->lpAnd = new restrictAnd;
		memset(lpDst->lpAnd,0,sizeof(restrictAnd));

		lpDst->lpAnd->__ptr = new restrictTable *[lpSrc->res.resAnd.cRes];
		memset(lpDst->lpAnd->__ptr, 0, sizeof(restrictTable*) * lpSrc->res.resAnd.cRes);
		lpDst->lpAnd->__size = lpSrc->res.resAnd.cRes;

		for(i=0;i<lpSrc->res.resAnd.cRes;i++) {
			hr = CopyMAPIRestrictionToSOAPRestriction(&lpDst->lpAnd->__ptr[i], &lpSrc->res.resAnd.lpRes[i]);

			if(hr != hrSuccess)
				goto exit;
		}
		break;

	case RES_BITMASK:
		lpDst->lpBitmask = new restrictBitmask;
		memset(lpDst->lpBitmask, 0, sizeof(restrictBitmask));

		lpDst->lpBitmask->ulMask = lpSrc->res.resBitMask.ulMask;
		lpDst->lpBitmask->ulPropTag = lpSrc->res.resBitMask.ulPropTag;
		lpDst->lpBitmask->ulType = lpSrc->res.resBitMask.relBMR;
		break;

	case RES_COMMENT:
		lpDst->lpComment = new restrictComment;
		memset(lpDst->lpComment, 0, sizeof(restrictComment));

		lpDst->lpComment->sProps.__ptr = new propVal[lpSrc->res.resComment.cValues];
		lpDst->lpComment->sProps.__size = lpSrc->res.resComment.cValues;
		for (unsigned int i=0; i < lpSrc->res.resComment.cValues; i++) {
			hr = CopyMAPIPropValToSOAPPropVal(&lpDst->lpComment->sProps.__ptr[i], &lpSrc->res.resComment.lpProp[i]);
			if(hr != hrSuccess)
				goto exit;
		}

		hr = CopyMAPIRestrictionToSOAPRestriction(&lpDst->lpComment->lpResTable, lpSrc->res.resComment.lpRes);
		if (hr != hrSuccess)
			goto exit;
		break;

	case RES_COMPAREPROPS:
		lpDst->lpCompare = new restrictCompare;
		memset(lpDst->lpCompare, 0, sizeof(restrictCompare));

		lpDst->lpCompare->ulPropTag1 = lpSrc->res.resCompareProps.ulPropTag1;
		lpDst->lpCompare->ulPropTag2 = lpSrc->res.resCompareProps.ulPropTag2;
		lpDst->lpCompare->ulType = lpSrc->res.resCompareProps.relop;
		break;

	case RES_CONTENT:
		lpDst->lpContent = new restrictContent;
		memset(lpDst->lpContent, 0, sizeof(restrictContent));

		if( (PROP_TYPE(lpSrc->res.resContent.lpProp->ulPropTag) != PT_BINARY &&
			PROP_TYPE(lpSrc->res.resContent.lpProp->ulPropTag) != PT_MV_BINARY &&
			PROP_TYPE(lpSrc->res.resContent.lpProp->ulPropTag) != PT_STRING8 &&
			PROP_TYPE(lpSrc->res.resContent.lpProp->ulPropTag) != PT_MV_STRING8) ||
			(PROP_TYPE(lpSrc->res.resContent.lpProp->ulPropTag) == PT_BINARY && lpSrc->res.resContent.lpProp->Value.bin.cb >0 && lpSrc->res.resContent.lpProp->Value.bin.lpb == NULL) ||
			(PROP_TYPE(lpSrc->res.resContent.lpProp->ulPropTag) == PT_STRING8 && lpSrc->res.resContent.lpProp->Value.lpszA == NULL) ) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		lpDst->lpContent->ulFuzzyLevel = lpSrc->res.resContent.ulFuzzyLevel;
		lpDst->lpContent->ulPropTag = lpSrc->res.resContent.ulPropTag;

		// NOTE: This is a workaround for version 4.1 backward compatibility
		if(PROP_TYPE(lpSrc->res.resContent.lpProp->ulPropTag) == PT_STRING8) {
			lpDst->lpContent->szSearchString = new char [strlen(lpSrc->res.resContent.lpProp->Value.lpszA)+1];
			strcpy(lpDst->lpContent->szSearchString, lpSrc->res.resContent.lpProp->Value.lpszA);
		}
		lpDst->lpContent->lpProp = new struct propVal;
		memset(lpDst->lpContent->lpProp, 0, sizeof(propVal));

		hr = CopyMAPIPropValToSOAPPropVal(lpDst->lpContent->lpProp, lpSrc->res.resContent.lpProp);
		if(hr != hrSuccess)
			goto exit;
		break;

	case RES_EXIST:
		lpDst->lpExist = new restrictExist;
		memset(lpDst->lpExist, 0, sizeof(restrictExist));

		lpDst->lpExist->ulPropTag = lpSrc->res.resExist.ulPropTag;
		break;

	case RES_NOT:
		lpDst->lpNot = new restrictNot;
		memset(lpDst->lpNot, 0, sizeof(restrictNot));
		hr = CopyMAPIRestrictionToSOAPRestriction(&lpDst->lpNot->lpNot, lpSrc->res.resNot.lpRes);
		if(hr != hrSuccess)
			goto exit;
		break;

	case RES_PROPERTY:
		lpDst->lpProp = new restrictProp;
		memset(lpDst->lpProp, 0, sizeof(restrictProp));

		lpDst->lpProp->ulType = lpSrc->res.resProperty.relop;
		lpDst->lpProp->lpProp = new struct propVal;
		memset(lpDst->lpProp->lpProp, 0, sizeof(propVal));
		lpDst->lpProp->ulPropTag = lpSrc->res.resProperty.ulPropTag;

		// This is needed since the lpProp is also converted to PT_STRING8
		if(PROP_TYPE(lpDst->lpProp->ulPropTag) == PT_UNICODE)
			lpDst->lpProp->ulPropTag = CHANGE_PROP_TYPE(lpDst->lpProp->ulPropTag, PT_STRING8);

		hr = CopyMAPIPropValToSOAPPropVal(lpDst->lpProp->lpProp, lpSrc->res.resProperty.lpProp);
		if(hr != hrSuccess)
			goto exit;
		break;

	case RES_SIZE:
		lpDst->lpSize = new restrictSize;
		memset(lpDst->lpSize, 0, sizeof(restrictSize));

		lpDst->lpSize->cb = lpSrc->res.resSize.cb;
		lpDst->lpSize->ulPropTag = lpSrc->res.resSize.ulPropTag;
		lpDst->lpSize->ulType = lpSrc->res.resSize.relop;
		break;

	case RES_SUBRESTRICTION:
		lpDst->lpSub = new restrictSub;
		memset(lpDst->lpSub, 0, sizeof(restrictSub));

		lpDst->lpSub->ulSubObject = lpSrc->res.resSub.ulSubObject;
		hr = CopyMAPIRestrictionToSOAPRestriction(&lpDst->lpSub->lpSubObject, lpSrc->res.resSub.lpRes);
		if(hr != hrSuccess)
			goto exit;
		break;

	default:
		hr = MAPI_E_INVALID_PARAMETER;
		if(hr != hrSuccess)
			goto exit;
		break;
	}

	*lppDst = lpDst;

exit:
	if(hr != hrSuccess && lpDst != NULL)
		FreeRestrictTable(lpDst);

	return hr;
}

HRESULT CopySOAPPropTagArrayToMAPIPropTagArray(struct propTagArray* lpsPropTagArray, LPSPropTagArray* lppPropTagArray, void* lpBase)
{
	HRESULT			hr = hrSuccess;
	LPSPropTagArray	lpPropTagArray = NULL;

	if(lpsPropTagArray == NULL || lppPropTagArray == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if(lpBase)
		hr = ECAllocateMore(CbNewSPropTagArray(lpsPropTagArray->__size), lpBase, (void**)&lpPropTagArray);
	else
		hr = ECAllocateBuffer(CbNewSPropTagArray(lpsPropTagArray->__size), (void**)&lpPropTagArray);
	if(hr != hrSuccess)
		goto exit;

	lpPropTagArray->cValues = lpsPropTagArray->__size;

	if(lpsPropTagArray->__size > 0)
		memcpy(lpPropTagArray->aulPropTag, lpsPropTagArray->__ptr, sizeof(unsigned int)*lpsPropTagArray->__size);

	*lppPropTagArray = lpPropTagArray;

exit:
	return hr;
}

HRESULT CopyABPropsFromSoap(struct propmapPairArray *lpsoapPropmap, struct propmapMVPairArray *lpsoapMVPropmap,
							SPROPMAP *lpPropmap, MVPROPMAP *lpMVPropmap, void *lpBase)
{
	HRESULT hr = hrSuccess;
	unsigned int nLen = 0;

	if (lpsoapPropmap != NULL) {
		lpPropmap->cEntries = lpsoapPropmap->__size;
		nLen = sizeof(*lpPropmap->lpEntries) * lpPropmap->cEntries;
		ECAllocateMore(nLen, lpBase, (void**)&lpPropmap->lpEntries);

		for (unsigned int i = 0; i < lpsoapPropmap->__size; i++) {
			lpPropmap->lpEntries[i].ulPropId = lpsoapPropmap->__ptr[i].ulPropId;

			nLen = strlen(lpsoapPropmap->__ptr[i].lpszValue)+1;
			ECAllocateMore(nLen, lpBase, (void**)&lpPropmap->lpEntries[i].lpszValue);
			strncpy(lpPropmap->lpEntries[i].lpszValue, lpsoapPropmap->__ptr[i].lpszValue, nLen);
		}
	}

	if (lpsoapMVPropmap != NULL) {
		lpMVPropmap->cEntries = lpsoapMVPropmap->__size;
		ECAllocateMore(sizeof(*lpMVPropmap->lpEntries) * lpMVPropmap->cEntries, lpBase, (void**)&lpMVPropmap->lpEntries);

		for (unsigned int i = 0; i < lpsoapMVPropmap->__size; i++) {
			lpMVPropmap->lpEntries[i].ulPropId = lpsoapMVPropmap->__ptr[i].ulPropId;

			lpMVPropmap->lpEntries[i].cValues = lpsoapMVPropmap->__ptr[i].sValues.__size;
			nLen = sizeof(*lpMVPropmap->lpEntries[i].lpszValues) * lpMVPropmap->lpEntries[i].cValues;
			ECAllocateMore(nLen, lpBase, (void**)&lpMVPropmap->lpEntries[i].lpszValues);

			for (int j = 0; j < lpsoapMVPropmap->__ptr[i].sValues.__size; j++) {
				nLen = strlen(lpsoapMVPropmap->__ptr[i].sValues.__ptr[j])+1;
				ECAllocateMore(nLen, lpBase, (void**)&lpMVPropmap->lpEntries[i].lpszValues[j]);
				strncpy(lpMVPropmap->lpEntries[i].lpszValues[j], lpsoapMVPropmap->__ptr[i].sValues.__ptr[j], nLen);
			}
		}
	}

	return hr;
}

HRESULT CopyABPropsToSoap(SPROPMAP *lpPropmap, MVPROPMAP *lpMVPropmap,
						  struct propmapPairArray **lppsoapPropmap, struct propmapMVPairArray **lppsoapMVPropmap)
{
	HRESULT hr = hrSuccess;
	struct propmapPairArray *soapPropmap = NULL;
	struct propmapMVPairArray *soapMVPropmap = NULL;


	if (lpPropmap->cEntries) {
		soapPropmap = new struct propmapPairArray;
		soapPropmap->__size = lpPropmap->cEntries;
		soapPropmap->__ptr = new struct propmapPair[soapPropmap->__size];
		for (unsigned int i = 0; i < soapPropmap->__size; i++) {
			soapPropmap->__ptr[i].ulPropId = lpPropmap->lpEntries[i].ulPropId;
			soapPropmap->__ptr[i].lpszValue = lpPropmap->lpEntries[i].lpszValue;
		}
	}

	if (lpMVPropmap->cEntries) {
		soapMVPropmap = new struct propmapMVPairArray;
		soapMVPropmap->__size = lpMVPropmap->cEntries;
		soapMVPropmap->__ptr = new struct propmapMVPair[soapMVPropmap->__size];

		for (unsigned int i = 0; i < soapMVPropmap->__size; i++) {
			soapMVPropmap->__ptr[i].ulPropId = lpMVPropmap->lpEntries[i].ulPropId;
			soapMVPropmap->__ptr[i].sValues.__size = lpMVPropmap->lpEntries[i].cValues;
			soapMVPropmap->__ptr[i].sValues.__ptr = new char*[soapMVPropmap->__ptr[i].sValues.__size];
			for (int j = 0; j < soapMVPropmap->__ptr[i].sValues.__size; j++)
				soapMVPropmap->__ptr[i].sValues.__ptr[j] = lpMVPropmap->lpEntries[i].lpszValues[j];
		}
	}

	if (lppsoapPropmap)
		*lppsoapPropmap = soapPropmap;
	if (lppsoapMVPropmap)
		*lppsoapMVPropmap = soapMVPropmap;

	return hr;
}

HRESULT FreeABProps(struct propmapPairArray *lpsoapPropmap, struct propmapMVPairArray *lpsoapMVPropmap)
{
	if (lpsoapPropmap) {
		delete [] lpsoapPropmap->__ptr;
		delete lpsoapPropmap;
	}

	if (lpsoapMVPropmap) {
		for (unsigned int i = 0; i < lpsoapMVPropmap->__size; i++)
			delete [] lpsoapMVPropmap->__ptr[i].sValues.__ptr;
		delete [] lpsoapMVPropmap->__ptr;
		delete lpsoapMVPropmap;
	}

	return hrSuccess;
}

HRESULT SoapUserToUser(struct user *lpUser, LPECUSER lpsUser, void *lpBase)
{
	HRESULT hr		= hrSuccess;
	int		nLen	= 0;

	if (lpUser == NULL || lpsUser == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if (lpBase == NULL)
		lpBase = lpsUser;

	memset(lpsUser, 0, sizeof(*lpsUser));

	nLen = strlen(lpUser->lpszUsername)+1;
	ECAllocateMore(nLen, lpBase, (void**)&lpsUser->lpszUsername);
	strncpy(lpsUser->lpszUsername, lpUser->lpszUsername, nLen);

	if (lpUser->lpszFullName != NULL) {
		nLen = strlen(lpUser->lpszFullName)+1;
		ECAllocateMore(nLen, lpBase, (void**)&lpsUser->lpszFullName);
		strncpy(lpsUser->lpszFullName, lpUser->lpszFullName, nLen);
	}

	if (lpUser->lpszMailAddress != NULL) {
		nLen = strlen(lpUser->lpszMailAddress)+1;
		ECAllocateMore(nLen, lpBase, (void**)&lpsUser->lpszFullEmail);
		strncpy(lpsUser->lpszFullEmail, lpUser->lpszMailAddress, nLen);
	}

	if (lpUser->lpszServername != NULL) {
		nLen = strlen(lpUser->lpszServername)+1;
		ECAllocateMore(nLen, lpBase, (void**)&lpsUser->lpszServername);
		strncpy(lpsUser->lpszServername, lpUser->lpszServername, nLen);
	}

	hr = CopyABPropsFromSoap(lpUser->lpsPropmap, lpUser->lpsMVPropmap,
							 &lpsUser->sPropmap, &lpsUser->sMVPropmap, lpBase);
	if (hr != hrSuccess)
		goto exit;

	hr = CopySOAPEntryIdToMAPIEntryId(&lpUser->sUserId, lpUser->ulUserId, (ULONG*)&lpsUser->sUserId.cb, (LPENTRYID*)&lpsUser->sUserId.lpb, lpBase);
	if (hr != hrSuccess)
		goto exit;

	lpsUser->ulIsAdmin		= lpUser->ulIsAdmin;
	lpsUser->ulIsABHidden	= lpUser->ulIsABHidden;
	lpsUser->ulCapacity		= lpUser->ulCapacity;

	/**
	 * If we're talking to a pre 6.40 server we won't get a object class,
	 * only an is-non-active flag. Luckily we don't have to support that.
	 * However, a 6.40.0 server will put the object class information in
	 * that is-non-active field. We (6.40.1 and up) expect the object class
	 * information in a dedicated object class field, and reverted the
	 * is-non-active field to its original usage.
	 *
	 * We can easily determine what's the case here:
	 *  If ulClass is missing (value == 0), we're dealing with a pre 6.40.1
	 *  server. In that case the ulIsNonActive either contains is-non-active
	 *  information or an object class. We can distinguish this since an
	 *  object class has data in the high 16-bit of its value, the
	 *  is-non-active field is either 0 or 1.
	 *  If we detect a class, we put the class in the ulClass field.
	 *  If we detect a is-non-active, we'll simply return an error since we're
	 *  not required to be able to communicate with a pre 6.40.
	 *  We could guess things here, but why bother?
	 */
	if (lpUser->ulObjClass == 0) {
		if (OBJECTCLASS_TYPE(lpUser->ulIsNonActive) != 0)
			lpsUser->ulObjClass = (objectclass_t)lpUser->ulIsNonActive;	// ulIsNonActive itself will be ignored by the offline server.

		else {
			hr = MAPI_E_UNABLE_TO_COMPLETE;
			goto exit;
		}
	} else
		lpsUser->ulObjClass = (objectclass_t)lpUser->ulObjClass;

exit:
	return hr;
}

HRESULT SoapUserArrayToUserArray(struct userArray* lpUserArray, ULONG *lpcUsers, LPECUSER* lppsUsers)
{
	HRESULT hr = hrSuccess;
	LPECUSER lpECUsers = NULL;
	unsigned int i = 0;

	if(lpUserArray == NULL || lpcUsers == NULL || lppsUsers == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	ECAllocateBuffer(sizeof(ECUSER) * lpUserArray->__size, (void**)&lpECUsers);
	memset(lpECUsers, 0, sizeof(ECUSER) * lpUserArray->__size);

	for(i=0; i < lpUserArray->__size; i++) {
		hr = SoapUserToUser(lpUserArray->__ptr + i, lpECUsers + i, lpECUsers);
		if (hr != hrSuccess)
			goto exit;
	}

	*lppsUsers = lpECUsers;
	*lpcUsers = lpUserArray->__size;

exit:
	return hr;
}

HRESULT SoapUserToUser(struct user *lpUser, LPECUSER *lppsUser)
{
	HRESULT		hr		= hrSuccess;
	LPECUSER	lpsUser	= NULL;

	if (lpUser == NULL || lppsUser == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = ECAllocateBuffer(sizeof *lpsUser, (void**)&lpsUser);
	if (hr != hrSuccess)
		goto exit;

	hr = SoapUserToUser(lpUser, lpsUser, NULL);
	if (hr != hrSuccess)
		goto exit;

	*lppsUser = lpsUser;
	lpsUser = NULL;

exit:
	if (lpsUser != NULL)
		ECFreeBuffer(lpsUser);

	return hr;
}

HRESULT SoapGroupToGroup(struct group *lpGroup, LPECGROUP lpsGroup, void *lpBase)
{
	HRESULT hr		= hrSuccess;

	if (lpGroup == NULL || lpsGroup == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if (lpBase == NULL)
		lpBase = lpsGroup;

	memset(lpsGroup, 0, sizeof(*lpsGroup));

	if (lpGroup->lpszGroupname == NULL)
	{
		hr = MAPI_E_INVALID_OBJECT;
		goto exit;
	}

	hr = MAPIAllocateMore(strlen(lpGroup->lpszGroupname)+1, lpBase, (void **)&lpsGroup->lpszGroupname);
	if(hr != hrSuccess)
		goto exit;

	strcpy(lpsGroup->lpszGroupname, lpGroup->lpszGroupname);

	if (lpGroup->lpszFullname) {
		hr = MAPIAllocateMore(strlen(lpGroup->lpszFullname)+1, lpBase, (void **)&lpsGroup->lpszFullname);
		if(hr != hrSuccess)
			goto exit;

		strcpy(lpsGroup->lpszFullname, lpGroup->lpszFullname);
	}

	if (lpGroup->lpszFullEmail) {
		hr = MAPIAllocateMore(strlen(lpGroup->lpszFullEmail)+1, lpBase, (void **)&lpsGroup->lpszFullEmail);
		if (hr != hrSuccess)
			goto exit;

		strcpy(lpsGroup->lpszFullEmail, lpGroup->lpszFullEmail);
	}

	hr = CopyABPropsFromSoap(lpGroup->lpsPropmap, lpGroup->lpsMVPropmap,
							 &lpsGroup->sPropmap, &lpsGroup->sMVPropmap, lpBase);
	if (hr != hrSuccess)
		goto exit;


	hr = CopySOAPEntryIdToMAPIEntryId(&lpGroup->sGroupId, lpGroup->ulGroupId, (ULONG*)&lpsGroup->sGroupId.cb, (LPENTRYID*)&lpsGroup->sGroupId.lpb);
	if (hr != hrSuccess)
		goto exit;

	lpsGroup->ulIsABHidden	= lpGroup->ulIsABHidden;

exit:
	return hr;
}

HRESULT SoapGroupArrayToGroupArray(struct groupArray* lpGroupArray, ULONG *lpcGroups, LPECGROUP *lppsGroups)
{
	HRESULT			hr = hrSuccess;
	unsigned int	i;
	LPECGROUP		lpECGroups = NULL;

	if(lpGroupArray == NULL || lpcGroups == NULL || lppsGroups == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	ECAllocateBuffer(sizeof(ECGROUP) * lpGroupArray->__size, (void**)&lpECGroups);
	memset(lpECGroups, 0, sizeof(ECGROUP) * lpGroupArray->__size);

	for(i=0; i < lpGroupArray->__size; i++) {
		hr = SoapGroupToGroup(lpGroupArray->__ptr + i, lpECGroups + i, lpECGroups);
		if (hr != hrSuccess)
			goto exit;
	}

	*lppsGroups = lpECGroups;
	*lpcGroups = lpGroupArray->__size;
	lpECGroups = NULL;

exit:
	if (lpECGroups != NULL)
		ECFreeBuffer(lpECGroups);

	return hr;
}

HRESULT SoapGroupToGroup(struct group *lpGroup, LPECGROUP *lppsGroup)
{
	HRESULT		hr			= hrSuccess;
	LPECGROUP	lpsGroup	= NULL;

	if (lpGroup == NULL || lppsGroup == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = ECAllocateBuffer(sizeof *lpsGroup, (void**)&lpsGroup);
	if (hr != hrSuccess)
		goto exit;

	hr = SoapGroupToGroup(lpGroup, lpsGroup, NULL);
	if (hr != hrSuccess)
		goto exit;

	*lppsGroup = lpsGroup;
	lpsGroup = NULL;

exit:
	if (lpsGroup != NULL)
		ECFreeBuffer(lpsGroup);

	return hr;
}

HRESULT SoapCompanyToCompany(struct company *lpCompany, LPECCOMPANY lpsCompany, void *lpBase)
{
	HRESULT hr		= hrSuccess;
	int		nLen	= 0;

	if (lpCompany == NULL || lpsCompany == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if (lpBase == NULL)
		lpBase = lpsCompany;

	memset(lpsCompany, 0, sizeof(*lpsCompany));

	nLen = strlen(lpCompany->lpszCompanyname) + 1;
	ECAllocateMore(nLen, lpBase, (void**)&lpsCompany->lpszCompanyname);
	strncpy(lpsCompany->lpszCompanyname, lpCompany->lpszCompanyname, nLen);

	if (lpCompany->lpszServername != NULL) {
		nLen = strlen(lpCompany->lpszServername) + 1;
		ECAllocateMore(nLen, lpBase, (void**)&lpsCompany->lpszServername);
		strncpy(lpsCompany->lpszServername, lpCompany->lpszServername, nLen);
	}

	hr = CopyABPropsFromSoap(lpCompany->lpsPropmap, lpCompany->lpsMVPropmap,
							 &lpsCompany->sPropmap, &lpsCompany->sMVPropmap, lpBase);
	if (hr != hrSuccess)
		goto exit;

	hr = CopySOAPEntryIdToMAPIEntryId(&lpCompany->sAdministrator, lpCompany->ulAdministrator, (ULONG*)&lpsCompany->sAdministrator.cb, (LPENTRYID*)&lpsCompany->sAdministrator.lpb, lpBase);
	if (hr != hrSuccess)
		goto exit;

	hr = CopySOAPEntryIdToMAPIEntryId(&lpCompany->sCompanyId, lpCompany->ulCompanyId, (ULONG*)&lpsCompany->sCompanyId.cb, (LPENTRYID*)&lpsCompany->sCompanyId.lpb, lpBase);
	if (hr != hrSuccess)
		goto exit;

	lpsCompany->ulIsABHidden	= lpCompany->ulIsABHidden;

exit:
	return hr;
}

HRESULT SoapCompanyArrayToCompanyArray(struct companyArray* lpCompanyArray, ULONG *lpcCompanies, LPECCOMPANY *lppsCompanies)
{
	HRESULT hr = hrSuccess;
	LPECCOMPANY lpECCompanies = NULL;

	if (lpCompanyArray == NULL || lpcCompanies == NULL || lppsCompanies == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	ECAllocateBuffer(sizeof(ECCOMPANY) * lpCompanyArray->__size, (void**)&lpECCompanies);
	memset(lpECCompanies, 0, sizeof(ECCOMPANY) * lpCompanyArray->__size);

	for (unsigned int i=0; i < lpCompanyArray->__size; i++) {
		hr = SoapCompanyToCompany(&lpCompanyArray->__ptr[i], lpECCompanies + i, lpECCompanies);
		if (hr != hrSuccess)
			goto exit;
	}

	*lppsCompanies = lpECCompanies;
	*lpcCompanies = lpCompanyArray->__size;
	lpECCompanies = NULL;

exit:
	if (lpECCompanies != NULL)
		ECFreeBuffer(lpECCompanies);

	return hr;
}

HRESULT SoapCompanyToCompany(struct company *lpCompany, LPECCOMPANY *lppsCompany)
{
	HRESULT		hr			= hrSuccess;
	LPECCOMPANY	lpsCompany	= NULL;

	if (lpCompany == NULL || lppsCompany == NULL)
	{
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = ECAllocateBuffer(sizeof *lpsCompany, (void**)&lpsCompany);
	if (hr != hrSuccess)
		goto exit;

	hr = SoapCompanyToCompany(lpCompany, lpsCompany, NULL);
	if (hr != hrSuccess)
		goto exit;

	*lppsCompany = lpsCompany;
	lpsCompany = NULL;

exit:
	if (lpsCompany != NULL)
		ECFreeBuffer(lpsCompany);

	return hr;
}

HRESULT SvrNameListToSoapMvString8(LPECSVRNAMELIST lpSvrNameList, struct mv_string8 **lppsSvrNameList)
{
	HRESULT				hr = hrSuccess;
	struct mv_string8	*lpsSvrNameList = NULL;
	unsigned			nLen = 0;

	if (lpSvrNameList == NULL || lppsSvrNameList == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	ECAllocateBuffer(sizeof *lpsSvrNameList, (void**)&lpsSvrNameList);
	memset(lpsSvrNameList, 0, sizeof *lpsSvrNameList);
	
	if (lpSvrNameList->cServers > 0) {
		lpsSvrNameList->__size = lpSvrNameList->cServers;
		ECAllocateMore(lpSvrNameList->cServers * sizeof *lpsSvrNameList->__ptr, lpsSvrNameList, (void**)&lpsSvrNameList->__ptr);
		memset(lpsSvrNameList->__ptr, 0, lpSvrNameList->cServers * sizeof *lpsSvrNameList->__ptr);
		
		for (unsigned i = 0; i < lpSvrNameList->cServers; ++i) {
			nLen = strlen(lpSvrNameList->lpszaServer[i]) + 1;
			ECAllocateMore(nLen, lpsSvrNameList, (void**)&lpsSvrNameList->__ptr[i]);
			memcpy(lpsSvrNameList->__ptr[i], lpSvrNameList->lpszaServer[i], nLen);
		}
	}
	
	*lppsSvrNameList = lpsSvrNameList;
	lpsSvrNameList = NULL;
	
exit:
	if (lpsSvrNameList && hr != hrSuccess)
		ECFreeBuffer(lpsSvrNameList);
		
	return hr;
}

HRESULT SoapServerListToServerList(struct serverList *lpsServerList, LPECSERVERLIST *lppServerList)
{
	HRESULT			hr = hrSuccess;
	LPECSERVERLIST	lpServerList = NULL;
	unsigned		nLen = 0;

	if (lpsServerList == NULL || lppServerList == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}
	
	ECAllocateBuffer(sizeof *lpServerList, (void**)&lpServerList);
	memset(lpServerList, 0, sizeof *lpServerList);
	
	if (lpsServerList->__size > 0 && lpsServerList->__ptr != NULL) {
		lpServerList->cServers = lpsServerList->__size;
		ECAllocateMore(lpsServerList->__size * sizeof *lpServerList->lpsaServer, lpServerList, (void**)&lpServerList->lpsaServer);
		memset(lpServerList->lpsaServer, 0, lpsServerList->__size * sizeof *lpServerList->lpsaServer);
		
		for (unsigned i = 0; i < lpsServerList->__size; ++i) {
			// Flags
			lpServerList->lpsaServer[i].ulFlags = lpsServerList->__ptr[i].ulFlags;
		
			// Name
			if (lpsServerList->__ptr[i].lpszName != NULL) {
				nLen = strlen(lpsServerList->__ptr[i].lpszName) + 1;
				if (nLen > 1) {
					ECAllocateMore(nLen, lpServerList, (void**)&lpServerList->lpsaServer[i].lpszName);
					memcpy(lpServerList->lpsaServer[i].lpszName, lpsServerList->__ptr[i].lpszName, nLen);
				}
			}
			
			// FilePath
			if (lpsServerList->__ptr[i].lpszFilePath != NULL) {
				nLen = strlen(lpsServerList->__ptr[i].lpszFilePath) + 1;
				if (nLen > 1) {
					ECAllocateMore(nLen, lpServerList, (void**)&lpServerList->lpsaServer[i].lpszFilePath);
					memcpy(lpServerList->lpsaServer[i].lpszFilePath, lpsServerList->__ptr[i].lpszFilePath, nLen);
				}
			}
			
			// HttpPath
			if (lpsServerList->__ptr[i].lpszHttpPath != NULL) {
				nLen = strlen(lpsServerList->__ptr[i].lpszHttpPath) + 1;
				if (nLen > 1) {
					ECAllocateMore(nLen, lpServerList, (void**)&lpServerList->lpsaServer[i].lpszHttpPath);
					memcpy(lpServerList->lpsaServer[i].lpszHttpPath, lpsServerList->__ptr[i].lpszHttpPath, nLen);
				}
			}
			
			// SslPath
			if (lpsServerList->__ptr[i].lpszSslPath != NULL) {
				nLen = strlen(lpsServerList->__ptr[i].lpszSslPath) + 1;
				if (nLen > 1) {
					ECAllocateMore(nLen, lpServerList, (void**)&lpServerList->lpsaServer[i].lpszSslPath);
					memcpy(lpServerList->lpsaServer[i].lpszSslPath, lpsServerList->__ptr[i].lpszSslPath, nLen);
				}
			}
			
			// PreferedPath
			if (lpsServerList->__ptr[i].lpszPreferedPath != NULL) {
				nLen = strlen(lpsServerList->__ptr[i].lpszPreferedPath) + 1;
				if (nLen > 1) {
					ECAllocateMore(nLen, lpServerList, (void**)&lpServerList->lpsaServer[i].lpszPreferedPath);
					memcpy(lpServerList->lpsaServer[i].lpszPreferedPath, lpsServerList->__ptr[i].lpszPreferedPath, nLen);
				}
			}
		}
	}
	
	*lppServerList = lpServerList;
	lpServerList = NULL;
	
exit:
	if (lpServerList)
		ECFreeBuffer(lpServerList);

	return hr;
}

// This function wraps the GSOAP fopen call to support "file:///var/run/socket" unix-socket URI's
int gsoap_connect_pipe(struct soap *soap, const char *endpoint, const char *host, int port)
{
	int fd;
	struct sockaddr_un saddr;
	memset(&saddr, 0, sizeof(struct sockaddr_un));

	// See stdsoap2.cpp:tcp_connect() function
	if (soap_valid_socket(soap->socket))
	    return SOAP_OK;

	soap->socket = SOAP_INVALID_SOCKET;

	if (strncmp(endpoint, "file://", 7) || strchr(endpoint+7,'/') == NULL)
   	{
      	return SOAP_EOF;
	}

	fd = socket(PF_UNIX, SOCK_STREAM, 0);

	saddr.sun_family = AF_UNIX;
	strcpy(saddr.sun_path, strchr(endpoint+7,'/'));

	connect(fd, (struct sockaddr *)&saddr, sizeof(struct sockaddr_un));

 	soap->sendfd = soap->recvfd = SOAP_INVALID_SOCKET;
	soap->socket = fd;

	// Because 'file:///var/run/file' will be parsed into host='', path='/var/run/file',
	// the gSOAP code doesn't set the soap->status variable. (see soap_connect_command in
	// stdsoap2.cpp:12278) This could possibly lead to
	// soap->status accidentally being set to SOAP_GET, which would break things. The
	// chances of this happening are, of course, small, but also very real.

	soap->status = SOAP_POST;

   	return SOAP_OK;
}

static int ssl_zvcb_index = -1;	// the index to get our custom data



int ssl_verify_callback_zarafa_silent(int ok, X509_STORE_CTX *store)
{
	int sslerr;

	if (ok == 0)
	{
		// Get the last ssl error
		sslerr = X509_STORE_CTX_get_error(store);
		switch (sslerr)
		{
			case X509_V_ERR_CERT_HAS_EXPIRED:
			case X509_V_ERR_CERT_NOT_YET_VALID:
			case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
				// always ignore these errors
				X509_STORE_CTX_set_error(store, X509_V_OK);
				ok = 1;
				break;
			default:
				break;
		}
	}

exit:
	if(!ok)
		TRACE_RELEASE("Server certificate rejected. Connect once with Outlook to verify the authenticity and select the option to remember the choice. Please make sure you do this for each server in your cluster.");
	return ok;
}


HRESULT CreateSoapTransport(ULONG ulUIFlags, sGlobalProfileProps sProfileProps, ZarafaCmd **lppCmd)
{
	HRESULT		hr = hrSuccess;
	ZarafaCmd*	lpCmd = NULL;

	if (sProfileProps.strServerPath.empty() || lppCmd == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	lpCmd = new ZarafaCmd();

	soap_imode(lpCmd->soap, SOAP_IO_KEEPALIVE);
	soap_omode(lpCmd->soap, SOAP_IO_KEEPALIVE | SOAP_XML_TREE);

	lpCmd->endpoint = strdup(sProfileProps.strServerPath.c_str());

#ifdef WITH_OPENSSL
	if (strncmp("https:", lpCmd->endpoint, 6) == 0) {
		// no need to add certificates to call, since soap also calls SSL_CTX_set_default_verify_paths()
		if(soap_ssl_client_context(lpCmd->soap,
								SOAP_SSL_DEFAULT | SOAP_SSL_SKIP_HOST_CHECK,
								sProfileProps.strSSLKeyFile.empty()? NULL : sProfileProps.strSSLKeyFile.c_str(), 
								sProfileProps.strSSLKeyPass.empty()? NULL : sProfileProps.strSSLKeyPass.c_str(),
								NULL, NULL,
								NULL)) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		// set connection string as callback information
		if (ssl_zvcb_index == -1) {
			ssl_zvcb_index = SSL_get_ex_new_index(0, NULL, NULL, NULL, NULL);
		}
		// callback data will be set right before tcp_connect()

		// set our own certificate check function
		lpCmd->soap->fsslverify = ssl_verify_callback_zarafa_silent;
		SSL_CTX_set_verify(lpCmd->soap->ctx, SSL_VERIFY_PEER, lpCmd->soap->fsslverify);
	}
#endif

	if(strncmp("file:", lpCmd->endpoint, 5) == 0) {
		lpCmd->soap->fconnect = gsoap_connect_pipe;
	}else {
		if ((sProfileProps.ulProxyFlags&EC_PROFILE_PROXY_FLAGS_USE_PROXY) && !sProfileProps.strProxyHost.empty()) {
			lpCmd->soap->proxy_host = strdup(sProfileProps.strProxyHost.c_str());
			lpCmd->soap->proxy_port = sProfileProps.ulProxyPort;

			if (!sProfileProps.strProxyUserName.empty())
				lpCmd->soap->proxy_userid = strdup(sProfileProps.strProxyUserName.c_str());

			if (!sProfileProps.strProxyPassword.empty())
				lpCmd->soap->proxy_passwd = strdup(sProfileProps.strProxyPassword.c_str());
		}

		lpCmd->soap->connect_timeout = sProfileProps.ulConnectionTimeOut;
	}


	*lppCmd = lpCmd;

exit:
	if (hr != hrSuccess && lpCmd) {
		free((void*)lpCmd->endpoint);	// because of strdup()
		delete lpCmd;
		lpCmd = NULL;
	}

	return hr;
}

VOID DestroySoapTransport(ZarafaCmd *lpCmd)
{
	if (!lpCmd)
		return;

	if (lpCmd->endpoint)
		free((void *)lpCmd->endpoint);   // because of strdup()

	if (lpCmd->soap->proxy_host)
		free((void *)lpCmd->soap->proxy_host);

	if (lpCmd->soap->proxy_userid)
		free((void *)lpCmd->soap->proxy_userid);

	if (lpCmd->soap->proxy_passwd)
		free((void *)lpCmd->soap->proxy_passwd);

	delete lpCmd;
}

// Wrap the server store entryid to client store entry. (Add a servername)
HRESULT WrapServerClientStoreEntry(const char* lpszServerName, entryId* lpsStoreId, ULONG* lpcbStoreID, LPENTRYID* lppStoreID)
{
	HRESULT		hr = hrSuccess;
	LPENTRYID	lpStoreID = NULL;
	ULONG		ulSize;

	if(lpsStoreId == NULL || lpszServerName == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// The new entryid size is, current size + servername size + 1 byte term 0 - 4 bytes padding
	ulSize = lpsStoreId->__size+strlen(lpszServerName)+1-4;

	hr = ECAllocateBuffer(ulSize, (void**)&lpStoreID);
	if(hr != hrSuccess)
		goto exit;

	memset(lpStoreID, 0, ulSize );

	//Copy the entryid without servername
	memcpy(lpStoreID, lpsStoreId->__ptr, lpsStoreId->__size);

	// Add the server name
	strcpy((char*)lpStoreID+(lpsStoreId->__size-4), lpszServerName);

	*lpcbStoreID = ulSize;
	*lppStoreID = lpStoreID;

exit:

	return hr;
}

// Un wrap the client store entryid to server store entry. (remove a servername)
HRESULT UnWrapServerClientStoreEntry(ULONG cbWrapStoreID, LPENTRYID lpWrapStoreID, ULONG* lpcbUnWrapStoreID, LPENTRYID* lppUnWrapStoreID)
{
	HRESULT	hr = hrSuccess;
	LPENTRYID lpUnWrapStoreID = NULL;
	PEID	peid = NULL;
	ULONG	ulSize = 0;

	if(lpWrapStoreID == NULL || lppUnWrapStoreID == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	//FIXME: check or it's an zarafa entry?

	peid = (PEID)lpWrapStoreID;

	if(peid->ulVersion == 0) {
		ulSize = sizeof(EID_V0);
	}else if(peid->ulVersion == 1) {
		ulSize = sizeof(EID);
	}else {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	if(cbWrapStoreID < ulSize) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	hr = ECAllocateBuffer(ulSize, (void**)&lpUnWrapStoreID);
	if(hr != hrSuccess)
		goto exit;

	memset(lpUnWrapStoreID, 0, ulSize);

	// Remove servername
	memcpy(lpUnWrapStoreID, lpWrapStoreID, ulSize-4);

	*lppUnWrapStoreID = lpUnWrapStoreID;
	*lpcbUnWrapStoreID = ulSize;

exit:
	return hr;
}

HRESULT UnWrapServerClientABEntry(ULONG cbWrapABID, LPENTRYID lpWrapABID, ULONG* lpcbUnWrapABID, LPENTRYID* lppUnWrapABID)
{
	HRESULT	hr = hrSuccess;
	LPENTRYID lpUnWrapABID = NULL;
	PABEID	pabeid = NULL;
	ULONG	ulSize = 0;

	if(lpWrapABID == NULL || lppUnWrapABID == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	// Check minimum size of EntryID
	if (cbWrapABID < sizeof(ABEID)) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	//FIXME: check or it's an zarafa entry?

	pabeid = (PABEID)lpWrapABID;

	if(pabeid->ulVersion == 0) {
		ulSize = sizeof(ABEID);
	}else if(pabeid->ulVersion == 1) {
		ulSize = CbABEID(pabeid);
	}else {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	if(cbWrapABID < ulSize) {
		hr = MAPI_E_INVALID_ENTRYID;
		goto exit;
	}

	hr = ECAllocateBuffer(ulSize, (void**)&lpUnWrapABID);
	if(hr != hrSuccess)
		goto exit;

	memset(lpUnWrapABID, 0, ulSize);

	// Remove servername
	memcpy(lpUnWrapABID, lpWrapABID, ulSize-4);

	*lppUnWrapABID = lpUnWrapABID;
	*lpcbUnWrapABID = ulSize;

exit:
	return hr;
}

HRESULT CopySOAPNotificationToMAPINotification(void *lpProvider, struct notification *lpSrc, LPNOTIFICATION *lppDst) {
	HRESULT hr = hrSuccess;
	LPNOTIFICATION lpNotification = NULL;
	int nLen;

	ECAllocateBuffer(sizeof(NOTIFICATION), (void**)&lpNotification);
	memset(lpNotification, 0, sizeof(NOTIFICATION));

	lpNotification->ulEventType = lpSrc->ulEventType;

	switch(lpSrc->ulEventType){
		case fnevCriticalError:// ERROR_NOTIFICATION
			hr = MAPI_E_INVALID_PARAMETER;
			break;
		case fnevNewMail://NEWMAIL_NOTIFICATION
			if(lpSrc->newmail->pEntryId) {
				// Ignore error now
				// FIXME: This must exist, so maybe give an error or skip them
				CopySOAPEntryIdToMAPIEntryId(lpSrc->newmail->pEntryId, &lpNotification->info.newmail.cbEntryID, &lpNotification->info.newmail.lpEntryID, (void **)lpNotification);
			}
			if(lpSrc->newmail->pParentId) {
				// Ignore error
				CopySOAPEntryIdToMAPIEntryId(lpSrc->newmail->pParentId, &lpNotification->info.newmail.cbParentID, &lpNotification->info.newmail.lpParentID, (void **)lpNotification);
			}


			if(lpSrc->newmail->lpszMessageClass != NULL) {
				nLen = strlen(lpSrc->newmail->lpszMessageClass)+1;
				ECAllocateMore(nLen, lpNotification, (void**)&lpNotification->info.newmail.lpszMessageClass);
				memcpy(lpNotification->info.newmail.lpszMessageClass, lpSrc->newmail->lpszMessageClass, nLen);
			}

			lpNotification->info.newmail.ulFlags = 0;
			lpNotification->info.newmail.ulMessageFlags = lpSrc->newmail->ulMessageFlags;

			break;
		case fnevObjectCreated:// OBJECT_NOTIFICATION
		case fnevObjectDeleted:
		case fnevObjectModified:
		case fnevObjectCopied:
		case fnevObjectMoved:
		case fnevSearchComplete:
			// FIXME for each if statement below, check the ELSE .. we can't send a TABLE_ROW_ADDED without lpProps for example ..
			lpNotification->info.obj.ulObjType = lpSrc->obj->ulObjType;

			// All errors of CopySOAPEntryIdToMAPIEntryId are ignored
			if(lpSrc->obj->pEntryId != NULL) {
				CopySOAPEntryIdToMAPIEntryId(lpSrc->obj->pEntryId, &lpNotification->info.obj.cbEntryID, &lpNotification->info.obj.lpEntryID, (void **)lpNotification);
			}

			if(lpSrc->obj->pParentId != NULL) {
				CopySOAPEntryIdToMAPIEntryId(lpSrc->obj->pParentId, &lpNotification->info.obj.cbParentID, &lpNotification->info.obj.lpParentID, (void **)lpNotification);
			}

			if(lpSrc->obj->pOldId != NULL) {
				CopySOAPEntryIdToMAPIEntryId(lpSrc->obj->pOldId, &lpNotification->info.obj.cbOldID, &lpNotification->info.obj.lpOldID, (void **)lpNotification);
			}

			if(lpSrc->obj->pOldParentId != NULL){
				CopySOAPEntryIdToMAPIEntryId(lpSrc->obj->pOldParentId, &lpNotification->info.obj.cbOldParentID, &lpNotification->info.obj.lpOldParentID, (void **)lpNotification);
			}

			if(lpSrc->obj->pPropTagArray) {
				// ignore errors
				CopySOAPPropTagArrayToMAPIPropTagArray(lpSrc->obj->pPropTagArray, &lpNotification->info.obj.lpPropTagArray, (void **)lpNotification);
			}
			break;
		case fnevTableModified:// TABLE_NOTIFICATION
			lpNotification->info.tab.ulTableEvent = lpSrc->tab->ulTableEvent;
			lpNotification->info.tab.propIndex.ulPropTag = lpSrc->tab->propIndex.ulPropTag;

			if(lpSrc->tab->propIndex.Value.bin){
				lpNotification->info.tab.propIndex.Value.bin.cb = lpSrc->tab->propIndex.Value.bin->__size;
				ECAllocateMore(lpNotification->info.tab.propIndex.Value.bin.cb, lpNotification, (void**)&lpNotification->info.tab.propIndex.Value.bin.lpb);

				memcpy(lpNotification->info.tab.propIndex.Value.bin.lpb, lpSrc->tab->propIndex.Value.bin->__ptr, lpSrc->tab->propIndex.Value.bin->__size);
			}

			lpNotification->info.tab.propPrior.ulPropTag = lpSrc->tab->propPrior.ulPropTag;

			if(lpSrc->tab->propPrior.Value.bin){
				lpNotification->info.tab.propPrior.Value.bin.cb = lpSrc->tab->propPrior.Value.bin->__size;
				ECAllocateMore(lpNotification->info.tab.propPrior.Value.bin.cb, lpNotification, (void**)&lpNotification->info.tab.propPrior.Value.bin.lpb);


				memcpy(lpNotification->info.tab.propPrior.Value.bin.lpb, lpSrc->tab->propPrior.Value.bin->__ptr, lpSrc->tab->propPrior.Value.bin->__size);
			}

			if(lpSrc->tab->pRow)
			{

				lpNotification->info.tab.row.cValues = lpSrc->tab->pRow->__size;
				ECAllocateMore(sizeof(SPropValue)*lpNotification->info.tab.row.cValues, lpNotification, (void**)&lpNotification->info.tab.row.lpProps);

				CopySOAPRowToMAPIRow(lpProvider, lpSrc->tab->pRow, lpNotification->info.tab.row.lpProps, (void **)lpNotification, lpSrc->tab->ulObjType);

			}
			break;
		case fnevStatusObjectModified: // STATUS_OBJECT_NOTIFICATION
			hr = MAPI_E_INVALID_PARAMETER;
			break;
		case fnevExtended: // EXTENDED_NOTIFICATION
			hr = MAPI_E_INVALID_PARAMETER;
			break;
		default:
			hr = MAPI_E_INVALID_PARAMETER;
			break;
	}

	if(hr != hrSuccess)
		goto exit;

	*lppDst = lpNotification;
	lpNotification = NULL;

exit:
	if (lpNotification)
		MAPIFreeBuffer(lpNotification);

	return hr;
}

HRESULT CopySOAPChangeNotificationToSyncState(struct notification *lpSrc, LPSBinary *lppDst, void *lpBase)
{
	HRESULT hr = hrSuccess;
	LPSBinary lpSBinary = NULL;

	if (lpSrc->ulEventType != fnevZarafaIcsChange) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	if (lpBase == NULL)
		ECAllocateBuffer(sizeof *lpSBinary, (void**)&lpSBinary);
	else
		ECAllocateMore(sizeof *lpSBinary, lpBase, (void**)&lpSBinary);
	memset(lpSBinary, 0, sizeof *lpSBinary);

	lpSBinary->cb = lpSrc->ics->pSyncState->__size;

	if (lpBase == NULL)
		ECAllocateMore(lpSBinary->cb, lpSBinary, (void**)&lpSBinary->lpb);
	else
		ECAllocateMore(lpSBinary->cb, lpBase, (void**)&lpSBinary->lpb);

	memcpy(lpSBinary->lpb, lpSrc->ics->pSyncState->__ptr, lpSBinary->cb);

	*lppDst = lpSBinary;
	lpSBinary = NULL;

exit:
    if (lpSBinary)
		MAPIFreeBuffer(lpSBinary);

	return hr;
}

HRESULT CopyMAPISourceKeyToSoapSourceKey(SBinary *lpsMAPISourceKey, struct xsd__base64Binary *lpsSoapSourceKey, void *lpBase)
{
	HRESULT						hr = hrSuccess;
	struct xsd__base64Binary	sSoapSourceKey = {0};

	if (lpsMAPISourceKey == NULL || lpsSoapSourceKey == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	sSoapSourceKey.__size = (int)lpsMAPISourceKey->cb;
	if (lpBase)
		hr = MAPIAllocateMore(lpsMAPISourceKey->cb, lpBase, (void**)&sSoapSourceKey.__ptr);
	else
		hr = MAPIAllocateBuffer(lpsMAPISourceKey->cb, (void**)&sSoapSourceKey.__ptr);
	if (hr != hrSuccess)
		goto exit;

	memcpy(sSoapSourceKey.__ptr, lpsMAPISourceKey->lpb, lpsMAPISourceKey->cb);
	*lpsSoapSourceKey = sSoapSourceKey;

exit:
	return hr;
}

HRESULT CopyICSChangeToSOAPSourceKeys(ULONG cbChanges, ICSCHANGE *lpsChanges, sourceKeyPairArray **lppsSKPA)
{
	HRESULT				hr = hrSuccess;
	sourceKeyPairArray	*lpsSKPA = NULL;

	if (lpsChanges == NULL || lppsSKPA == NULL) {
		hr = MAPI_E_INVALID_PARAMETER;
		goto exit;
	}

	hr = MAPIAllocateBuffer(sizeof *lpsSKPA, (void**)&lpsSKPA);
	if (hr != hrSuccess)
		goto exit;
	memset(lpsSKPA, 0, sizeof *lpsSKPA);

	if (cbChanges > 0) {
		lpsSKPA->__size = cbChanges;

		hr = MAPIAllocateMore(cbChanges * sizeof *lpsSKPA->__ptr, lpsSKPA, (void**)&lpsSKPA->__ptr);
		if (hr != hrSuccess)
			goto exit;
		memset(lpsSKPA->__ptr, 0, cbChanges * sizeof *lpsSKPA->__ptr);

		for (unsigned i = 0; i < cbChanges; ++i) {
			hr = CopyMAPISourceKeyToSoapSourceKey(&lpsChanges[i].sSourceKey, &lpsSKPA->__ptr[i].sObjectKey, lpsSKPA);
			if (hr != hrSuccess)
				goto exit;

			hr = CopyMAPISourceKeyToSoapSourceKey(&lpsChanges[i].sParentSourceKey, &lpsSKPA->__ptr[i].sParentKey, lpsSKPA);
			if (hr != hrSuccess)
				goto exit;
		}
	}

	*lppsSKPA = lpsSKPA;
	lpsSKPA = NULL;

exit:
	if (lpsSKPA)
		MAPIFreeBuffer(lpsSKPA);

	return hr;
}
