/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <platform.h>

#include <stdio.h>
#include <stdlib.h>

#include <sstream>
#include <iostream>
#include <algorithm>

#include <mapi.h>
#include <mapix.h>
#include <mapicode.h>
#include <mapidefs.h>
#include <mapiutil.h>
#include <mapiguid.h>
#include <ECDefs.h>
#include <CommonUtil.h>
#include <ECTags.h>
#include <ECIConv.h>
#include <inetmapi.h>
#include <mapiext.h>
#include <vector>
#include <list>
#include <set>
#include <algorithm>
#include <base64.h>
#include <options.h>

#include "edkmdb.h"
#include "stringutil.h"

#include "IMAP.h"
using namespace std;

string strMonth[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};
string strDay[] = {
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

/**
 * Returns TRUE if the given string starts with strPrefix
 *
 * @param[in]	strInput	string to find prefix in
 * @param[in]	strPrefix	test if input starts with this string
 */
bool Prefix(const std::string &strInput, std::string strPrefix)
{
    return (strInput.compare(0, strPrefix.size(), strPrefix) == 0);
}

IMAP::IMAP(ECChannel *lpChan, char *szServerPath, ECLogger *lpLog, ECConfig *lpConf, bool bMailFoldersOnly, bool bPublicFolder) {
	lpChannel = lpChan;
	lpLogger = lpLog;
	lpConfig = lpConf;

	lpSession = NULL;
	lpStore = NULL;
	lpAddrBook = NULL;
	lpPublicStore = NULL;

	m_strPath = szServerPath;
	imopt_default_delivery_options(&dopt);

	bOnlyMailFolders = bMailFoldersOnly;
	bShowPublicFolder = bPublicFolder;

	m_ulLastUid = 0;

	m_bIdleMode = false;
	m_lpIdleAdviseSink = NULL;
	m_ulIdleAdviseConnection = 0;
	m_lpIdleTable = NULL;
	
	m_ulCacheUID = 0;
	
	pthread_mutex_init(&m_mIdleLock, NULL);
}

IMAP::~IMAP() {
	// Free/release all new and allocated memory
	if (lpPublicStore)
		lpPublicStore->Release();

	if (lpStore)
		lpStore->Release();

	if (lpAddrBook)
		lpAddrBook->Release();

	if (lpSession)
		lpSession->Release();

	// idle cleanup
	if (m_lpIdleAdviseSink)
		m_lpIdleAdviseSink->Release();

	if (m_lpIdleTable)
		m_lpIdleTable->Release();

	pthread_mutex_destroy(&m_mIdleLock);
}

bool IMAP::isLoggedIn() {
	return (lpSession != NULL);
}

void IMAP::ToUpper(string &strString) {
	transform(strString.begin(), strString.end(), strString.begin(), ::toupper);
}

bool IMAP::CaseCompare(string strA, string strB)
{
    ToUpper(strA);
    ToUpper(strB);
    
    return strA == strB;
}

HRESULT IMAP::HrSplitPath(string &strInput, vector<string> &lstPaths) {
    // Strip leading and trailing /
    if(strInput[0] == IMAP_HIERARCHY_DELIMITER)
		strInput.erase(0,1);
        
    if(strInput[strInput.size()-1] == IMAP_HIERARCHY_DELIMITER)
		strInput.erase(strInput.size()-1, 1);
        
    // split into parts
    lstPaths = tokenize(strInput, IMAP_HIERARCHY_DELIMITER);
    
    return hrSuccess;    
}

HRESULT IMAP::HrUnsplitPath(vector<string> &lstPaths, string &strInput) {
    strInput.clear();
        
    for(unsigned int i = 0; i < lstPaths.size(); i++) {
        strInput += lstPaths[i];
        if(i != lstPaths.size()-1)
            strInput += IMAP_HIERARCHY_DELIMITER;
    }
    
    return hrSuccess;
}

HRESULT IMAP::HrSplitInput(string &strInput, vector<string> &lstWords) {
	unsigned int uSpecialCount = 0;
	string::size_type beginPos = 0;
	string::size_type currentPos = 0;
	string::size_type findPos = strInput.find_first_of("\"()[] ", currentPos);
	string::size_type specialPos = string::npos;

	while (findPos != string::npos) {
		if (uSpecialCount == 0 && strInput[findPos] == '"') {
			// find corresponding " and add the string
			specialPos = findPos;
			do {
				specialPos = strInput.find_first_of("\"", specialPos + 1);
			} while (specialPos != string::npos && strInput[specialPos-1] == '\\');

			if (specialPos != string::npos) {
				lstWords.push_back(strInput.substr(findPos + 1, specialPos - findPos - 1));
				findPos = specialPos;
				beginPos = findPos + 1;
			}
		} else if (strInput[findPos] == '(' || strInput[findPos] == '[') {
			uSpecialCount++;
		} else if (strInput[findPos] == ')' || strInput[findPos] == ']') {
			if (uSpecialCount > 0)
				uSpecialCount--;
		} else if (uSpecialCount == 0) {
			if (findPos > beginPos) {
				lstWords.push_back(strInput.substr(beginPos, findPos - beginPos));
			}
			beginPos = findPos + 1;
		}

		currentPos = findPos + 1;
		findPos = strInput.find_first_of("\"()[] ", currentPos);
	}

	if (beginPos < strInput.size()) {
		lstWords.push_back(strInput.substr(beginPos));
	}

	return hrSuccess;
}

HRESULT IMAP::HrGetCommand(string &strCommand, IMAP_Command &eCommand) {
	HRESULT hr = hrSuccess;

	ToUpper(strCommand);
	if (strCommand == "CAPABILITY") {
		eCommand = IMAP_CMD_CAPABILITY;
	} else if (strCommand == "NOOP") {
		eCommand = IMAP_CMD_NOOP;
	} else if (strCommand == "LOGOUT") {
		eCommand = IMAP_CMD_LOGOUT;
	} else if (strCommand == "STARTTLS") {
		eCommand = IMAP_CMD_STARTTLS;
	} else if (strCommand == "AUTHENTICATE") {
		eCommand = IMAP_CMD_AUTHENTICATE;
	} else if (strCommand == "LOGIN") {
		eCommand = IMAP_CMD_LOGIN;
	} else if (strCommand == "SELECT") {
		eCommand = IMAP_CMD_SELECT;
	} else if (strCommand == "EXAMINE") {
		eCommand = IMAP_CMD_EXAMINE;
	} else if (strCommand == "CREATE") {
		eCommand = IMAP_CMD_CREATE;
	} else if (strCommand == "DELETE") {
		eCommand = IMAP_CMD_DELETE;
	} else if (strCommand == "RENAME") {
		eCommand = IMAP_CMD_RENAME;
	} else if (strCommand == "SUBSCRIBE") {
		eCommand = IMAP_CMD_SUBSCRIBE;
	} else if (strCommand == "UNSUBSCRIBE") {
		eCommand = IMAP_CMD_UNSUBSCRIBE;
	} else if (strCommand == "LIST") {
		eCommand = IMAP_CMD_LIST;
	} else if (strCommand == "LSUB") {
		eCommand = IMAP_CMD_LSUB;
	} else if (strCommand == "STATUS") {
		eCommand = IMAP_CMD_STATUS;
	} else if (strCommand == "APPEND") {
		eCommand = IMAP_CMD_APPEND;
	} else if (strCommand == "CHECK") {
		eCommand = IMAP_CMD_CHECK;
	} else if (strCommand == "CLOSE") {
		eCommand = IMAP_CMD_CLOSE;
	} else if (strCommand == "EXPUNGE") {
		eCommand = IMAP_CMD_EXPUNGE;
	} else if (strCommand == "SEARCH") {
		eCommand = IMAP_CMD_SEARCH;
	} else if (strCommand == "FETCH") {
		eCommand = IMAP_CMD_FETCH;
	} else if (strCommand == "STORE") {
		eCommand = IMAP_CMD_STORE;
	} else if (strCommand == "COPY") {
		eCommand = IMAP_CMD_COPY;
	} else if (strCommand == "XAOL-MOVE") {
		eCommand = IMAP_CMD_XAOL_MOVE;
	} else if (strCommand == "UID") {
		eCommand = IMAP_CMD_UID;
	} else if (strCommand == "IDLE") {
		eCommand = IMAP_CMD_IDLE;
	} else if (strCommand.compare("NAMESPACE") == 0) {
		eCommand = IMAP_CMD_NAMESPACE;
	} else {
		hr = MAPI_E_CALL_FAILED;
	}

	return hr;
}

HRESULT IMAP::HrCmdCapability(string strTag) {
	std::string strCapabilities = "CAPABILITY IMAP4rev1 CHILDREN XAOL-OPTION NAMESPACE";
	char *idle = lpConfig->GetSetting("imap_capability_idle");

	if (idle && strcmp(idle, "yes") == 0)
		strCapabilities += " IDLE";
	
	if (!lpChannel->UsingSsl() && lpChannel->sslctx())
		strCapabilities += " STARTTLS";

	HrResponse(strCapabilities);
	HrResultOk(strTag, "CAPABILITY Completed");
	return hrSuccess;
}

HRESULT IMAP::HrCmdNoop(string strTag) {
	HrRefreshFolderMails(false, !bCurrentFolderReadOnly, false, NULL);
	HrResultOk(strTag, "NOOP completed");
	return hrSuccess;
}

HRESULT IMAP::HrCmdLogout(string strTag) {
	HrResponse("BYE Zarafa server logging out");
	HrResultOk(strTag, "LOGOUT completed");
	return hrSuccess;
}

HRESULT IMAP::HrCmdStarttls(string strTag) {
	HRESULT hr = hrSuccess;

	if (!lpChannel->sslctx()) {
		HrResultNo(strTag, "STARTTLS error in ssl context");
		goto exit;
	}

	if (lpChannel->UsingSsl()) {
		HrResultNo(strTag, "STARTTLS already using SSL/TLS");
		goto exit;
	}

	HrResultOk(strTag, "Begin TLS negotiation now");

	hr = lpChannel->HrEnableTLS();
	if (hr != hrSuccess) {
		HrResponse("BAD [ALERT] Error switching to secure SSL/TLS connection");
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Error switching to SSL in STARTTLS");
		goto exit;
	}

	if (lpChannel->UsingSsl()) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, "Using SSL now");
	}

exit:
	return hr;
}

HRESULT IMAP::HrCmdAuthenticate(string strTag, string strAuthMethod) {
	HrResultNo(strTag, "AUTHENTICATE method not supported");
	return hrSuccess;
}

HRESULT IMAP::HrCmdLogin(string strTag, string strUser, string strPass) {
	HRESULT hr = hrSuccess;
	char szResponse[IMAP_RESP_MAX + 1];
	string strUsername;
	size_t i;

	// strUser isn't sent in imap style utf-7, but \ is escaped, so strip those
	for (i=0; i < strUser.length(); i++) {
		if (strUser[i] == '\\' && i+1 < strUser.length() && (strUser[i+1] == '"' || strUser[i+1] == '\\'))
			i++;
		strUsername += strUser[i];
	}	

	if (lpSession != NULL) {
		lpLogger->Log(EC_LOGLEVEL_INFO, "Ignoring to login TWICE for username \"%s\"", strUsername.c_str());
		HrResultNo(strTag, "LOGIN Can't login twice");
		// hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	// do not disable notifications for imap conncetions, may be idle and sessions on the zarafa server will disappear.
	hr = HrOpenECSession(&lpSession, (char*)strUsername.c_str(), (char*)strPass.c_str(), (char*)m_strPath.c_str(), EC_PROFILE_FLAGS_NO_COMPRESSION, NULL, NULL);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to login from %s with invalid username \"%s\" or wrong password. Error: 0x%08X",
					  lpChannel->GetIPAddress().c_str(), strUsername.c_str(), hr);

		snprintf(szResponse, IMAP_RESP_MAX, "LOGIN wrong username or password");
		HrResultNo(strTag, szResponse);
		goto exit;
	}

	hr = HrOpenDefaultStore(lpSession, &lpStore);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Failed to open default store");
		HrResultNo(strTag, "LOGIN can't open default store");
		goto exit;
	}

	hr = HrGetSubscribedList();
	// ignore error, empty list of subscribed folder

	hr = lpSession->OpenAddressBook(0, NULL, AB_NO_DIALOG, &lpAddrBook);
	if (hr != hrSuccess)
		lpAddrBook = NULL;

	if(bShowPublicFolder){
		hr = HrOpenECPublicStore(lpSession, &lpPublicStore);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_WARNING, "Failed to open public store");
			lpPublicStore = NULL;
		}
	}

	hr = HrMakeSpecialsList();
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, "Failed to find special folder properties");
		HrResultNo(strTag, "LOGIN can't find special folder properties");
		goto exit;
	}

	lpLogger->Log(EC_LOGLEVEL_ERROR, "IMAP Login from %s for user %s", lpChannel->GetIPAddress().c_str(), strUsername.c_str());
	HrResultOk(strTag, "LOGIN completed");

exit:
	return hr;
}

/**
 * Make the strFolder the current working folder.
 *
 * @param[in]	strTag	IMAP command tag
 * @param[in]	strFolder	IMAP folder name in UTF-7 something charset
 * @param[in]	bReadOnly	The EXAMINE command was given instead of the SELECT command
 */
HRESULT IMAP::HrCmdSelect(string strTag, string strFolder, bool bReadOnly) {
	HRESULT hr = hrSuccess;
	char szResponse[IMAP_RESP_MAX + 1];
	unsigned int ulUnseen = 0;
	string command = "SELECT";

	if (bReadOnly)
		command = "EXAMINE";
	
	if (!lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, command+" error no session");
		goto exit;
	}

	// Apple mail client does this request, so we need to block it.
	if (strFolder.empty()) {
		HrResultNo(strTag, command+" invalid folder name");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	strCurrentFolder = IMAP2MAPICharset(strFolder);

	bCurrentFolderReadOnly = bReadOnly;
	hr = HrRefreshFolderMails(true, !bCurrentFolderReadOnly, false, &ulUnseen);
	if (hr != hrSuccess) {
		HrResultNo(strTag, command+" error getting mails in folder");
		goto exit;
	}

	// \Seen = PR_MESSAGE_FLAGS MSGFLAG_READ
	// \Answered = PR_MSG_STATUS MSGSTATUS_ANSWERED  //PR_LAST_VERB_EXECUTED EXCHIVERB_REPLYTOSENDER EXCHIVERB_REPLYTOALL
	// \Draft = PR_MSG_STATUS MSGSTATUS_DRAFT        //PR_MESSAGE_FLAGS MSGFLAG_UNSENT
	// \Flagged = PR_FLAG_STATUS
	// \Deleted = PR_MSG_STATUS MSGSTATUS_DELMARKED
	// \Recent = ??? (arrived after last command/login)
	HrResponse("FLAGS (\\Seen \\Draft \\Deleted \\Flagged \\Answered)");
	HrResponse("OK [PERMANENTFLAGS (\\Seen \\Flagged \\Draft \\Deleted \\Answered)] Permanent flags");
	snprintf(szResponse, IMAP_RESP_MAX, "OK [UIDNEXT %u] Predicted next UID", m_ulLastUid + 1);
	HrResponse(szResponse);
	if(ulUnseen) {
    	snprintf(szResponse, IMAP_RESP_MAX, "OK [UNSEEN %u] First unseen message", ulUnseen);
    	HrResponse(szResponse);
    }
	snprintf(szResponse, IMAP_RESP_MAX, "OK [UIDVALIDITY %u] UIDVALIDITY value", 1);
	HrResponse(szResponse);
	if (bReadOnly)
		HrResultOk(strTag, "[READ-ONLY] EXAMINE completed");
	else
		HrResultOk(strTag, "[READ-WRITE] SELECT completed");

exit:
	return hr;
}

HRESULT IMAP::HrCmdCreate(string strTag, string strFolder) {
	HRESULT hr = hrSuccess;
	IMAPIFolder *lpFolder = NULL;
	IMAPIFolder *lpSubFolder = NULL;
	vector<string> strPaths;
	vector<string>::iterator iPath;
	string strPath;
	SPropValue sFolderClass;

	if (!lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "CREATE error no session");
		goto exit;
	}

	if (strFolder.empty()) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "CREATE error no folder");
		goto exit;
	}

	strFolder = IMAP2MAPICharset(strFolder);
	if (strFolder[0] == IMAP_HIERARCHY_DELIMITER) {
		// courier and dovecot also block this
        HrResultNo(strTag, "CREATE invalid folder name");
        goto exit;
	}

    hr = HrFindFolderPartial(strFolder, &lpFolder, &strPath);
    if(hr != hrSuccess) {
        HrResultNo(strTag, "CREATE error opening destination folder");
        goto exit;
    }

	if (strPath.empty()) {
		HrResultNo(strTag, "CREATE folder already exists");
		goto exit;
	}

	strPaths = tokenize(strPath, IMAP_HIERARCHY_DELIMITER);

	for (iPath = strPaths.begin(); iPath != strPaths.end(); iPath++) {
		hr = lpFolder->CreateFolder(FOLDER_GENERIC, (TCHAR *) iPath->c_str(), NULL, NULL, 0, &lpSubFolder);
		if (hr != hrSuccess) {
			if (hr == MAPI_E_COLLISION)
				HrResultNo(strTag, "CREATE folder already exists");
			else
				HrResultNo(strTag, "CREATE can't create folder");
			goto exit;
		}

		sFolderClass.ulPropTag = PR_CONTAINER_CLASS_A;
		sFolderClass.Value.lpszA = "IPF.Note";
		hr = HrSetOneProp(lpSubFolder, &sFolderClass);
		if (hr != hrSuccess)
			goto exit;

		lpFolder->Release();
		lpFolder = lpSubFolder;
		lpSubFolder = NULL;
	}

	HrResultOk(strTag, "CREATE completed");

exit:
	if (lpFolder)
		lpFolder->Release();

	if (lpSubFolder)
		lpSubFolder->Release();

	return hr;
}

HRESULT IMAP::HrCmdDelete(string strTag, string strFolder) {
	HRESULT hr = hrSuccess;
	list<SFolder>::iterator lpFolder;
	list<SFolder>::iterator lpDelFolder;
	IMAPIFolder *lpParentFolder = NULL;
	ULONG cbEntryID;
	LPENTRYID lpEntryID = NULL;
	LPSPropValue lpPropVal = NULL;

	if (!lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "DELETE error no session");
		goto exit;
	}

	strFolder = IMAP2MAPICharset(strFolder);
	ToUpper(strFolder);

	if (strFolder == "INBOX") {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "DELETE error deleting INBOX is not allowed");
		goto exit;
	}

	hr = HrFindFolderEntryID(strFolder, &cbEntryID, &lpEntryID);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "DELETE error folder not found");
		goto exit;
	}

	if (IsSpecialFolder(cbEntryID, lpEntryID)) {
		HrResultNo(strTag, "DELETE special folder may not be deleted");
		goto exit;
	}

	hr = HrOpenParentFolder(cbEntryID, lpEntryID, &lpParentFolder);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "DELETE error opening parent folder");
		goto exit;
	}

	hr = lpParentFolder->DeleteFolder(cbEntryID, lpEntryID, 0, NULL, DEL_FOLDERS | DEL_MESSAGES);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "DELETE error deleting folder");
		goto exit;
	}

	if (strCurrentFolder == strFolder) {
	    strCurrentFolder.clear();
    }

	HrResultOk(strTag, "DELETE completed");

exit:
    if (lpEntryID)
        MAPIFreeBuffer(lpEntryID);
        
	if (lpParentFolder)
		lpParentFolder->Release();

	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	return hr;
}

HRESULT IMAP::HrCmdRename(string strTag, string strExistingFolder, string strNewFolder) {
	HRESULT hr = hrSuccess;
	string strFolder;
	string strPath;
	LPSPropValue lpPropVal = NULL;
	LPSPropValue lppvFromEntryID = NULL;
	LPSPropValue lppvDestEntryID = NULL;
	IMAPIFolder *lpParentFolder = NULL;
	IMAPIFolder *lpMakeFolder = NULL;
	IMAPIFolder *lpSubFolder = NULL;
	ULONG ulObjType = 0;
	string::size_type deliPos;
	ULONG cbMovFolder = 0;
	LPENTRYID lpMovFolder = NULL;
	SPropValue sFolderClass;

	if (!lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "RENAME error no session");
		goto exit;
	}

	strExistingFolder = IMAP2MAPICharset(strExistingFolder);
	strNewFolder = IMAP2MAPICharset(strNewFolder);

	ToUpper(strExistingFolder);
	if (strExistingFolder == "INBOX") {
		// FIXME, rfc text:
		// 
		//       Renaming INBOX is permitted, and has special behavior.  It moves
		//       all messages in INBOX to a new mailbox with the given name,
		//       leaving INBOX empty.  If the server implementation supports
		//       inferior hierarchical names of INBOX, these are unaffected by a
		//       rename of INBOX.

		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	hr = HrFindFolderEntryID(strExistingFolder, &cbMovFolder, &lpMovFolder);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "RENAME error source folder not found");
		goto exit;
	}

	if (IsSpecialFolder(cbMovFolder, lpMovFolder)) {
		HrResultNo(strTag, "RENAME special folder may not be moved or renamed");
		goto exit;
	}

    hr = HrOpenParentFolder(cbMovFolder, lpMovFolder, &lpParentFolder);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "RENAME error opening parent folder");
		goto exit;
	}

	// Find the folder as far as we can
	hr = HrFindFolderPartial(strNewFolder, &lpMakeFolder, &strPath);
	if(hr != hrSuccess) {
	    HrResultNo(strTag, "RENAME error opening destination folder");
	    goto exit;
    }
    
    if(strPath.empty()) {
        HrResultNo(strTag, "RENAME destination already exists");
        goto exit;
    }
    
    // strPath now contains subfolder we want to create (eg sub/new). So now we have to
    // mkdir -p all the folder leading up to the last (if any)
	do {
		deliPos = strPath.find(IMAP_HIERARCHY_DELIMITER);
		if (deliPos == string::npos) {
			strFolder = strPath;
		} else {
			strFolder = strPath.substr(0, deliPos);
			strPath = strPath.substr(deliPos + 1);

			if (!strFolder.empty())
				hr = lpMakeFolder->CreateFolder(FOLDER_GENERIC, (TCHAR *) strFolder.c_str(), NULL, NULL, OPEN_IF_EXISTS, &lpSubFolder);

			if (hr != hrSuccess) {
				HrResultNo(strTag, "RENAME error creating folder");
				goto exit;
			}

			sFolderClass.ulPropTag = PR_CONTAINER_CLASS_A;
			sFolderClass.Value.lpszA = "IPF.Note";
			hr = HrSetOneProp(lpSubFolder, &sFolderClass);
			if (hr != hrSuccess)
				goto exit;

			lpMakeFolder->Release();
			lpMakeFolder = lpSubFolder;
			lpSubFolder = NULL;
		}
	} while (deliPos != string::npos);

	if (HrGetOneProp(lpParentFolder, PR_ENTRYID, &lppvFromEntryID) != hrSuccess || HrGetOneProp(lpMakeFolder, PR_ENTRYID, &lppvDestEntryID) != hrSuccess) {
		HrResultNo(strTag, "RENAME error opening source or destination");
		goto exit;
	}

	// When moving in the same folder, just rename
	if (lppvFromEntryID->Value.bin.cb != lppvDestEntryID->Value.bin.cb || memcmp(lppvFromEntryID->Value.bin.lpb, lppvDestEntryID->Value.bin.lpb, lppvDestEntryID->Value.bin.cb) != 0) {
	    // Do the real move
		hr = lpParentFolder->CopyFolder(cbMovFolder, lpMovFolder, &IID_IMAPIFolder, lpMakeFolder,
										(TCHAR *) strFolder.c_str(), 0, NULL, FOLDER_MOVE);
	} else {
		// from is same as dest folder, use SetProps(PR_DISPLAY_NAME)
		SPropValue propName;
		propName.ulPropTag = PR_DISPLAY_NAME;
		propName.Value.lpszA = (char*)strFolder.c_str();

		hr = lpSession->OpenEntry(cbMovFolder, lpMovFolder, &IID_IMAPIFolder, MAPI_MODIFY,
								&ulObjType,	(LPUNKNOWN *) &lpSubFolder);
		if (hr != hrSuccess) {
			HrResultNo(strTag, "RENAME error opening folder");
			goto exit;
		}

		hr = lpSubFolder->SetProps(1, &propName, NULL);
	}

	if (hr != hrSuccess) {
		HrResultNo(strTag, "RENAME error moving folder");
		goto exit;
	}

	HrResultOk(strTag, "RENAME completed");

exit:
	if (lppvFromEntryID)
		MAPIFreeBuffer(lppvFromEntryID);

	if (lppvDestEntryID)
		MAPIFreeBuffer(lppvDestEntryID);

	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	if (lpMovFolder)
		MAPIFreeBuffer(lpMovFolder);

	if (lpParentFolder)
		lpParentFolder->Release();

	if (lpMakeFolder)
		lpMakeFolder->Release();

	if (lpSubFolder)
		lpSubFolder->Release();

	return hr;
}

// SUBSCRIBE and UNSUBSCRIBE commands
HRESULT IMAP::HrCmdSubscribe(string strTag, string strFolder, bool bSubscribe) {
	HRESULT hr = hrSuccess;
	string strAction;
	vector<BinaryArray>::iterator iFolder;
	bool bChanged = false;
	ULONG cbEntryID = 0;
	LPENTRYID lpEntryID = NULL;

	if (bSubscribe)
		strAction = "SUBSCRIBE";
	else
		strAction = "UNSUBSCRIBE";

	if (!lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, strAction+" error no session");
		goto exit;
	}

	strFolder = IMAP2MAPICharset(strFolder);

	hr = HrFindFolderEntryID(strFolder, &cbEntryID, &lpEntryID);
	if (hr != hrSuccess) {
		// folder not found, but not error, so thunderbird updates view correctly.
		HrResultOk(strTag, strAction+" folder not found");
		goto exit;
	}

	if (IsSpecialFolder(cbEntryID, lpEntryID)) {
		if (!bSubscribe)
			HrResultNo(strTag, strAction+" cannot unsubscribe this special folder");
		else
			HrResultOk(strTag, strAction+" completed");
		goto exit;
	}		

	iFolder = find(m_vSubscriptions.begin(), m_vSubscriptions.end(), BinaryArray((BYTE *)lpEntryID, cbEntryID, true));
	if (iFolder == m_vSubscriptions.end()) {
		if (bSubscribe) {
			m_vSubscriptions.push_back(BinaryArray((BYTE *)lpEntryID, cbEntryID));
			bChanged = true;
		}
	} else {
		if (!bSubscribe) {
			m_vSubscriptions.erase(iFolder);
			bChanged = true;
		}
	}

	if (bChanged) {
		hr = HrSetSubscribedList();
		if (hr != hrSuccess) {
			HrResultNo(strTag, strAction+" writing subscriptions to server failed");
			goto exit;
		}
	}

	HrResultOk(strTag, strAction+" completed");

exit:
    if (lpEntryID)
        MAPIFreeBuffer(lpEntryID);


	return hr;
}


// LIST and LSUB commands
HRESULT IMAP::HrCmdList(string strTag, string strReferenceFolder, string strFindFolder, bool bSubscribedOnly) {
	HRESULT hr = hrSuccess;
	string strAction;
	string strResponse;
	string strPattern;
	string strListProps;
	string strCompare;
	string strFolderPath;
	list<SFolder> lstFolders;

	if (bSubscribedOnly)
		strAction = "LSUB";
	else
		strAction = "LIST";

	if (!lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, strAction+" error no session");
		goto exit;
	}

	if (strFindFolder.empty()) {
		strResponse = strAction+" (\\Noselect) \"";
		strResponse += IMAP_HIERARCHY_DELIMITER;
		strResponse += "\" \"\"";
		HrResponse(strResponse);
		HrResultOk(strTag, strAction+" completed");
		goto exit;
	}

	HrGetSubscribedList();
	// ignore error
	
	// add trailing delimiter to strReferenceFolder
	if (!strReferenceFolder.empty() && strReferenceFolder[strReferenceFolder.length()-1] != IMAP_HIERARCHY_DELIMITER) {
		strReferenceFolder += IMAP_HIERARCHY_DELIMITER;
	}

	strPattern = IMAP2MAPICharset(strReferenceFolder + strFindFolder);
	ToUpper(strPattern);
	
	// Get all folders
	hr = HrGetFolderList(lstFolders);
	if(hr != hrSuccess)
	    goto exit;

	// Loop through all folders to see if they match
	for (list<SFolder>::iterator iFld = lstFolders.begin(); iFld != lstFolders.end(); iFld++) {
		if (bSubscribedOnly && !iFld->bActive && !iFld->bSpecialFolder) {
		    // Folder is not subscribed to
		    continue;
        }

		// Get full path name
		strFolderPath.clear();
		// if path is empty, we're probably dealing the IPM_SUBTREE entry
		if(HrGetFolderPath(iFld, lstFolders, strFolderPath) != hrSuccess || strFolderPath.empty())
		    continue;
		    
		if (!strFolderPath.empty())
			strFolderPath.erase(0,1); // strip / from start of string

        if (MatchFolderPath(strFolderPath, strPattern)) {
			strResponse = (string)"\"" + MAPI2IMAPCharset(strFolderPath) + "\""; // add "" around folder name
			strResponse = (string)"\"" + IMAP_HIERARCHY_DELIMITER + "\" " + strResponse; // prepend folder delimiter

			strListProps = strAction+" (";
			if (!iFld->bMailFolder) {
				strListProps += "\\Noselect ";
			}
			if (!bSubscribedOnly) {
				// don't list flag on LSUB command
				if (iFld->bHasSubfolders) {
					strListProps += "\\HasChildren";
				} else {
					strListProps += "\\HasNoChildren";
				}
			}
			strListProps += ") ";

			strResponse = strListProps + strResponse;

			HrResponse(strResponse);
		}
	}

	HrResultOk(strTag, strAction+" completed");

exit:
	return hr;
}

HRESULT IMAP::HrCmdStatus(string strTag, string strFolder, string strStatusData) {
	HRESULT hr = hrSuccess;
	list<SFolder>::iterator lpFolder;
	IMAPIFolder *lpStatusFolder = NULL;
	vector<string> lstStatusData;
	string strData;
	string strResponse;
	char szBuffer[11];
	ULONG ulCounter;
	ULONG ulMessages = 0;
	ULONG ulUnseen = 0;
	ULONG ulRecent = 0;
	ULONG cStatusData = 0;
	IMAPITable *lpTable = NULL;
	ULONG cValues;
	SizedSPropTagArray(2, sPropsFolderCounters) = { 2, { PR_CONTENT_COUNT, PR_CONTENT_UNREAD } };
	LPSPropValue lpPropCounters = NULL;
	string strIMAPFolder;
	SPropValue sPropMaxID;
	LPSPropValue lpPropMaxID = NULL;
    SRestriction sRestrictRecent;
    
	if (!lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "STATUS error no session");
		goto exit;
	}

	strIMAPFolder = IMAP2MAPICharset(strFolder);

	ToUpper(strStatusData);
	ToUpper(strIMAPFolder);

	hr = HrFindFolder(strIMAPFolder, false, &lpStatusFolder);
	if(hr != hrSuccess) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "STATUS error finding folder");
		goto exit;
	}

	if (!IsMailFolder(lpStatusFolder)) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "STATUS error no mail folder");
		goto exit;
	}

	hr = lpStatusFolder->GetProps((LPSPropTagArray)&sPropsFolderCounters, 0, &cValues, &lpPropCounters);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "STATUS error fetching folder content counters");
		goto exit;
	}
	hr = hrSuccess;

	if (lpPropCounters[0].ulPropTag == PR_CONTENT_COUNT)
		ulMessages = lpPropCounters[0].Value.ul;

	if (lpPropCounters[1].ulPropTag == PR_CONTENT_UNREAD)
		ulUnseen = lpPropCounters[1].Value.ul;

	// split statusdata
	if (strStatusData.size() > 1 && strStatusData.compare(0, 1, "(") == 0) {
		strStatusData.erase(0,1);
		strStatusData.resize(strStatusData.size()-1);
	}

	HrSplitInput(strStatusData, lstStatusData);

	// loop statusdata
	cStatusData = lstStatusData.size();
	strResponse = "STATUS \"";
	strResponse += strFolder;	// strFolder is in upper case, works with all clients?
	strResponse += "\" (";
	for (ulCounter = 0; ulCounter < cStatusData; ulCounter++) {
		strData = lstStatusData[ulCounter];
		strResponse += strData;
		strResponse += " ";

		if (strData == "MESSAGES") {
			snprintf(szBuffer, 10, "%u", ulMessages);
			strResponse += szBuffer;
		} else if (strData == "RECENT") {
            // Get 'recent' count from the table
            HrGetOneProp(lpStatusFolder, PR_EC_IMAP_MAX_ID, &lpPropMaxID);

            if(lpPropMaxID) {
                hr = lpStatusFolder->GetContentsTable(MAPI_DEFERRED_ERRORS, &lpTable);
                if(hr != hrSuccess) {
                    HrResultNo(strTag, "STATUS error getting contents");
                    goto exit;
                }

                sPropMaxID.ulPropTag = PR_EC_IMAP_ID;
                sPropMaxID.Value.ul = lpPropMaxID->Value.ul;
                
                sRestrictRecent.rt = RES_PROPERTY;
                sRestrictRecent.res.resProperty.ulPropTag = PR_EC_IMAP_ID;
                sRestrictRecent.res.resProperty.relop = RELOP_GT;
                sRestrictRecent.res.resProperty.lpProp = &sPropMaxID;
                
                hr = lpTable->Restrict(&sRestrictRecent, TBL_BATCH);
                if(hr != hrSuccess) {
                    HrResultNo(strTag, "STATUS error getting recent");
                    goto exit;
                }
                
                hr = lpTable->GetRowCount(0, &ulRecent);
                if(hr != hrSuccess) {
                    HrResultNo(strTag, "STATUS error getting row count for recent");
                    goto exit;
                }
            } else {
                // No max id, all messages are Recent
                ulRecent = ulMessages;
            }
            
			snprintf(szBuffer, 10, "%u", ulRecent);
			strResponse += szBuffer;
		} else if (strData == "UIDNEXT") {
			snprintf(szBuffer, 10, "%u", m_ulLastUid + 1);
			strResponse += szBuffer;
		} else if (strData == "UIDVALIDITY") {
			snprintf(szBuffer, 10, "%u", 1);
			strResponse += szBuffer;
		} else if (strData == "UNSEEN") {
			snprintf(szBuffer, 10, "%u", ulUnseen);
			strResponse += szBuffer;
		} else {
			strResponse += "NIL";
		}
		
		if (ulCounter+1 < cStatusData)
			strResponse += " ";
	}

	strResponse += ")";
	HrResponse(strResponse);
	HrResultOk(strTag, "STATUS completed");

exit:
    if (lpPropMaxID)
        MAPIFreeBuffer(lpPropMaxID);
        
    if (lpTable)
        lpTable->Release();
        
	if (lpPropCounters)
		MAPIFreeBuffer(lpPropCounters);

	if (lpStatusFolder)
		lpStatusFolder->Release();

	return hr;
}

HRESULT IMAP::HrCmdAppend(string strTag, string strFolder, string strData, string strFlags, string strTime) {
	HRESULT hr = hrSuccess;
	list<SFolder>::iterator lpFolder;
	IMAPIFolder *lpAppendFolder = NULL;
	LPMESSAGE lpMessage = NULL;
	vector<string> lstFlags;
	ULONG ulCounter;
	string strFlag;
	LPSPropValue lpPropVal = NULL;

	if (!lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "APPEND error no session");
		goto exit;
	}

	// strTime and strFlags can be switched when no flags are given by the client
	if (strTime.empty() && strFlags.find_first_of("0123456789 ") == 0) {
		strTime = strFlags;
		strFlags.clear();
	}

	strFolder = IMAP2MAPICharset(strFolder);

	hr = HrFindFolder(strFolder, false, &lpAppendFolder);
	if(hr != hrSuccess) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "[TRYCREATE] APPEND error finding folder");
		goto exit;
	}

	if (!IsMailFolder(lpAppendFolder)) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "APPEND error not a mail folder");
		goto exit;
	}

	hr = lpAppendFolder->CreateMessage(NULL, 0, &lpMessage);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "APPEND error creating message");
		goto exit;
	}

	hr = IMToMAPI(lpSession, lpStore, lpAddrBook, lpMessage, strData, dopt, lpLogger);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "APPEND error converting message");
		goto exit;
	}

	if (strFlags.size() > 2 && strFlags[0] == '(') {
		// remove () around flags
		strFlags.erase(0, 1);
		strFlags.erase(strFlags.size()-1, 1);
	}

	ToUpper(strFlags);
	HrSplitInput(strFlags, lstFlags);

	for (ulCounter = 0; ulCounter < lstFlags.size(); ulCounter++) {
		strFlag = lstFlags[ulCounter];

		if (strFlag == "\\SEEN") {
			if (lpPropVal)
				MAPIFreeBuffer(lpPropVal);
			lpPropVal = NULL;

			if (HrGetOneProp(lpMessage, PR_MESSAGE_FLAGS, &lpPropVal) != hrSuccess) {
				hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) & lpPropVal);
				if (hr != hrSuccess)
					goto exit;

				lpPropVal->ulPropTag = PR_MESSAGE_FLAGS;
				lpPropVal->Value.ul = 0;
			}

			lpPropVal->Value.ul |= MSGFLAG_READ;
			HrSetOneProp(lpMessage, lpPropVal);
		} else if (strFlag == "\\DRAFT") {
			if (lpPropVal)
				MAPIFreeBuffer(lpPropVal);
			lpPropVal = NULL;

			if (HrGetOneProp(lpMessage, PR_MSG_STATUS, &lpPropVal) != hrSuccess) {
				hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) & lpPropVal);
				if (hr != hrSuccess)
					goto exit;

				lpPropVal->ulPropTag = PR_MSG_STATUS;
				lpPropVal->Value.ul = 0;
			}

			lpPropVal->Value.ul |= MSGSTATUS_DRAFT;
			HrSetOneProp(lpMessage, lpPropVal);

			// When saving a draft, also mark it as UNSENT, so webaccess opens the editor, not the viewer
			MAPIFreeBuffer(lpPropVal);
			lpPropVal = NULL;

			if (HrGetOneProp(lpMessage, PR_MESSAGE_FLAGS, &lpPropVal) != hrSuccess) {
				hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) & lpPropVal);
				if (hr != hrSuccess)
					goto exit;

				lpPropVal->ulPropTag = PR_MESSAGE_FLAGS;
				lpPropVal->Value.ul = 0;
			}

			lpPropVal->Value.ul |= MSGFLAG_UNSENT;
			HrSetOneProp(lpMessage, lpPropVal);
		} else if (strFlag == "\\FLAGGED") {
			if (lpPropVal == NULL) {
				hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) & lpPropVal);
				if (hr != hrSuccess)
					goto exit;
			}

			lpPropVal->ulPropTag = PR_FLAG_STATUS;
			lpPropVal->Value.ul = 2;
			HrSetOneProp(lpMessage, lpPropVal);
		} else if (strFlag == "\\ANSWERED") {
			if (lpPropVal)
				MAPIFreeBuffer(lpPropVal);
			lpPropVal = NULL;

			if (HrGetOneProp(lpMessage, PR_MSG_STATUS, &lpPropVal) != hrSuccess) {
				hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) & lpPropVal);
				if (hr != hrSuccess)
					goto exit;

				lpPropVal->ulPropTag = PR_MSG_STATUS;
				lpPropVal->Value.ul = 0;
			}

			lpPropVal->Value.ul |= MSGSTATUS_ANSWERED;
			HrSetOneProp(lpMessage, lpPropVal);

			if (lpPropVal)
				MAPIFreeBuffer(lpPropVal);
			lpPropVal = NULL;

			hr = MAPIAllocateBuffer(sizeof(SPropValue) * 3, (LPVOID *) & lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpPropVal[0].ulPropTag = PR_LAST_VERB_EXECUTED;
			lpPropVal[0].Value.ul = NOTEIVERB_REPLYTOSENDER;
			lpPropVal[1].ulPropTag = PR_LAST_VERB_EXECUTION_TIME;
			GetSystemTimeAsFileTime(&lpPropVal[1].Value.ft);
			lpPropVal[2].ulPropTag = PR_ICON_INDEX;
			lpPropVal[2].Value.ul = ICON_MAIL_REPLIED;

			hr = lpMessage->SetProps(3, lpPropVal, NULL);
			if (hr != hrSuccess)
				goto exit;
		} else if (strFlag == "\\DELETED") {
			if (lpPropVal)
				MAPIFreeBuffer(lpPropVal);
			lpPropVal = NULL;

			if (HrGetOneProp(lpMessage, PR_MSG_STATUS, &lpPropVal) != hrSuccess) {
				hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) & lpPropVal);
				if (hr != hrSuccess)
					goto exit;

				lpPropVal->ulPropTag = PR_MSG_STATUS;
				lpPropVal->Value.ul = 0;
			}

			lpPropVal->Value.ul |= MSGSTATUS_DELMARKED;
			HrSetOneProp(lpMessage, lpPropVal);
		}
	}

	// set time
	if (lpPropVal == NULL) {
		hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) & lpPropVal);
		if (hr != hrSuccess)
			goto exit;
	}

	if (!strTime.empty()) {
		lpPropVal->Value.ft = StringToFileTime(strTime);
		lpPropVal->ulPropTag = PR_MESSAGE_DELIVERY_TIME;
		HrSetOneProp(lpMessage, lpPropVal);
		
		lpPropVal->Value.ft = StringToFileTime(strTime, true);
		lpPropVal->ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
		HrSetOneProp(lpMessage, lpPropVal);
	}


	hr = lpMessage->SaveChanges(KEEP_OPEN_READWRITE | FORCE_SAVE);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "APPEND error saving message");
		goto exit;
	}

	if (strCurrentFolder == strFolder) {
	    // Fixme, add the appended message instead of HrRefreshFolderMails; the message is now seen as Recent
		HrRefreshFolderMails(false, !bCurrentFolderReadOnly, false, NULL);
	}

	HrResultOk(strTag, "APPEND completed");

exit:
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	if (lpMessage)
		lpMessage->Release();

	if (lpAppendFolder)
		lpAppendFolder->Release();

	return hr;
}

HRESULT IMAP::HrCmdCheck(string strTag) {
    HRESULT hr = hrSuccess;
    
	hr = HrRefreshFolderMails(false, !bCurrentFolderReadOnly, false, NULL);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "CHECK error reading folder messages");
		goto exit;
	}
    
	HrResultOk(strTag, "CHECK completed");

exit:
	return hr;
}

HRESULT IMAP::HrCmdClose(string strTag) {
	HRESULT hr = hrSuccess;
	IMAPIFolder *lpFolder = NULL;
	ULONG ulMailnr;
	ENTRYLIST sEntryList;
	LPSRestriction lpRootRestrict = NULL;
	LPSRestriction lpRestriction;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	enum { EID, NUM_COLS };
	SizedSPropTagArray(NUM_COLS, spt) = { NUM_COLS, {PR_ENTRYID} };

	sEntryList.lpbin = NULL;

	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "CLOSE error no folder");
		goto exit;
	}

	if (bCurrentFolderReadOnly) {
	    strCurrentFolder.clear();
		HrResultOk(strTag, "CLOSE completed");
		goto exit;
	}
	
	hr = HrFindFolder(strCurrentFolder, bCurrentFolderReadOnly, &lpFolder);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "CLOSE error opening folder");
		goto exit;
	}
	hr = lpFolder->GetContentsTable(MAPI_DEFERRED_ERRORS, &lpTable);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "CLOSE error opening folder contents");
		goto exit;
	}

	hr = MAPIAllocateBuffer(sizeof(SRestriction), (LPVOID *) &lpRootRestrict);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "CLOSE error Allocating buffer");
		goto exit;
	}
	hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpRestriction);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "CLOSE error Allocating buffer");
		goto exit;
	}
	lpRootRestrict->rt = RES_AND;
	lpRootRestrict->res.resAnd.cRes = 2;
	lpRootRestrict->res.resAnd.lpRes = lpRestriction;

	lpRestriction[0].rt = RES_EXIST;
	lpRestriction[0].res.resExist.ulReserved1 = 0;
	lpRestriction[0].res.resExist.ulReserved2 = 0;
	lpRestriction[0].res.resExist.ulPropTag = PR_MSG_STATUS;
	lpRestriction[1].rt = RES_BITMASK;
	lpRestriction[1].res.resBitMask.relBMR = BMR_NEZ;
	lpRestriction[1].res.resBitMask.ulPropTag = PR_MSG_STATUS;
	lpRestriction[1].res.resBitMask.ulMask = MSGSTATUS_DELMARKED;

	hr = HrQueryAllRows(lpTable, (LPSPropTagArray) &spt, lpRootRestrict, NULL, 0, &lpRows);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "CLOSE error queryring all rows");
		goto exit;
	}
	if (lpRows->cRows == 0) {
		HrResultOk(strTag, "CLOSE completed");
		goto exit;
	}
	sEntryList.cValues = lpRows->cRows;
	MAPIAllocateBuffer(sizeof(SBinary) * lpRows->cRows, (LPVOID *) &sEntryList.lpbin);

	for (ulMailnr = 0; ulMailnr < lpRows->cRows; ulMailnr++) {
		sEntryList.lpbin[ulMailnr] = lpRows->aRow[ulMailnr].lpProps[EID].Value.bin;
	}

	hr = lpFolder->DeleteMessages(&sEntryList, 0, NULL, 0);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "CLOSE error deleting messages");
		goto exit;
	}

	HrResultOk(strTag, "CLOSE completed");

exit:
	if (sEntryList.lpbin)
		MAPIFreeBuffer(sEntryList.lpbin);

	if (lpRootRestrict)
		MAPIFreeBuffer(lpRootRestrict);

	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	if (lpFolder)
		lpFolder->Release();

	return hr;
}

HRESULT IMAP::HrCmdExpunge(string strTag) {
	HRESULT hr = hrSuccess;
	IMAPIFolder *lpFolder = NULL;
	ULONG ulMailnr;
	ENTRYLIST sEntryList;
	LPSRestriction lpRootRestrict = NULL;
	LPSRestriction lpRestriction;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	enum { EID, NUM_COLS };
	SizedSPropTagArray(NUM_COLS, spt) = { NUM_COLS, {PR_ENTRYID} };

	sEntryList.lpbin = NULL;

	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "EXPUNGE error no folder");
		goto exit;
	}

	if (bCurrentFolderReadOnly) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "EXPUNGE error folder read only");
		goto exit;
	}

	hr = HrFindFolder(strCurrentFolder, bCurrentFolderReadOnly, &lpFolder);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "EXPUNGE error opening folder");
		goto exit;
	}

	hr = lpFolder->GetContentsTable(MAPI_DEFERRED_ERRORS , &lpTable);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "EXPUNGE error opening folder contents");
		goto exit;
	}

	hr = MAPIAllocateBuffer(sizeof(SRestriction), (LPVOID *) &lpRootRestrict);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "EXPUNGE error Allocating buffer");
		goto exit;
	}

	hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpRestriction);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "EXPUNGE error Allocating buffer");
		goto exit;
	}

	lpRootRestrict->rt = RES_AND;
	lpRootRestrict->res.resAnd.cRes = 2;
	lpRootRestrict->res.resAnd.lpRes = lpRestriction;

	lpRestriction[0].rt = RES_EXIST;
	lpRestriction[0].res.resExist.ulReserved1 = 0;
	lpRestriction[0].res.resExist.ulReserved2 = 0;
	lpRestriction[0].res.resExist.ulPropTag = PR_MSG_STATUS;
	lpRestriction[1].rt = RES_BITMASK;
	lpRestriction[1].res.resBitMask.relBMR = BMR_NEZ;
	lpRestriction[1].res.resBitMask.ulPropTag = PR_MSG_STATUS;
	lpRestriction[1].res.resBitMask.ulMask = MSGSTATUS_DELMARKED;

	hr = HrQueryAllRows(lpTable, (LPSPropTagArray) &spt, lpRootRestrict, NULL, 0, &lpRows);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "EXPUNGE error queryring rows");
		goto exit;
	}

	if(lpRows->cRows) {
        sEntryList.cValues = 0;
        MAPIAllocateBuffer(sizeof(SBinary) * lpRows->cRows, (LPVOID *) &sEntryList.lpbin);

        for (ulMailnr = 0; ulMailnr < lpRows->cRows; ulMailnr++) {
            // FIXME used to unmark deleted
            sEntryList.lpbin[sEntryList.cValues++] = lpRows->aRow[ulMailnr].lpProps[EID].Value.bin;
        }

        hr = lpFolder->DeleteMessages(&sEntryList, 0, NULL, 0);
        if (hr != hrSuccess) {
            HrResultNo(strTag, "EXPUNGE error deleting messages");
            goto exit;
        }
    }
    
	// Let HrRefreshFolderMails output the actual EXPUNGEs
	HrRefreshFolderMails(false, !bCurrentFolderReadOnly, false, NULL);

	HrResultOk(strTag, "EXPUNGE completed");

exit:
	if (sEntryList.lpbin)
		MAPIFreeBuffer(sEntryList.lpbin);

	if (lpRootRestrict)
		MAPIFreeBuffer(lpRootRestrict);

	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	if (lpFolder)
		lpFolder->Release();

	return hr;
}

HRESULT IMAP::HrCmdSearch(string strTag, vector<string> &lstSearchCriteria, bool bUidMode) {
	HRESULT hr = hrSuccess;
	list<ULONG> lstMailnr;
	list<ULONG>::iterator lpMailnr;
	ULONG ulCriterianr = 0;
	string strResponse;
	char szBuffer[33];
	ECIConv *iconv = NULL;
	string strMode;

	if (bUidMode)
		strMode = "UID ";

	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, strMode+"SEARCH error no folder");
		goto exit;
	}

	// don't support other charsets
	if (lstSearchCriteria[0] == "CHARSET") {
		if (lstSearchCriteria[1] != "WINDOWS-1252") {
			iconv = new ECIConv("windows-1252", lstSearchCriteria[1]);
			if (!iconv->canConvert()) {
				HrResultNo(strTag, "[BADCHARSET WINDOWS-1252] "+strMode+"SEARCH charset not supported");
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}
		}
		ulCriterianr += 2;
	}

	hr = HrSearch(lstSearchCriteria, ulCriterianr, lstMailnr, iconv);
	if (hr != hrSuccess) {
		HrResultNo(strTag, strMode+"SEARCH error");
		goto exit;
	}

	strResponse = "SEARCH";

	for (lpMailnr = lstMailnr.begin(); lpMailnr != lstMailnr.end(); lpMailnr++) {
		snprintf(szBuffer, 32, " %u", bUidMode ? lstFolderMailEIDs[*lpMailnr].ulUid : (*lpMailnr) + 1);
		strResponse += szBuffer;
	}

	HrResponse(strResponse);
	HrResultOk(strTag, strMode+"SEARCH completed");

exit:
	if (iconv)
		delete iconv;

	return hr;
}

HRESULT IMAP::HrCmdFetch(string strTag, string strSeqSet, string strMsgDataItemNames, bool bUidMode) {
	HRESULT hr = hrSuccess;
	vector<string> lstDataItems;
	list<ULONG> lstMails;
	ULONG ulCurrent = 0;
	bool bFound = false;
	string strMode;

	if (bUidMode)
		strMode = "UID ";

	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, strMode+"FETCH error no folder");
		goto exit;
	}

	HrGetDataItems(strMsgDataItemNames, lstDataItems);
	if (bUidMode) {
		for (ulCurrent = 0; !bFound && ulCurrent < lstDataItems.size(); ulCurrent++) {
			if (lstDataItems[ulCurrent] == "UID")
				bFound = true;
		}
		if (!bFound)
			lstDataItems.push_back("UID");
	}

	if (bUidMode)
		hr = HrParseSeqUidSet(strSeqSet, lstMails);
	else
		hr = HrParseSeqSet(strSeqSet, lstMails);
	if (hr != hrSuccess) {
		HrResultNo(strTag, strMode+"FETCH sequence parse error in: " + strSeqSet);
		goto exit;
	}

	hr = HrPropertyFetch(lstMails, lstDataItems);
	if (hr != hrSuccess)
		HrResultNo(strTag, strMode+"FETCH failed");
	else
		HrResultOk(strTag, strMode+"FETCH completed");

exit:
	return hr;
}

HRESULT IMAP::HrCmdStore(string strTag, string strSeqSet, string strMsgDataItemName, string strMsgDataItemValue, bool bUidMode) {
	HRESULT hr = hrSuccess;
	list<ULONG> lstMails;
	vector<string> lstDataItems;
	string strMode;

	if (bUidMode)
		strMode = "UID ";

	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, strMode+"STORE error no folder");
		goto exit;
	}

	if (bCurrentFolderReadOnly) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, strMode+"STORE error folder read only");
		goto exit;
	}

	lstDataItems.push_back("FLAGS");
	if (bUidMode)
		lstDataItems.push_back("UID");

	if (bUidMode)
		hr = HrParseSeqUidSet(strSeqSet, lstMails);
	else
		hr = HrParseSeqSet(strSeqSet, lstMails);
	if (hr != hrSuccess) {
		HrResultNo(strTag, strMode+"STORE sequence parse error in: " + strSeqSet);
		goto exit;
	}

	hr = HrStore(lstMails, strMsgDataItemName, strMsgDataItemValue);
	if (hr != MAPI_E_NOT_ME) {
		HrPropertyFetch(lstMails, lstDataItems);
		hr = hrSuccess;
	}

	HrResultOk(strTag, strMode+"STORE completed");

exit:
	return hr;
}

HRESULT IMAP::HrCmdCopy(string strTag, string strSeqSet, string strFolder, bool bUidMode) {
	HRESULT hr = hrSuccess;
	list<ULONG> lstMails;
	string strMode;

	if (bUidMode)
		strMode = "UID ";

	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, strMode+"COPY error no folder");
		goto exit;
	}

	if (bUidMode)
		hr = HrParseSeqUidSet(strSeqSet, lstMails);
	else
		hr = HrParseSeqSet(strSeqSet, lstMails);
	if (hr != hrSuccess) {
		HrResultNo(strTag, strMode+"COPY sequence parse error in: " + strSeqSet);
		goto exit;
	}

	hr = HrCopy(lstMails, strFolder);
	if (hr == MAPI_E_NOT_FOUND) {
		HrResultNo(strTag, "[TRYCREATE] "+strMode+"COPY folder not found");
		goto exit;
	} else if (hr != hrSuccess) {
		HrResultNo(strTag, strMode+"COPY error");
		goto exit;
	}

	HrResultOk(strTag, strMode+"COPY completed");

exit:
	return hr;
}

HRESULT IMAP::HrCmdUidXaolMove(string strTag, string strSeqSet, string strFolder) {
	HRESULT hr = hrSuccess;
	list<ULONG> lstMails;

	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		HrResultNo(strTag, "UID XAOL-MOVE error no folder");
		goto exit;
	}

	hr = HrParseSeqUidSet(strSeqSet, lstMails);
	if (hr != hrSuccess) {
		HrResultNo(strTag, "UID XAOL-MOVE sequence parse error in: " + strSeqSet);
		goto exit;
	}

	hr = HrCopy(lstMails, strFolder, true);
	if (hr == MAPI_E_NOT_FOUND) {
		HrResultNo(strTag, "[TRYCREATE] UID XAOL-MOVE folder not found");
		goto exit;
	} else if (hr != hrSuccess) {
		HrResultNo(strTag, "UID XAOL-MOVE error");
		goto exit;
	}

	// Let HrRefreshFolderMails output the actual EXPUNGEs
	HrRefreshFolderMails(false, !bCurrentFolderReadOnly, false, NULL);

	HrResultOk(strTag, "UID XAOL-MOVE completed");

exit:
	return hr;
}

std::string IMAP::PropsToFlags(LPSPropValue lpProps, unsigned int cValues, bool bRecent) {
	string strFlags;
	LPSPropValue lpMessageFlags = PpropFindProp(lpProps, cValues, PR_MESSAGE_FLAGS);
	LPSPropValue lpFlagStatus = PpropFindProp(lpProps, cValues, PR_FLAG_STATUS);
	LPSPropValue lpMsgStatus = PpropFindProp(lpProps, cValues, PR_MSG_STATUS);
	LPSPropValue lpLastVerb = PpropFindProp(lpProps, cValues, PR_LAST_VERB_EXECUTED);

	if (lpMessageFlags && lpMessageFlags->Value.ul & MSGFLAG_READ) {
		strFlags += "\\Seen ";
	}

	if (lpFlagStatus && lpFlagStatus->Value.ul != 0) {
		strFlags += "\\Flagged ";
	}

	if (lpLastVerb) {
		if ((lpLastVerb->Value.ul == NOTEIVERB_REPLYTOSENDER || lpLastVerb->Value.ul == NOTEIVERB_REPLYTOALL)) {
			strFlags += "\\Answered ";
		}

		// there is no flag in imap for forwards. thunderbird uses the custom flag $Forwarded,
		// but we currently do not support custom flags
		/*
		if (lpLastVerb->Value.ul == NOTEIVERB_FORWARD) {
			strFlags += "$Forwarded ";
		}
		*/
	}

	if (lpMsgStatus) {
		if (lpMsgStatus->Value.ul & MSGSTATUS_DRAFT) {
			strFlags += "\\Draft ";
		}

		if (!lpLastVerb && lpMsgStatus->Value.ul & MSGSTATUS_ANSWERED) {
			strFlags += "\\Answered ";
		}

		if (lpMsgStatus->Value.ul & MSGSTATUS_DELMARKED) {
			strFlags += "\\Deleted ";
		}
	}
	
	if (bRecent)
	    strFlags += "\\Recent ";
	
	// strip final space
	if (!strFlags.empty()) {
		strFlags.resize(strFlags.size() - 1);
	}

	return strFlags;
}

std::string IMAP::PropsToEmailAddress(LPSPropValue lpProps, unsigned int cValues, ULONG ulEmailPropTag, ULONG ulNamePropTag)
{
    LPSPropValue lpPropEmail = PpropFindProp(lpProps, cValues, ulEmailPropTag);
    LPSPropValue lpPropName = PpropFindProp(lpProps, cValues, ulNamePropTag);
    std::string strAddress;
    
    if(lpPropEmail && lpPropName && strcmp(lpPropEmail->Value.lpszA, lpPropName->Value.lpszA) == 0) {
        strAddress = (string)"<" + lpPropEmail->Value.lpszA + ">";
    } else if(lpPropEmail && lpPropName) {
        strAddress = (string)"\"" + lpPropName->Value.lpszA + "\" <" + lpPropEmail->Value.lpszA + ">";
    } else if(lpPropEmail) {
        strAddress = (string)"<" + lpPropEmail->Value.lpszA + ">";
    } else {
        // Only a name ??
    }
    
    return strAddress;
}

LONG __stdcall IMAPIdleAdviseCallback(void *lpContext, ULONG cNotif, LPNOTIFICATION lpNotif) {
	IMAP *lpIMAP = (IMAP*)lpContext;
	string strFlags;
	ULONG ulMailNr = 0;
	ULONG ulRecent = 0;
	bool bReload = false;
	vector<IMAP::SMail> oldMails;
	vector<IMAP::SMail>::iterator iterMail;
	vector<IMAP::SMail>::reverse_iterator riterOld, riterNew;
	enum { EID, IKEY, IMAPID, MESSAGE_FLAGS, FLAG_STATUS, MSG_STATUS, LAST_VERB, NUM_COLS };
	IMAP::SMail sMail;

	{
		// modify:
		// * [mailnr] FETCH (FLAGS (\Seen \Recent \Flagged))
		// delete:
		// * [mailnr] EXPUNGE
		// new mail/add:
		// * [total] exists (optional?)
		// * 1 recent
	}

	if (!lpIMAP)
		return MAPI_E_CALL_FAILED;

	pthread_mutex_lock(&lpIMAP->m_mIdleLock);

	if (!lpIMAP->m_bIdleMode) {
		pthread_mutex_unlock(&lpIMAP->m_mIdleLock);
		return MAPI_E_CALL_FAILED;
	}

	for (ULONG i = 0; i < cNotif && !bReload; i++) {

		if (lpNotif[i].ulEventType != fnevTableModified)
			continue;

		switch (lpNotif[i].info.tab.ulTableEvent) {
		case TABLE_ROW_ADDED:
			sMail.sEntryID = BinaryArray(lpNotif[i].info.tab.row.lpProps[EID].Value.bin);
			sMail.sInstanceKey = BinaryArray(lpNotif[i].info.tab.propIndex.Value.bin);
			sMail.bRecent = true;

			if (lpNotif[i].info.tab.row.lpProps[IMAPID].ulPropTag == PR_EC_IMAP_ID) {
				sMail.ulUid = lpNotif[i].info.tab.row.lpProps[IMAPID].Value.ul;
			}

			sMail.strFlags = lpIMAP->PropsToFlags(lpNotif[i].info.tab.row.lpProps, lpNotif[i].info.tab.row.cValues, true);
			lpIMAP->lstFolderMailEIDs.push_back(sMail);
			lpIMAP->m_ulLastUid = max(lpIMAP->m_ulLastUid, sMail.ulUid);
			ulRecent++;
			break;

		case TABLE_ROW_DELETED:
			// find number and print N EXPUNGE
			if (lpNotif[i].info.tab.propIndex.ulPropTag == PR_INSTANCE_KEY) {
			    for(iterMail = lpIMAP->lstFolderMailEIDs.begin(); iterMail != lpIMAP->lstFolderMailEIDs.end(); iterMail++) {
			        if(iterMail->sInstanceKey == BinaryArray(lpNotif[i].info.tab.propIndex.Value.bin))
			            break;
			    }
			    
				if (iterMail != lpIMAP->lstFolderMailEIDs.end()) {
					ulMailNr = iterMail - lpIMAP->lstFolderMailEIDs.begin();

					// remove mail from list
					lpIMAP->HrResponse(stringify(ulMailNr+1) + " EXPUNGE");

					lpIMAP->lstFolderMailEIDs.erase(iterMail);
				}
			}
			break;

		case TABLE_ROW_MODIFIED:
			// find number and print N FETCH (FLAGS (flags...))
			strFlags.clear();

			if (lpNotif[i].info.tab.row.lpProps[IMAPID].ulPropTag == PR_EC_IMAP_ID) {
				iterMail = find(lpIMAP->lstFolderMailEIDs.begin(), lpIMAP->lstFolderMailEIDs.end(), lpNotif[i].info.tab.row.lpProps[IMAPID].Value.ul);
				// not found probably means the client needs to sync
				if (iterMail != lpIMAP->lstFolderMailEIDs.end()) {
					ulMailNr = iterMail - lpIMAP->lstFolderMailEIDs.begin();

					strFlags = lpIMAP->PropsToFlags(lpNotif[i].info.tab.row.lpProps, lpNotif[i].info.tab.row.cValues, iterMail->bRecent);

					lpIMAP->HrResponse(stringify(ulMailNr+1) + " FETCH (FLAGS ("+strFlags+"))");
				}
			}
			break;

		case TABLE_RELOAD:
			// TABLE_RELOAD is unused in Zarafa
		case TABLE_CHANGED:
            lpIMAP->HrRefreshFolderMails(false, !lpIMAP->bCurrentFolderReadOnly, false, NULL);
		    break;
		};
	}

	if (ulRecent) {
		lpIMAP->HrResponse(stringify(ulRecent) + " RECENT");
		lpIMAP->HrResponse(stringify(lpIMAP->lstFolderMailEIDs.size()) + " EXISTS");
	}

	pthread_mutex_unlock(&lpIMAP->m_mIdleLock);

	return S_OK;
}

HRESULT IMAP::HrCmdIdle(string strTag) {
	HRESULT hr = hrSuccess;
	LPMAPIFOLDER lpFolder = NULL;
	enum { EID, IKEY, IMAPID, MESSAGE_FLAGS, FLAG_STATUS, MSG_STATUS, LAST_VERB, NUM_COLS };
	SizedSPropTagArray(NUM_COLS, spt) = { NUM_COLS, {PR_ENTRYID, PR_INSTANCE_KEY, PR_EC_IMAP_ID, PR_MESSAGE_FLAGS, PR_FLAG_STATUS, PR_MSG_STATUS, PR_LAST_VERB_EXECUTED} };


	// Outlook (express) IDLEs without selecting a folder
	// when sending an error from this command, Outlook loops on the IDLE command forever :(
	// therefor, we can never return an HrResultBad() or ...No() here, so we always "succeed"

	m_strIdleTag = strTag;
	m_bIdleMode = true;

	if (strCurrentFolder.empty() || !lpSession) {
		hr = HrResponseContinuation("empty idle, nothing is going to happen");
		goto exit;
	}

	hr = HrFindFolder(strCurrentFolder, bCurrentFolderReadOnly, &lpFolder);
	if (hr != hrSuccess) {
		HrResponseContinuation("Can't open selected folder to idle in");
		goto exit;
	}

	hr = lpFolder->GetContentsTable(0, &m_lpIdleTable);
	if (hr != hrSuccess) {
		HrResponseContinuation("Can't open selected contents table to idle in");
		goto exit;
	}

	hr = m_lpIdleTable->SetColumns((LPSPropTagArray) &spt, 0);
	if (hr != hrSuccess) {
		HrResponseContinuation("Can't select colums on selected contents table for idle information");
		goto exit;
	}

	hr = HrAllocAdviseSink(IMAPIdleAdviseCallback, (void*)this, &m_lpIdleAdviseSink);
	if (hr != hrSuccess) {
		HrResponseContinuation("Can't allocate memory to idle");
		goto exit;
	}

	pthread_mutex_lock(&m_mIdleLock);

	hr = m_lpIdleTable->Advise(fnevTableModified, m_lpIdleAdviseSink, &m_ulIdleAdviseConnection);
	if (hr != hrSuccess) {
		HrResponseContinuation("Can't advise on current selected folder");
		pthread_mutex_unlock(&m_mIdleLock);
		goto exit;
	}

	// \o/ we really succeeded this time
	hr = HrResponseContinuation("waiting for notifications");

	pthread_mutex_unlock(&m_mIdleLock);


exit:
	if (hr != hrSuccess) {
		if (m_ulIdleAdviseConnection && m_lpIdleTable) {
			m_lpIdleTable->Unadvise(m_ulIdleAdviseConnection);
			m_ulIdleAdviseConnection = 0;
		}
		if (m_lpIdleAdviseSink) {
			m_lpIdleAdviseSink->Release();
			m_lpIdleAdviseSink = NULL;
		}
		if (m_lpIdleTable) {
			m_lpIdleTable->Release();
			m_lpIdleTable = NULL;
		}
	}

	if (lpFolder)
		lpFolder->Release();

	return hr;
}

HRESULT IMAP::HrDone(bool bSendResponse) {
	HRESULT hr = hrSuccess;

	// TODO: maybe add sleep here, so thunderbird gets all notifications?

	pthread_mutex_lock(&m_mIdleLock);

	if (bSendResponse) {
		if (m_bIdleMode)
			hr = HrResultOk(m_strIdleTag, "IDLE complete");
		else
			hr = HrResultBad(m_strIdleTag, "was not idling");
	}

	if (m_ulIdleAdviseConnection && m_lpIdleTable) {
		m_lpIdleTable->Unadvise(m_ulIdleAdviseConnection);
		m_ulIdleAdviseConnection = 0;
	}

	if (m_lpIdleAdviseSink)
		m_lpIdleAdviseSink->Release();
	m_lpIdleAdviseSink = NULL;

	m_ulIdleAdviseConnection = 0;
	m_bIdleMode = false;
	m_strIdleTag.clear();

	if (m_lpIdleTable)
		m_lpIdleTable->Release();
	m_lpIdleTable = NULL;

	pthread_mutex_unlock(&m_mIdleLock);

	return hr;
}

/** 
 * @brief Handles the NAMESPACE command
 *
 * We currently only support the user namespace.
 * 
 * @param[in] strTag the IMAP tag for this command
 * 
 * @return MAPI Error code
 */
HRESULT IMAP::HrCmdNamespace(const string &strTag) {
	HRESULT hr = hrSuccess;

	hr = HrResponse(string("NAMESPACE ((\"\" \"") + IMAP_HIERARCHY_DELIMITER + "\")) NIL NIL");
	if (hr != hrSuccess)
		goto exit;

	hr = HrResultOk(strTag, "NAMESPACE Completed");
	if (hr != hrSuccess)
		goto exit;

exit:
	return hr;
}

/** 
 * Returns the idle state.
 * 
 * @return If the last client command was IDLE
 */
bool IMAP::isIdle() {
	return m_bIdleMode;
}

HRESULT IMAP::HrResponse(string strResponse) {
	strResponse = (string) "* " + strResponse;
    // Early cutoff of debug messages. This means the current process's config
    // determines if we log debug info (so HUP will only affect new processes if
    // you want debug output)
    if(lpLogger->Log(EC_LOGLEVEL_DEBUG))
      	lpLogger->Log(EC_LOGLEVEL_DEBUG, "> %s", strResponse.c_str());
	return lpChannel->HrWriteLine(strResponse);
}

HRESULT IMAP::HrResponseContinuation(string strResponse) {
	strResponse = (string) "+ " + strResponse;
    if(lpLogger->Log(EC_LOGLEVEL_DEBUG))
    	lpLogger->Log(EC_LOGLEVEL_DEBUG, "> %s", strResponse.c_str());
	return lpChannel->HrWriteLine(strResponse);
}

HRESULT IMAP::HrResultOk(string strTag, string strResponse) {
	strResponse = strTag + " OK " + strResponse;
    if(lpLogger->Log(EC_LOGLEVEL_DEBUG))
    	lpLogger->Log(EC_LOGLEVEL_DEBUG, "> %s", strResponse.c_str());
	return lpChannel->HrWriteLine(strResponse);
}

HRESULT IMAP::HrResultNo(string strTag, string strResponse) {
	strResponse = strTag + " NO " + strResponse;
    if(lpLogger->Log(EC_LOGLEVEL_DEBUG))
    	lpLogger->Log(EC_LOGLEVEL_DEBUG, "> %s", strResponse.c_str());
	return lpChannel->HrWriteLine(strResponse);
}

HRESULT IMAP::HrResultBad(string strTag, string strResponse) {
	strResponse = strTag + " BAD " + strResponse;
    if(lpLogger->Log(EC_LOGLEVEL_DEBUG))
    	lpLogger->Log(EC_LOGLEVEL_DEBUG, "> %s", strResponse.c_str());
	return lpChannel->HrWriteLine(strResponse);
}

HRESULT IMAP::HrGetFolderList(list<SFolder> &lstFolders) {
	HRESULT hr = hrSuccess;
	LPSPropValue lpPropVal = NULL;
	ULONG cbEntryID;
	LPENTRYID lpEntryID = NULL;
	list<SFolder>::iterator lpFolder;

	lstFolders.clear();

	hr = HrGetOneProp(lpStore, PR_IPM_SUBTREE_ENTRYID, &lpPropVal);
	if (hr != hrSuccess) {
		goto exit;
	}

	// make folders list from IPM_SUBTREE
	hr = HrGetSubTree(lstFolders, lpPropVal->Value.bin, "", lstFolders.end());
	if (hr != hrSuccess) {
		goto exit;
	}

	hr = lpStore->GetReceiveFolder("IPM", 0, &cbEntryID, &lpEntryID, NULL);
	if (hr != hrSuccess) {
		goto exit;
	}

	// find the inbox, and name it INBOX
	for (lpFolder = lstFolders.begin(); lpFolder != lstFolders.end(); lpFolder++) {
		if (cbEntryID == lpFolder->sEntryID.cb && memcmp(lpFolder->sEntryID.lpb, lpEntryID, cbEntryID) == 0) {
			lpFolder->strFolderName = "INBOX";
			break;
		}
	}
	
	if (lpPropVal){
		MAPIFreeBuffer(lpPropVal);
		lpPropVal = NULL;
	}

	if(!lpPublicStore)
		goto exit;

	hr = HrGetOneProp(lpPublicStore, PR_IPM_PUBLIC_FOLDERS_ENTRYID, &lpPropVal);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, "Public store is enabled in configuation, but Public Folders inside public store could not be found.");
		hr = hrSuccess;
		goto exit;
	}

	// make public folder folders list
	hr = HrGetSubTree(lstFolders, lpPropVal->Value.bin, PUBLIC_FOLDERS_NAME, --lstFolders.end());
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_WARNING, "Public store is enabled in configuation, but Public Folders inside public store could not be found.");
		hr = hrSuccess;
	}

exit:
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	if (lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	return hr;
}

bool IMAP::IsSubscribed(BinaryArray sEntryId) {
	vector<BinaryArray>::iterator iFld;

	iFld = find(m_vSubscriptions.begin(), m_vSubscriptions.end(), sEntryId);
	return iFld != m_vSubscriptions.end();
}

HRESULT IMAP::HrGetSubscribedList() {
	HRESULT hr = hrSuccess;
	LPSTREAM lpStream = NULL;
	ULONG size, i;
	ULONG read;
	ULONG cb = 0;
	BYTE *lpb = NULL;
	
	m_vSubscriptions.clear();

	hr = lpStore->OpenProperty(PR_EC_IMAP_SUBSCRIBED, &IID_IStream, 0, 0, (LPUNKNOWN*)&lpStream);
	if (hr != hrSuccess)
		goto exit;

	hr = lpStream->Read(&size, sizeof(ULONG), &read);
	if (hr != hrSuccess || read != sizeof(ULONG))
		goto exit;

	if (size == 0)
		goto exit;

	for (i = 0; i < size; i++) {
		hr = lpStream->Read(&cb, sizeof(ULONG), &read);
		if (hr != hrSuccess || read != sizeof(ULONG))
			goto exit;

		lpb = new BYTE[cb];

		hr = lpStream->Read(lpb, cb, &read);
		if (hr != hrSuccess || read != cb) {
		    hr = MAPI_E_NOT_FOUND;
			goto exit;
        }

		m_vSubscriptions.push_back(BinaryArray(lpb, cb));
		
		delete [] lpb;
		lpb = NULL;
	}

exit:
    if (lpb)
        delete [] lpb;
        
	if (lpStream)
		lpStream->Release();

	return hr;
}

HRESULT IMAP::HrSetSubscribedList() {
	HRESULT hr = hrSuccess;
	LPSTREAM lpStream = NULL;
	vector<BinaryArray>::iterator iFolders;
	ULONG written;
	ULONG size;
	ULARGE_INTEGER liZero = {{0, 0}};

	hr = lpStore->OpenProperty(PR_EC_IMAP_SUBSCRIBED, &IID_IStream, STGM_TRANSACTED, MAPI_CREATE | MAPI_MODIFY, (LPUNKNOWN*)&lpStream);
	if (hr != hrSuccess)
		goto exit;
		
    lpStream->SetSize(liZero);

	size = m_vSubscriptions.size();
	hr = lpStream->Write(&size, sizeof(ULONG), &written);
	if (hr != hrSuccess)
		goto exit;

	for (iFolders = m_vSubscriptions.begin(); iFolders != m_vSubscriptions.end(); iFolders++) {
		hr = lpStream->Write(&(iFolders->cb), sizeof(ULONG), &written);
		if (hr != hrSuccess)
			goto exit;

		hr = lpStream->Write(iFolders->lpb, iFolders->cb, &written);
		if (hr != hrSuccess)
			goto exit;
	}

	hr = lpStream->Commit(0);

exit:
	if (lpStream)
		lpStream->Release();

	return hr;
}

HRESULT IMAP::HrMakeSpecialsList() {
	HRESULT			hr = hrSuccess;
	ULONG			cbEntryID = 0;
	LPENTRYID		lpEntryID = NULL;
	LPMAPIFOLDER	lpInbox = NULL;
	ULONG			ulObjType = 0;
	ULONG			cValues = 0;
	LPSPropValue	lpPropArrayStore = NULL;
	LPSPropValue	lpPropArrayInbox = NULL;
	LPSPropValue	lpPropVal = NULL;
	SizedSPropTagArray(3, sPropsStore) = { 3,
		{ PR_IPM_OUTBOX_ENTRYID, PR_IPM_SENTMAIL_ENTRYID, PR_IPM_WASTEBASKET_ENTRYID } };
	SizedSPropTagArray(6, sPropsInbox) = { 6,
		{ PR_IPM_APPOINTMENT_ENTRYID, PR_IPM_CONTACT_ENTRYID, PR_IPM_DRAFTS_ENTRYID,
		  PR_IPM_JOURNAL_ENTRYID, PR_IPM_NOTE_ENTRYID, PR_IPM_TASK_ENTRYID } };

	hr = lpStore->GetProps((LPSPropTagArray)&sPropsStore, 0, &cValues, &lpPropArrayStore);
	if (hr != hrSuccess)
		goto exit;

	for (ULONG i = 0; i < cValues; i++) {
		if (PROP_TYPE(lpPropArrayStore[i].ulPropTag) == PT_BINARY)
			lstSpecialEntryIDs.insert(BinaryArray(lpPropArrayStore[i].Value.bin.lpb, lpPropArrayStore[i].Value.bin.cb));
	}

	hr = lpStore->GetReceiveFolder("IPM", 0, &cbEntryID, &lpEntryID, NULL);
	if (hr != hrSuccess)
		goto exit;

	// inbox is special too
	lstSpecialEntryIDs.insert(BinaryArray((LPBYTE)lpEntryID, cbEntryID));

	hr = lpSession->OpenEntry(cbEntryID, lpEntryID, &IID_IMAPIFolder, 0, &ulObjType, (LPUNKNOWN*)&lpInbox);
	if (hr != hrSuccess)
		goto exit;

	hr = lpInbox->GetProps((LPSPropTagArray)&sPropsInbox, 0, &cValues, &lpPropArrayInbox);
	if (hr != hrSuccess)
		goto exit;

	for (ULONG i = 0; i < cValues; i++) {
		if (PROP_TYPE(lpPropArrayInbox[i].ulPropTag) == PT_BINARY)
			lstSpecialEntryIDs.insert(BinaryArray(lpPropArrayInbox[i].Value.bin.lpb, lpPropArrayInbox[i].Value.bin.cb));
	}

	if (HrGetOneProp(lpInbox, PR_ADDITIONAL_REN_ENTRYIDS, &lpPropVal) == hrSuccess &&
		lpPropVal->Value.MVbin.cValues >= 5 && lpPropVal->Value.MVbin.lpbin[4].cb != 0)
	{
		lstSpecialEntryIDs.insert(BinaryArray(lpPropVal->Value.MVbin.lpbin[4].lpb, lpPropVal->Value.MVbin.lpbin[4].cb));
	}
	
	if (lpPropVal){
		MAPIFreeBuffer(lpPropVal);
		lpPropVal = NULL;
	}

	if(!lpPublicStore)
		goto exit;

	if (HrGetOneProp(lpPublicStore, PR_IPM_PUBLIC_FOLDERS_ENTRYID, &lpPropVal) == hrSuccess)
		lstSpecialEntryIDs.insert(BinaryArray(lpPropVal->Value.bin.lpb, lpPropVal->Value.bin.cb));

exit:
	if (lpPropArrayStore)
		MAPIFreeBuffer(lpPropArrayStore);

	if (lpPropArrayInbox)
		MAPIFreeBuffer(lpPropArrayInbox);

	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	if (lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	if (lpInbox)
		lpInbox->Release();

	return hr;
}

bool IMAP::IsSpecialFolder(ULONG cb, LPENTRYID lpEntryID) {
	if(lstSpecialEntryIDs.find(BinaryArray((BYTE *)lpEntryID, cb, true)) == lstSpecialEntryIDs.end())
        return false;
        
    return true;
}

HRESULT IMAP::HrRefreshFolderMails(bool bInitialLoad, bool bResetRecent, bool bShowUID, unsigned int *lpulUnseen) {
	HRESULT hr = hrSuccess;
	IMAPIFolder *lpFolder = NULL;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	ULONG ulMailnr = 0;
	ULONG ulMaxUID = 0;
	ULONG ulRecent = 0;
	int n = 0;
	SMail sMail;
	bool bNewMail = false;
	enum { EID, IKEY, IMAPID, FLAGS, FLAGSTATUS, MSGSTATUS, LAST_VERB, NUM_COLS };
	SizedSPropTagArray(NUM_COLS, spt) = { NUM_COLS, {PR_ENTRYID, PR_INSTANCE_KEY, PR_EC_IMAP_ID, PR_MESSAGE_FLAGS, PR_FLAG_STATUS, PR_MSG_STATUS, PR_LAST_VERB_EXECUTED} };
	SizedSSortOrderSet(1, sSortUID) = { 1, 0, 0, { { PR_EC_IMAP_ID, TABLE_SORT_ASCEND } } };
	vector<SMail>::iterator iterMail;
	map<unsigned int, unsigned int> mapUIDs; // Map UID -> ID
	map<unsigned int, unsigned int>::iterator iterUID;
	LPSPropValue lpsPropMax = NULL;
	SPropValue sPropMax;
	unsigned int ulUnseen = 0;
	
	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	hr = HrFindFolder(strCurrentFolder, bCurrentFolderReadOnly, &lpFolder);
	if (hr != hrSuccess)
		goto exit;
		
    if(HrGetOneProp(lpFolder, PR_EC_IMAP_MAX_ID, &lpsPropMax) != hrSuccess) {
        ulMaxUID = 0;
    } else {
        ulMaxUID = lpsPropMax->Value.ul;
    }

	hr = lpFolder->GetContentsTable(MAPI_DEFERRED_ERRORS, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns((LPSPropTagArray) &spt, TBL_BATCH);
	if (hr != hrSuccess)
		goto exit;

    hr = lpTable->SortTable((LPSSortOrderSet)&sSortUID, TBL_BATCH);
    if (hr != hrSuccess)
        goto exit;

    // Remember UIDs if needed
    if(!bInitialLoad) {
        for(iterMail = lstFolderMailEIDs.begin(); iterMail != lstFolderMailEIDs.end(); iterMail++) {
            mapUIDs[iterMail->ulUid] = n;
            n++;
        }
    }
    
    if(bInitialLoad)
        lstFolderMailEIDs.clear();

    iterMail = lstFolderMailEIDs.begin();

    // Scan MAPI for new and existing messages
	while(1) {
		hr = lpTable->QueryRows(ROWS_PER_REQUEST, 0, &lpRows);
		if (hr != hrSuccess)
            goto exit;
            
        if(lpRows->cRows == 0)
            break;
            
		for(ulMailnr = 0; ulMailnr < lpRows->cRows; ulMailnr++) {
            if (lpRows->aRow[ulMailnr].lpProps[EID].ulPropTag != spt.aulPropTag[EID] ||
                lpRows->aRow[ulMailnr].lpProps[IKEY].ulPropTag != spt.aulPropTag[IKEY] ||
                lpRows->aRow[ulMailnr].lpProps[IMAPID].ulPropTag != spt.aulPropTag[IMAPID] ||
                lpRows->aRow[ulMailnr].lpProps[FLAGS].ulPropTag != spt.aulPropTag[FLAGS])
                continue;

            iterUID = mapUIDs.find(lpRows->aRow[ulMailnr].lpProps[IMAPID].Value.ul);
		    if(iterUID == mapUIDs.end()) {
		        // There is a new message
                sMail.sEntryID = BinaryArray(lpRows->aRow[ulMailnr].lpProps[EID].Value.bin);
                sMail.sInstanceKey = BinaryArray(lpRows->aRow[ulMailnr].lpProps[IKEY].Value.bin);
                sMail.ulUid = lpRows->aRow[ulMailnr].lpProps[IMAPID].Value.ul;

                // Mark as recent if the message has a UID higher than the last highest read UID
                // in this folder. This means that this session is the only one to see the message
                // as recent.
                sMail.bRecent = sMail.ulUid > ulMaxUID;

                // Remember flags
                sMail.strFlags = PropsToFlags(lpRows->aRow[ulMailnr].lpProps, lpRows->aRow[ulMailnr].cValues, sMail.bRecent);

                // Put message on the end of our message
    			lstFolderMailEIDs.push_back(sMail);

                m_ulLastUid = max(sMail.ulUid, m_ulLastUid);
                bNewMail = true;
                
                // Remember the first unseen message
                if(ulUnseen == 0) {
                    if((lpRows->aRow[ulMailnr].lpProps[FLAGS].Value.ul & MSGFLAG_READ) == 0) {
                        ulUnseen = lstFolderMailEIDs.size()-1+1; // size()-1 = last offset, mail ID = position + 1
                    }
                }
            } else {
                // Check flags
                std::string strFlags = PropsToFlags(lpRows->aRow[ulMailnr].lpProps, lpRows->aRow[ulMailnr].cValues, lstFolderMailEIDs[iterUID->second].bRecent);
                
                if(lstFolderMailEIDs[iterUID->second].strFlags != strFlags) {
                    // Flags have changed, notify it
                    if(bShowUID)
                        HrResponse(stringify(iterUID->second+1) + " FETCH (UID " + stringify(lpRows->aRow[ulMailnr].lpProps[IMAPID].Value.ul) + " FLAGS (" + strFlags + "))");
                    else
                        HrResponse(stringify(iterUID->second+1) + " FETCH (FLAGS (" + strFlags + "))");
                    lstFolderMailEIDs[iterUID->second].strFlags = strFlags;
                }
                    
                // We already had this message, remove it from setUIDs
                mapUIDs.erase(iterUID);
            }
        
		}

		if (lpRows)
			FreeProws(lpRows);
		lpRows = NULL;
    }

    // All messages left in mapUIDs have been deleted, so loop through the current list so we can
    // send the correct EXPUNGE calls; At the same time, count RECENT messages.
    ulMailnr = 0;
    while(ulMailnr < lstFolderMailEIDs.size()) {
        if(mapUIDs.find(lstFolderMailEIDs[ulMailnr].ulUid) != mapUIDs.end()) {
            HrResponse(stringify(ulMailnr+1) + " EXPUNGE");
            lstFolderMailEIDs.erase(lstFolderMailEIDs.begin() + ulMailnr);
        } else {
            if(lstFolderMailEIDs[ulMailnr].bRecent)
                ulRecent++;
                
            iterMail++;
            ulMailnr++;
        }
    }
    
    if (bNewMail || bInitialLoad) {
        HrResponse(stringify(lstFolderMailEIDs.size()) + " EXISTS");
        HrResponse(stringify(ulRecent) + " RECENT");
    }

	sort(lstFolderMailEIDs.begin(), lstFolderMailEIDs.end());
	
    // Save the max UID so that other session will not see the items as \Recent
    if(bResetRecent && ulRecent) {
    	sPropMax.ulPropTag = PR_EC_IMAP_MAX_ID;
    	sPropMax.Value.ul = m_ulLastUid;
    	HrSetOneProp(lpFolder, &sPropMax);
    }
    
    if(lpulUnseen)
        *lpulUnseen = ulUnseen;

exit:
    if (lpsPropMax)
        MAPIFreeBuffer(lpsPropMax);
        
	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	if (lpFolder)
		lpFolder->Release();

	return hr;
}

HRESULT IMAP::HrGetFolderPath(list<SFolder>::iterator lpFolder, list<SFolder> &lstFolders, string &strPath) {
	HRESULT hr = hrSuccess;

	if (lpFolder == lstFolders.end()) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	if (lpFolder->lpParentFolder != lstFolders.end()) {
		hr = HrGetFolderPath(lpFolder->lpParentFolder, lstFolders, strPath);
		if (hr != hrSuccess)
			goto exit;
		strPath += IMAP_HIERARCHY_DELIMITER + lpFolder->strFolderName;
	}

exit:
	return hr;
}

HRESULT IMAP::HrGetSubTree(list<SFolder> &lstFolders, SBinary &sEntryID, string strFolderName, list<SFolder>::iterator lpParentFolder,
						   bool bSubfolders, bool bMailFolder) {
	HRESULT hr = hrSuccess;
	IMAPIFolder *lpFolder = NULL;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	ULONG ulObjType;
	SFolder sFolder;
	SBinary sChildEntryID;
	list<SFolder>::iterator lpNewParent;
	string strClass;
	vector<BinaryArray>::iterator iFolder;

	enum { EID, NAME, IMAPID, SUBFOLDERS, CONTAINERCLASS, NUM_COLS };
	SizedSPropTagArray(NUM_COLS, spt) = { NUM_COLS, {PR_ENTRYID, PR_DISPLAY_NAME, PR_EC_IMAP_ID,
													 PR_SUBFOLDERS, PR_CONTAINER_CLASS } };

	if (!lpSession) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	while (strFolderName.find(IMAP_HIERARCHY_DELIMITER) != string::npos) {
		strFolderName.erase(strFolderName.find(IMAP_HIERARCHY_DELIMITER), 1);
	}

	iFolder = find(m_vSubscriptions.begin(), m_vSubscriptions.end(), BinaryArray(sEntryID));
	sFolder.bActive = !(iFolder == m_vSubscriptions.end());
	sFolder.bSpecialFolder = IsSpecialFolder(sEntryID.cb, (LPENTRYID)sEntryID.lpb);
	sFolder.bMailFolder = true;
	sFolder.lpParentFolder = lpParentFolder;
	sFolder.strFolderName = strFolderName;
	sFolder.sEntryID = sEntryID;
	sFolder.bMailFolder = bMailFolder;
	sFolder.bHasSubfolders = bSubfolders;

	lstFolders.push_front(sFolder);
	lpNewParent = lstFolders.begin();
	
	if(!bSubfolders)
		goto exit;

	hr = lpSession->OpenEntry(sEntryID.cb, (LPENTRYID) sEntryID.lpb, &IID_IMAPIFolder, 0, &ulObjType, (LPUNKNOWN *) &lpFolder);
	if (hr != hrSuccess)
		goto exit;

	hr = lpFolder->GetHierarchyTable(0, &lpTable);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->SetColumns((LPSPropTagArray) &spt, 0);
	if (hr != hrSuccess)
		goto exit;

	hr = lpTable->QueryRows(-1, 0, &lpRows);
	if (hr != hrSuccess)
		goto exit;

	for (ULONG i = 0; i < lpRows->cRows; i++) {
		if (PROP_TYPE(lpRows->aRow[i].lpProps[EID].ulPropTag) != PT_BINARY)
			continue;			// no entryid, no folder

		if (PROP_TYPE(lpRows->aRow[i].lpProps[IMAPID].ulPropTag) != PT_LONG) {
		    lpLogger->Log(EC_LOGLEVEL_FATAL, "Server does not support PR_EC_IMAP_ID. Please update zarafa server.");
		    break;
        }

		char *foldername = "";
		bSubfolders = true;
		bMailFolder = true;

		if (PROP_TYPE(lpRows->aRow[i].lpProps[NAME].ulPropTag) == PT_STRING8)
			foldername = lpRows->aRow[i].lpProps[NAME].Value.lpszA;

		if (PROP_TYPE(lpRows->aRow[i].lpProps[SUBFOLDERS].ulPropTag) == PT_BOOLEAN)
			bSubfolders = lpRows->aRow[i].lpProps[SUBFOLDERS].Value.b;

		if (PROP_TYPE(lpRows->aRow[i].lpProps[CONTAINERCLASS].ulPropTag) == PT_STRING8){
			strClass =  lpRows->aRow[i].lpProps[CONTAINERCLASS].Value.lpszA;

			ToUpper(strClass);
			
			if (!strClass.empty() && strClass.compare(0, 3, "IPM") != 0 && strClass.compare("IPF.NOTE") != 0){
				if (bOnlyMailFolders)
					continue;
				bMailFolder = false;
			}
		}

		sChildEntryID = lpRows->aRow[i].lpProps[EID].Value.bin;
		HrGetSubTree(lstFolders, sChildEntryID, foldername, lpNewParent, bSubfolders, bMailFolder);
	}

exit:
	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	if (lpFolder)
		lpFolder->Release();

	return hr;
}

/** 
 * Extends IMAP shortcuts into real full IMAP proptags, and returns an
 * vector of all separate and capitalized items.
 * 
 * @param[in] strMsgDataItemNames String of wanted data items from a FETCH command.
 * @param[out] lstDataItems Vector of all separate items in the string in uppercase.
 * 
 * @return hrSuccess
 */
HRESULT IMAP::HrGetDataItems(string strMsgDataItemNames, vector<string> &lstDataItems) {
	HRESULT hr = hrSuccess;

	// translate macro's
	ToUpper(strMsgDataItemNames);
	if (strMsgDataItemNames == "ALL") {
		strMsgDataItemNames = "FLAGS INTERNALDATE RFC822.SIZE ENVELOPE";
	} else if (strMsgDataItemNames == "FAST") {
		strMsgDataItemNames = "FLAGS INTERNALDATE RFC822.SIZE";
	} else if (strMsgDataItemNames == "FULL") {
		strMsgDataItemNames = "FLAGS INTERNALDATE RFC822.SIZE ENVELOPE BODY";
	}

	// split data items
	if (strMsgDataItemNames.size() > 1 && strMsgDataItemNames[0] == '(') {
		strMsgDataItemNames.erase(0,1);
		strMsgDataItemNames.erase(strMsgDataItemNames.size()-1, 1);
	}

	hr = HrSplitInput(strMsgDataItemNames, lstDataItems);

//exit:
	return hr;
}

HRESULT IMAP::HrSemicolonToComma(string &strData) {
	string::size_type ulPos = strData.find(";");

	while (ulPos != string::npos) {
		strData.replace(ulPos, 1, ",");
		ulPos = strData.find(";", ulPos + 1);
	}

	return hrSuccess;
}

// Do a FETCH based on table data for a specific list of messages
HRESULT IMAP::HrPropertyFetch(list<ULONG> &lstMails, vector<string> &lstDataItems) {
	HRESULT hr = hrSuccess;
	IMAPIFolder *lpFolder = NULL;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	LPSRow lpRow = NULL;
	ULONG ulDataItemNr;
	string strDataItem;
	SRestriction sRestriction;
	SPropValue sPropVal;
	string strResponse;
	LPSPropTagArray lpPropTags = NULL;
	list<ULONG>::iterator lpMail;
    set<ULONG> setProps;
    set<ULONG>::iterator iterProps;
    int n;
    LPSPropValue lpProps;
    ULONG cValues;
    unsigned int ulReadAhead = 0;
    SizedSSortOrderSet (1, sSortUID) = { 1, 0, 0, { { PR_EC_IMAP_ID, TABLE_SORT_ASCEND } } };
    
	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	// Setup the readahead length
	ulReadAhead = lstMails.size() > 50 ? 50 : lstMails.size();

	// Find out which properties we will be needing from the table. This should be kept in-sync
	// with the properties that are used in HrPropertyFetchRow()

	for (ulDataItemNr = 0; ulDataItemNr < lstDataItems.size(); ulDataItemNr++) {
		strDataItem = lstDataItems[ulDataItemNr];

		if (strDataItem == "FLAGS") {
		    setProps.insert(PR_MESSAGE_FLAGS);
		    setProps.insert(PR_FLAG_STATUS);
		    setProps.insert(PR_MSG_STATUS);
		    setProps.insert(PR_LAST_VERB_EXECUTED);
		} else if (strDataItem == "XAOL.SIZE") {
		    setProps.insert(PR_MESSAGE_SIZE);
		} else if (strDataItem == "INTERNALDATE") {
		    setProps.insert(PR_MESSAGE_DELIVERY_TIME);
		    setProps.insert(PR_CLIENT_SUBMIT_TIME);
		} else if (strDataItem == "ENVELOPE") {
		    // No properties needed, entryid is available outside table
		} else if (Prefix(strDataItem, "BODY[HEADER") || Prefix(strDataItem, "RFC822.HEADER") || Prefix(strDataItem, "BODY.PEEK[HEADER")) {
		    setProps.insert(PR_TRANSPORT_MESSAGE_HEADERS);
        }
	}

	if(!setProps.empty()) {
        // Build an LPSPropTagArray
        hr = MAPIAllocateBuffer(CbNewSPropTagArray(setProps.size()+1), (void **)&lpPropTags);
        if(hr != hrSuccess)
            goto exit;

        // Always get UID
        lpPropTags->aulPropTag[0] = PR_INSTANCE_KEY;
        n = 1;
        for(iterProps = setProps.begin(); iterProps != setProps.end(); iterProps++) {
            lpPropTags->aulPropTag[n] = *iterProps;
            n++;
        }
        lpPropTags->cValues = setProps.size()+1;

        // Open the folder in question
        hr = HrFindFolder(strCurrentFolder, bCurrentFolderReadOnly, &lpFolder);
        if (hr != hrSuccess)
            goto exit;

        // Don't let the server cap the contents to 255 bytes, so our PR_TRANSPORT_MESSAGE_HEADERS is complete in the table
        hr = lpFolder->GetContentsTable(EC_TABLE_NOCAP | MAPI_DEFERRED_ERRORS, &lpTable);
        if (hr != hrSuccess)
            goto exit;

        // Request our columns
        hr = lpTable->SetColumns(lpPropTags, TBL_BATCH);
        if (hr != hrSuccess)
            goto exit;
            
        // Messages are usually requested in UID order, so sort the table in UID order too. This improves
        // the row prefetch hit ratio.
        hr = lpTable->SortTable((LPSSortOrderSet)&sSortUID, TBL_BATCH);
        if(hr != hrSuccess)
            goto exit;

        // Setup a find restriction that we modify for each row
        sRestriction.rt = RES_PROPERTY;
        sRestriction.res.resProperty.relop = RELOP_EQ;
        sRestriction.res.resProperty.ulPropTag = PR_INSTANCE_KEY;
        sRestriction.res.resProperty.lpProp = &sPropVal;
        sPropVal.ulPropTag = PR_INSTANCE_KEY;
    }
    
	// Loop through all requested rows, and get the data for each (FIXME: slow for large requests)
	for (lpMail = lstMails.begin(); lpMail != lstMails.end(); lpMail++) {
		sPropVal.Value.bin.cb = lstFolderMailEIDs[*lpMail].sInstanceKey.cb;
		sPropVal.Value.bin.lpb = lstFolderMailEIDs[*lpMail].sInstanceKey.lpb;

        // We use a read-ahead mechanism here, reading 50 rows at a time.		
		if (lpTable) {
            // First, see if the next row is somewhere in our already-read data
            lpRow = NULL;
            if(lpRows) {
                for(unsigned int i=0; i<lpRows->cRows;i++) {
                    if(lpRows->aRow[i].lpProps[0].ulPropTag == PR_INSTANCE_KEY && BinaryArray(lpRows->aRow[i].lpProps[0].Value.bin) == BinaryArray(sPropVal.Value.bin)) {
                        lpRow = &lpRows->aRow[i];
                    }
                }
            }
                        
            if(lpRow == NULL) {
                if(lpRows)
                    FreeProws(lpRows);
                lpRows = NULL;
                
                // Row was not found in our current data, request new data
                if(lpTable->FindRow(&sRestriction, BOOKMARK_BEGINNING, 0) == hrSuccess && lpTable->QueryRows(ulReadAhead, 0, &lpRows) == hrSuccess) {
					if (lpRows->cRows != 0) {
						// The row we want is the first returned row
						lpRow = &lpRows->aRow[0];
					}
                }
            }
            
		    // Pass the row data for conversion
		    if(lpRow) {
    		    cValues = lpRow->cValues;
	    	    lpProps = lpRow->lpProps;
            } else {
                cValues = 0;
                lpProps = NULL;
            }
        } else {
            // If the row is unavailable, or the table is not needed, do not pass any properties
            cValues = 0;
            lpProps = NULL;
        }
        
        // Fetch the row data
        if (HrPropertyFetchRow(lpProps, cValues, strResponse, *lpMail, lstDataItems) != hrSuccess) {
            lpLogger->Log(EC_LOGLEVEL_ERROR, "{?} Error fetching mail");
        } else {
            HrResponse(strResponse);
        }
	}

exit:
    if (lpPropTags)
        MAPIFreeBuffer(lpPropTags);
        
	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	if (lpFolder)
		lpFolder->Release();

	return hr;
}

// Does a FETCH based on row-data from a MAPI table
HRESULT IMAP::HrPropertyFetchRow(LPSPropValue lpProps, ULONG cValues, string &strResponse, ULONG ulMailnr, vector < string > &lstDataItems)
{
	HRESULT hr = hrSuccess;
	string strDataItem;
	string strParts;
	string strHeaders;
	string::size_type ulPos;
	size_t ulDataItemNr;
	LPSPropValue lpProp = NULL; // non-free
	char szBuffer[IMAP_RESP_MAX + 1];
	IMessage *lpMessage = NULL;
	ULONG ulObjType = 0;
	sending_options sopt;
	imopt_default_sending_options(&sopt);
	sopt.no_recipients_workaround = true;	// do not stop processing mail on empty recipient table
	sopt.alternate_boundary = "=_ZG_static";
	string strMessage;
	bool bForceFlags = false;
	string strMessagePart;
	unsigned int ulCount = 0;
	bool bSkipOpen = false;
	ostringstream os;
	
	// Response always starts with "<id> FETCH ("
	snprintf(szBuffer, IMAP_RESP_MAX, "%u FETCH (", ulMailnr + 1);
	strResponse = szBuffer;

	for (ulDataItemNr = 0; ulDataItemNr < lstDataItems.size(); ulDataItemNr++) {
		strDataItem = lstDataItems[ulDataItemNr];
		
        // Open message if we need it later
		if ((Prefix(strDataItem, "BODY") || Prefix(strDataItem,"RFC822") || strDataItem == "ENVELOPE") && lpMessage == NULL) {
		    bSkipOpen = false;
		    
		    // Exception: headers requested and PR_TRANSPORT_MESSAGE_HEADERS available
            if(strDataItem == "RFC822.HEADER" || Prefix(strDataItem, "BODY[HEADER") || Prefix(strDataItem, "BODY.PEEK[HEADER")) {
                lpProp = PpropFindProp(lpProps, cValues, PR_TRANSPORT_MESSAGE_HEADERS);
                
                if(lpProp)
                    bSkipOpen = true;

				// Exception on exception, if we need to re-generate the same as the full generation
				if (parseBool(lpConfig->GetSetting("imap_always_generate")) == true)
					bSkipOpen = false;
			} else {
                // Exception: body requested but cached
                if(lstFolderMailEIDs[ulMailnr].ulUid == m_ulCacheUID)
                    bSkipOpen = true;
            }

            if(!bSkipOpen) {
                if(lpSession->OpenEntry(lstFolderMailEIDs[ulMailnr].sEntryID.cb, (LPENTRYID) lstFolderMailEIDs[ulMailnr].sEntryID.lpb, &IID_IMessage, MAPI_DEFERRED_ERRORS, &ulObjType, (LPUNKNOWN *) &lpMessage) != hrSuccess)
                    lpMessage = NULL;
            }
		}

		// Handle requested properties
		if (strDataItem == "FLAGS") {
		    if(lpMessage) {
                // If the message is open, get flags from there, since they may have changed due to setting the \Seen flag
                HrGetMessageFlags(strResponse, lpMessage, lstFolderMailEIDs[ulMailnr].bRecent);
            } else {
	    		strResponse += "FLAGS (";
                strResponse += PropsToFlags(lpProps, cValues, lstFolderMailEIDs[ulDataItemNr].bRecent);
		    	strResponse += ")";
            }
            bForceFlags = false;
		} else if (strDataItem == "XAOL.SIZE") {
		    lpProp = PpropFindProp(lpProps, cValues, PR_MESSAGE_SIZE);
		    
		    strResponse += "XAOL.SIZE ";
		    if(lpProp) {
		        strResponse += stringify(lpProp->Value.ul);
            } else {
                strResponse += "NIL";
            }
		} else if (strDataItem == "INTERNALDATE") {
			strResponse += "INTERNALDATE ";

		    lpProp = PpropFindProp(lpProps, cValues, PR_MESSAGE_DELIVERY_TIME);
		    
			if (lpProp) {
				strResponse += "\"";
				strResponse += FileTimeToString(lpProp->Value.ft);
				strResponse += "\"";
			} else {
			    // Fallback to client delivery time
			    lpProp = PpropFindProp(lpProps, cValues, PR_CLIENT_SUBMIT_TIME);
			    if(lpProp) {
    				strResponse += "\"";
	    			strResponse += FileTimeToString(lpProp->Value.ft);;
		    		strResponse += "\"";
                } else {
                    strResponse += "NIL";
                }
			}
		} else if (strDataItem == "UID") {
			snprintf(szBuffer, IMAP_RESP_MAX, "UID %u", lstFolderMailEIDs[ulMailnr].ulUid);
			strResponse += szBuffer;
        } else if (Prefix(strDataItem, "BODY") || Prefix(strDataItem, "RFC822")) {
            // Do the conversion to RFC822

			// For some clients, headers need to match body contents, so we cannot
			// optimize here to get headers from PR_TRANSPORT_MESSAGE_HEADERS,
			// since they are from the incoming message, not from a
			// vmime generated message.

            // Optimization: request headers only from inetmapi if that's all we need
			if (strDataItem == "RFC822.HEADER" || Prefix(strDataItem, "BODY[HEADER") || Prefix(strDataItem, "BODY.PEEK[HEADER")) {
				sopt.headers_only = true;
			} else {
				sopt.headers_only = false;
			}

            strMessage.clear();

			if(sopt.headers_only && parseBool(lpConfig->GetSetting("imap_always_generate")) == false) {
				lpProp = PpropFindProp(lpProps, cValues, PR_TRANSPORT_MESSAGE_HEADERS);
				if(lpProp) {
					strMessage = lpProp->Value.lpszA;
					FixupBoundary(strMessage);
				}
			}

            if (!sopt.headers_only && m_ulCacheUID == lstFolderMailEIDs[ulMailnr].ulUid) {
                strMessage = m_strCache;
            } else if (strMessage.empty() && lpMessage) {

                // Get message from cache if possible
    			if (IMToINet(lpSession, lpAddrBook, lpMessage, os, sopt, lpLogger) != hrSuccess) {
	    			HrResponse("BAD [PARSE] error converting mail");
		    		continue;
                }
                
                strMessage = os.str();
                
                // Cache the generated message
				if(!sopt.headers_only) {
					m_ulCacheUID = lstFolderMailEIDs[ulMailnr].ulUid;
					m_strCache = strMessage;
				}
            } else {
				ASSERT(FALSE);	// logic error
			}
			
			if (strDataItem == "BODY" || strDataItem == "BODYSTRUCTURE") {
			    /* RFC 3501 6.4.5:
			     * BODY
			              Non-extensible form of BODYSTRUCTURE.
			     * BODYSTRUCTURE
			              The [MIME-IMB] body structure of the message.
			     */         
			              
				strResponse += strDataItem;
				strResponse += " ";

				if (HrGetBodyStructure(strDataItem == "BODYSTRUCTURE", strMessagePart, strMessage) == hrSuccess) {
					strResponse += strMessagePart;
				} else {
					strResponse += "NIL";
				}
			} else {
			    /* RFC 3501 6.4.5:
			     * BODY[<section>]<<partial>>
			     *        The text of a particular body section, without boundaries.
			     * BODY.PEEK[<section>]<<partial>>
			     *        An alternate form of BODY[<section>] that does not implicitly
                 *        set the \Seen flag.
			     *                  
                 * RFC822
                 *        Functionally equivalent to BODY[]
                 * RFC822.HEADER
                 *        Functionally equivalent to BODY.PEEK[HEADER]
                 * RFC822.SIZE
                 *        The [RFC-2822] size of the message.
                 * RFC822.TEXT
                 *        Functionally equivalent to BODY[TEXT]
                 */
                
				if (!bCurrentFolderReadOnly && !Prefix(strDataItem, "BODY.PEEK") && strDataItem != "RFC822.HEADER" && strDataItem != "RFC822.SIZE") {
				    if(lpMessage) {
                        hr = lpMessage->SetReadFlag(SUPPRESS_RECEIPT);
	    				if (hr != hrSuccess)
		    				goto exit;
						
                        bForceFlags = true;
                    }
				}

				if (strDataItem == "RFC822.SIZE") {
					// We must return the real size, since clients use this when using chunked mode to download the full message
					strResponse.append("RFC822.SIZE " + stringify(strMessage.size()));
				} else {
				    if (strDataItem == "RFC822" || strDataItem.find("[]") != string::npos) {
                        // Handle BODY[] and RFC822 (entire message)
                        strMessagePart = strMessage;
                        if(strDataItem == "RFC822") {
                            strResponse += "RFC822";
                        } else if(Prefix(strDataItem, "BODY")) {
                            strResponse += "BODY[]";
                        }
                    } else {
                        // Handle BODY[subpart], RFC822.TEXT and RFC822.HEADER
                        if(Prefix(strDataItem, "BODY")) {
                            // BODY[subpart], strParts = <subpart> (so "1.2.3" or "3.HEADER" or "TEXT" etc)
                            ulPos = strDataItem.find("[");
                            if (ulPos != string::npos)
                                strParts = strDataItem.substr(ulPos + 1);

                            ulPos = strParts.find("]");
                            if (ulPos != string::npos)
                                strParts = strParts.substr(0, ulPos);
                        } else {
                            // RFC822.<subpart> (so "TEXT" or "HEADER")
                            ulPos = strDataItem.find(".");
                            if(ulPos != string::npos)
                                strParts = strDataItem.substr(ulPos + 1);
                        }

                        if(Prefix(strDataItem, "BODY")) {
                            strResponse += (string)"BODY[" + strParts + "]";
                        } else {
                            strResponse += (string)"RFC822." + strParts;
                        }
                        
                        // Get the correct message part (1.2.3, TEXT, HEADER, 1.2.3.TEXT, 1.2.3.HEADER)
                        HrGetMessagePart(strMessagePart, strMessage, strParts);
                    }

                    // Process byte-part request ( <12345.12345> ) for BODY
                    if (Prefix(strDataItem, "BODY")) {
                        if (strDataItem[strDataItem.size() - 1] == '>') {
                            ulPos = strDataItem.rfind('<');
                            if (ulPos != string::npos) {
                                strParts = strDataItem.substr(ulPos + 1, strDataItem.size() - ulPos - 2);

                                ulPos = strParts.find('.');
                                if (ulPos != string::npos) {
                                    ulCount = strtoul(strParts.substr(0, ulPos).c_str(), NULL, 0);
                                    ulPos = strtoul(strParts.substr(ulPos + 1).c_str(), NULL, 0);
                                } else {
                                    ulCount = strtoul(strParts.c_str(), NULL, 0);
                                    ulPos = strMessagePart.size();
                                }

                                if (ulCount > strMessagePart.size()) {
                                    strMessagePart.clear();
                                } else if (ulCount + ulPos > strMessagePart.size()) {
                                    // strMessagePart = strMessagePart.substr(ulCount);
									strMessagePart.erase(0, ulCount);
                                } else {
                                    // strMessagePart = strMessagePart.substr(ulCount, ulPos);
									strMessagePart.erase(0, ulCount);
									strMessagePart.erase(ulPos);
                                }

                                snprintf(szBuffer, IMAP_RESP_MAX, "<%u>", ulCount);
                                strResponse += szBuffer; // FIXME: szBuffer attached here, AND later (10 lines below)
                            }
                        }
                    }
                    
                    if (strMessagePart.empty()) {
                        strResponse += " NIL";
                    } else {
                        // Output actual data
                        snprintf(szBuffer, IMAP_RESP_MAX, " {%u}\r\n", (ULONG)strMessagePart.size());
                        strResponse += szBuffer;
                        strResponse += strMessagePart;
                    }
                }
			}
		} else if (strDataItem == "ENVELOPE") {
		    string strEnvelope;
		    if(lpMessage == NULL)
		        strResponse += " NIL";
            else {
    			HrGetMessageEnvelope(strEnvelope, lpMessage);
	    		strResponse += strEnvelope;
            }
		} else {
			strResponse += strDataItem;
			strResponse += " NIL";
		}

		strResponse += " ";
	}

	// Output flags if modified
	if(bForceFlags && lpMessage) {
	    HrGetMessageFlags(strResponse, lpMessage, lstFolderMailEIDs[ulMailnr].bRecent);
	}
	
	if (strResponse.compare(strResponse.size()-1, 1, " ") == 0) {
		strResponse.resize(strResponse.size()-1);
	}

	strResponse += ")";

exit:
    if(lpMessage)
        lpMessage->Release();

	return hr;
}

HRESULT IMAP::HrGetMessageEnvelope(string &strResponse, LPMESSAGE lpMessage) {
	HRESULT hr = hrSuccess;
	LPSPropValue lpPropVal = NULL;
	string strPart;
	string::size_type ulPos;
	ULONG ulCount;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	enum { EMAIL_ADDRESS, DISPLAY_NAME, RECIPIENT_TYPE, NUM_COLS };
	SizedSPropTagArray(NUM_COLS, spt) = { NUM_COLS, {PR_EMAIL_ADDRESS, PR_DISPLAY_NAME, PR_RECIPIENT_TYPE} };

	if (!lpMessage) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	strResponse += "ENVELOPE (";
	// date string
	if (HrGetOneProp(lpMessage, PR_CLIENT_SUBMIT_TIME, &lpPropVal) == hrSuccess
		|| HrGetOneProp(lpMessage, PR_MESSAGE_DELIVERY_TIME, &lpPropVal) == hrSuccess) {
		strResponse += "\"";
		strResponse += FileTimeToString(lpPropVal->Value.ft);
		strResponse += "\" ";
	} else {
		strResponse += "NIL ";
	}

	// subject
	strResponse += "\"";
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	if (HrGetOneProp(lpMessage, PR_SUBJECT, &lpPropVal) == hrSuccess) {
		strResponse += EscapeString(lpPropVal->Value.lpszA);
	}

	// from
	strResponse += "\" ((\"";
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	if (HrGetOneProp(lpMessage, PR_SENT_REPRESENTING_NAME, &lpPropVal) == hrSuccess) {
		strResponse += EscapeString(lpPropVal->Value.lpszA);
	}

	strResponse += "\" NIL \"";
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	if (HrGetOneProp(lpMessage, PR_SENT_REPRESENTING_EMAIL_ADDRESS, &lpPropVal) == hrSuccess) {
		strPart = lpPropVal->Value.lpszA;
	} else {
		strPart = "@";
	}

	ulPos = strPart.find("@", 0);
	if (ulPos != string::npos) {
		strResponse += EscapeStringQT(strPart.substr(0, ulPos));
		strResponse += "\" \"";
		strResponse += EscapeStringQT(strPart.substr(ulPos + 1));
	} else {
		strResponse += EscapeStringQT(strPart);
		strResponse += "\" \"";
	}

	// sender
	strResponse += "\")) ((\"";
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	if (HrGetOneProp(lpMessage, PR_SENDER_NAME, &lpPropVal) == hrSuccess) {
		strResponse += EscapeString(lpPropVal->Value.lpszA);
	}

	strResponse += "\" NIL \"";
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	if (HrGetOneProp(lpMessage, PR_SENDER_EMAIL_ADDRESS, &lpPropVal) == hrSuccess) {
		strPart = lpPropVal->Value.lpszA;
	} else {
		strPart = "@";
	}

	ulPos = strPart.find("@", 0);
	if (ulPos != string::npos) {
		strResponse += EscapeStringQT(strPart.substr(0, ulPos));
		strResponse += "\" \"";
		strResponse += EscapeStringQT(strPart.substr(ulPos + 1));
	} else {
		strResponse += EscapeStringQT(strPart);
		strResponse += "\" \"";
	}

	// reply-to
	strResponse += "\")) ((\"";
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	if (HrGetOneProp(lpMessage, PR_SENT_REPRESENTING_NAME, &lpPropVal) == hrSuccess) {
		strResponse += EscapeString(lpPropVal->Value.lpszA);
	}

	strResponse += "\" NIL \"";
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	if (HrGetOneProp(lpMessage, PR_SENT_REPRESENTING_EMAIL_ADDRESS, &lpPropVal) == hrSuccess) {
		strPart = EscapeStringQT(lpPropVal->Value.lpszA);
	} else {
		strPart = "@";
	}

	ulPos = strPart.find("@", 0);
	if (ulPos != string::npos) {
		strResponse += EscapeStringQT(strPart.substr(0, ulPos));
		strResponse += "\" \"";
		strResponse += EscapeStringQT(strPart.substr(ulPos + 1));
	} else {
		strResponse += EscapeStringQT(strPart);
		strResponse += "\" \"";
	}

	strResponse += "\")) ";
	hr = lpMessage->GetRecipientTable(0, &lpTable);
	if (hr != hrSuccess)
		goto recipientsdone;

	hr = lpTable->SetColumns((LPSPropTagArray) &spt, 0);
	if (hr != hrSuccess)
		goto recipientsdone;

	hr = lpTable->QueryRows(-1, 0, &lpRows);
	if (hr != hrSuccess)
		goto recipientsdone;

	strResponse += "(";
	// to
	for (ulCount = 0; ulCount < lpRows->cRows; ulCount++) {
		if (lpRows->aRow[ulCount].lpProps[RECIPIENT_TYPE].Value.ul == MAPI_TO) {
			strResponse += "(\"";
			if (lpRows->aRow[ulCount].lpProps[DISPLAY_NAME].ulPropTag == PR_DISPLAY_NAME) {
				strResponse += EscapeString(lpRows->aRow[ulCount].lpProps[DISPLAY_NAME].Value.lpszA);
			}

			strResponse += "\" NIL \"";
			if (lpRows->aRow[ulCount].lpProps[EMAIL_ADDRESS].ulPropTag == PR_EMAIL_ADDRESS) {
				strPart = lpRows->aRow[ulCount].lpProps[EMAIL_ADDRESS].Value.lpszA;
			} else {
				strPart = "@";
			}

			ulPos = strPart.find("@");
			if (ulPos != string::npos) {
				strResponse += EscapeStringQT(strPart.substr(0, ulPos));
				strResponse += "\" \"";
				strResponse += EscapeStringQT(strPart.substr(ulPos + 1));
			} else {
				strResponse += EscapeStringQT(strPart);
				strResponse += "\" \"";
			}

			strResponse += "\") ";
		}
	}

	if (strResponse.compare(strResponse.size() - 1, 1, " ") == 0) {
		strResponse.resize(strResponse.size() - 1);
		strResponse += ") (";
	} else {
		strResponse.resize(strResponse.size() - 1);
		strResponse += "NIL (";
	}

	// cc
	for (ulCount = 0; ulCount < lpRows->cRows; ulCount++) {
		if (lpRows->aRow[ulCount].lpProps[RECIPIENT_TYPE].Value.ul == MAPI_CC) {
			strResponse += "(\"";
			if (lpRows->aRow[ulCount].lpProps[DISPLAY_NAME].ulPropTag == PR_DISPLAY_NAME) {
				strResponse += EscapeString(lpRows->aRow[ulCount].lpProps[DISPLAY_NAME].Value.lpszA);
			}

			strResponse += "\" NIL \"";
			if (lpRows->aRow[ulCount].lpProps[EMAIL_ADDRESS].ulPropTag == PR_EMAIL_ADDRESS) {
				strPart = lpRows->aRow[ulCount].lpProps[EMAIL_ADDRESS].Value.lpszA;
			} else {
				strPart = "@";
			}

			ulPos = strPart.find("@");
			if (ulPos != string::npos) {
				strResponse += EscapeStringQT(strPart.substr(0, ulPos));
				strResponse += "\" \"";
				strResponse += EscapeStringQT(strPart.substr(ulPos + 1));
			} else {
				strResponse += EscapeStringQT(strPart);
				strResponse += "\" \"";
			}
			strResponse += "\") ";
		}
	}

	if (strResponse.compare(strResponse.size() - 1, 1, " ") == 0) {
        strResponse.resize(strResponse.size() - 1);
        strResponse += ") (";
    } else {
        strResponse.resize(strResponse.size() - 1);
        strResponse += "NIL (";
    }

	// bcc
	for (ulCount = 0; ulCount < lpRows->cRows; ulCount++) {
		if (lpRows->aRow[ulCount].lpProps[RECIPIENT_TYPE].Value.ul == MAPI_BCC) {
			strResponse += "(\"";
			if (lpRows->aRow[ulCount].lpProps[DISPLAY_NAME].ulPropTag == PR_DISPLAY_NAME) {
				strResponse += EscapeString(lpRows->aRow[ulCount].lpProps[DISPLAY_NAME].Value.lpszA);
			}

			strResponse += "\" NIL \"";
			if (lpRows->aRow[ulCount].lpProps[EMAIL_ADDRESS].ulPropTag == PR_EMAIL_ADDRESS) {
				strPart = lpRows->aRow[ulCount].lpProps[EMAIL_ADDRESS].Value.lpszA;
			} else {
				strPart = "@";
			}

			ulPos = strPart.find("@");
			if (ulPos != string::npos) {
				strResponse += EscapeStringQT(strPart.substr(0, ulPos));
				strResponse += "\" \"";
				strResponse += EscapeStringQT(strPart.substr(ulPos + 1));
			} else {
				strResponse += EscapeStringQT(strPart);
				strResponse += "\" \"";
			}
			strResponse += "\") ";
		}
	}

    if (strResponse.compare(strResponse.size() - 1, 1, " ") == 0) {
        strResponse.resize(strResponse.size() - 1);
        strResponse += ") ";
    } else {
        strResponse.resize(strResponse.size() - 1);
        strResponse += "NIL ";
    }

recipientsdone:
	if (hr != hrSuccess) {
		strResponse += "NIL NIL NIL ";
		hr = hrSuccess;
	}

	// in reply to
	strResponse += "\"";
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	if (HrGetOneProp(lpMessage, PR_IN_REPLY_TO_ID, &lpPropVal) == hrSuccess) {
		strResponse += lpPropVal->Value.lpszA;
	}

	// internet message id
	strResponse += "\" \"";
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);
	lpPropVal = NULL;

	if (HrGetOneProp(lpMessage, PR_INTERNET_MESSAGE_ID, &lpPropVal) == hrSuccess) {
		strResponse += lpPropVal->Value.lpszA;
	}

	strResponse += "\")";

exit:
	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	return hr;
}

HRESULT IMAP::HrGetMessageFlags(string &strResponse, LPMESSAGE lpMessage, bool bRecent) {
	HRESULT hr = hrSuccess;
	LPSPropValue lpProps = NULL;
	ULONG cValues;
	SizedSPropTagArray(4, sptaFlagProps) = { 4, { PR_MESSAGE_FLAGS, PR_FLAG_STATUS, PR_MSG_STATUS, PR_LAST_VERB_EXECUTED } };
	
	if (!lpMessage) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	hr = lpMessage->GetProps((LPSPropTagArray)&sptaFlagProps, 0, &cValues, &lpProps);
	if (FAILED(hr))
		goto exit;
	hr = hrSuccess;

	strResponse += "FLAGS (";
	strResponse += PropsToFlags(lpProps, cValues, bRecent);
	strResponse += ")";

exit:
	if (lpProps)
		MAPIFreeBuffer(lpProps);

	return hr;
}

HRESULT IMAP::HrGetMessageInterDate(string &strResponse, LPMESSAGE lpMessage) {
	HRESULT hr = hrSuccess;
	LPSPropValue lpPropVal = NULL;

	if (!lpMessage) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	strResponse += "INTERNALDATE ";
	if (HrGetOneProp(lpMessage, PR_MESSAGE_DELIVERY_TIME, &lpPropVal) == hrSuccess
		|| HrGetOneProp(lpMessage, PR_CLIENT_SUBMIT_TIME, &lpPropVal) == hrSuccess) {
		strResponse += "\"";
		strResponse += FileTimeToString(lpPropVal->Value.ft);
		strResponse += "\"";
	} else {
		strResponse += "NIL";
	}

exit:
	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	return hr;
}

/*
 * RFC 3501, section 6.4.5:
 *
 *       BODY[<section>]<<partial>>
 *
 *       The text of a particular body section.  The section
 *       specification is a set of zero or more part specifiers
 *       delimited by periods.  A part specifier is either a part number
 *       or one of the following: HEADER, HEADER.FIELDS, 
 *       HEADER.FIELDS.NOT, MIME, and TEXT.  An empty section
 *       specification refers to the entire message, including the
 *       header.
 *
 *       --- end of RFC text
 *
 *		 Please note that there is a difference between getting MIME and HEADER:
 *		 - HEADER and TEXT are only used in the top-level object OR an embedded message/rfc822 message
 *		 - MIME are the Content-* headers for the MIME part, NEVER the header for a message/rfc822 message
 *		 - The 'whole' part is HEADER + TEXT for toplevel and message/rfc822 parts
 *		 - The 'whole' part does NOT include the MIME-IMB headers part
 *
 */
HRESULT IMAP::HrGetMessagePart(string &strMessagePart, string &strMessage, string strPartName) {
	HRESULT hr = hrSuccess;
	string::size_type ulPos;
	unsigned long int ulPartnr;
	string strHeaders;
	string strBoundary;
	size_t ulHeaderBegin;
	size_t ulHeaderEnd;
	size_t ulCounter;

	if (strPartName.find_first_of("123456789") == 0) {
	    std::string strNextPart;
	    // BODY[1] or BODY[1.2] etc
	    
	    // Find subsection
		ulPos = strPartName.find(".");
		if (ulPos == string::npos) { // first section
			ulPartnr = strtoul(strPartName.c_str(), NULL, 0);
		} else { // sub section
			ulPartnr = strtoul(strPartName.substr(0, ulPos).c_str(), NULL, 0);
		}

		// Find the correct part
		ulHeaderBegin = strMessage.find("boundary=\"");
		if (ulHeaderBegin != string::npos) {
			ulHeaderBegin += strlen("boundary=\"");;
			ulHeaderEnd = strMessage.find("\"", ulHeaderBegin);

			if (ulHeaderEnd != string::npos) {
			    // strBoundary is the boundary we are looking for
				strBoundary = strMessage.substr(ulHeaderBegin, ulHeaderEnd - ulHeaderBegin);
				// strHeaders is what we are looking for
				strHeaders = (string) "--" + strBoundary + "\r\n"; //Skip always the end header
				ulHeaderBegin = strMessage.find(strHeaders, ulHeaderBegin);

				// Find the section/part by looking for the Nth boundary string
				for (ulCounter = 0; ulCounter < ulPartnr && ulHeaderBegin != string::npos; ulCounter++) {
					ulHeaderBegin += strHeaders.size();
					ulHeaderEnd = ulHeaderBegin;
					ulHeaderBegin = strMessage.find(strHeaders, ulHeaderBegin);
				}

				if (ulHeaderBegin != string::npos) {
				    // Found it, discard data after and before the part we want
				    strMessage.erase(ulHeaderBegin, strMessage.size()-ulHeaderBegin);
				    strMessage.erase(0, ulHeaderEnd);
				} else {
				    // Didn't find it, see if we can find the trailing boundary
                    ulHeaderBegin = strMessage.find((string) "--" + strBoundary + "--\r\n", ulHeaderEnd);
                    if(ulHeaderBegin != string::npos) {
                        // If found, output everything up to the trailing boundary
                        strMessage.erase(ulHeaderBegin, strMessage.size()-ulHeaderBegin);
                        strMessage.erase(0, ulHeaderEnd);
                    } else {
                        // Otherwise, treat the rest of the message as data
                        strMessage.erase(0, ulHeaderEnd);
                    }
				}
			}
		}

		// We now have the entire MIME part in strMessage, decide what to do with it
		if (ulPos != string::npos) {
			// There are sub sections, see what we want to do
			strNextPart = strPartName.substr(ulPos+1);
			if(strNextPart == "MIME") {
			    // Handle MIME request
                ulPos = strMessage.find("\r\n\r\n");
                if (ulPos != string::npos) {
                    strMessagePart = strMessage.substr(0, ulPos+4); // include trailing \r\n\r\n (+4)
                } else {
                    // Only headers in the message
                    strMessagePart = strMessage + "\r\n\r\n";
                }
                
                // All done
                goto exit;
			} else if(strNextPart.find_first_of("123456789") == 0) {
			    // Handle Subpart
    			HrGetMessagePart(strMessagePart, strMessage, strNextPart);
    			
    			// All done
    			goto exit;
            }
        }
        
        // Handle any other request (HEADER, TEXT or 'empty'). This means we first skip the MIME-IMB headers
        // and process the rest from there.
        ulPos = strMessage.find("\r\n\r\n");
        if (ulPos != string::npos) {
            strMessage.erase(0, ulPos + 4);
        } else {
            // The message only has headers ?
            strMessage.clear();
        }
        // Handle HEADER and TEXT if requested        
        if (!strNextPart.empty())
            HrGetMessagePart(strMessagePart, strMessage, strNextPart);
        else
            // Swap to conserve memory: Original: strMessagePart = strMessage
            swap(strMessagePart, strMessage);
            
	} else if (strPartName == "TEXT") {
	    // Everything except for the headers
		ulPos = strMessage.find("\r\n\r\n");
		if (ulPos != string::npos) {
		    // Swap for less memory usage. Original: strMessagePart = strMessage.substr(ulPos+4)
		    strMessage.erase(0,ulPos+4);
		    swap(strMessage, strMessagePart);
		} else {
		    // The message only has headers ?
			strMessagePart.clear();
		}
	} else if (strPartName == "HEADER") {
	    // Only the headers
		ulPos = strMessage.find("\r\n\r\n");
		if (ulPos != string::npos) {
		    // Swap for less memory usage. Original: strMessagePart = strMessage.substr(0, ulPos+4);
		    strMessage.erase(ulPos+4, strMessage.size() - (ulPos+4));
		    swap(strMessagePart, strMessage);
		} else {
		    // Only headers in the message
			strMessagePart = strMessage + "\r\n\r\n";
		}
	} else if (Prefix(strPartName, "HEADER.FIELDS")) {
	    /* RFC 3501, section 6.4.5
	     *
	     * HEADER.FIELDS and HEADER.FIELDS.NOT are followed by a list of
         * field-name (as defined in [RFC-2822]) names, and return a
         * subset of the header.
         *
         * eg. HEADER.FIELDS (SUBJECT TO)
         * eg. HEADER.FIELDS.NOT (SUBJECT)
         */
        bool bNot = Prefix(strPartName, "HEADER.FIELDS.NOT");
        list<pair<string, string> > lstFields;
        list<pair<string, string> >::iterator iterField;
        string strFields;
        
        // Parse headers in message
        HrParseHeaders(strMessage, lstFields);
        
        // Get (<fields>)
        HrGetSubString(strFields, strPartName, "(", ")");
        
        strMessagePart.clear();
        
        if(bNot) {
            set<string> setFields;
            set<string>::iterator iterSet;

            // Get fields as set
            HrTokenize(setFields, strFields);
            
            // Output all headers except those specified
            for(iterField = lstFields.begin(); iterField != lstFields.end(); iterField++) {
                string strFieldUpper = iterField->first;
                ToUpper(strFieldUpper);
                if(setFields.find(strFieldUpper) == setFields.end()) {
                    strMessagePart += iterField->first;
                    strMessagePart += ": ";
                    strMessagePart += iterField->second;
                    strMessagePart += "\r\n";
                }
            }    
        } else {
            list<string> lstReqFields;
            list<string>::iterator iterReqField;

            // Get fields as list
            HrTokenize(lstReqFields, strFields);
            
            // Output headers specified, in order of field set
            for(iterReqField = lstReqFields.begin(); iterReqField != lstReqFields.end(); iterReqField++) {
                for(iterField = lstFields.begin(); iterField != lstFields.end(); iterField++) {
                    if(CaseCompare(*iterReqField, iterField->first)) {
                        strMessagePart += iterField->first;
                        strMessagePart += ": ";
                        strMessagePart += iterField->second;
                        strMessagePart += "\r\n";
						break;
                    }
                }
            }
        }
	} else {
		strMessagePart = "NIL";
	}
	
exit:
	return hr;
}

/** 
 * Convert a sequence number to it's actual number. It will either
 * return a number or a UID, depending on the input.
 * A special treatment for 
 * 
 * @param[in] szNr a number of the sequence input
 * @param[in] bUID sequence input are UID numbers or not
 * 
 * @return the number corresponding to the input.
 */
ULONG IMAP::LastOrNumber(const char *szNr, bool bUID)
{
	if (*szNr != '*')
		return atoui(szNr);

	if (!bUID)
		return lstFolderMailEIDs.size();

	if (lstFolderMailEIDs.empty())
		return 0;				// special case: return an "invalid" number
	else
		return lstFolderMailEIDs.back().ulUid;
}

/** 
 * Convert an IMAP sequence set to a flat list of email numbers. See
 * RFC-3501 paragraph 9 for the syntax.  This function will return the
 * closest range of requested items, and thus may return an empty
 * list.
 * 
 * @param[in] strSeqSet IMAP sequence set of UID numbers
 * @param[out] lstMails flat list of email numbers
 * 
 * @return MAPI Error code
 */
HRESULT IMAP::HrParseSeqUidSet(const string &strSeqSet, list<ULONG> &lstMails) {
	HRESULT hr = hrSuccess;
	vector<string> vSequences;
	string::size_type ulPos = 0;
	ULONG ulMailnr;
	ULONG ulBeginMailnr;

	// split different sequence parts into a vector
	vSequences = tokenize(strSeqSet, ',');

	for (ULONG i = 0; i < vSequences.size(); i++) {
		ulPos = vSequences[i].find(':');
		if (ulPos == string::npos) {
			// single number
			ulMailnr = LastOrNumber(vSequences[i].c_str(), true);

			vector<SMail>::iterator i = find(lstFolderMailEIDs.begin(), lstFolderMailEIDs.end(), ulMailnr);
			if (i != lstFolderMailEIDs.end())
				lstMails.push_back(std::distance(lstFolderMailEIDs.begin(), i));
		} else {
			// range
			ulBeginMailnr = LastOrNumber(vSequences[i].c_str(), true);
			ulMailnr = LastOrNumber(vSequences[i].c_str() + ulPos + 1, true);

			if (ulBeginMailnr > ulMailnr)
				swap(ulBeginMailnr, ulMailnr);

			vector<SMail>::iterator b = std::lower_bound(lstFolderMailEIDs.begin(), lstFolderMailEIDs.end(), ulBeginMailnr);
			vector<SMail>::iterator e = std::upper_bound(lstFolderMailEIDs.begin(), lstFolderMailEIDs.end(), ulMailnr);

			for (vector<SMail>::iterator i = b; i != e; i++)
				lstMails.push_back(std::distance(lstFolderMailEIDs.begin(), i));
		}
	}

	lstMails.sort();
	lstMails.unique();

	return hr;
}

/** 
 * Convert an IMAP sequence set to a flat list of email numbers. See
 * RFC-3501 paragraph 9 for the syntax.  These exact numbers requested
 * must be present in the folder, otherwise an error will be returned.
 *
 * @param[in] strSeqSet IMAP sequence set of direct numbers
 * @param[out] lstMails flat list of email numbers
 * 
 * @return MAPI Error code
 */
HRESULT IMAP::HrParseSeqSet(const string &strSeqSet, list<ULONG> &lstMails) {
	HRESULT hr = hrSuccess;
	vector<string> vSequences;
	string::size_type ulPos = 0;
	ULONG ulMailnr;
	ULONG ulBeginMailnr;

	if(lstFolderMailEIDs.size() == 0) {
	    hr = MAPI_E_NOT_FOUND;
	    goto exit;
    }

	// split different sequence parts into a vector
	vSequences = tokenize(strSeqSet, ',');

	for (ULONG i = 0; i < vSequences.size(); i++) {
		ulPos = vSequences[i].find(':');
		if (ulPos == string::npos) {
			// single number
			ulMailnr = LastOrNumber(vSequences[i].c_str(), false) - 1;
			if (ulMailnr > lstFolderMailEIDs.size()) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}
			lstMails.push_back(ulMailnr);
		} else {
			// range
			ulBeginMailnr = LastOrNumber(vSequences[i].c_str(), false) - 1;
			ulMailnr = LastOrNumber(vSequences[i].c_str() + ulPos + 1, false) - 1;

			if (ulBeginMailnr > ulMailnr)
				swap(ulBeginMailnr, ulMailnr);

			if (ulBeginMailnr > lstFolderMailEIDs.size() || ulMailnr > lstFolderMailEIDs.size()) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			for (ULONG j = ulBeginMailnr; j <= ulMailnr; j++)
				lstMails.push_back(j);
		}
	}

	lstMails.sort();
	lstMails.unique();

exit:
	return hr;
}

HRESULT IMAP::HrStore(list<ULONG> &lstMails, string strMsgDataItemName, string strMsgDataItemValue) {
	HRESULT hr = hrSuccess;
	vector<string> lstFlags;
	ULONG ulCurrent;
	LPMESSAGE lpMessage = NULL;
	LPSPropValue lpPropVal = NULL;
	LPSPropTagArray lpPropTagArray = NULL;
	ULONG cValues;
	ULONG ulObjType;
	list<ULONG>::iterator lpMail;
	string strNewFlags;

	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	ToUpper(strMsgDataItemName);
	ToUpper(strMsgDataItemValue);
	if (strMsgDataItemValue.size() > 1 && strMsgDataItemValue[0] == '(') {
		strMsgDataItemValue.erase(0, 1);
		strMsgDataItemValue.erase(strMsgDataItemValue.size() - 1, 1);
	}
	HrSplitInput(strMsgDataItemValue, lstFlags);

	for (lpMail = lstMails.begin(); lpMail != lstMails.end(); lpMail++) {
		hr = lpSession->OpenEntry(lstFolderMailEIDs[(*lpMail)].sEntryID.cb, (LPENTRYID) lstFolderMailEIDs[(*lpMail)].sEntryID.lpb,
								&IID_IMessage, MAPI_MODIFY, &ulObjType, (LPUNKNOWN *) &lpMessage);
		if (hr != hrSuccess) {
			goto exit;
		}

		// FLAGS, FLAGS.SILENT, +FLAGS, +FLAGS.SILENT, -FLAGS, -FLAGS.SILENT
		if (strMsgDataItemName.compare(0, 5, "FLAGS") == 0) {

			if (strMsgDataItemValue.find("\\SEEN") == string::npos) {
				hr = lpMessage->SetReadFlag(CLEAR_READ_FLAG);
				if (hr != hrSuccess)
					goto exit;
			} else {
				hr = lpMessage->SetReadFlag(SUPPRESS_RECEIPT);
				if (hr != hrSuccess)
					goto exit;
			}

			if (lpPropTagArray)
				MAPIFreeBuffer(lpPropTagArray);
			lpPropTagArray = NULL;

			hr = MAPIAllocateBuffer(CbNewSPropTagArray(5), (void**)&lpPropTagArray);
			if (hr != hrSuccess)
				goto exit;

			lpPropTagArray->cValues = 5;
			lpPropTagArray->aulPropTag[0] = PR_MSG_STATUS;
			lpPropTagArray->aulPropTag[1] = PR_FLAG_STATUS;
			lpPropTagArray->aulPropTag[2] = PR_ICON_INDEX;
			lpPropTagArray->aulPropTag[3] = PR_LAST_VERB_EXECUTED;
			lpPropTagArray->aulPropTag[4] = PR_LAST_VERB_EXECUTION_TIME;

			hr = lpMessage->GetProps(lpPropTagArray, 0, &cValues, &lpPropVal);
			if (FAILED(hr))
				goto exit;
			cValues = 5;

			lpPropVal[1].ulPropTag = PR_FLAG_STATUS;
			if (strMsgDataItemValue.find("\\FLAGGED") == string::npos) {
				lpPropVal[1].Value.ul = 0;
			} else {
				lpPropVal[1].Value.ul = 2;
			}

			if (lpPropVal[2].ulPropTag != PR_ICON_INDEX) {
				lpPropVal[2].ulPropTag = PR_ICON_INDEX;
				lpPropVal[2].Value.l = ICON_FOLDER_DEFAULT;
			}

			if (lpPropVal[0].ulPropTag != PR_MSG_STATUS) {
				lpPropVal[0].ulPropTag = PR_MSG_STATUS;
				lpPropVal[0].Value.ul = 0;
			}

			if (strMsgDataItemValue.find("\\ANSWERED") == string::npos) {
				lpPropVal[0].Value.ul &= ~MSGSTATUS_ANSWERED;
				cValues -= 2;	// leave PR_LAST_VERB_EXECUTED properties
			} else {
				lpPropVal[0].Value.ul |= MSGSTATUS_ANSWERED;

				lpPropVal[3].ulPropTag = PR_LAST_VERB_EXECUTED;
				lpPropVal[3].Value.ul = NOTEIVERB_REPLYTOSENDER;

				lpPropVal[4].ulPropTag = PR_LAST_VERB_EXECUTION_TIME;
				GetSystemTimeAsFileTime(&lpPropVal[4].Value.ft);
			}

			if (strMsgDataItemValue.find("\\DELETED") == string::npos) {
				lpPropVal[0].Value.ul &= ~MSGSTATUS_DELMARKED;
			} else {
				lpPropVal[0].Value.ul |= MSGSTATUS_DELMARKED;
			}

			hr = lpMessage->SetProps(cValues, lpPropVal, NULL);
			if (hr != hrSuccess)
				goto exit;

			hr = lpMessage->SaveChanges(KEEP_OPEN_READWRITE | FORCE_SAVE);
			if (hr != hrSuccess)
				goto exit;
		} else if (strMsgDataItemName.compare(0, 6, "+FLAGS") == 0) {
			for (ulCurrent = 0; ulCurrent < lstFlags.size(); ulCurrent++) {
				if (lstFlags[ulCurrent] == "\\SEEN") {
					hr = lpMessage->SetReadFlag(SUPPRESS_RECEIPT);
					if (hr != hrSuccess)
						goto exit;
				} else if (lstFlags[ulCurrent] == "\\DRAFT") {
					// not allowed
				} else if (lstFlags[ulCurrent] == "\\FLAGGED") {
					if (lpPropVal)
						MAPIFreeBuffer(lpPropVal);

					hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) &lpPropVal);
					if (hr != hrSuccess)
						goto exit;

					lpPropVal->ulPropTag = PR_FLAG_STATUS;
					lpPropVal->Value.ul = 2; // 0: none, 1: green ok mark, 2: red flag
					HrSetOneProp(lpMessage, lpPropVal);
					// TODO: set PR_FLAG_ICON here too?
				} else if (lstFlags[ulCurrent] == "\\ANSWERED") {
					if (lpPropVal)
						MAPIFreeBuffer(lpPropVal);
					lpPropVal = NULL;

					if (lpPropTagArray)
						MAPIFreeBuffer(lpPropTagArray);
					lpPropTagArray = NULL;

					hr = MAPIAllocateBuffer(CbNewSPropTagArray(4), (void**)&lpPropTagArray);
					if (hr != hrSuccess)
						goto exit;

					lpPropTagArray->cValues = 4;
					lpPropTagArray->aulPropTag[0] = PR_MSG_STATUS;
					lpPropTagArray->aulPropTag[1] = PR_ICON_INDEX;
					lpPropTagArray->aulPropTag[2] = PR_LAST_VERB_EXECUTED;
					lpPropTagArray->aulPropTag[3] = PR_LAST_VERB_EXECUTION_TIME;

					hr = lpMessage->GetProps(lpPropTagArray, 0, &cValues, &lpPropVal);
					if (FAILED(hr))
						goto exit;
					cValues = 4;

					if (lpPropVal[0].ulPropTag != PR_MSG_STATUS) {
						lpPropVal[0].ulPropTag = PR_MSG_STATUS;
						lpPropVal[0].Value.ul = 0;
					}
					lpPropVal->Value.ul |= MSGSTATUS_ANSWERED;

					lpPropVal[1].ulPropTag = PR_ICON_INDEX;
					lpPropVal[1].Value.l = ICON_MAIL_REPLIED;
					lpPropVal[2].ulPropTag = PR_LAST_VERB_EXECUTED;
					lpPropVal[2].Value.ul = NOTEIVERB_REPLYTOSENDER;
					lpPropVal[3].ulPropTag = PR_LAST_VERB_EXECUTION_TIME;
					GetSystemTimeAsFileTime(&lpPropVal[3].Value.ft);

					hr = lpMessage->SetProps(cValues, lpPropVal, NULL);
					if (hr != hrSuccess)
						goto exit;
				} else if (lstFlags[ulCurrent] == "\\DELETED") {
					if (lpPropVal)
						MAPIFreeBuffer(lpPropVal);
					lpPropVal = NULL;

					if (HrGetOneProp(lpMessage, PR_MSG_STATUS, &lpPropVal) != hrSuccess) {
						hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) &lpPropVal);
						if (hr != hrSuccess)
							goto exit;
						lpPropVal->ulPropTag = PR_MSG_STATUS;
						lpPropVal->Value.ul = 0;
					}

					lpPropVal->Value.ul |= MSGSTATUS_DELMARKED;
					HrSetOneProp(lpMessage, lpPropVal);
				}
				lpMessage->SaveChanges(KEEP_OPEN_READWRITE | FORCE_SAVE);
			}
		} else if (strMsgDataItemName.compare(0, 6, "-FLAGS") == 0) {
			for (ulCurrent = 0; ulCurrent < lstFlags.size(); ulCurrent++) {
				if (lstFlags[ulCurrent] == "\\SEEN") {
					hr = lpMessage->SetReadFlag(CLEAR_READ_FLAG);
					if (hr != hrSuccess)
						goto exit;
				} else if (lstFlags[ulCurrent] == "\\DRAFT") {
					// not allowed
				} else if (lstFlags[ulCurrent] == "\\FLAGGED") {
					if (lpPropVal == NULL) {
						hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) &lpPropVal);
						if (hr != hrSuccess)
							goto exit;
					}

					lpPropVal->ulPropTag = PR_FLAG_STATUS;
					lpPropVal->Value.ul = 0;
					HrSetOneProp(lpMessage, lpPropVal);
				} else if (lstFlags[ulCurrent] == "\\ANSWERED") {
					if (lpPropVal)
						MAPIFreeBuffer(lpPropVal);
					lpPropVal = NULL;

					if (lpPropTagArray)
						MAPIFreeBuffer(lpPropTagArray);
					lpPropTagArray = NULL;

					hr = MAPIAllocateBuffer(CbNewSPropTagArray(4), (void**)&lpPropTagArray);
					if (hr != hrSuccess)
						goto exit;

					lpPropTagArray->cValues = 4;
					lpPropTagArray->aulPropTag[0] = PR_MSG_STATUS;
					lpPropTagArray->aulPropTag[1] = PR_ICON_INDEX;
					lpPropTagArray->aulPropTag[2] = PR_LAST_VERB_EXECUTED;
					lpPropTagArray->aulPropTag[3] = PR_LAST_VERB_EXECUTION_TIME;

					hr = lpMessage->GetProps(lpPropTagArray, 0, &cValues, &lpPropVal);
					if (FAILED(hr))
						goto exit;
					cValues = 4;

					if (lpPropVal[0].ulPropTag != PR_MSG_STATUS) {
						lpPropVal[0].ulPropTag = PR_MSG_STATUS;
						lpPropVal[0].Value.ul = 0;
					}
					lpPropVal->Value.ul &= ~MSGSTATUS_ANSWERED;

					lpPropVal[1].ulPropTag = PR_ICON_INDEX;
					lpPropVal[1].Value.l = ICON_FOLDER_DEFAULT;
					lpPropVal[2].ulPropTag = PR_LAST_VERB_EXECUTED;
					lpPropVal[2].Value.ul = NOTEIVERB_OPEN;
					lpPropVal[3].ulPropTag = PR_LAST_VERB_EXECUTION_TIME;
					GetSystemTimeAsFileTime(&lpPropVal[3].Value.ft);

					hr = lpMessage->SetProps(cValues, lpPropVal, NULL);
					if (hr != hrSuccess)
						goto exit;
				} else if (lstFlags[ulCurrent] == "\\DELETED") {
					if (lpPropVal)
						MAPIFreeBuffer(lpPropVal);
					lpPropVal = NULL;

					if (HrGetOneProp(lpMessage, PR_MSG_STATUS, &lpPropVal) != hrSuccess) {
						hr = MAPIAllocateBuffer(sizeof(SPropValue), (LPVOID *) &lpPropVal);
						if (hr != hrSuccess)
							goto exit;
						lpPropVal->ulPropTag = PR_MSG_STATUS;
						lpPropVal->Value.ul = 0;
					}

					lpPropVal->Value.ul &= ~MSGSTATUS_DELMARKED;
					HrSetOneProp(lpMessage, lpPropVal);
				}
				lpMessage->SaveChanges(KEEP_OPEN_READWRITE | FORCE_SAVE);
			}
		}

		// Get the newly updated flags
        hr = HrGetMessageFlags(strNewFlags, lpMessage, lstFolderMailEIDs[*lpMail].bRecent);
        if(hr != hrSuccess)
            goto exit;
            
        // Update our internal flag status
        lstFolderMailEIDs[*lpMail].strFlags = strNewFlags;

		if (lpMessage)
			lpMessage->Release();
		lpMessage = NULL;
	} // loop on mails

	if (strMsgDataItemName.size() > 7 && strMsgDataItemName.compare(strMsgDataItemName.size() - 7, 7, ".SILENT") == 0) {
		hr = MAPI_E_NOT_ME;		// abuse error code that will not be used elsewhere from this function
	}

exit:
	if (lpPropTagArray)
		MAPIFreeBuffer(lpPropTagArray);

	if (lpPropVal)
		MAPIFreeBuffer(lpPropVal);

	if (lpMessage)
		lpMessage->Release();

	return hr;
}

HRESULT IMAP::HrCopy(list<ULONG> &lstMails, string strFolder, bool bMove) {
	HRESULT hr = hrSuccess;
	IMAPIFolder *lpFromFolder = NULL;
	IMAPIFolder *lpDestFolder = NULL;
	list<ULONG>::iterator lpMail;
	list<SFolder>::iterator iDestFolder;
	ULONG ulCount;
	ENTRYLIST sEntryList;

	sEntryList.lpbin = NULL;

	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	hr = HrFindFolder(strCurrentFolder, bCurrentFolderReadOnly, &lpFromFolder);
	if (hr != hrSuccess)
		goto exit;

	// get dest folder
	strFolder = IMAP2MAPICharset(strFolder);

	hr = HrFindFolder(strFolder, false, &lpDestFolder);
	if (hr != hrSuccess)
		goto exit;

	sEntryList.cValues = lstMails.size();
	MAPIAllocateBuffer(sizeof(SBinary) * lstMails.size(), (LPVOID *) &sEntryList.lpbin);
	ulCount = 0;

	for (lpMail = lstMails.begin(); lpMail != lstMails.end(); lpMail++) {
		sEntryList.lpbin[ulCount].cb = lstFolderMailEIDs[(*lpMail)].sEntryID.cb;
		sEntryList.lpbin[ulCount].lpb = lstFolderMailEIDs[(*lpMail)].sEntryID.lpb;
		ulCount++;
	}

	hr = lpFromFolder->CopyMessages(&sEntryList, NULL, lpDestFolder, 0, NULL, bMove ? MESSAGE_MOVE : 0);

exit:
	if (sEntryList.lpbin)
		MAPIFreeBuffer(sEntryList.lpbin);

	if (lpFromFolder)
		lpFromFolder->Release();

	if (lpDestFolder)
		lpDestFolder->Release();

	return hr;
}

HRESULT IMAP::HrSearch(vector<string> &lstSearchCriteria, ULONG &ulStartCriteria, list<ULONG> &lstMailnr, ECIConv *iconv) {
	HRESULT hr = hrSuccess;
	string strSearchCriterium;
	vector<string> vSubSearch;
	vector<LPSRestriction> lstRestrictions;
	list<ULONG> lstMails;
	list<ULONG>::iterator lpMail;
	LPSRestriction lpRootRestrict = NULL;
	LPSRestriction lpRestriction, lpExtraRestriction;
	IMAPIFolder *lpFolder = NULL;
	LPMAPITABLE lpTable = NULL;
	LPSRowSet lpRows = NULL;
	LPSPropValue lpPropVal;
	ULONG ulMailnr, ulRownr, ulCount;
	FILETIME ft;
	char *szBuffer;
	enum { EID, NUM_COLS };
	SizedSPropTagArray(NUM_COLS, spt) = { NUM_COLS, {PR_EC_IMAP_ID} };
	vector<SMail>::iterator iterFolderMails;
	map<unsigned int, unsigned int> mapUIDs;
	map<unsigned int, unsigned int>::iterator iterUID;
	int n = 0;
	SRestriction sAndRestriction;
	SRestriction sPropertyRestriction;
	
	if (strCurrentFolder.empty() || !lpSession) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	// don't search if only search for uid, sequence set, all, recent, new or old
	strSearchCriterium = lstSearchCriteria[ulStartCriteria];
	ToUpper(strSearchCriterium);
	if (lstSearchCriteria.size() - ulStartCriteria == 2 && strSearchCriterium == "UID") {
		hr = HrParseSeqUidSet(lstSearchCriteria[ulStartCriteria + 1], lstMailnr);
		goto exit;
	}

	if (lstSearchCriteria.size() - ulStartCriteria == 1) {
		if (strSearchCriterium.find_first_of("123456789*") == 0) {
			hr = HrParseSeqSet(lstSearchCriteria[ulStartCriteria], lstMailnr);
			goto exit;
		} else if (strSearchCriterium == "ALL") {
			for (ulMailnr = 0; ulMailnr < lstFolderMailEIDs.size(); ulMailnr++) {
				lstMailnr.push_back(ulMailnr);
			}
			goto exit;
		} else if (strSearchCriterium == "RECENT") {
			for (ulMailnr = 0; ulMailnr < lstFolderMailEIDs.size(); ulMailnr++) {
			    if(lstFolderMailEIDs[ulMailnr].bRecent)
    				lstMailnr.push_back(ulMailnr);
			}
			goto exit;
		} else if (strSearchCriterium == "NEW") {
			for (ulMailnr = 0; ulMailnr < lstFolderMailEIDs.size(); ulMailnr++) {
			    if(lstFolderMailEIDs[ulMailnr].bRecent && lstFolderMailEIDs[ulMailnr].strFlags.find("Seen") == std::string::npos)
    				lstMailnr.push_back(ulMailnr);
			}
			goto exit;
		} else if (strSearchCriterium == "OLD") {
			for (ulMailnr = 0; ulMailnr < lstFolderMailEIDs.size(); ulMailnr++) {
			    if(!lstFolderMailEIDs[ulMailnr].bRecent)
    				lstMailnr.push_back(ulMailnr);
			}
			goto exit;
		}
	}
	
	// Make a map of UID->ID
	for(iterFolderMails = lstFolderMailEIDs.begin(); iterFolderMails != lstFolderMailEIDs.end(); iterFolderMails++) {
	    mapUIDs[iterFolderMails->ulUid] = n;
	    n++;
	}

	hr = HrFindFolder(strCurrentFolder, bCurrentFolderReadOnly, &lpFolder);
	if (hr != hrSuccess) {
		goto exit;
	}

	hr = lpFolder->GetContentsTable(MAPI_DEFERRED_ERRORS, &lpTable);
	if (hr != hrSuccess) {
		goto exit;
	}

	hr = MAPIAllocateBuffer(sizeof(SRestriction), (LPVOID *) &lpRootRestrict);
	if (hr != hrSuccess) {
		goto exit;
	}

	lstRestrictions.push_back(lpRootRestrict);

	// Thunderbird searches:
	// or:
	// 12 uid SEARCH UNDELETED (OR SUBJECT "Undelivered" TO "henk")
	// 15 uid SEARCH UNDELETED (OR (OR SUBJECT "sender" SUBJECT "mail") SUBJECT "returned")
	// 17 uid SEARCH UNDELETED (OR SUBJECT "Undelivered" NOT TO "henk")
	// and:
	// 14 uid SEARCH UNDELETED SUBJECT "Undelivered" TO "henk"
	// 16 uid SEARCH UNDELETED SUBJECT "Undelivered" NOT TO "henk"
	// both cases, ulStartCriteria == 3: UNDELETED

	// this breaks to following search:
	//   (or subject henk subject kees) (or to henk from kees)
	// since this will translate in all or's, not: and(or(subj:henk,subj:kees),or(to:henk,from:kees))
	// however, thunderbird cannot build such a query, so we don't care currently.

	while (ulStartCriteria < lstSearchCriteria.size())
	{
		if (lstSearchCriteria[ulStartCriteria][0] == '(') {
			// remove all () and [], and resplit.
			strSearchCriterium.clear();
			for (string::iterator i = lstSearchCriteria[ulStartCriteria].begin(); i != lstSearchCriteria[ulStartCriteria].end(); i++) {
				if (*i == '(' || *i == ')' || *i == '[' || *i == ']')
					continue;
				strSearchCriterium += *i;
			}
			vSubSearch.clear();
			HrSplitInput(strSearchCriterium, vSubSearch);

			// replace in list.
			lstSearchCriteria.erase(lstSearchCriteria.begin() + ulStartCriteria);
			lstSearchCriteria.insert(lstSearchCriteria.begin() + ulStartCriteria, vSubSearch.begin(), vSubSearch.end());
		}

		strSearchCriterium = lstSearchCriteria[ulStartCriteria];
		ToUpper(strSearchCriterium);

		if (lstRestrictions.size() == 1) {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpRestriction);
			if (hr != hrSuccess)
				goto exit;

			lstRestrictions[0]->rt = RES_AND; // RES_AND / RES_OR
			lstRestrictions[0]->res.resAnd.cRes = 2;
			lstRestrictions[0]->res.resAnd.lpRes = lpRestriction;
			lstRestrictions[0] = &lpRestriction[1];
		} else {
			lpRestriction = lstRestrictions[lstRestrictions.size() - 1];
			lstRestrictions.pop_back();
		}

		if (strSearchCriterium.find_first_of("123456789*") == 0) {	// sequence set
			lstMails.clear();

			hr = HrParseSeqSet(strSearchCriterium, lstMails);
			if (hr != hrSuccess)
				goto exit;

			hr = MAPIAllocateMore(sizeof(SRestriction) * lstMails.size(), lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_OR;
			lpRestriction->res.resOr.cRes = lstMails.size();
			lpRestriction->res.resOr.lpRes = lpExtraRestriction;
			ulCount = 0;

			for (lpMail = lstMails.begin(); lpMail != lstMails.end(); lpMail++) {
				hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
				if (hr != hrSuccess)
					goto exit;

				lpPropVal->ulPropTag = PR_EC_IMAP_ID;
				lpPropVal->Value.ul = lstFolderMailEIDs[(*lpMail)].ulUid;
				lpExtraRestriction[ulCount].rt = RES_PROPERTY;
				lpExtraRestriction[ulCount].res.resProperty.relop = RELOP_EQ;
				lpExtraRestriction[ulCount].res.resProperty.ulPropTag = PR_EC_IMAP_ID;
				lpExtraRestriction[ulCount].res.resProperty.lpProp = lpPropVal;
				ulCount++;
			}

			ulStartCriteria++;
		} else if (strSearchCriterium == "ALL" || strSearchCriterium == "NEW" || strSearchCriterium == "RECENT") {	// do nothing
			lpRestriction->rt = RES_EXIST;
			lpRestriction->res.resExist.ulReserved1 = 0;
			lpRestriction->res.resExist.ulReserved2 = 0;
			lpRestriction->res.resExist.ulPropTag = PR_ENTRYID;
			ulStartCriteria++;
		} else if (strSearchCriterium == "ANSWERED") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_MSG_STATUS;
			lpExtraRestriction[1].rt = RES_BITMASK;
			lpExtraRestriction[1].res.resBitMask.relBMR = BMR_NEZ;
			lpExtraRestriction[1].res.resBitMask.ulPropTag = PR_MSG_STATUS;
			lpExtraRestriction[1].res.resBitMask.ulMask = MSGSTATUS_ANSWERED;
			ulStartCriteria++;
			// TODO: find also in PR_LAST_VERB_EXECUTED
		} else if (strSearchCriterium == "BEFORE") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
			lpPropVal->ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
			ft = StringToFileTime(lstSearchCriteria[ulStartCriteria + 1].c_str());
			lpPropVal->Value.ft = ft;
			lpExtraRestriction[1].rt = RES_PROPERTY;
			lpExtraRestriction[1].res.resProperty.relop = RELOP_LT;
			lpExtraRestriction[1].res.resProperty.ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
			lpExtraRestriction[1].res.resProperty.lpProp = lpPropVal;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "BODY") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_BODY;
			lpExtraRestriction[1].rt = RES_CONTENT;
			lpExtraRestriction[1].res.resContent.ulFuzzyLevel = lstSearchCriteria[ulStartCriteria+1].size() > 0 ? (FL_SUBSTRING | FL_IGNORECASE) : FL_FULLSTRING;
			lpExtraRestriction[1].res.resContent.ulPropTag = PR_BODY;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			if (iconv)
				iconv->convert(lstSearchCriteria[ulStartCriteria+1]);

			hr = MAPIAllocateMore(sizeof(char) * (lstSearchCriteria[ulStartCriteria + 1].size() + 1), lpRootRestrict,
								  (LPVOID *) &szBuffer);
			if (hr != hrSuccess)
				goto exit;

			memcpy(szBuffer, lstSearchCriteria[ulStartCriteria + 1].c_str(), lstSearchCriteria[ulStartCriteria + 1].size() + 1);
			lpPropVal->ulPropTag = PR_BODY;
			lpPropVal->Value.lpszA = szBuffer;
			lpExtraRestriction[1].res.resContent.lpProp = lpPropVal;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "DELETED") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_MSG_STATUS;
			lpExtraRestriction[1].rt = RES_BITMASK;
			lpExtraRestriction[1].res.resBitMask.relBMR = BMR_NEZ;
			lpExtraRestriction[1].res.resBitMask.ulPropTag = PR_MSG_STATUS;
			lpExtraRestriction[1].res.resBitMask.ulMask = MSGSTATUS_DELMARKED;
			ulStartCriteria++;
		} else if (strSearchCriterium == "DRAFT") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_MSG_STATUS;
			lpExtraRestriction[1].rt = RES_BITMASK;
			lpExtraRestriction[1].res.resBitMask.relBMR = BMR_NEZ;
			lpExtraRestriction[1].res.resBitMask.ulPropTag = PR_MSG_STATUS;
			lpExtraRestriction[1].res.resBitMask.ulMask = MSGSTATUS_DRAFT;
			ulStartCriteria++;
			// FIXME: add restriction to find PR_MESSAGE_FLAGS with MSGFLAG_UNSENT on
		} else if (strSearchCriterium == "FLAGGED") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_FLAG_STATUS;
			lpExtraRestriction[1].rt = RES_BITMASK;
			lpExtraRestriction[1].res.resBitMask.relBMR = BMR_NEZ;
			lpExtraRestriction[1].res.resBitMask.ulPropTag = PR_FLAG_STATUS;
			lpExtraRestriction[1].res.resBitMask.ulMask = 65535;
			ulStartCriteria++;
		} else if (strSearchCriterium == "FROM") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_OR;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_CONTENT;
			lpExtraRestriction[0].res.resContent.ulFuzzyLevel = lstSearchCriteria[ulStartCriteria+1].size() > 0 ? (FL_SUBSTRING | FL_IGNORECASE) : FL_FULLSTRING;
			lpExtraRestriction[0].res.resContent.ulPropTag = PR_SENT_REPRESENTING_NAME;
			lpExtraRestriction[1].rt = RES_CONTENT;
			lpExtraRestriction[1].res.resContent.ulFuzzyLevel = lstSearchCriteria[ulStartCriteria+1].size() > 0 ? (FL_SUBSTRING | FL_IGNORECASE) : FL_FULLSTRING;
			lpExtraRestriction[1].res.resContent.ulPropTag = PR_SENT_REPRESENTING_EMAIL_ADDRESS;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			if (iconv)
				iconv->convert(lstSearchCriteria[ulStartCriteria+1]);

			hr = MAPIAllocateMore(sizeof(char) * (lstSearchCriteria[ulStartCriteria + 1].size() + 1), lpRootRestrict,
								  (LPVOID *) &szBuffer);
			if (hr != hrSuccess)
				goto exit;

			memcpy(szBuffer, lstSearchCriteria[ulStartCriteria + 1].c_str(), lstSearchCriteria[ulStartCriteria + 1].size() + 1);
			lpPropVal->ulPropTag = PR_SENT_REPRESENTING_NAME;
			lpPropVal->Value.lpszA = szBuffer;
			lpExtraRestriction[0].res.resContent.lpProp = lpPropVal;
			lpExtraRestriction[1].res.resContent.lpProp = lpPropVal;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "KEYWORD") {
			lpRestriction->rt = RES_BITMASK;
			lpRestriction->res.resBitMask.relBMR = BMR_NEZ;
			lpRestriction->res.resBitMask.ulPropTag = PR_ENTRYID;
			lpRestriction->res.resBitMask.ulMask = 0;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "LARGER") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpPropVal->ulPropTag = PR_MESSAGE_SIZE;
			lpPropVal->Value.ul = strtoul(lstSearchCriteria[ulStartCriteria + 1].c_str(), NULL, 0);
			lpRestriction->rt = RES_PROPERTY;
			lpRestriction->res.resProperty.relop = RELOP_GT;
			lpRestriction->res.resProperty.ulPropTag = PR_MESSAGE_SIZE;
			lpRestriction->res.resProperty.lpProp = lpPropVal;
			ulStartCriteria += 2;
			// NEW done with ALL
		} else if (strSearchCriterium == "NOT") {
			hr = MAPIAllocateMore(sizeof(SRestriction), lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_NOT;
			lpRestriction->res.resNot.ulReserved = 0;
			lpRestriction->res.resNot.lpRes = lpExtraRestriction;
			lstRestrictions.push_back(lpExtraRestriction);
			ulStartCriteria++;
		} else if (strSearchCriterium == "OLD") {	// none?
			lpRestriction->rt = RES_BITMASK;
			lpRestriction->res.resBitMask.relBMR = BMR_NEZ;
			lpRestriction->res.resBitMask.ulPropTag = PR_ENTRYID;
			lpRestriction->res.resBitMask.ulMask = 0;
			ulStartCriteria++;
		} else if (strSearchCriterium == "ON") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SRestriction) * 3, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 3;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
			lpPropVal->ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
			ft = StringToFileTime(lstSearchCriteria[ulStartCriteria + 1].c_str());
			lpPropVal->Value.ft = ft;
			lpExtraRestriction[1].rt = RES_PROPERTY;
			lpExtraRestriction[1].res.resProperty.relop = RELOP_GE;
			lpExtraRestriction[1].res.resProperty.lpProp = lpPropVal;
			lpExtraRestriction[1].res.resProperty.ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpPropVal->ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
			ft = StringToFileTime(lstSearchCriteria[ulStartCriteria + 1].c_str());
			lpPropVal->Value.ft = AddDay(ft);
			lpExtraRestriction[2].rt = RES_PROPERTY;
			lpExtraRestriction[2].res.resProperty.relop = RELOP_LT;
			lpExtraRestriction[2].res.resProperty.ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
			lpExtraRestriction[2].res.resProperty.lpProp = lpPropVal;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "OR") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_OR;
			lpRestriction->res.resOr.cRes = 2;
			lpRestriction->res.resOr.lpRes = lpExtraRestriction;
			lstRestrictions.push_back(&lpExtraRestriction[1]);
			lstRestrictions.push_back(&lpExtraRestriction[0]);
			ulStartCriteria++;
			// RECENT done with ALL
		} else if (strSearchCriterium == "SEEN") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_MESSAGE_FLAGS;
			lpExtraRestriction[1].rt = RES_BITMASK;
			lpExtraRestriction[1].res.resBitMask.relBMR = BMR_NEZ;
			lpExtraRestriction[1].res.resBitMask.ulPropTag = PR_MESSAGE_FLAGS;
			lpExtraRestriction[1].res.resBitMask.ulMask = MSGFLAG_READ;
			ulStartCriteria++;
		} else if (strSearchCriterium == "SENTBEFORE") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
			lpPropVal->ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
			ft = StringToFileTime(lstSearchCriteria[ulStartCriteria + 1].c_str());
			lpPropVal->Value.ft = ft;
			lpExtraRestriction[1].rt = RES_PROPERTY;
			lpExtraRestriction[1].res.resProperty.relop = RELOP_LT;
			lpExtraRestriction[1].res.resProperty.ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
			lpExtraRestriction[1].res.resProperty.lpProp = lpPropVal;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "SENTON") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SRestriction) * 3, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 3;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
			lpPropVal->ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
			ft = StringToFileTime(lstSearchCriteria[ulStartCriteria + 1].c_str());
			lpPropVal->Value.ft = ft;
			lpExtraRestriction[1].rt = RES_PROPERTY;
			lpExtraRestriction[1].res.resProperty.relop = RELOP_GE;
			lpExtraRestriction[1].res.resProperty.lpProp = lpPropVal;
			lpExtraRestriction[1].res.resExist.ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpPropVal->ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
			ft = StringToFileTime(lstSearchCriteria[ulStartCriteria + 1].c_str());
			lpPropVal->Value.ft = AddDay(ft);
			lpExtraRestriction[2].rt = RES_PROPERTY;
			lpExtraRestriction[2].res.resProperty.relop = RELOP_LT;
			lpExtraRestriction[2].res.resProperty.ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
			lpExtraRestriction[2].res.resProperty.lpProp = lpPropVal;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "SENTSINCE") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
			lpPropVal->ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
			ft = StringToFileTime(lstSearchCriteria[ulStartCriteria + 1].c_str());
			lpPropVal->Value.ft = ft;
			lpExtraRestriction[1].rt = RES_PROPERTY;
			lpExtraRestriction[1].res.resProperty.relop = RELOP_GE;
			lpExtraRestriction[1].res.resProperty.ulPropTag = PR_EC_CLIENT_SUBMIT_DATE;
			lpExtraRestriction[1].res.resProperty.lpProp = lpPropVal;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "SINCE") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
			lpPropVal->ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
			ft = StringToFileTime(lstSearchCriteria[ulStartCriteria + 1].c_str());
			lpPropVal->Value.ft = ft;
			lpExtraRestriction[1].rt = RES_PROPERTY;
			lpExtraRestriction[1].res.resProperty.relop = RELOP_GE;
			lpExtraRestriction[1].res.resProperty.ulPropTag = PR_EC_MESSAGE_DELIVERY_DATE;
			lpExtraRestriction[1].res.resProperty.lpProp = lpPropVal;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "SMALLER") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpPropVal->ulPropTag = PR_MESSAGE_SIZE;
			lpPropVal->Value.ul = strtoul(lstSearchCriteria[ulStartCriteria + 1].c_str(), NULL, 0);
			lpRestriction->rt = RES_PROPERTY;
			lpRestriction->res.resProperty.relop = RELOP_LT;
			lpRestriction->res.resProperty.ulPropTag = PR_MESSAGE_SIZE;
			lpRestriction->res.resProperty.lpProp = lpPropVal;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "SUBJECT") {
		    // Handle SUBJECT <s>
		    char *szSearch;
		    
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			szSearch  = (char *)lstSearchCriteria[ulStartCriteria+1].c_str();

			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_AND;
			lpRestriction->res.resAnd.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_SUBJECT;
			lpExtraRestriction[1].rt = RES_CONTENT;

            lpExtraRestriction[1].res.resContent.ulFuzzyLevel = strlen(szSearch) > 0 ? (FL_SUBSTRING | FL_IGNORECASE) : FL_FULLSTRING;
			lpExtraRestriction[1].res.resContent.ulPropTag = PR_SUBJECT;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			lpPropVal->ulPropTag = PR_SUBJECT;
			lpPropVal->Value.lpszA = szSearch;
			lpExtraRestriction[1].res.resContent.lpProp = lpPropVal;

            ulStartCriteria += 2;
                
		} else if (strSearchCriterium == "TEXT") {
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_OR;
			lpRestriction->res.resOr.cRes = 2;
			lpRestriction->res.resAnd.lpRes = lpExtraRestriction;
			lpRestriction = lpExtraRestriction;

			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction[0].rt = RES_AND;
			lpRestriction[0].res.resAnd.cRes = 2;
			lpRestriction[0].res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_BODY;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			if (iconv)
				iconv->convert(lstSearchCriteria[ulStartCriteria+1]);

			hr = MAPIAllocateMore(sizeof(char) * (lstSearchCriteria[ulStartCriteria + 1].size() + 1), lpRootRestrict,
								  (LPVOID *) &szBuffer);
			if (hr != hrSuccess)
				goto exit;

			lpExtraRestriction[1].rt = RES_CONTENT;
			lpExtraRestriction[1].res.resContent.ulFuzzyLevel = lstSearchCriteria[ulStartCriteria+1].size() > 0 ? (FL_SUBSTRING | FL_IGNORECASE) : FL_FULLSTRING;
			lpExtraRestriction[1].res.resContent.ulPropTag = PR_BODY;
			memcpy(szBuffer, lstSearchCriteria[ulStartCriteria + 1].c_str(), lstSearchCriteria[ulStartCriteria + 1].size() + 1);
			lpPropVal->ulPropTag = PR_BODY;
			lpPropVal->Value.lpszA = szBuffer;
			lpExtraRestriction[1].res.resContent.lpProp = lpPropVal;

			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction[1].rt = RES_AND;
			lpRestriction[1].res.resAnd.cRes = 2;
			lpRestriction[1].res.resAnd.lpRes = lpExtraRestriction;
			lpExtraRestriction[0].rt = RES_EXIST;
			lpExtraRestriction[0].res.resExist.ulReserved1 = 0;
			lpExtraRestriction[0].res.resExist.ulReserved2 = 0;
			lpExtraRestriction[0].res.resExist.ulPropTag = PR_TRANSPORT_MESSAGE_HEADERS;

			hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
			if (hr != hrSuccess)
				goto exit;

			if (iconv)
				iconv->convert(lstSearchCriteria[ulStartCriteria+1]);

			hr = MAPIAllocateMore(sizeof(char) * (lstSearchCriteria[ulStartCriteria + 1].size() + 1), lpRootRestrict,
								  (LPVOID *) &szBuffer);
			if (hr != hrSuccess)
				goto exit;

			lpExtraRestriction[1].rt = RES_CONTENT;
			lpExtraRestriction[1].res.resContent.ulFuzzyLevel = lstSearchCriteria[ulStartCriteria+1].size() > 0 ? (FL_SUBSTRING | FL_IGNORECASE) : FL_FULLSTRING;
			lpExtraRestriction[1].res.resContent.ulPropTag = PR_TRANSPORT_MESSAGE_HEADERS;
			memcpy(szBuffer, lstSearchCriteria[ulStartCriteria + 1].c_str(), lstSearchCriteria[ulStartCriteria + 1].size() + 1);
			lpPropVal->ulPropTag = PR_TRANSPORT_MESSAGE_HEADERS;
			lpPropVal->Value.lpszA = szBuffer;
			lpExtraRestriction[1].res.resContent.lpProp = lpPropVal;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "TO" || strSearchCriterium == "CC" || strSearchCriterium == "BCC") {
		    
/*		    SPropValue sPropRecipType, sPropDisplay, sPropEmail;
		    
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}

            // Make a subrestriction looking for (TYPE == MAPI_TO && (DISPLAY_NAME == VALUE || EMAIL_ADDRESS == VALUE))
			lpRestriction->rt = RES_SUBRESTRICTION;
            lpRestriction->res.resSub.ulSubObject = PR_MESSAGE_RECIPIENTS;
            
            hr = MAPIAllocateMore(sizeof(SRestriction), lpRootRestrict, (void **)&lpRestriction->res.resSub.lpRes);
            if(hr != hrSuccess)
                goto exit;
                
            lpRestriction->res.resSub.lpRes->rt = RES_AND;
            lpRestriction->res.resSub.lpRes->res.resAnd.cRes = 2;
            
            hr = MAPIAllocateMore(sizeof(SRestriction)*2, lpRootRestrict, (void **)&lpRestriction->res.resSub.lpRes->res.resAnd.lpRes);
            if(hr != hrSuccess)
                goto exit;
                
                
            sPropRecipType.ulPropTag = PR_RECIPIENT_TYPE;
            if(strSearchCriterium == "TO")
                sPropRecipType.Value.ul = MAPI_TO;
            else if(strSearchCriterium == "CC")
                sPropRecipType.Value.ul = MAPI_CC;
            else
                sPropRecipType.Value.ul = MAPI_BCC;

            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[0].rt = RES_PROPERTY;
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[0].res.resProperty.ulPropTag = PR_RECIPIENT_TYPE;
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[0].res.resProperty.lpProp = &sPropRecipType;
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[0].res.resProperty.relop = RELOP_EQ;
            
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].rt = RES_OR;
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].res.resOr.cRes = 2;
            hr = MAPIAllocateMore(sizeof(SRestriction)*2, lpRootRestrict, (void **)&lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].res.resOr.lpRes);
            if(hr != hrSuccess)
                goto exit;
                
            
            sPropDisplay.ulPropTag = PR_DISPLAY_NAME;
            sPropDisplay.Value.lpszA = (char *)lstSearchCriteria[ulStartCriteria + 1].c_str();

            sPropEmail.ulPropTag = PR_EMAIL_ADDRESS;
            sPropEmail.Value.lpszA = (char *)lstSearchCriteria[ulStartCriteria + 1].c_str();

            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].res.resOr.lpRes[0].rt = RES_CONTENT;
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].res.resOr.lpRes[0].res.resContent.ulPropTag = PR_DISPLAY_NAME;
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].res.resOr.lpRes[0].res.resContent.lpProp = &sPropDisplay;
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].res.resOr.lpRes[0].res.resContent.ulFuzzyLevel = FL_SUBSTRING | FL_IGNORECASE;

            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].res.resOr.lpRes[1].rt = RES_CONTENT;
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].res.resOr.lpRes[1].res.resContent.ulPropTag = PR_EMAIL_ADDRESS;
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].res.resOr.lpRes[1].res.resContent.lpProp = &sPropEmail;
            lpRestriction->res.resSub.lpRes->res.resAnd.lpRes[1].res.resOr.lpRes[1].res.resContent.ulFuzzyLevel = FL_SUBSTRING | FL_IGNORECASE;
            
			ulStartCriteria += 2;
			*/
			
		    char *szField;
		    char *szSearch;
		    
			if (lstSearchCriteria.size() - ulStartCriteria <= 1) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}
			
			szField = (char *)strSearchCriterium.c_str();
			szSearch = (char *)lstSearchCriteria[ulStartCriteria+1].c_str();
			
            // Search for "^HEADER:.*DATA" in PR_TRANSPORT_HEADERS
            string strSearch = (string)"^" + szField + ":.*" + szSearch;
            
            lpRestriction->rt = RES_PROPERTY;
            lpRestriction->res.resProperty.ulPropTag = PR_TRANSPORT_MESSAGE_HEADERS;
            lpRestriction->res.resProperty.relop = RELOP_RE;

            hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (void **)&lpPropVal);
            if(hr != hrSuccess)
                goto exit;
                
            lpPropVal->ulPropTag = PR_TRANSPORT_MESSAGE_HEADERS;
            
            hr = MAPIAllocateMore(strSearch.size()+1, lpRootRestrict, (void **)&lpPropVal->Value.lpszA);
            strcpy(lpPropVal->Value.lpszA, strSearch.c_str());
            
            lpRestriction->res.resProperty.lpProp = lpPropVal;

			ulStartCriteria += 2;
		} else if (strSearchCriterium == "UID") {
			lstMails.clear();

			hr = HrParseSeqUidSet(lstSearchCriteria[ulStartCriteria + 1], lstMails);
			if (hr != hrSuccess)
				goto exit;

			hr = MAPIAllocateMore(sizeof(SRestriction) * lstMails.size(), lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_OR;
			lpRestriction->res.resOr.cRes = lstMails.size();
			lpRestriction->res.resOr.lpRes = lpExtraRestriction;
			ulCount = 0;

			for (lpMail = lstMails.begin(); lpMail != lstMails.end(); lpMail++) {
				hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (LPVOID *) &lpPropVal);
				if (hr != hrSuccess)
					goto exit;

				lpPropVal->ulPropTag = PR_EC_IMAP_ID;
				lpPropVal->Value.ul = lstFolderMailEIDs[(*lpMail)].ulUid;
				lpExtraRestriction[ulCount].rt = RES_PROPERTY;
				lpExtraRestriction[ulCount].res.resProperty.relop = RELOP_EQ;
				lpExtraRestriction[ulCount].res.resProperty.ulPropTag = PR_EC_IMAP_ID;
				lpExtraRestriction[ulCount].res.resProperty.lpProp = lpPropVal;
				ulCount++;
			}

			ulStartCriteria += 2;
		} else if (strSearchCriterium == "UNANSWERED") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_OR;
			lpRestriction->res.resOr.cRes = 2;
			lpRestriction->res.resOr.lpRes = lpExtraRestriction;

			hr = MAPIAllocateMore(sizeof(SRestriction), lpRootRestrict, (LPVOID *) &lpRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpExtraRestriction[0].rt = RES_NOT;
			lpExtraRestriction[0].res.resNot.ulReserved = 0;
			lpExtraRestriction[0].res.resNot.lpRes = lpRestriction;

			lpRestriction[0].rt = RES_EXIST;
			lpRestriction[0].res.resExist.ulReserved1 = 0;
			lpRestriction[0].res.resExist.ulReserved2 = 0;
			lpRestriction[0].res.resExist.ulPropTag = PR_MSG_STATUS;

			lpExtraRestriction[1].rt = RES_BITMASK;
			lpExtraRestriction[1].res.resBitMask.relBMR = BMR_EQZ;
			lpExtraRestriction[1].res.resBitMask.ulPropTag = PR_MSG_STATUS;
			lpExtraRestriction[1].res.resBitMask.ulMask = MSGSTATUS_ANSWERED;
			ulStartCriteria++;
			// TODO: also find in PR_LAST_VERB_EXECUTED
		} else if (strSearchCriterium == "UNDELETED") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_OR;
			lpRestriction->res.resOr.cRes = 2;
			lpRestriction->res.resOr.lpRes = lpExtraRestriction;

			hr = MAPIAllocateMore(sizeof(SRestriction), lpRootRestrict, (LPVOID *) &lpRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpExtraRestriction[0].rt = RES_NOT;
			lpExtraRestriction[0].res.resNot.ulReserved = 0;
			lpExtraRestriction[0].res.resNot.lpRes = lpRestriction;

			lpRestriction[0].rt = RES_EXIST;
			lpRestriction[0].res.resExist.ulReserved1 = 0;
			lpRestriction[0].res.resExist.ulReserved2 = 0;
			lpRestriction[0].res.resExist.ulPropTag = PR_MSG_STATUS;

			lpExtraRestriction[1].rt = RES_BITMASK;
			lpExtraRestriction[1].res.resBitMask.relBMR = BMR_EQZ;
			lpExtraRestriction[1].res.resBitMask.ulPropTag = PR_MSG_STATUS;
			lpExtraRestriction[1].res.resBitMask.ulMask = MSGSTATUS_DELMARKED;
			ulStartCriteria++;
		} else if (strSearchCriterium == "UNDRAFT") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_OR;
			lpRestriction->res.resOr.cRes = 2;
			lpRestriction->res.resOr.lpRes = lpExtraRestriction;

			hr = MAPIAllocateMore(sizeof(SRestriction), lpRootRestrict, (LPVOID *) &lpRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpExtraRestriction[0].rt = RES_NOT;
			lpExtraRestriction[0].res.resNot.ulReserved = 0;
			lpExtraRestriction[0].res.resNot.lpRes = lpRestriction;

			lpRestriction[0].rt = RES_EXIST;
			lpRestriction[0].res.resExist.ulReserved1 = 0;
			lpRestriction[0].res.resExist.ulReserved2 = 0;
			lpRestriction[0].res.resExist.ulPropTag = PR_MSG_STATUS;

			lpExtraRestriction[1].rt = RES_BITMASK;
			lpExtraRestriction[1].res.resBitMask.relBMR = BMR_EQZ;
			lpExtraRestriction[1].res.resBitMask.ulPropTag = PR_MSG_STATUS;
			lpExtraRestriction[1].res.resBitMask.ulMask = MSGSTATUS_DRAFT;
			ulStartCriteria++;
			// FIXME: add restrictin to find PR_MESSAGE_FLAGS with MSGFLAG_UNSENT off
		} else if (strSearchCriterium == "UNFLAGGED") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_OR;
			lpRestriction->res.resOr.cRes = 2;
			lpRestriction->res.resOr.lpRes = lpExtraRestriction;

			hr = MAPIAllocateMore(sizeof(SRestriction), lpRootRestrict, (LPVOID *) &lpRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpExtraRestriction[0].rt = RES_NOT;
			lpExtraRestriction[0].res.resNot.ulReserved = 0;
			lpExtraRestriction[0].res.resNot.lpRes = lpRestriction;

			lpRestriction[0].rt = RES_EXIST;
			lpRestriction[0].res.resExist.ulReserved1 = 0;
			lpRestriction[0].res.resExist.ulReserved2 = 0;
			lpRestriction[0].res.resExist.ulPropTag = PR_FLAG_STATUS;

			lpExtraRestriction[1].rt = RES_BITMASK;
			lpExtraRestriction[1].res.resBitMask.relBMR = BMR_EQZ;
			lpExtraRestriction[1].res.resBitMask.ulPropTag = PR_FLAG_STATUS;
			lpExtraRestriction[1].res.resBitMask.ulMask = 65535;
			ulStartCriteria++;
		} else if (strSearchCriterium == "UNKEYWORD") {
			lpRestriction->rt = RES_EXIST;
			lpRestriction->res.resExist.ulReserved1 = 0;
			lpRestriction->res.resExist.ulReserved2 = 0;
			lpRestriction->res.resExist.ulPropTag = PR_ENTRYID;
			ulStartCriteria += 2;
		} else if (strSearchCriterium == "UNSEEN") {
			hr = MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (LPVOID *) &lpExtraRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpRestriction->rt = RES_OR;
			lpRestriction->res.resOr.cRes = 2;
			lpRestriction->res.resOr.lpRes = lpExtraRestriction;

			hr = MAPIAllocateMore(sizeof(SRestriction), lpRootRestrict, (LPVOID *) &lpRestriction);
			if (hr != hrSuccess)
				goto exit;

			lpExtraRestriction[0].rt = RES_NOT;
			lpExtraRestriction[0].res.resNot.ulReserved = 0;
			lpExtraRestriction[0].res.resNot.lpRes = lpRestriction;

			lpRestriction[0].rt = RES_EXIST;
			lpRestriction[0].res.resExist.ulReserved1 = 0;
			lpRestriction[0].res.resExist.ulReserved2 = 0;
			lpRestriction[0].res.resExist.ulPropTag = PR_MESSAGE_FLAGS;

			lpExtraRestriction[1].rt = RES_BITMASK;
			lpExtraRestriction[1].res.resBitMask.relBMR = BMR_EQZ;
			lpExtraRestriction[1].res.resBitMask.ulPropTag = PR_MESSAGE_FLAGS;
			lpExtraRestriction[1].res.resBitMask.ulMask = MSGFLAG_READ;
			ulStartCriteria++;
		} else if (strSearchCriterium == "HEADER") {
		    char *szField;
		    char *szSearch;
		    
			if (lstSearchCriteria.size() - ulStartCriteria <= 2) {
				hr = MAPI_E_CALL_FAILED;
				goto exit;
			}
			
			szField = (char *)lstSearchCriteria[ulStartCriteria+1].c_str();
			szSearch = (char *)lstSearchCriteria[ulStartCriteria+2].c_str();
			
            // Search for "^HEADER:.*DATA" in PR_TRANSPORT_HEADERS
            string strSearch = (string)"^" + szField + ":.*" + szSearch;
            
            lpRestriction->rt = RES_PROPERTY;
            lpRestriction->res.resProperty.ulPropTag = PR_TRANSPORT_MESSAGE_HEADERS;
            lpRestriction->res.resProperty.relop = RELOP_RE;

            hr = MAPIAllocateMore(sizeof(SPropValue), lpRootRestrict, (void **)&lpPropVal);
            if(hr != hrSuccess)
                goto exit;
                
            lpPropVal->ulPropTag = PR_TRANSPORT_MESSAGE_HEADERS;
            
            hr = MAPIAllocateMore(strSearch.size()+1, lpRootRestrict, (void **)&lpPropVal->Value.lpszA);
            strcpy(lpPropVal->Value.lpszA, strSearch.c_str());
            
            lpRestriction->res.resProperty.lpProp = lpPropVal;

			ulStartCriteria += 3;
		} else {
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
	}

	while (lstRestrictions.size() > 0) {
		lpRestriction = lstRestrictions[lstRestrictions.size() - 1];
		lpRestriction->rt = RES_EXIST;
		lpRestriction->res.resExist.ulReserved1 = 0;
		lpRestriction->res.resExist.ulReserved2 = 0;
		lpRestriction->res.resExist.ulPropTag = PR_ENTRYID;
		lstRestrictions.pop_back();
	}
		
    // Wrap the resulting restriction (lpRootRestrict) in an AND:
    // AND (lpRootRestrict, EXIST(PR_INSTANCE_KEY)) to make sure that the query
    // will not be passed to the indexer
    
	sPropertyRestriction.rt = RES_EXIST;
	sPropertyRestriction.res.resExist.ulPropTag = PR_INSTANCE_KEY;
	
	sAndRestriction.rt = RES_AND;
	sAndRestriction.res.resAnd.cRes = 2;
	MAPIAllocateMore(sizeof(SRestriction) * 2, lpRootRestrict, (void **)&sAndRestriction.res.resAnd.lpRes);
	sAndRestriction.res.resAnd.lpRes[0] = sPropertyRestriction;
	sAndRestriction.res.resAnd.lpRes[1] = *lpRootRestrict;
		
	hr = HrQueryAllRows(lpTable, (LPSPropTagArray) &spt, &sAndRestriction, NULL, 0, &lpRows);
	if (hr != hrSuccess)
		goto exit;

	if (lpRows->cRows == 0)
		goto exit;

    for (ulRownr = 0; ulRownr < lpRows->cRows; ulRownr++) {
        iterUID = mapUIDs.find(lpRows->aRow[ulRownr].lpProps[0].Value.ul);
        if(iterUID == mapUIDs.end()) {
            // Found a match for a message that is not in our message list .. skip it
            continue;
        }
        
        lstMailnr.push_back(iterUID->second);
    }

	lstMailnr.sort();

exit:
	if (lpRootRestrict)
		MAPIFreeBuffer(lpRootRestrict);

	if (lpRows)
		FreeProws(lpRows);

	if (lpTable)
		lpTable->Release();

	if (lpFolder)
		lpFolder->Release();

	return hr;
}

string IMAP::GetHeaderValue(const string &strMessage, const string &strHeader, const string &strDefault) {
	string::size_type posStart, posEnd;

	posStart = strMessage.find(strHeader);
	if (posStart == string::npos)
		return strDefault;

	posStart += strHeader.length();
	posEnd = strMessage.find("\r\n", posStart);
	if (posEnd == string::npos)
		return strDefault;

	// don't care about uppercase, it's all good.
	return "\"" + strMessage.substr(posStart, posEnd - posStart) + "\"";
}

// Create a bodystructure (RFC 3501)
HRESULT IMAP::HrGetBodyStructure(bool bExtended, string &strBodyStructure, const string& strMessage) {
	HRESULT hr = hrSuccess;
	string::size_type ulContentTypePos, ulEndPos, ulEndLinePos, ulBoundaryPos, ulBoundaryPosEnd, ulDispositionPos, ulNamePos, ulFilenamePos;
	ULONG ulLineCount;
	string strSubStructure, strBoundary, strMessageType, strBoundaryTMP, strDisposition, strTag, strFileName;
	char szBuffer[33];

	ulContentTypePos = strMessage.find("Content-Type: ");
	if (ulContentTypePos == string::npos) {
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	ulContentTypePos += 14;	 // += strlen("Content-Type: ");
	while (ulContentTypePos != string::npos &&
		   (strMessage[ulContentTypePos] == '\r' || strMessage[ulContentTypePos] == '\n' ||
			strMessage[ulContentTypePos] == '\t' || strMessage[ulContentTypePos] == ' ')) { ulContentTypePos++; }
	if (strMessage.compare(ulContentTypePos, 9, "multipart") == 0) { // multipart body part
		ulContentTypePos += 10;	// += strlen("multipart")

		// Strip message type
		ulEndPos = strMessage.find(";", ulContentTypePos);

		if (ulEndPos != string::npos) {
			strMessageType = strMessage.substr(ulContentTypePos, ulEndPos - ulContentTypePos);
		} else {
			strMessageType = "alternative";
		}

		ToUpper(strMessageType);
		
		// find the boundary and add them in strBoundary
		ulBoundaryPos = strMessage.find("boundary=\"", ulContentTypePos);

		if (ulBoundaryPos == string::npos) {
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		ulBoundaryPos += 10;
		ulEndPos = strMessage.find("\"", ulBoundaryPos);

		if (ulEndPos == string::npos) {
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		strBoundary = strMessage.substr(ulBoundaryPos, ulEndPos - ulBoundaryPos);
		strBoundaryTMP = strBoundary;
		strBoundary = (string) "--" + strBoundary; //part start sign + strBoundary

		// Find the begin of the first boundary part
		ulBoundaryPos = strMessage.find(strBoundary, ulEndPos);
		ulBoundaryPos += strBoundary.size() + 2;

		//Find the end boundary, for "speedup" the loops
		ulBoundaryPosEnd = strMessage.find(strBoundary+"--", ulBoundaryPos);

		//////////////////////////////////////////////
		// Start formatting multipart body structure

		strBodyStructure = "(";

		while (ulBoundaryPos != string::npos && ulBoundaryPos < ulBoundaryPosEnd) {
			
			ulEndPos = ulBoundaryPos;
			do {
				ulEndPos = strMessage.find(strBoundary, ulEndPos);
				
				// Check or this is the real boundary and not a part of a boundary
				string endtag = strMessage.substr(ulEndPos+strBoundary.length(), 2);
				if(ulEndPos != string::npos && (endtag.compare("\r\n") == 0 || endtag.compare("--") == 0))
					break;
				else if(ulEndPos != string::npos)
					ulEndPos+=strBoundary.length();

			}while(ulEndPos != string::npos && ulEndPos < ulBoundaryPosEnd);

			if (ulEndPos != string::npos) {
				strSubStructure.clear();

				hr = HrGetBodyStructure(bExtended, strSubStructure, strMessage.substr(ulBoundaryPos, ulEndPos - ulBoundaryPos));
				if (hr != hrSuccess) {
					// strBodyStructure += "()";
				} else {
					strBodyStructure += strSubStructure;
				}

				ulBoundaryPos = ulEndPos + strBoundary.size() + 2;
			} else {
				ulBoundaryPos = ulEndPos;
			}
		}

		strBodyStructure += " \"" + strMessageType + "\"";

		if (bExtended) {
			strBodyStructure += " (\"BOUNDARY\" \"" + strBoundaryTMP + "\")";

			// Some servers also add the follow fields (RFC 3501) :
			// disposition language
			strBodyStructure += " NIL NIL";
		}

		strBodyStructure += ")";
	} else { // Non-multipart body part
		strBodyStructure = "(\"";
		// type
		ulEndPos = strMessage.find("/", ulContentTypePos);
		if (ulEndPos == string::npos) {
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		strMessageType = strMessage.substr(ulContentTypePos, ulEndPos - ulContentTypePos);
		ToUpper(strMessageType);
		strBodyStructure += strMessageType;
		strBodyStructure += "\" \"";

		// subtype
		ulContentTypePos = ulEndPos + 1;
		ulEndPos = strMessage.find(";", ulContentTypePos);
		ulEndLinePos = strMessage.find("\r\n", ulContentTypePos);

		if (ulEndPos != string::npos) {
			if (ulEndLinePos != string::npos && ulEndLinePos < ulEndPos) {
				ulEndPos = ulEndLinePos;
			}
		} else if (ulEndLinePos != string::npos) {
			ulEndPos = ulEndLinePos;
		} else {
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		strSubStructure = strMessage.substr(ulContentTypePos, ulEndPos - ulContentTypePos);
		ToUpper(strSubStructure);
		strBodyStructure += strSubStructure + "\" ";


		if (strMessageType == "TEXT") {
			// add info about text typed body or attachment
			// include: charset
			ulEndPos = strMessage.find("charset=", ulEndPos);
			if (ulEndPos != string::npos) {
				ulEndPos += 8;
				ulEndLinePos = strMessage.find("\r\n", ulEndPos);
				if (ulEndLinePos != string::npos) {
					strSubStructure = strMessage.substr(ulEndPos, ulEndLinePos - ulEndPos);
					ToUpper(strSubStructure);
					strBodyStructure += (string) "(\"CHARSET\" \"" + strSubStructure + "\")";
				} else {
					strSubStructure = strMessage.substr(ulEndPos);
					ToUpper(strSubStructure);
					strBodyStructure += (string) "(\"CHARSET\" \"" + strSubStructure + "\")";
				}
			} else {
				strBodyStructure += "(\"CHARSET\" \"WINDOWS-1252\")";
			}
		} else if (strMessageType == "MESSAGE") {
			// TODO: msg-in-msg info?
			strBodyStructure += "NIL";
		} else {
			// add info about data typed attachment
			// include: "name" parameter of Content-Type header

			// must only be added from the Content-Type header line, so find the end of this header.
			// since we know vmime generated the mail, we can search for '\r\n '.
			ulEndLinePos = strMessage.find("\r\n", ulContentTypePos);
			// ulEndLinePos is not going to be npos-1
			while (ulEndLinePos != string::npos && strMessage[ulEndLinePos+2] == ' ')
				ulEndLinePos = strMessage.find("\r\n", ulEndLinePos+2);

			ulEndPos = strMessage.find("name=", ulContentTypePos);
			if (ulEndPos != string::npos && ulEndLinePos != string::npos && ulEndPos < ulEndLinePos) {
				ulEndPos += 5;
				// Skip leading quote
				if(strMessage[ulEndPos] == '"')
				    ulEndPos++;
				    
				ulEndLinePos = strMessage.find("\r\n", ulEndPos);
				if (ulEndLinePos != string::npos) {
				    // Strip trailing ; (used in multiple parameters for attachment names)
				    if(strMessage[ulEndLinePos-1] == ';')
				        ulEndLinePos--;
				    // Strip trailing quote
				    if(strMessage[ulEndLinePos-1] == '"')
				        ulEndLinePos--;
					strBodyStructure += (string) "(\"NAME\" \"" + strMessage.substr(ulEndPos, ulEndLinePos - ulEndPos) + "\")";
				} else {
					strBodyStructure += (string) "(\"NAME\" \"" + strMessage.substr(ulEndPos) + "\")";
				}
			} else {
				strBodyStructure += "NIL";
			}
		}
		strBodyStructure += " ";

		// Content-Id header value
		strBodyStructure += GetHeaderValue(strMessage, "Content-Id: ", "NIL");
		strBodyStructure += " ";

		// Content-Description header value
		strBodyStructure += GetHeaderValue(strMessage, "Content-Description: ", "NIL");
		strBodyStructure += " ";

		// Content-Transfer-Encoding header value
		strBodyStructure += GetHeaderValue(strMessage, "Content-Transfer-Encoding: ", "7bit");
		strBodyStructure += " ";

		// size
		ulEndLinePos = strMessage.find("\r\n\r\n");

		if (ulEndLinePos != string::npos) {
			snprintf(szBuffer, 32, "%u", (ULONG)strMessage.size() - (ULONG)ulEndLinePos - 4);
			strBodyStructure += szBuffer;
		} else {
			strBodyStructure += "0";
		}

		if (strMessageType == "MESSAGE") {	// envelope structure, body structure, size in textlines
			// TODO: add info
		} else if (strMessageType == "TEXT") {	// size in textlines
			ulLineCount = 0;

			if (ulEndLinePos != string::npos) {
				ulEndLinePos += 2;

				while (ulEndLinePos != string::npos) {
					ulEndLinePos += 2;
					ulEndLinePos = strMessage.find("\r\n", ulEndLinePos);
					ulLineCount++;
				}
				ulLineCount--;
			}

			snprintf(szBuffer, 32, " %u", ulLineCount);
			strBodyStructure += szBuffer;
		}

		// RFC3501: Extension data is never returned with the BODY
		// 			fetch, but can be returned with a BODYSTRUCTURE fetch.
		// MD5 disposition language location

		// we need this for sugarcrm imap to work with attachments.
		if (bExtended && strMessageType != "MESSAGE") {
			// if bodystructure (==extended)

			ulDispositionPos = strMessage.find("Content-Disposition: ");
			if (ulDispositionPos != string::npos) {
				ulDispositionPos += 21; // strlen("Content-Disposition: ")
				while (ulDispositionPos != string::npos &&
					   (strMessage[ulDispositionPos] == '\r' || strMessage[ulDispositionPos] == '\n' ||
						strMessage[ulDispositionPos] == '\t' || strMessage[ulDispositionPos] == ' ')) { ulDispositionPos++; }

				// since we're going to send info on the disposition, we add the NIL for MD5sum too
				strBodyStructure += " NIL "; // no md5

				ulEndPos = strMessage.find(";", ulDispositionPos);
				ulEndLinePos = strMessage.find("\r\n", ulDispositionPos);

				if (ulEndPos != string::npos) {
					// we have a ; so chop off the word in between and add
					strDisposition = strMessage.substr(ulDispositionPos, ulEndPos - ulDispositionPos);
				} else if (ulEndLinePos != string::npos) {
					strDisposition = strMessage.substr(ulDispositionPos, ulEndLinePos - ulDispositionPos);
				} else {
					strDisposition = "attachment";
				}

				strBodyStructure += "(\"" + strDisposition + "\" ";

				// either name or filename
				ulNamePos = strMessage.find("name=", ulDispositionPos);
				if (ulNamePos != string::npos) {
					// tag = knip
					ulEndPos = ulNamePos + 4;
					ulNamePos = strMessage.rfind(" ", ulNamePos) + 1; // find prefix space, and move to after space
					strTag = strMessage.substr(ulNamePos, ulEndPos - ulNamePos);

					// filename = knip
					ulFilenamePos = ulEndPos + 1; // skip =
					ulEndPos = strMessage.find_first_of(";\r\n", ulEndPos);
					if (strMessage[ulFilenamePos] == '"') {
						strFileName = strMessage.substr(ulFilenamePos + 1, ulEndPos - ulFilenamePos - 2); // with strip of ""
					} else {
						strFileName = strMessage.substr(ulFilenamePos, ulEndPos - ulFilenamePos); // without strip of ""
					}

					strBodyStructure += "(\"" + strTag + "\" \"" + strFileName + "\")";
				} else {
					strBodyStructure += "NIL";
				}
				strBodyStructure += ")";

				strBodyStructure += " NIL"; // no language
			} else {
				// courier/dovecot here print also NIL NIL NIL (no, not the location NIL)
				strBodyStructure += " NIL NIL NIL";
			}
		}

		strBodyStructure += ")";
	}

exit:
	return hr;
}

string IMAP::FileTimeToRFC822String(FILETIME sFileTime) {
	string strTime;
	char szBuffer[31];
	time_t sTime;
	struct tm ptr;

	sTime = FileTimeToUnixTime(sFileTime.dwHighDateTime, sFileTime.dwLowDateTime);
	gmtime_safe(&sTime, &ptr);
	// month must be one of the following:

	strTime += strDay[ptr.tm_wday];
	snprintf(szBuffer, 30, ", %d ", ptr.tm_mday);
	strTime += szBuffer;
	strTime += strMonth[ptr.tm_mon];
	strftime(szBuffer, 30, " %Y %H:%M:%S +0000", &ptr);
	strTime += szBuffer;

	return strTime;
}

string IMAP::FileTimeToString(FILETIME sFileTime) {
	string strTime;
	char szBuffer[31];
	time_t sTime;
	struct tm ptr;

	sTime = FileTimeToUnixTime(sFileTime.dwHighDateTime, sFileTime.dwLowDateTime);
	gmtime_safe(&sTime, &ptr);
	// month must be one of the following:
	strftime(szBuffer, 30, "%d-", &ptr);
	strTime += szBuffer;
	strTime += strMonth[ptr.tm_mon];
	strftime(szBuffer, 30, "-%Y %H:%M:%S +0000", &ptr);
	strTime += szBuffer;
	return strTime;
}

FILETIME IMAP::StringToFileTime(string strTime, bool bDateOnly) {
	FILETIME sFileTime;
	struct tm sTm;
	ULONG ulMonth;
	time_t sTime;
	
	sTm.tm_mday = 1;
	sTm.tm_mon = 0;
	sTm.tm_year = 100;			// years since 1900
	sTm.tm_hour = 0;
	sTm.tm_min = 0;
	sTm.tm_sec = 0;
	sTm.tm_isdst = -1;			// daylight saving time off

	// 01-Jan-2006 00:00:00 +0000
	if (strTime.size() < 2)
		goto done;
	
	if(strTime.at(1) == '-'){
		strTime = " " + strTime;
	}

	// day of month
	if(strTime.at(0) == ' '){
		sTm.tm_mday = atoi(strTime.substr(1, 1).c_str());
	}else{
		sTm.tm_mday = atoi(strTime.substr(0, 2).c_str());
	}
	// month name 3 chars
	if (strTime.size() < 6)
		goto done;

	sTm.tm_mon = 0;
	for (ulMonth = 0; ulMonth < 12; ulMonth++) {
		if (CaseCompare(strMonth[ulMonth],strTime.substr(3, 3))) {
			sTm.tm_mon = ulMonth;
		}
	}
	
	    
	if (strTime.size() < 11)
		goto done;

	sTm.tm_year = atoi(strTime.substr(7, 4).c_str()) - 1900;	// year 4 chars
	if (strTime.size() < 14)
		goto done;

	if (bDateOnly)
	    goto done;

	sTm.tm_hour = atoi(strTime.substr(12, 2).c_str());	// hours
	if (strTime.size() < 17)
		goto done;

	sTm.tm_min = atoi(strTime.substr(15, 2).c_str());	// minutes
	if (strTime.size() < 20)
		goto done;

	sTm.tm_sec = atoi(strTime.substr(18, 2).c_str());	// seconds
	if (strTime.size() < 26)
		goto done;

	if (strTime.substr(21, 1) == "+") {
		sTm.tm_hour -= atoi(strTime.substr(22, 2).c_str());
		sTm.tm_min -= atoi(strTime.substr(24, 2).c_str());
	} else if (strTime.substr(21, 1) == "-") {
		sTm.tm_hour += atoi(strTime.substr(22, 2).c_str());
		sTm.tm_min += atoi(strTime.substr(24, 2).c_str());
	}

done:
	sTime = timegm(&sTm);
	UnixTimeToFileTime(sTime, &sFileTime);
	return sFileTime;
}

FILETIME IMAP::AddDay(FILETIME sFileTime) {
	FILETIME sFT;

	// add 24 hour in seconds = 24*60*60 seconds
	UnixTimeToFileTime(FileTimeToUnixTime(sFileTime.dwHighDateTime, sFileTime.dwLowDateTime) + 24 * 60 * 60, &sFT);
	return sFT;
}

ULONG IMAP::BinaryToLong(LPSBinary lpBinary) {
	ULONG ulBinary = 0;
	ULONG ulBytenr;

	if (lpBinary == NULL)
		return 0;

	for (ulBytenr = 0; ulBytenr < 4 && ulBytenr < lpBinary->cb; ulBytenr++) {
		// shift every byte into his position and add them to the long
		//  << 3 equals * 8 = size of byte
		ulBinary += ((ULONG) (BYTE) (lpBinary->lpb[ulBytenr])) << (ulBytenr << 3);
	}

	return ulBinary;
}

// This is a kind of cheat, as we *always* encode " and \ with the QP values, which isn't
// really necessary for QP, but it *is* necessary for IMAP. So we do it in this function anyway
string IMAP::ToQP(char *input) {
	string s;
	char hex[] = "0123456789ABCDEF";

	while (input && *input) {
		if (*input >= 0x21 && *input <= 0x7e && *input != '=' && *input != '\\' && *input != '"')
			s += *input;
		else {
			s += '=';
			s += hex[((*input) >> 4) & 0xf];
			s += hex[(*input) & 0xf];
		}

		input++;
	}

	return s;
}

string IMAP::EscapeString(char *input, char *charset, char *encoding) {
	string s;

	if (stricmp(encoding, "quoted-printable") != 0)
		return "unsupported encoding";

	if (!input || !*input)
		return s;

	// Q = quoted-printable
	s = (string) "=?" + charset + "?Q?" + ToQP(input) + "?=";

	return s;
}

string IMAP::EscapeStringQT(string input) {
	string s;
	unsigned int i;

	/*
	 * qtext           =       NO-WS-CTL /     ; Non white space controls
	 * %d33 /          ; The rest of the US-ASCII
	 * %d35-91 /       ;  characters not including "\"
	 * %d93-126        ;  or the quote character
	 */

	// We quote NO-WS-CTL anyway, just to be sure
	for (i = 0; i < input.length(); i++) {
		if (input[i] == 33 || (input[i] >= 35 && input[i] <= 91) || (input[i] >= 93 && input[i] <= 126))
			s += input[i];
		else {
			s += "\\";
			s += input[i];
		}
	}

	return s;
}

// imap folder encoding is a modified form of utf-7 (+ becomes &, so & is "escaped",
// utf-7 is a modifed form of base64, based from the utf16 character
// I'll use the iconv convertor for this, per character .. sigh

string IMAP::MAPI2IMAPCharset(string input) {
	string output;
	size_t i;
	ECIConv iconv("utf-7","windows-1252");

	for (i=0; i < input.length(); i++) {
		if ( (input[i] >= 0x20 && input[i] <= 0x25) || (input[i] >= 0x27 && input[i] <= 0x7e) ) {
			if (input[i] == '"' || input[i] == '\\')
				output += '\\';
			output += input[i];
		} else if (input[i] == 0x26) {
			output += "&-";		// & is encoded as &-
		} else {
			string conv;
			conv = input[i];
			while (i+1 < input.length() && (input[i+1] < 0x20 || input[i+1] >= 0x7f))
				conv += input[++i];
			conv = iconv.convert(conv);
			conv[0] = '&';		// convert + to &
			for (size_t j=0; j<conv.size(); j++)
				if (conv[j] == '/')
					conv[j] = ','; // convert / from base64 to ,
			output += conv;	// also contains the terminating -
		}
	}
	return output;
}

string IMAP::IMAP2MAPICharset(string input) {
	string output;
	size_t i;
	ECIConv iconv("windows-1252","utf-7");

	for (i=0; i < input.length(); i++) {
		if (input[i] == '&') {
			if (i+1 >= input.length()) {
				// premature end of string
				output += input[i];
				break;
			}
			if (input[i+1] == '-') {
				output += '&';
				i++;			// skip '-'
			} else {
				string conv = "+";
				i++;			// skip imap '&', is a '+' in utf-7
				while (i < input.length() && input[i] != '-') {
					if (input[i] == ',')
						conv += '/'; // , -> / for utf-7
					else
						conv += input[i++];
				}
				conv = iconv.convert(conv);
				output += conv;
			}
		} else {
			if (input[i] == '\\' && i+1 < input.length() && (input[i+1] == '"' || input[i+1] == '\\'))
				i++;
			output += input[i];
		}
	}
	return output;
}

// Replaces boundary="xxxx" with boundary="=_ZG_static"
void IMAP::FixupBoundary(string &strHeaders)
{
    size_t begin, end;
    
    begin = strHeaders.find("boundary=\"");
    if(begin != string::npos) {
        end = strHeaders.find("\"", begin+strlen("boundary=\""));
        if(end == string::npos)
            return;
    } else {
        begin = strHeaders.find("boundary=");
        if(begin == string::npos)
            return;
        end = strHeaders.find_first_of(" ;\t\r\n", begin+strlen("boundary="));
        if(end == string::npos)
            return;
        end--; // don't replace the found char
    }
        
    strHeaders = strHeaders.substr(0, begin) + "boundary=\"=_ZG_static\"" + strHeaders.substr(end+1);
}

bool IMAP::MatchFolderPath(string strFolder, string strPattern)
{
    bool bMatch = false;
    int f = 0;
    int p = 0;
    
    ToUpper(strFolder);
    ToUpper(strPattern);
    
    while(1) {
        if(f == (int)strFolder.size() && p == (int)strPattern.size()) {
            // Reached the end of the folder and the pattern strings, so match
            bMatch = true;
            break;
        }
        
        if(strPattern[p] == '*') {
            // Match 0-n chars, try longest match first
            for(int i = strFolder.size(); i >= f; i--) {
                // Try matching the rest of the string from position i in the string
                if(MatchFolderPath(strFolder.substr(i), strPattern.substr(p+1))) {
                    // Match OK, apply the 'skip i' chars
                    bMatch = true;
                    goto exit;
                }
            }
            
            // No match found, failed
            bMatch = false;
            break;
        } else if(strPattern[p] == '%') {
            // Match 0-n chars excluding '/', try longest match first
            size_t slash = strFolder.find('/', f);
            if(slash == std::string::npos)
                slash = strFolder.size();

            for(int i = slash; i >= f; i--) {
                // Try matching the rest of the string from position i in the string
                if(MatchFolderPath(strFolder.substr(i), strPattern.substr(p+1))) {
                    // Match OK, apply the 'skip i' chars
                    bMatch = true;
                    goto exit;
                }
            }

            // No match found, failed
            bMatch = false;
            break;
        } else {
            // Match normal string
            if(strFolder[f] != strPattern[p])
                break;
            f++;
            p++;
        }
    }
    
exit:
    return bMatch;
}

void IMAP::HrParseHeaders(string &strHeaders, list<pair<string, string> > &lstHeaders)
{
    size_t pos = 0;
    size_t end = 0;
    string strLine;
    string strField;
    string strData;
    list<pair<string, string> >::iterator iterLast;
    
    lstHeaders.clear();
    iterLast = lstHeaders.end();
    
    while(1) {
        end = strHeaders.find("\r\n", pos);
        
        if(end == string::npos) {
            strLine = strHeaders.substr(pos);
        } else {
            strLine = strHeaders.substr(pos, end-pos);
        }
        
        if((strLine[0] == ' ' || strLine[0] == '\t') && iterLast != lstHeaders.end()) {
            // Continuation of previous header
            iterLast->second += "\r\n";
            iterLast->second += strLine;
        } else {
            size_t colon = strLine.find(":");
            
            if(colon != string::npos) {
                // Get field name
                strField = strLine.substr(0, colon);
            
                strData = strLine.substr(colon+1);
                
                // Remove leading spaces
                while(strData[0] == ' ') {
                    strData.erase(0,1);
                }
                
                lstHeaders.push_back(pair<string, string>(strField, strData));
                iterLast = --lstHeaders.end();
            } else {
                // Broken header ? (no :)
            }
        }    
        
        if(end == string::npos)
            break;
        
        pos = end + 2; // Go to next line (+2 = \r\n)
    }
}

/**
 * Find a substring in strInput that starts with strBegin, and
 * possebly ends with strEnd. Only the data between the strBegin and
 * strEnd is returned.
 *
 * @param[out]	strOutput	The found substring in strInput
 * @param[in]	strInput	Input string
 * @param[in]	strBegin	Substring should start with
 * @param[in]	strEnd	Substring should end with
 */
void IMAP::HrGetSubString(string &strOutput, string &strInput, string strBegin, string strEnd)
{
    size_t begin;
    size_t end;

    strOutput.clear();

    begin = strInput.find(strBegin);
    if(begin == string::npos)
        return;
        
    end = strInput.find(strEnd, begin+1);
    if(end == string::npos)
        strOutput = strInput.substr(begin+1);
    else
        strOutput = strInput.substr(begin+1, end-begin-1);
}

void IMAP::HrTokenize(set<string> &setTokens, string &strInput)
{
    vector<string> lstTokens = tokenize(strInput, " ");
    vector<string>::iterator iterTokens;
    
    setTokens.clear();
    
    for(iterTokens = lstTokens.begin(); iterTokens != lstTokens.end(); iterTokens++) {
        setTokens.insert(*iterTokens);
    }
}

void IMAP::HrTokenize(list<string> &lstTokens, string &strInput)
{
    vector<string> vecTokens = tokenize(strInput, " ");
    vector<string>::iterator iterTokens;
    
    lstTokens.clear();
    
    for(iterTokens = vecTokens.begin(); iterTokens != vecTokens.end(); iterTokens++) {
        lstTokens.push_back(*iterTokens);
    }
}

HRESULT IMAP::HrFindFolder(string strFolder, bool bReadOnly, IMAPIFolder **lppFolder)
{
    HRESULT hr = hrSuccess;
    ULONG cbEntryID = 0;
    LPENTRYID lpEntryID = NULL;
    ULONG ulObjType = 0;
    IMAPIFolder *lpFolder = NULL;
	ULONG ulFlags = 0;

	if (!bReadOnly)
		ulFlags |= MAPI_MODIFY;

    hr = HrFindFolderEntryID(strFolder, &cbEntryID, &lpEntryID);
    if(hr != hrSuccess)
        goto exit;
        
    hr = lpSession->OpenEntry(cbEntryID, lpEntryID, NULL, ulFlags, &ulObjType, (IUnknown **)&lpFolder);
    if(hr != hrSuccess)
        goto exit;
        
    if(ulObjType != MAPI_FOLDER) {
        hr = MAPI_E_INVALID_PARAMETER;
        goto exit;
    }
    
    *lppFolder = lpFolder;
    
exit:
    if(lpEntryID)
        MAPIFreeBuffer(lpEntryID);
        
    return hr;
}

HRESULT IMAP::HrFindFolderEntryID(string strFolder, ULONG *lpcbEntryID, LPENTRYID *lppEntryID)
{
    HRESULT hr = hrSuccess;
    vector<string> vFolders;
    ULONG cbEntryID = 0;
    LPENTRYID lpEntryID = NULL;
    IMAPIFolder *lpFolder = NULL;
    ULONG ulObjType = 0;
    
    hr = HrSplitPath(strFolder, vFolders);
    if(hr != hrSuccess)
        goto exit;
        
    for(unsigned int i = 0; i < vFolders.size(); i++) {
        hr = HrFindSubFolder(lpFolder, vFolders[i], &cbEntryID, &lpEntryID);
        if(hr != hrSuccess)
            goto exit;
            
        if(i == vFolders.size() - 1)
            break;
            
        if(lpFolder)
            lpFolder->Release();
        lpFolder = NULL;
            
        hr = lpSession->OpenEntry(cbEntryID, lpEntryID, NULL, MAPI_MODIFY, &ulObjType, (IUnknown **)&lpFolder);
        if(hr != hrSuccess)
            goto exit;
            
        if(ulObjType != MAPI_FOLDER) {
            hr = MAPI_E_INVALID_PARAMETER;
            goto exit;
        }
            
        MAPIFreeBuffer(lpEntryID);
        lpEntryID = NULL;
    }
    
    *lpcbEntryID = cbEntryID;
    *lppEntryID = lpEntryID;
    lpEntryID = NULL;
            
exit:
    if(lpEntryID)
        MAPIFreeBuffer(lpEntryID);
        
    if(lpFolder)
        lpFolder->Release();
        
    return hr;
}

HRESULT IMAP::HrFindSubFolder(IMAPIFolder *lpFolder, string strName, ULONG *lpcbEntryID, LPENTRYID *lppEntryID)
{
    HRESULT hr = hrSuccess;
    IMAPITable *lpTable = NULL;
    SRestriction sRestrict;
    SPropValue sProp;
    SizedSPropTagArray(1, sptaCols) = {1, { PR_ENTRYID }};
    LPENTRYID lpEntryID = NULL;
    ULONG cbEntryID = 0;
    LPSRowSet lpRowSet = NULL;
    LPSPropValue lpProp = NULL;
    IMAPIFolder *lpSubTree = NULL;
    ULONG ulObjType = 0;
    
    sRestrict.rt = RES_PROPERTY;
    sRestrict.res.resProperty.ulPropTag = PR_DISPLAY_NAME;
    sRestrict.res.resProperty.relop = RELOP_EQ;
    sRestrict.res.resProperty.lpProp = &sProp;
    
    sProp.ulPropTag = PR_DISPLAY_NAME;
    sProp.Value.lpszA = (char *)strName.c_str();
    
    // lpFolder is NULL when we're referring to the IMAP root. The IMAP root contains
    // INBOX, the public folder container, and all folders under the users IPM_SUBTREE.
    
    if(lpFolder == NULL) {
        if(stricmp(strName.c_str(), "INBOX") == 0) {
            // Inbox request, we know where that is.
            hr = lpStore->GetReceiveFolder("IPM", 0, lpcbEntryID, lppEntryID, NULL);
            goto exit;            
        } else if(stricmp(strName.c_str(), PUBLIC_FOLDERS_NAME) == 0) {
            // Public folders requested, we know where that is too
			if(!lpPublicStore) {
				hr = MAPI_E_NOT_FOUND;
				goto exit;
			}

            hr = HrGetOneProp(lpPublicStore, PR_IPM_PUBLIC_FOLDERS_ENTRYID, &lpProp);
            if(hr != hrSuccess)
                goto exit;
                
            cbEntryID = lpProp->Value.bin.cb;
            hr = MAPIAllocateBuffer(cbEntryID, (void **)&lpEntryID);
            if(hr != hrSuccess)
                goto exit;
                
            memcpy(lpEntryID, lpProp->Value.bin.lpb, cbEntryID);
            
            *lppEntryID = lpEntryID;
            *lpcbEntryID = cbEntryID;
            
            goto exit;
        } else {
            // Other folder in the root requested, use normal search algorithm to find it
            // under IPM_SUBTREE
            
            hr = HrGetOneProp(lpStore, PR_IPM_SUBTREE_ENTRYID, &lpProp);
            if(hr != hrSuccess)
                goto exit;
                
            hr = lpStore->OpenEntry(lpProp->Value.bin.cb, (LPENTRYID)lpProp->Value.bin.lpb, NULL, 0, &ulObjType, (IUnknown **)&lpSubTree);
            if(hr != hrSuccess)
                goto exit;
                
            lpFolder = lpSubTree;
            // Fall through to normal folder lookup code
        }
    }
    
    // Use a restriction to find the folder in the hierarchy table
    hr = lpFolder->GetHierarchyTable(MAPI_DEFERRED_ERRORS, &lpTable);
    if(hr != hrSuccess)
        goto exit;
     
    hr = lpTable->SetColumns((LPSPropTagArray)&sptaCols, 0);
    if(hr != hrSuccess)
        goto exit;
        
    hr = lpTable->Restrict(&sRestrict, TBL_BATCH);
    if(hr != hrSuccess)
        goto exit;
        
    hr = lpTable->QueryRows(1, 0, &lpRowSet);
    if(hr != hrSuccess)
        goto exit;
        
    if(lpRowSet->cRows == 0) {
        hr = MAPI_E_NOT_FOUND;
        goto exit;
    }
    
    if(lpRowSet->aRow[0].lpProps[0].ulPropTag != PR_ENTRYID) {
        hr = MAPI_E_INVALID_PARAMETER;
        goto exit;
    }
    
    cbEntryID = lpRowSet->aRow[0].lpProps[0].Value.bin.cb;

    hr = MAPIAllocateBuffer(cbEntryID, (void **)&lpEntryID);
    if(hr != hrSuccess)
        goto exit;
        
    memcpy(lpEntryID, lpRowSet->aRow[0].lpProps[0].Value.bin.lpb, cbEntryID);
    
    *lppEntryID = lpEntryID;
    *lpcbEntryID = cbEntryID;
    
exit:
    if(lpSubTree)
        lpSubTree->Release();
    if(lpProp)
        MAPIFreeBuffer(lpProp);
    if(lpRowSet)
        FreeProws(lpRowSet);
    if(lpTable)
        lpTable->Release();
         
    return hr;
}

/**
 * Find the deepest folder in strFolder and return that IMAPIFolder
 * and the remaining folders which were not found.
 * If no folders are found from strFolder at all, return the IPM subtree.
 *
 * @param[in]	strFolder	Folder string complete path to find deepest IMAPIFolder for
 * @param[out]	lppFolder	Last IMAPIFolder found in strFolder
 * @param[out]	strNotFound	Folders not found in strFolder
 */
HRESULT IMAP::HrFindFolderPartial(string strFolder, IMAPIFolder **lppFolder, string *strNotFound)
{
    HRESULT hr = hrSuccess;
    vector<string> vFolders;
    ULONG cbEntryID = 0;
    LPENTRYID lpEntryID = NULL;
    IMAPIFolder *lpFolder = NULL;
    ULONG ulObjType = 0;
    LPSPropValue lpTree = NULL;
    unsigned int i = 0;
    
    hr = HrSplitPath(strFolder, vFolders);
    if(hr != hrSuccess)
        goto exit;
        
    // Loop through all the path parts until we find a part that we can't find
    for(i = 0; i < vFolders.size(); i++) {
        hr = HrFindSubFolder(lpFolder, vFolders[i], &cbEntryID, &lpEntryID);
        if(hr != hrSuccess) {
            hr = hrSuccess; // Not an error
            break;
        }
        
        if(lpFolder)
            lpFolder->Release();
        lpFolder = NULL;
            
        hr = lpSession->OpenEntry(cbEntryID, lpEntryID, NULL, MAPI_MODIFY, &ulObjType, (IUnknown **)&lpFolder);
        if(hr != hrSuccess)
            goto exit;
            
        MAPIFreeBuffer(lpEntryID);
        lpEntryID = NULL;
        
    }
    
    // Remove parts that we already have processed
    vFolders.erase(vFolders.begin(),vFolders.begin()+i);

    // The remaining path parts are the relative path that we could not find    
    hr = HrUnsplitPath(vFolders, *strNotFound);
    if(hr != hrSuccess)
        goto exit;

    if(lpFolder == NULL) {
        hr = HrGetOneProp(lpStore, PR_IPM_SUBTREE_ENTRYID, &lpTree);
        if(hr != hrSuccess)
            goto exit;
            
        hr = lpSession->OpenEntry(lpTree->Value.bin.cb, (LPENTRYID)lpTree->Value.bin.lpb, NULL, MAPI_MODIFY, &ulObjType, (IUnknown **)&lpFolder);
        if(hr != hrSuccess)
            goto exit;
    }
            
    *lppFolder = lpFolder;
    lpFolder = NULL;
    
exit:
    if(lpTree)
        MAPIFreeBuffer(lpTree);
    if(lpFolder)
        lpFolder->Release();
        
    if(lpEntryID)
        MAPIFreeBuffer(lpEntryID);
        
    return hr;
}

bool IMAP::IsSpecialFolder(IMAPIFolder *lpFolder)
{
    bool result = false;
    LPSPropValue lpProp = NULL;
    HRESULT hr = hrSuccess;
    
    hr = HrGetOneProp(lpFolder, PR_ENTRYID, &lpProp);
    if(hr != hrSuccess)
        goto exit;
        
    result = IsSpecialFolder(lpProp->Value.bin.cb, (LPENTRYID)lpProp->Value.bin.lpb);
    
exit:
    if(lpProp)
        MAPIFreeBuffer(lpProp);
        
    return result;
}

bool IMAP::IsMailFolder(IMAPIFolder *lpFolder)
{
    bool result = false;
    LPSPropValue lpProp = NULL;
    HRESULT hr = hrSuccess;
    
    hr = HrGetOneProp(lpFolder, PR_CONTAINER_CLASS, &lpProp);
    if(hr != hrSuccess) {
		// if the property is missing, treat it as an email folder
		result = true;
        goto exit;
	}
        
    if(strlen(lpProp->Value.lpszA) > 0 && stricmp(lpProp->Value.lpszA, "IPM") != 0 && stricmp(lpProp->Value.lpszA, "IPF.NOTE") != 0) 
        result = false;
    else
        result = true;
    
exit:
    if(lpProp)
        MAPIFreeBuffer(lpProp);
        
    return result;
}

HRESULT IMAP::HrOpenParentFolder(ULONG cbEntryID, LPENTRYID lpEntryID, IMAPIFolder **lppFolder)
{
    HRESULT hr = hrSuccess;
    IMAPIFolder *lpFolder = NULL;
    ULONG ulObjType = 0;

    hr = lpSession->OpenEntry(cbEntryID, lpEntryID, NULL, MAPI_MODIFY, &ulObjType, (IUnknown **)&lpFolder);
    if(hr != hrSuccess)
        goto exit;
        
    hr = HrOpenParentFolder(lpFolder, lppFolder);

exit:
    if(lpFolder)
        lpFolder->Release();
        
    return hr;  
}

HRESULT IMAP::HrOpenParentFolder(IMAPIFolder *lpFolder, IMAPIFolder **lppFolder)
{
    HRESULT hr = hrSuccess;
    LPSPropValue lpParent = NULL;
    ULONG ulObjType = 0;
    
    hr = HrGetOneProp(lpFolder, PR_PARENT_ENTRYID, &lpParent);
    if(hr != hrSuccess)
        goto exit;
        
    hr = lpSession->OpenEntry(lpParent->Value.bin.cb, (LPENTRYID)lpParent->Value.bin.lpb, NULL, MAPI_MODIFY, &ulObjType, (IUnknown **)lppFolder);
    if(hr != hrSuccess)
        goto exit;
    
exit:
    if(lpParent)
        MAPIFreeBuffer(lpParent);
        
    return hr;
}

